!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   input_mod
!> @brief   input parameters and data for analysis
!! @authors Norio Takase (NT)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../../config.h"
#endif

module input_mod

  use input_str_mod
  use fileio_control_mod
  use fileio_gpr_mod
  use fileio_pdb_mod
  use fileio_psf_mod
  use string_mod
  use messages_mod

  implicit none
  private

  ! structures
  type, public :: s_inp_info
    character(MaxFilename) :: psffile  = ''
    character(MaxFilename) :: reffile  = ''
    character(MaxFilename) :: coorfile = ''
    character(MaxFilename) :: velfile  = ''
    character(MaxFilename) :: xscfile  = ''
    character(MaxFilename) :: dcdfile  = ''
    character(MaxFilename) :: remfile  = ''
    character(MaxFilename) :: pdbfile  = ''
    character(MaxFilename) :: gprfile  = ''
    character(MaxFilename) :: rstfile  = ''
  end type s_inp_info

  public  :: show_ctrl_input
  public  :: read_ctrl_input
  public  :: setup_input
  public  :: input_files

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_input
  !> @brief        show control parameters in INPUT section
  !! @authors      NT
  !! @param[in]    keys : comma separated keys for show
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_input(keys)

    ! formal arguments
    character(*),            intent(in)    :: keys


    if (len_trim(keys) == 0) &
      return

    write(MsgOut,'(A)') '[INPUT]'

    if (index(keys,'psf') /= 0) &
      write(MsgOut,'(A)') 'psffile        = input.psf       # protein structure file'

    if (index(keys,'ref') /= 0) &
      write(MsgOut,'(A)') 'reffile        = input.pdb       # PDB file'

    if (index(keys,'coor') /= 0) &
      write(MsgOut,'(A)') 'coorfile       = input.coor      # coordinate restart file'

    if (index(keys,'vel') /= 0) &
      write(MsgOut,'(A)') 'velfile        = input.vel       # velocity restart file'

    if (index(keys,'xsc') /= 0) &
      write(MsgOut,'(A)') 'xscfile        = input.xsc       # extended system file'

    if (index(keys,'dcd') /= 0) &
      write(MsgOut,'(A)') 'dcdfile        = input.dcd       # DCD file'

    if (index(keys,'DCD') /= 0) &
      write(MsgOut,'(A)') 'dcdfile        = input{}.dcd     # DCD file'

    if (index(keys,'rem') /= 0) &
      write(MsgOut,'(A)') 'remfile        = input.rem       # REMD parameter ID file'
    
    if (index(keys,'REM') /= 0) &
      write(MsgOut,'(A)') 'remfile        = input{}.rem     # REMD parameter ID file'
    
    if (index(keys,'pdb') /= 0) &
      write(MsgOut,'(A)') 'pdbfile        = input.pdb       # PDB file'

!    if (index(keys,'gpr') /= 0) &
!      write(MsgOut,'(A)') 'gprfile        = input.gpr       # GPR file'

    if (index(keys,'rst') /= 0) &
      write(MsgOut,'(A)') 'rstfile        = input.rst       # restart file'

    write(MsgOut,'(A)') ' '

    return

  end subroutine show_ctrl_input

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_ctrl_input
  !> @brief        read control parameters in INPUT section
  !! @authors      NT
  !! @param[in]    handle   : unit number of control file
  !! @param[out]   inp_info : INPUT section control parameters information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine read_ctrl_input(handle, inp_info)
  
    ! parameters
    character(*),            parameter :: Section = 'Input'
    
    ! formal argments
    integer,                 intent(in)    :: handle
    type(s_inp_info),        intent(inout) :: inp_info


    ! read parameters from control file
    ! 
    call read_ctrlfile_string(handle, Section, 'Psffile', inp_info%psffile)
    call read_ctrlfile_string(handle, Section, 'Reffile', inp_info%reffile)
    call read_ctrlfile_string(handle, Section, 'Coorfile',inp_info%coorfile)
    call read_ctrlfile_string(handle, Section, 'Velfile', inp_info%velfile)
    call read_ctrlfile_string(handle, Section, 'Xscfile', inp_info%xscfile)
    call read_ctrlfile_string(handle, Section, 'Dcdfile', inp_info%dcdfile)
    call read_ctrlfile_string(handle, Section, 'Remfile', inp_info%remfile)
    call read_ctrlfile_string(handle, Section, 'Pdbfile', inp_info%pdbfile)
    call read_ctrlfile_string(handle, Section, 'gprfile', inp_info%gprfile)
    call read_ctrlfile_string(handle, Section, 'Rstfile', inp_info%rstfile)


    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_Ctrl_Input> Input Files'
    if (len_trim(inp_info%psffile) > 0) &
      write(MsgOut,'(A20,A)') '  psffile         = ', trim(inp_info%psffile)
    if (len_trim(inp_info%reffile) > 0) &
      write(MsgOut,'(A20,A)') '  reffile         = ', trim(inp_info%reffile)
    if (len_trim(inp_info%coorfile) > 0) &
      write(MsgOut,'(A20,A)') '  coorfile        = ', trim(inp_info%coorfile)
    if (len_trim(inp_info%velfile) > 0) &
      write(MsgOut,'(A20,A)') '  velfile         = ', trim(inp_info%velfile)
    if (len_trim(inp_info%xscfile) > 0) &
      write(MsgOut,'(A20,A)') '  xscfile         = ', trim(inp_info%xscfile)
    if (len_trim(inp_info%dcdfile) > 0) &
      write(MsgOut,'(A20,A)') '  dcdfile         = ', trim(inp_info%dcdfile)
    if (len_trim(inp_info%remfile) > 0) &
      write(MsgOut,'(A20,A)') '  remfile         = ', trim(inp_info%remfile)
    if (len_trim(inp_info%pdbfile) > 0) &
      write(MsgOut,'(A20,A)') '  pdbfile         = ', trim(inp_info%pdbfile)
    if (len_trim(inp_info%gprfile) > 0) &
      write(MsgOut,'(A20,A)') '  gprfile         = ', trim(inp_info%gprfile)
    if (len_trim(inp_info%rstfile) > 0) &
      write(MsgOut,'(A20,A)') '  rstfile         = ', trim(inp_info%rstfile)
    write(MsgOut,'(A)') ' '

    return

  end subroutine read_ctrl_input

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_input
  !> @brief        setup input files
  !! @authors      NT
  !! @param[in]    inp_info : INPUT section control parameters information
  !! @param[out]   input    : input information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_input(inp_info, input)

    ! formal argments
    type(s_inp_info),        intent(in)    :: inp_info
    type(s_input),           intent(inout) :: input


    input%psffile  = inp_info%psffile
    input%reffile  = inp_info%reffile
    input%coorfile = inp_info%coorfile
    input%velfile  = inp_info%velfile
    input%xscfile  = inp_info%xscfile
    input%dcdfile  = inp_info%dcdfile
    input%remfile  = inp_info%remfile
    input%pdbfile  = inp_info%pdbfile
    input%gprfile  = inp_info%gprfile
    input%rstfile  = inp_info%rstfile

    return

  end subroutine setup_input

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    input_files
  !> @brief        read input data from files
  !! @authors      NT
  !! @param[in]    inp_info : INPUT section control parameters information
  !! @param[out]   psf      : PSF information [optional]
  !! @param[out]   ref      : coordinate data for reference [optional]
  !! @param[out]   gpr      : GPR information [optional]
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine input_files(inp_info, psf, ref, gpr)

    ! formal arguments
    type(s_inp_info),        intent(in)    :: inp_info
    type(s_psf),   optional, intent(inout) :: psf
    type(s_pdb),   optional, intent(inout) :: ref
    type(s_gpr),   optional, intent(inout) :: gpr


    if (len_trim(inp_info%psffile) > 0 .and. present(psf)) then
      call input_psf(inp_info%psffile, psf)
    end if

    if (len_trim(inp_info%reffile) > 0 .and. present(ref)) then
      call input_pdb(inp_info%reffile, ref)
    end if

    if (len_trim(inp_info%gprfile) > 0 .and. present(gpr)) then
      call input_gpr(inp_info%gprfile, gpr)
    end if

    return

  end subroutine input_files

end module input_mod
