!--------1---------2---------3---------4---------5---------6---------7---------8
! 
!> Module   cg_convert_mod
!! @brief   module for convert
!! @authors Takaharu Mori (TM)
!!
!! (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../../../config.h"
#endif

module cg_convert_mod

  use cg_option_str_mod
  use output_str_mod
  use input_str_mod
  use fileio_psf_mod
  use fileio_pdb_mod
  use fileio_gpr_mod
  use messages_mod
  use constants_mod

  implicit none
  private

  public  :: convert
  private :: make_gpr_nonpair
  private :: make_gpr_pair
  private :: make_gpr_bond
  private :: make_gpr_angl
  private :: make_gpr_dihe
  private :: make_gpr_dihe_cago
  private :: make_gpr_impr

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    convert
  !> @brief        convert CHARMM to GO parameters
  !! @authors      TM
  !! @param[in]    input  : input information
  !! @param[in]    output : output information
  !! @param[inout] option : option information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine convert(input, output, option)

    ! formal arguments
    type(s_input),           intent(in)    :: input
    type(s_output),          intent(in)    :: output
    type(s_option),          intent(inout) :: option

    ! local variables
    type(s_psf)              :: psf
    type(s_pdb)              :: pdb
    type(s_gpr)              :: gpr


    ! check check-only
    !
    if (option%check_only) return


    ! Read pdb and psf files
    !
    call input_pdb(input%pdbfile, pdb)
    call input_psf(input%psffile, psf)


    ! make GPR data
    !
    call init_gpr(gpr)
    call make_gpr_nonpair(option, gpr)
    call make_gpr_pair   (option, pdb, psf, gpr)
    call make_gpr_bond   (option, pdb, psf, gpr)
    call make_gpr_angl   (option, pdb, psf, gpr)

    if (option%forcefield == ForceFieldAAGO) then

      call make_gpr_dihe(option, pdb, psf, gpr)
      call make_gpr_impr(option, pdb, psf, gpr)

    end if

    ! output GPR file
    !
    call output_gpr(output%gprfile, gpr)

    ! deallocate array
    !
    call dealloc_pdb_all(pdb)
    call dealloc_psf_all(psf)
    call dealloc_gpr_all(gpr)

    return

  end subroutine convert

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_gpr_nonpair
  !> @brief        make gpr nonpair
  !! @authors      TM
  !! @param[in]    boundary :
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine make_gpr_nonpair(option, gpr)

    ! formal arguments
    type(s_option),          intent(in)    :: option
    type(s_gpr),             intent(inout) :: gpr


    gpr%nonpair_eps      = option%e_nc
    gpr%nonpair_dist_min = option%s_nc


    return

  end subroutine make_gpr_nonpair

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_gpr_pair
  !> @brief        make gpr pair
  !! @authors      TM
  !! @param[in]    boundary :
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine make_gpr_pair(option, pdb, psf, gpr)

    ! formal arguments
    type(s_option),          intent(inout) :: option
    type(s_pdb),     target, intent(in)    :: pdb
    type(s_psf),             intent(in)    :: psf
    type(s_gpr),             intent(inout) :: gpr

    ! local variables
    real(wp)                 :: dx12, dy12, dz12, r12
    real(wp)                 :: eps_c
    integer                  :: i, j, k, found, num_contact
    logical                  :: exclude

    real(wp),        pointer :: coord(:,:)


    coord => pdb%atom_coord


    ! count the number of native contact
    !
    found = 0
    do i = 1, psf%num_atoms - 1
      do j = i + 1, psf%num_atoms

        dx12 = coord(1,i) - coord(1,j)
        dy12 = coord(2,i) - coord(2,j)
        dz12 = coord(3,i) - coord(3,j)
        r12 = sqrt(dx12*dx12 + dy12*dy12 + dz12*dz12)

        if (r12 <= option%cutoff) then

          ! 1-2 exclusion
          !
          exclude = .false.
          do k = 1, psf%num_bonds
            if (i == psf%bond_list(1,k) .and. j == psf%bond_list(2,k) .or. &
                i == psf%bond_list(2,k) .and. j == psf%bond_list(1,k)) then
              exclude = .true.
            end if
          end do

          ! 1-3 exclusion
          !
          do k = 1, psf%num_angles
            if (i == psf%angl_list(1,k) .and. j == psf%angl_list(3,k) .or. &
                i == psf%angl_list(3,k) .and. j == psf%angl_list(1,k)) then
              exclude = .true.
            end if
          end do
          
          ! 1-4 exclusion
          !
          do k = 1, psf%num_dihedrals
            if (i == psf%dihe_list(1,k) .and. j == psf%dihe_list(4,k) .or. &
                i == psf%dihe_list(4,k) .and. j == psf%dihe_list(1,k)) then
              exclude = .true.
            end if
          end do
          do k = 1, psf%num_impropers
            if (i == psf%impr_list(1,k) .and. j == psf%impr_list(4,k) .or. &
                i == psf%impr_list(4,k) .and. j == psf%impr_list(1,k)) then
              exclude = .true.
            end if
          end do

          ! count the number of native contact
          !
          if (.not. exclude) then
            if (psf%segment_name(i) == psf%segment_name(j)) then
              if (abs(psf%residue_no(i) - psf%residue_no(j)) &
                                                    >= option%min_dist) then
                found = found + 1
               end if
            else
              found = found + 1
            end if
          end if

        end if

      end do
    end do
    num_contact = found


    ! allocatation for PAIRS
    !
    gpr%num_pairs = num_contact
    call alloc_gpr(gpr, GprPair, gpr%num_pairs)


    if (option%forcefield == ForceFieldAAGO) then
      if (num_contact /= 0) then
        eps_c = (option%r_cd/(1.0_wp+option%r_cd)) * psf%num_atoms / num_contact
      else
        eps_c = 1.0_wp
        option%r_cd = 0.0_wp
      end if
    end if


    found = 0
    do i = 1, psf%num_atoms - 1
      do j = i + 1, psf%num_atoms

        dx12 = coord(1,i) - coord(1,j)
        dy12 = coord(2,i) - coord(2,j)
        dz12 = coord(3,i) - coord(3,j)
        r12 = sqrt(dx12*dx12 + dy12*dy12 + dz12*dz12)

        if (r12 <= option%cutoff) then
          exclude = .false.
          do k = 1, psf%num_bonds
            if (i == psf%bond_list(1,k) .and. j == psf%bond_list(2,k) .or. &
                i == psf%bond_list(2,k) .and. j == psf%bond_list(1,k)) then
               exclude = .true.
            end if
          end do
          do k = 1, psf%num_angles
            if (i == psf%angl_list(1,k) .and. j == psf%angl_list(3,k) .or. &
                i == psf%angl_list(3,k) .and. j == psf%angl_list(1,k)) then
               exclude = .true.
            end if
          end do
          do k = 1, psf%num_dihedrals
            if (i == psf%dihe_list(1,k) .and. j == psf%dihe_list(4,k) .or. &
                i == psf%dihe_list(4,k) .and. j == psf%dihe_list(1,k)) then
               exclude = .true.
            end if
          end do
          do k = 1, psf%num_impropers
            if (i == psf%impr_list(1,k) .and. j == psf%impr_list(4,k) .or. &
                i == psf%impr_list(4,k) .and. j == psf%impr_list(1,k)) then
               exclude = .true.
            end if
          end do

          if (.not. exclude) then
            if (psf%segment_name(i) == psf%segment_name(j)) then
              if (abs(psf%residue_no(i) - psf%residue_no(j)) &
                                                    >= option%min_dist) then
                found = found + 1
                gpr%pair_list(1,found) = i
                gpr%pair_list(2,found) = j
                gpr%pair_eps(found)    = eps_c
                gpr%pair_dist_min(found) = r12
              end if
            else
              found = found + 1
              gpr%pair_list(1,found) = i
              gpr%pair_list(2,found) = j
              gpr%pair_eps(found)    = eps_c
              gpr%pair_dist_min(found) = r12
            end if
          end if

        end if

      end do
    end do

    return

  end subroutine make_gpr_pair

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_gpr_bond
  !> @brief        make gpr bond
  !! @authors      TM
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine make_gpr_bond(option, pdb, psf, gpr)

    ! formal arguments
    type(s_option),          intent(in)    :: option
    type(s_pdb),     target, intent(in)    :: pdb
    type(s_psf),     target, intent(in)    :: psf
    type(s_gpr),             intent(inout) :: gpr

    ! local variables
    real(wp)                 :: dx12, dy12, dz12, r12
    integer                  :: i, num_bonds

    real(wp),        pointer :: coord(:,:)
    integer,         pointer :: list(:,:)


    list      => psf%bond_list
    coord     => pdb%atom_coord
    num_bonds =  psf%num_bonds


    ! allocation for BONDS
    !
    gpr%num_bonds = num_bonds
    call alloc_gpr(gpr, GprBond, gpr%num_bonds)

    do i = 1, num_bonds

      dx12 = coord(1,list(1,i)) - coord(1,list(2,i))
      dy12 = coord(2,list(1,i)) - coord(2,list(2,i))
      dz12 = coord(3,list(1,i)) - coord(3,list(2,i))
      r12 = sqrt(dx12*dx12 + dy12*dy12 + dz12*dz12)

      gpr%bond_list(1,i) = list(1,i)
      gpr%bond_list(2,i) = list(2,i)
      gpr%bond_dist_min(i) = r12
      gpr%bond_force_const(i) = option%k_bond

    end do

    return

  end subroutine make_gpr_bond

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_gpr_angle
  !> @brief        make gpr angle
  !! @authors      TM
  !! @param[in]    
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine make_gpr_angl(option, pdb, psf, gpr)

    ! formal arguments
    type(s_option),          intent(in)    :: option
    type(s_pdb),     target, intent(in)    :: pdb
    type(s_psf),     target, intent(in)    :: psf
    type(s_gpr),             intent(inout) :: gpr

    ! local variables
    real(wp)                 :: dx12, dy12, dz12, r12, r12r32
    real(wp)                 :: dx32, dy32, dz32, r32
    real(wp)                 :: cost, theta
    integer                  :: i, num_angl

    real(wp),        pointer :: coord(:,:)
    integer,         pointer :: list(:,:)


    list     => psf%angl_list
    coord    => pdb%atom_coord
    num_angl =  psf%num_angles


    ! allocation for ANGLES
    !
    gpr%num_angles = num_angl
    call alloc_gpr(gpr, GprAngl, gpr%num_angles)


    do i = 1, num_angl

      dx12 = coord(1,list(1,i)) - coord(1,list(2,i))
      dy12 = coord(2,list(1,i)) - coord(2,list(2,i))
      dz12 = coord(3,list(1,i)) - coord(3,list(2,i))
      dx32 = coord(1,list(3,i)) - coord(1,list(2,i))
      dy32 = coord(2,list(3,i)) - coord(2,list(2,i))
      dz32 = coord(3,list(3,i)) - coord(3,list(2,i))
      r12 = dx12*dx12 + dy12*dy12 + dz12*dz12
      r32 = dx32*dx32 + dy32*dy32 + dz32*dz32
      r12r32 = sqrt( r12*r32 )
      cost   = dx12*dx32 + dy12*dy32 + dz12*dz32
      r12r32 = 1.0_wp / r12r32
      r12    = 1.0_wp / r12
      r32    = 1.0_wp / r32
      cost   = cost * r12r32
      if (cost >  1.0_wp) cost = 1.0_wp
      if (cost < -1.0_wp) cost = -1.0_wp
      theta = acos(cost)

      gpr%angl_list(1,i) = list(1,i)
      gpr%angl_list(2,i) = list(2,i)
      gpr%angl_list(3,i) = list(3,i)
      gpr%angl_theta_min(i)   = theta/RAD
      gpr%angl_force_const(i) = option%k_angle

    end do

    return

  end subroutine make_gpr_angl

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_gpr_dihe
  !> @brief        make gpr dihe
  !! @authors      TM
  !! @param[in]  
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine make_gpr_dihe(option, pdb, psf, gpr)
    
    ! formal arguments
    type(s_option),          intent(in)    :: option
    type(s_pdb),     target, intent(in)    :: pdb
    type(s_psf),     target, intent(in)    :: psf
    type(s_gpr),             intent(inout) :: gpr

    ! local variables
    real(wp)                 :: dx21, dy21, dz21, dx32, dy32, dz32
    real(wp)                 :: dx43, dy43, dz43
    real(wp)                 :: px12, py12, pz12, px23, py23, pz23
    real(wp)                 :: px123, py123, pz123
    real(wp)                 :: r12, r23, r12r23, s1223, s32123
    real(wp)                 :: cosp, phi
    real(wp)                 :: e_bb, e_sc, sum_e_bb, sum_e_sc
    real(wp)                 :: scale_dihe
    integer                  :: i, j, m, n, num_atom, num_dihe, found
    integer                  :: an2i, an3i, an2j, an3j
    character(4)             :: ac2i, ac3i
    logical                  :: backbone

    real(wp),        pointer :: coord(:,:)
    integer,         pointer :: list(:,:)
    character(4),    pointer :: atom_name(:)


    list      => psf%dihe_list
    atom_name => psf%atom_name
    coord     => pdb%atom_coord
    num_atom  =  psf%num_atoms
    num_dihe  =  psf%num_dihedrals


    ! allocation for DIHE
    !
    gpr%num_dihedrals = num_dihe*2
    call alloc_gpr(gpr, GprDihe, gpr%num_dihedrals)


    ! compute total e_bb and e_sc
    !
    sum_e_bb = 0.0_wp
    sum_e_sc = 0.0_wp
    e_sc = 1.0_wp
    e_bb = option%r_bbsc * e_sc
    do i = 1, num_dihe

      ! find same 2-3 definition
      !
      backbone = .false.
      an2i = list(2,i)
      an3i = list(3,i)
      found = 0
      do j = 1, num_dihe
        an2j = list(2,j)
        an3j = list(3,j)

        if ( (an2i == an2j .and. an3i == an3j) .or.      &
             (an2i == an3j .and. an3i == an2j)    ) then
          found = found + 1
        end if
      end do
      ac2i = atom_name(an2i)
      ac3i = atom_name(an3i)
      if ( (ac2i == 'CA  ' .and. ac3i == 'C   ') .or. &
           (ac2i == 'C   ' .and. ac3i == 'CA  ') .or. &
           (ac2i == 'N   ' .and. ac3i == 'CA  ') .or. &
           (ac2i == 'CA  ' .and. ac3i == 'N   ') ) then
        backbone = .true.
      end if

      if(backbone) then
        sum_e_bb = sum_e_bb + e_bb/real(found,wp)
      else
        sum_e_sc = sum_e_sc + e_sc/real(found,wp)
      end if
    end do


    ! set scale factor
    !
    if (gpr%num_pairs == 0) then
      scale_dihe = 1.0_wp
    else
      scale_dihe = &
           (real(num_atom,wp)/(1.0_wp + option%r_cd))/(sum_e_bb + sum_e_sc)
    end if


    do i = 1, num_dihe

      ! find same 2-3 definition
      !
      backbone = .false.
      an2i = list(2,i)
      an3i = list(3,i)
      found = 0
      do j = 1, num_dihe
        an2j = list(2,j)
        an3j = list(3,j)

        if ( (an2i == an2j .and. an3i == an3j) .or. &
             (an2i == an3j .and. an3i == an2j) ) then
          found = found + 1
        end if
      end do
      ac2i = atom_name(an2i)
      ac3i = atom_name(an3i)
      if ( (ac2i == 'CA  ' .and. ac3i == 'C   ') .or. &
           (ac2i == 'C   ' .and. ac3i == 'CA  ') .or. &
           (ac2i == 'N   ' .and. ac3i == 'CA  ') .or. &
           (ac2i == 'CA  ' .and. ac3i == 'N   ') ) then
        backbone = .true.
      end if


      dx21 = coord(1,list(2,i)) - coord(1,list(1,i))
      dy21 = coord(2,list(2,i)) - coord(2,list(1,i))
      dz21 = coord(3,list(2,i)) - coord(3,list(1,i))
      dx32 = coord(1,list(3,i)) - coord(1,list(2,i))
      dy32 = coord(2,list(3,i)) - coord(2,list(2,i))
      dz32 = coord(3,list(3,i)) - coord(3,list(2,i))
      dx43 = coord(1,list(4,i)) - coord(1,list(3,i))
      dy43 = coord(2,list(4,i)) - coord(2,list(3,i))
      dz43 = coord(3,list(4,i)) - coord(3,list(3,i))

      px12 = dy21*dz32 - dy32*dz21
      py12 = dz21*dx32 - dz32*dx21
      pz12 = dx21*dy32 - dx32*dy21
      px23 = dy32*dz43 - dy43*dz32
      py23 = dz32*dx43 - dz43*dx32
      pz23 = dx32*dy43 - dx43*dy32
      r12  = px12*px12 + py12*py12 + pz12*pz12
      r23  = px23*px23 + py23*py23 + pz23*pz23
      r12r23 = sqrt( r12 * r23 )

      if ( r12r23 >  0.0_wp ) then

        r12r23    = 1.0_wp / r12r23
        s1223     = px12*px23 + py12*py23 + pz12*pz23
        cosp      = s1223 * r12r23
        if (cosp > 1.0_wp  ) cosp = 1.0_wp
        if (cosp < -1.0_wp ) cosp = -1.0_wp

        px123     = py12*pz23 - py23*pz12
        py123     = pz12*px23 - pz23*px12
        pz123     = px12*py23 - px23*py12
        s32123    = dx32*px123 + dy32*py123 + dz32*pz123

        phi  = acos(cosp)
        if (s32123 < 0.0_wp) then
          phi = ( 2.0_wp*PI ) - phi
        end if

      end if


      m = 2*i - 1
      n = 2*i
      gpr%dihe_list(1,m) = list(1,i)
      gpr%dihe_list(2,m) = list(2,i)
      gpr%dihe_list(3,m) = list(3,i)
      gpr%dihe_list(4,m) = list(4,i)
      gpr%dihe_phase(m)  = PHI/RAD + 180.0_wp
      gpr%dihe_periodicity(m) = 1

      gpr%dihe_list(1,n) = list(1,i)
      gpr%dihe_list(2,n) = list(2,i)
      gpr%dihe_list(3,n) = list(3,i)
      gpr%dihe_list(4,n) = list(4,i)
      gpr%dihe_phase(n)  = 3.0_wp*PHI/RAD + 180.0_wp
      gpr%dihe_periodicity(n) = 3

      if (backbone) then
        gpr%dihe_force_const(m) = scale_dihe*e_bb/real(found,wp)
        gpr%dihe_force_const(n) = scale_dihe*0.5_wp*e_bb/real(found,wp)
      else
        gpr%dihe_force_const(m) = scale_dihe*e_sc/real(found,wp)
        gpr%dihe_force_const(n) = scale_dihe*0.5_wp*e_sc/real(found,wp)
      end if

    end do

    return

  end subroutine make_gpr_dihe

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_gpr_dihe_cago
  !> @brief        make gpr dihe cago
  !! @authors      TM
  !! @param[in]  
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine make_gpr_dihe_cago(option, pdb, psf, gpr)
    
    ! formal arguments
    type(s_option),          intent(in)    :: option
    type(s_pdb),     target, intent(in)    :: pdb
    type(s_psf),     target, intent(in)    :: psf
    type(s_gpr),             intent(inout) :: gpr

    ! local variables
    real(wp)                 :: dx21, dy21, dz21, dx32, dy32, dz32
    real(wp)                 :: dx43, dy43, dz43
    real(wp)                 :: px12, py12, pz12, px23, py23, pz23
    real(wp)                 :: px123, py123, pz123
    real(wp)                 :: r12, r23, r12r23, s1223, s32123
    real(wp)                 :: cosp, phi
    real(wp)                 :: e_bb, e_sc, sum_e_bb, sum_e_sc
    real(wp)                 :: scale_dihe
    integer                  :: i, j, m, n, num_atom, num_dihe, found
    integer                  :: an2i, an3i, an2j, an3j
    character(4)             :: ac2i, ac3i
    logical                  :: backbone

    real(wp),        pointer :: coord(:,:)
    integer,         pointer :: list(:,:)
    character(4),    pointer :: atom_name(:)


    list      => psf%dihe_list
    atom_name => psf%atom_name
    coord     => pdb%atom_coord
    num_atom  =  psf%num_atoms
    num_dihe  =  psf%num_dihedrals


    ! allocation for DIHE
    !
    gpr%num_dihedrals = num_dihe*2
    call alloc_gpr(gpr, GprDihe, gpr%num_dihedrals)

    do i = 1, num_dihe
      dx21 = coord(1,list(2,i)) - coord(1,list(1,i))
      dy21 = coord(2,list(2,i)) - coord(2,list(1,i))
      dz21 = coord(3,list(2,i)) - coord(3,list(1,i))
      dx32 = coord(1,list(3,i)) - coord(1,list(2,i))
      dy32 = coord(2,list(3,i)) - coord(2,list(2,i))
      dz32 = coord(3,list(3,i)) - coord(3,list(2,i))
      dx43 = coord(1,list(4,i)) - coord(1,list(3,i))
      dy43 = coord(2,list(4,i)) - coord(2,list(3,i))
      dz43 = coord(3,list(4,i)) - coord(3,list(3,i))

      px12 = dy21*dz32 - dy32*dz21
      py12 = dz21*dx32 - dz32*dx21
      pz12 = dx21*dy32 - dx32*dy21
      px23 = dy32*dz43 - dy43*dz32
      py23 = dz32*dx43 - dz43*dx32
      pz23 = dx32*dy43 - dx43*dy32
      r12  = px12*px12 + py12*py12 + pz12*pz12
      r23  = px23*px23 + py23*py23 + pz23*pz23
      r12r23 = sqrt( r12 * r23 )

      if ( r12r23 >  0.0_wp ) then

        r12r23    = 1.0_wp / r12r23
        s1223     = px12*px23 + py12*py23 + pz12*pz23
        cosp      = s1223 * r12r23
        if (cosp > 1.0_wp  ) cosp = 1.0_wp
        if (cosp < -1.0_wp ) cosp = -1.0_wp

        px123     = py12*pz23 - py23*pz12
        py123     = pz12*px23 - pz23*px12
        pz123     = px12*py23 - px23*py12
        s32123    = dx32*px123 + dy32*py123 + dz32*pz123

        phi  = acos(cosp)
        if(s32123 < 0.0_wp) then
          phi = ( 2.0_wp*PI ) - phi
        endif

      end if


      m = 2*i - 1
      n = 2*i
      gpr%dihe_list(1,m) = list(1,i)
      gpr%dihe_list(2,m) = list(2,i)
      gpr%dihe_list(3,m) = list(3,i)
      gpr%dihe_list(4,m) = list(4,i)
      gpr%dihe_phase(m)  = PHI/RAD + 180.0_wp
      gpr%dihe_periodicity(m) = 1

      gpr%dihe_list(1,n) = list(1,i)
      gpr%dihe_list(2,n) = list(2,i)
      gpr%dihe_list(3,n) = list(3,i)
      gpr%dihe_list(4,n) = list(4,i)
      gpr%dihe_phase(n)  = 3.0_wp*PHI/RAD + 180.0_wp
      gpr%dihe_periodicity(n) = 3

      gpr%dihe_force_const(m) = option%e_dihedral
      gpr%dihe_force_const(n) = option%e_dihedral*0.5_wp

    end do

    return

  end subroutine make_gpr_dihe_cago

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    make_gpr_impr
  !> @brief        make gpr impr
  !! @note         copied from energy_dihedrals.f90
  !! @authors      TM
  !! @param[in]  
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine make_gpr_impr(option, pdb, psf, gpr)

    ! formal arguments
    type(s_option),          intent(in)    :: option
    type(s_pdb),     target, intent(in)    :: pdb
    type(s_psf),     target, intent(in)    :: psf
    type(s_gpr),             intent(inout) :: gpr

    ! local variables
    real(wp)                 :: dx21, dy21, dz21, dx32, dy32, dz32
    real(wp)                 :: dx43, dy43, dz43
    real(wp)                 :: px12, py12, pz12, px23, py23, pz23
    real(wp)                 :: px123, py123, pz123
    real(wp)                 :: r12,  r23,  r12r23, s1223,s32123
    real(wp)                 :: cosp, phi
    integer                  :: i, j, num_impr, found
    integer                  :: an1i, an2i, an3i, an4i
    integer                  :: an1j, an2j, an3j, an4j
    logical                  :: ring

    real(wp),        pointer :: coord(:,:)
    integer,         pointer :: list(:,:)
    character(4),    pointer :: atm_name(:), res_name(:)


    list      => psf%impr_list
    atm_name  => psf%atom_name
    res_name  => psf%residue_name
    coord     => pdb%atom_coord
    num_impr  =  psf%num_impropers


    ! allocation for IMPR
    !
    gpr%num_impropers = num_impr
    call alloc_gpr(gpr, GprImpr, gpr%num_impropers)


    do i = 1, num_impr

      ! find the same definition
      !
      an1i = list(1,i)
      an2i = list(2,i)
      an3i = list(3,i)
      an4i = list(4,i)
      found = 0

      if ((atm_name(an1i) == 'CB  ' .and. atm_name(an2i) == 'CA  ' .and. &
           atm_name(an3i) == 'C   ' .and. atm_name(an4i) == 'N   ') .or. &
           atm_name(an1i) == 'O   ' .and. atm_name(an2i) == 'CA  ' .and. &
           atm_name(an3i) == 'C   ' .and. atm_name(an4i) == 'N   ') then
        found = 1

      else
        do j = 1, num_impr
          an1j = list(1,j)
          an2j = list(2,j)
          an3j = list(3,j)
          an4j = list(4,j)
          if ( (an2i == an2j .and. an3i == an3j) .or. &
               (an2i == an3j .and. an3i == an2j) ) then
            found = found + 1
          end if
        end do
      end if

      ring = .false.
      ! ring or not
      !
      if (res_name(an2i) == 'HIS ' .and. res_name(an3i) == 'HIS ') then
        if ( (atm_name(an2i) == 'CG  ' .and. atm_name(an3i) == 'ND1 ') .or. &
             (atm_name(an3i) == 'CG  ' .and. atm_name(an2i) == 'ND1 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'ND1 ' .and. atm_name(an3i) == 'CE1 ') .or. &
             (atm_name(an3i) == 'ND1 ' .and. atm_name(an2i) == 'CE1 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CE1 ' .and. atm_name(an3i) == 'NE2 ') .or. &
             (atm_name(an3i) == 'CE1 ' .and. atm_name(an2i) == 'NE2 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'NE2 ' .and. atm_name(an3i) == 'CD2 ') .or. &
             (atm_name(an3i) == 'NE2 ' .and. atm_name(an2i) == 'CD2 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CD2 ' .and. atm_name(an3i) == 'CG  ') .or. &
             (atm_name(an3i) == 'CD2 ' .and. atm_name(an2i) == 'CG  ') ) then
          ring = .true.
        end if

      else if (res_name(an2i) == 'PHE ' .and. res_name(an3i) == 'PHE ') then
        if ( (atm_name(an2i) == 'CG  ' .and. atm_name(an3i) == 'CD1 ') .or. &
             (atm_name(an3i) == 'CG  ' .and. atm_name(an2i) == 'CD1 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CD1 ' .and. atm_name(an3i) == 'CE1 ') .or. &
             (atm_name(an3i) == 'CD1 ' .and. atm_name(an2i) == 'CE1 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CE1 ' .and. atm_name(an3i) == 'CZ  ') .or. &
             (atm_name(an3i) == 'CE1 ' .and. atm_name(an2i) == 'CZ  ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CZ  ' .and. atm_name(an3i) == 'CE2 ') .or. &
             (atm_name(an3i) == 'CZ  ' .and. atm_name(an2i) == 'CE2 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CE2 ' .and. atm_name(an3i) == 'CD2 ') .or. &
             (atm_name(an3i) == 'CE2 ' .and. atm_name(an2i) == 'CD2 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CD2 ' .and. atm_name(an3i) == 'CG  ') .or. &
             (atm_name(an3i) == 'CD2 ' .and. atm_name(an2i) == 'CG  ') ) then
          ring = .true.
        end if

      else if (res_name(an2i) == 'TYR ' .and. res_name(an3i) == 'TYR ') then
        if ( (atm_name(an2i) == 'CG  ' .and. atm_name(an3i) == 'CD1 ') .or. &
             (atm_name(an3i) == 'CG  ' .and. atm_name(an2i) == 'CD1 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CD1 ' .and. atm_name(an3i) == 'CE1 ') .or. &
             (atm_name(an3i) == 'CD1 ' .and. atm_name(an2i) == 'CE1 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CE1 ' .and. atm_name(an3i) == 'CZ  ') .or. &
             (atm_name(an3i) == 'CE1 ' .and. atm_name(an2i) == 'CZ  ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CZ  ' .and. atm_name(an3i) == 'CE2 ') .or. &
             (atm_name(an3i) == 'CZ  ' .and. atm_name(an2i) == 'CE2 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CE2 ' .and. atm_name(an3i) == 'CD2 ') .or. &
             (atm_name(an3i) == 'CE2 ' .and. atm_name(an2i) == 'CD2 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CD2 ' .and. atm_name(an3i) == 'CG  ') .or. &
             (atm_name(an3i) == 'CD2 ' .and. atm_name(an2i) == 'CG  ') ) then
          ring = .true.
        end if

      else if (res_name(an2i) == 'TRP ' .and. res_name(an3i) == 'TRP ') then
        if ( (atm_name(an2i) == 'CG  ' .and. atm_name(an3i) == 'CD1 ') .or. &
             (atm_name(an3i) == 'CG  ' .and. atm_name(an2i) == 'CD1 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CD1 ' .and. atm_name(an3i) == 'NE1 ') .or. &
             (atm_name(an3i) == 'CD1 ' .and. atm_name(an2i) == 'NE1 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'NE1 ' .and. atm_name(an3i) == 'CE2 ') .or. &
             (atm_name(an3i) == 'NE1 ' .and. atm_name(an2i) == 'CE2 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CE2 ' .and. atm_name(an3i) == 'CD2 ') .or. &
             (atm_name(an3i) == 'CE2 ' .and. atm_name(an2i) == 'CD2 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CD2 ' .and. atm_name(an3i) == 'CG  ') .or. &
             (atm_name(an3i) == 'CD2 ' .and. atm_name(an2i) == 'CG  ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CD2 ' .and. atm_name(an3i) == 'CE3 ') .or. &
             (atm_name(an3i) == 'CD2 ' .and. atm_name(an2i) == 'CE3 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CE3 ' .and. atm_name(an3i) == 'CZ3 ') .or. &
             (atm_name(an3i) == 'CE3 ' .and. atm_name(an2i) == 'CZ3 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CZ3 ' .and. atm_name(an3i) == 'CH2 ') .or. &
             (atm_name(an3i) == 'CZ3 ' .and. atm_name(an2i) == 'CH2 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CH2 ' .and. atm_name(an3i) == 'CZ2 ') .or. &
             (atm_name(an3i) == 'CH2 ' .and. atm_name(an2i) == 'CZ2 ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CZ2 ' .and. atm_name(an3i) == 'CE2 ') .or. &
             (atm_name(an3i) == 'CZ2 ' .and. atm_name(an2i) == 'CE2 ') ) then
          ring = .true.
        end if

      else if (res_name(an2i) == 'PRO ' .and. res_name(an3i) == 'PRO ') then
        if ( (atm_name(an2i) == 'N   ' .and. atm_name(an3i) == 'CD  ') .or. &
             (atm_name(an3i) == 'N   ' .and. atm_name(an2i) == 'CD  ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CD  ' .and. atm_name(an3i) == 'CG  ') .or. &
             (atm_name(an3i) == 'CD  ' .and. atm_name(an2i) == 'CG  ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CG  ' .and. atm_name(an3i) == 'CB  ') .or. &
             (atm_name(an3i) == 'CG  ' .and. atm_name(an2i) == 'CB  ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CB  ' .and. atm_name(an3i) == 'CA  ') .or. &
             (atm_name(an3i) == 'CB  ' .and. atm_name(an2i) == 'CA  ') ) then
          ring = .true.
        end if
        if ( (atm_name(an2i) == 'CA  ' .and. atm_name(an3i) == 'N   ') .or. &
             (atm_name(an3i) == 'CA  ' .and. atm_name(an2i) == 'N   ') ) then
          ring = .true.
        end if

      end if


      dx21 = coord(1,list(2,i)) - coord(1,list(1,i))
      dy21 = coord(2,list(2,i)) - coord(2,list(1,i))
      dz21 = coord(3,list(2,i)) - coord(3,list(1,i))
      dx32 = coord(1,list(3,i)) - coord(1,list(2,i))
      dy32 = coord(2,list(3,i)) - coord(2,list(2,i))
      dz32 = coord(3,list(3,i)) - coord(3,list(2,i))
      dx43 = coord(1,list(4,i)) - coord(1,list(3,i))
      dy43 = coord(2,list(4,i)) - coord(2,list(3,i))
      dz43 = coord(3,list(4,i)) - coord(3,list(3,i))

      px12 = dy21*dz32 - dy32*dz21
      py12 = dz21*dx32 - dz32*dx21
      pz12 = dx21*dy32 - dx32*dy21
      px23 = dy32*DZ43 - dy43*dz32
      py23 = dz32*DX43 - dz43*dx32
      pz23 = dx32*DY43 - dx43*dy32
      r12  = px12*px12 + py12*py12 + pz12*pz12
      r23  = px23*px23 + py23*py23 + pz23*pz23
      r12r23 = sqrt( r12 * r23 )

      if ( r12r23 > 0.0_wp ) then

        r12r23    = 1.0_wp / r12r23
        s1223     = px12*px23 + py12*py23 + pz12*pz23
        cosp      = s1223 * r12r23
        if (cosp > 1.0_wp  ) cosp = 1.0_wp
        if (cosp < -1.0_wp ) cosp = -1.0_wp

        px123     = py12*pz23 - py23*pz12
        py123     = pz12*px23 - pz23*px12
        pz123     = px12*py23 - px23*py12
        s32123    = dx32*px123 + dy32*py123 + dz32*pz123

        phi  = acos(cosp)
        if (s32123 < 0.000) then
          phi = ( 2.0_wp*PI ) - phi
        end if

      end if
      

      gpr%impr_list(1,i) = list(1,i)
      gpr%impr_list(2,i) = list(2,i)
      gpr%impr_list(3,i) = list(3,i)
      gpr%impr_list(4,i) = list(4,i)
      gpr%impr_phase(i)  = phi/RAD

      if (ring) then
        gpr%impr_force_const(i) = &
             option%k_improper*option%r_scale/real(found,wp)
      else
        gpr%impr_force_const(i) = &
             option%k_improper/real(found,wp) 
      end if

    end do

    return

  end subroutine make_gpr_impr

end module cg_convert_mod
