!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   cg_option_mod
!> @brief   module for conversion options
!! @authors Takaharu Mori (TM), Norio Takase (NT)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../../../config.h"
#endif

module cg_option_mod

  use cg_option_str_mod
  use fileio_control_mod
  use messages_mod
  use constants_mod

  implicit none
  private

  ! structures
  type, public :: s_opt_info
    logical    :: check_only   = .true.
    integer    :: forcefield   = ForceFieldAAGO
    real(wp)   :: k_bond       = 200.0_wp
    real(wp)   :: k_angle      =  40.0_wp
    real(wp)   :: k_improper   =  10.0_wp
    real(wp)   :: e_dihedral   =   1.0_wp
    real(wp)   :: e_c          =   1.0_wp
    real(wp)   :: e_nc         =  0.01_wp
    real(wp)   :: s_nc         =   2.5_wp
    real(wp)   :: r_cd         =   2.0_wp
    real(wp)   :: r_bbsc       =   2.0_wp
    real(wp)   :: cutoff       =   4.0_wp
    integer    :: min_dist     =   4
    real(wp)   :: r_scale      =   4.0_wp
  end type s_opt_info

  ! subroutines
  public  :: show_ctrl_option
  public  :: read_ctrl_option
  public  :: setup_option

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_option
  !> @brief        show control parameters in OPTION section
  !! @authors      NT
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_option


    write(MsgOut,'(A)') '[OPTION]'
    write(MsgOut,'(A)') 'check_only     = YES    # (YES/NO)'
    write(MsgOut,'(A)') 'forcefield     = AAGO   # (AAGO)'
    write(MsgOut,'(A)') ' '
    write(MsgOut,'(A)') '# AAGO options'
    write(MsgOut,'(A)') 'k_bond         = 200    # bond force constant'
    write(MsgOut,'(A)') 'k_angle        = 40     # angle force constant'
    write(MsgOut,'(A)') 'k_improper     = 10     # improper force constant'
    write(MsgOut,'(A)') 'e_nc           = 0.01   # repulsive weight (epsilon) in the excluded volume term'
    write(MsgOut,'(A)') 's_nc           = 2.5    # repulsive distance (sigma) in the excluded volume term'
    write(MsgOut,'(A)') 'r_cd           = 2.0    # total contact to total dihedral energy ratio'
    write(MsgOut,'(A)') 'r_bbsc         = 2.0    # ratio of backbone to side chain dihedral strength'
    write(MsgOut,'(A)') 'cutoff         = 4.0    # cutoff distance for native contacts'
    write(MsgOut,'(A)') 'min_dist       = 4      # minimum distance (in sequence) of native contacts'
    write(MsgOut,'(A)') 'r_scale        = 4.0    # scale factor for ring improper torsion'
    write(MsgOut,'(A)') ' '

    return

  end subroutine show_ctrl_option

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_ctrl_option
  !> @brief        read control parameters in OPTION section
  !! @authors      TM
  !! @param[in]    handle   : unit number of control file
  !! @param[inout] opt_info : OPTION section control parameters information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine read_ctrl_option(handle, opt_info)
  
    ! parameters
    character(*),            parameter     :: Section = 'Option'

    ! formal argments
    integer,                 intent(in)    :: handle
    type(s_opt_info),        intent(inout) :: opt_info


    ! read parameters
    !
    call read_ctrlfile_logical(handle, Section,                  &
                              'check_only', opt_info%check_only)
    call read_ctrlfile_type   (handle, Section,                  &
                              'Forcefield', opt_info%forcefield,ForceFieldTypes)
    call read_ctrlfile_real   (handle, Section,                  &
                              'k_bond', opt_info%k_bond)
    call read_ctrlfile_real   (handle, Section,                  &
                              'k_angle', opt_info%k_angle)
    call read_ctrlfile_real   (handle, Section,                  &
                              'k_improper', opt_info%k_improper)
    call read_ctrlfile_real   (handle, Section,                  &
                              'e_dihedral', opt_info%e_dihedral)
    call read_ctrlfile_real   (handle, Section,                  &
                              'e_c', opt_info%e_c)
    call read_ctrlfile_real   (handle, Section,                  &
                              'e_nc', opt_info%e_nc)
    call read_ctrlfile_real   (handle, Section,                  &
                              's_nc', opt_info%s_nc)
    call read_ctrlfile_real   (handle, Section,                  &
                              'r_cd', opt_info%r_cd)
    call read_ctrlfile_real   (handle, Section,                  &
                              'r_bbsc', opt_info%r_bbsc)
    call read_ctrlfile_real   (handle, Section,                  &
                              'cutoff', opt_info%cutoff)
    call read_ctrlfile_integer(handle, Section,                  &
                              'min_dist', opt_info%min_dist)
    call read_ctrlfile_real   (handle, Section,                  &
                              'r_scale', opt_info%r_scale)


    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_Ctrl_Option> Parameters of Options'
    if (opt_info%check_only) then
      write(MsgOut,'(A20,A3)') '  check only      = ', 'yes'
    else
      write(MsgOut,'(A20,A2)') '  check only      = ', 'no'
    end if
    write(MsgOut,'(A20,A10)')   '  forcefield      = ', &
         ForceFieldTypes(opt_info%forcefield)

    if (opt_info%forcefield == ForceFieldAAGO) then
      write(MsgOut,'(A20,F10.3)') '  k_bond          = ', opt_info%k_bond
      write(MsgOut,'(A20,F10.3)') '  k_angle         = ', opt_info%k_angle
      write(MsgOut,'(A20,F10.3)') '  k_improper      = ', opt_info%k_improper
      write(MsgOut,'(A20,F10.3)') '  e_nc            = ', opt_info%e_nc
      write(MsgOut,'(A20,F10.3)') '  s_nc            = ', opt_info%s_nc
      write(MsgOut,'(A20,F10.3)') '  r_cd            = ', opt_info%r_cd
      write(MsgOut,'(A20,F10.3)') '  r_bbsc          = ', opt_info%r_bbsc
      write(MsgOut,'(A20,F10.3)') '  cutoff          = ', opt_info%cutoff
      write(MsgOut,'(A20,I10)')   '  min_dist        = ', opt_info%min_dist
      write(MsgOut,'(A20,F10.3)') '  r_scale         = ', opt_info%r_scale
    end if
    write(MsgOut,'(A)') ' '

    return

  end subroutine read_ctrl_option

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_option
  !> @brief        setup option information
  !! @authors      TM
  !! @param[in]    opt_info : OPTION section control parameters information
  !! @param[inout] option   : option information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine setup_option(opt_info, option)
  
    ! formal argments
    type(s_opt_info),        intent(in)    :: opt_info
    type(s_option),          intent(inout) :: option


    option%check_only  = opt_info%check_only
    option%forcefield  = opt_info%forcefield
    option%k_bond      = opt_info%k_bond
    option%k_angle     = opt_info%k_angle
    option%k_improper  = opt_info%k_improper
    option%e_dihedral  = opt_info%e_dihedral
    option%e_c         = opt_info%e_c
    option%e_nc        = opt_info%e_nc
    option%s_nc        = opt_info%s_nc
    option%r_cd        = opt_info%r_cd
    option%r_bbsc      = opt_info%r_bbsc
    option%cutoff      = opt_info%cutoff
    option%min_dist    = opt_info%min_dist
    option%r_scale     = opt_info%r_scale

    return

  end subroutine setup_option

end module cg_option_mod
