!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   rs_option_mod
!> @brief   module for conversion options
!! @authors Norio Takase (NT)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../../../config.h"
#endif

module rs_option_mod

  use rs_option_str_mod
  use output_str_mod
  use fileio_control_mod
  use string_mod
  use messages_mod
  use constants_mod

  implicit none
  private

  ! structures
  type, public :: s_opt_info

    logical                     :: check_only        = .true.
    integer                     :: rstin_format      = RstFormatNAMD
    integer                     :: rstout_format     = RstFormatGENESIS
    integer                     :: rstout_type       = RstTypeAuto

    real(wp)                    :: min_energy        = 0.0_wp
    real(wp)                    :: min_delta_r       = 0.0_wp

    integer                     :: md_iseed          = 0
    integer                     :: step              = 0
    integer                     :: md_dof            = 0
    real(wp)                    :: md_thermo_moment  = 0.0_wp
    real(wp)                    :: md_baro_moment(3) = 0.0_wp
    real(wp)                    :: origin_x      = 0.0_wp
    real(wp)                    :: origin_y      = 0.0_wp
    real(wp)                    :: origin_z      = 0.0_wp

  end type s_opt_info

  public  :: show_ctrl_option
  public  :: read_ctrl_option
  public  :: setup_option

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_option
  !> @brief        show control parameters in OPTION section
  !! @authors      NT
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_option

    write(MsgOut,'(A)') '[OPTION]'
    write(MsgOut,'(A)') 'check_only     = YES             # (YES/NO)'
    write(MsgOut,'(A)') 'rstin_format   = NAMD            # (NAMD/GENESIS)'
    write(MsgOut,'(A)') 'rstout_format  = GENESIS         # (NAMD/GENESIS)'
    write(MsgOut,'(A)') 'rstout_type    = AUTO            # (AUTO/MIN/MD)'
    write(MsgOut,'(A)') 'min_energy     = 0.0             # energy'
    write(MsgOut,'(A)') 'min_delta_r    = 0.0             # delta-r'
    write(MsgOut,'(A)') 'md_iseed       = 0               # random seed'
    write(MsgOut,'(A)') 'md_dof         = 0               # number of degree of freedom'
    write(MsgOut,'(A)') 'md_thermo_moment = 0.0           # thermostat momentum'
    write(MsgOut,'(A)') 'md_baro_moment   = 0.0 0.0 0.0   # barostat momentum'
    write(MsgOut,'(A)') 'md_baro_moment   = 0.0 0.0 0.0   # barostat momentum'
    write(MsgOut,'(A)') 'namd_step        = 0             # step'
    write(MsgOut,'(A)') 'namd_originx     = 0.0           # box origin (x)'
    write(MsgOut,'(A)') 'namd_originy     = 0.0           # box origin (y)'
    write(MsgOut,'(A)') 'namd_originz     = 0.0           # box origin (z)'
    write(MsgOut,'(A)') ' '

    return

  end subroutine show_ctrl_option

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_ctrl_option
  !> @brief        read control parameters in OPTION section
  !! @authors      NT
  !! @param[in]    handle   : unit number of control file
  !! @param[out]   opt_info : OPTION section control parameters information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine read_ctrl_option(handle, opt_info)
  
    ! parameters
    character(*),            parameter     :: Section = 'Option'

    ! formal argments
    integer,                 intent(in)    :: handle
    type(s_opt_info),        intent(inout) :: opt_info

    character(Maxline)                     :: line


    ! read parameters
    !

    call read_ctrlfile_logical(handle, Section, 'check_only', &
                               opt_info%check_only)

    call read_ctrlfile_type   (handle, Section, 'rstin_format', &
                               opt_info%rstin_format, RstFormatTypes)

    call read_ctrlfile_type   (handle, Section, 'rstout_format', &
                               opt_info%rstout_format, RstFormatTypes)

    call read_ctrlfile_type   (handle, Section, 'rstout_type', &
                               opt_info%rstout_type, RstTypeTypes)

    call read_ctrlfile_real   (handle, Section, 'min_energy', &
                               opt_info%min_energy)

    call read_ctrlfile_real   (handle, Section, 'min_delta_r', &
                               opt_info%min_delta_r)

    call read_ctrlfile_integer(handle, Section, 'md_iseed' , &
                               opt_info%md_iseed)

    call read_ctrlfile_integer(handle, Section, 'md_dof' , &
                               opt_info%md_dof)

    call read_ctrlfile_real   (handle, Section, 'md_thermo_moment', &
                               opt_info%md_thermo_moment)

    if (opt_info%rstout_format == RstFormatGENESIS) then
      call read_ctrlfile_string(handle, Section, 'md_baro_moment', line)
  
      read(line,*,err=900,end=900) opt_info%md_baro_moment
    endif

    if (opt_info%rstout_format == opt_info%rstin_format)  then
      call error_msg                                                         &
           ('Read_Ctrl_Option> Formats of input/output should be different')
    endif

    call read_ctrlfile_integer(handle, Section, 'namd_step', &
                               opt_info%step)

    call read_ctrlfile_real   (handle, Section, 'namd_originX', &
                               opt_info%origin_x)

    call read_ctrlfile_real   (handle, Section, 'namd_originY', &
                               opt_info%origin_y)

    call read_ctrlfile_real   (handle, Section, 'namd_originZ', &
                               opt_info%origin_z)


    ! write parameters to MsgOut
    !
    write(MsgOut,'(A)') 'Read_Ctrl_Option> Parameters of Options'
    if (opt_info%check_only) then
      write(MsgOut,'(A20,A3)') '  check only      = ', 'yes'
    else
      write(MsgOut,'(A20,A2)') '  check only      = ', 'no'
    end if
    write(MsgOut,'(A20,A10)')   '  rst in format   = ', &
         RstFormatTypes(opt_info%rstin_format)
    write(MsgOut,'(A20,A10)')   '  rst out format  = ', &
         RstFormatTypes(opt_info%rstout_format)
    write(MsgOut,'(A20,A10)')   '  rst out type    = ', &
         RstTypeTypes(opt_info%rstout_type)
    write(MsgOut,'(A20,F10.3)') '  min: energy     = ', &
         opt_info%min_energy
    write(MsgOut,'(A20,F10.3)') '  min: delta-r    = ', &
         opt_info%min_delta_r
    write(MsgOut,'(A20,I16)')   '  md: iseed       = ', &
         opt_info%md_iseed
    write(MsgOut,'(A20,I16)')   '  md: num DOF     = ', &
         opt_info%md_dof
    write(MsgOut,'(A20,F16.9)') '  md: thermo moment = ', &
         opt_info%md_thermo_moment
    write(MsgOut,'(A20,F16.9)') '  md: baro moment   = ', &
         opt_info%md_baro_moment(1)
    write(MsgOut,'(A20,F16.9)') '                    ', &
         opt_info%md_baro_moment(2)
    write(MsgOut,'(A20,F16.9)') '                    ', &
         opt_info%md_baro_moment(3)
    write(MsgOut,'(A20,I16)')   'namd: step          = ', &
         opt_info%step
    write(MsgOut,'(A20,F16.9)') 'namd: origin_x      = ', &
         opt_info%origin_x
    write(MsgOut,'(A20,F16.9)') 'namd: origin_y      = ', &
         opt_info%origin_y
    write(MsgOut,'(A20,F16.9)') 'namd: origin_z      = ', &
         opt_info%origin_z
    write(MsgOut,'(A)') ' '

    return

900 call error_msg('Read_Ctrl_Option> format error : "md_baro_moment"')

  end subroutine read_ctrl_option

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_option
  !> @brief        setup option information
  !! @authors      NT
  !! @param[in]    opt_info : OPTION section control parameters information
  !! @param[out]   option   : option information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine setup_option(opt_info, option)
  
    ! formal argments
    type(s_opt_info),        intent(in)    :: opt_info
    type(s_option),          intent(inout) :: option


    ! check only
    option%check_only    = opt_info%check_only

    ! restart input format
    option%rstin_format = opt_info%rstin_format

    ! restart output format
    option%rstout_format = opt_info%rstout_format

    ! restart output type
    option%rstout_type = opt_info%rstout_type

    ! minimization parameters
    option%min_energy  = opt_info%min_energy
    option%min_delta_r = opt_info%min_delta_r

    ! MD parameters
    option%md_iseed         = opt_info%md_iseed
    option%md_dof           = opt_info%md_dof
    option%md_thermo_moment = opt_info%md_thermo_moment
    option%md_baro_moment   = opt_info%md_baro_moment

    ! NAMD parameters
    option%step             = opt_info%step
    option%origin_x         = opt_info%origin_x
    option%origin_y         = opt_info%origin_y
    option%origin_z         = opt_info%origin_z

    return

  end subroutine setup_option

end module rs_option_mod
