!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_boundary_mod
!> @brief   utilities for boundary conditions
!! @authors Takaharu Mori (TM), Takashi Imai (TI), Jaewoon Jung (JJ), 
!!          Norio Takase (NT), Yuji Sugita (YS)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_boundary_mod

  use at_boundary_str_mod
  use molecules_str_mod
  use fileio_rst_mod
  use fileio_control_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod

  implicit none
  private

  real(wp), parameter   :: TableWaterMargin = 2.0_wp
  real(wp), parameter   :: FakeDefault      = 999999_wp

  ! structures
  type, public :: s_pbc_info
    real(wp)            :: box_size_x       = FakeDefault
    real(wp)            :: box_size_y       = FakeDefault
    real(wp)            :: box_size_z       = FakeDefault
    logical             :: wrap_all         = .false.
  end type s_pbc_info

  type, public :: s_boundary_info
    integer             :: type             = BoundaryTypePBC
    type(s_pbc_info)    :: pbc_info
    real(wp)            :: origin_x         = 0.0_wp
    real(wp)            :: origin_y         = 0.0_wp
    real(wp)            :: origin_z         = 0.0_wp
  end type s_boundary_info

  ! subroutines
  public  :: show_ctrl_boundary
  public  :: read_ctrl_boundary
  public  :: setup_boundary
  public  :: update_boundary
  private :: update_boundary_pbc
  public  :: wrap_molecules
  private :: wrap_all

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_boundary
  !> @brief        show BOUNDARY section usage
  !! @authors      NT, TM
  !! @param[in]    show_all : show all usage or not
  !! @param[in]    run_mode : run mode string : "md", "min", "remd"
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_boundary(show_all, run_mode)

    ! formal arguments
    logical,                 intent(in)    :: show_all
    character(*),            intent(in)    :: run_mode


    if (show_all) then

      select case (run_mode)

      case ('md', 'min', 'remd')

        write(MsgOut,'(A)') '[BOUNDARY]'
        write(MsgOut,'(A)') 'type          = PBC       # [NOBC,PBC]'
        write(MsgOut,'(A)') 'box_size_x    = 999999    # box size (x) in [PBC]'
        write(MsgOut,'(A)') 'box_size_y    = 999999    # box size (y) in [PBC]'
        write(MsgOut,'(A)') 'box_size_z    = 999999    # box size (z) in [PBC]'
        write(MsgOut,'(A)') ' '


      end select

    else

      select case (run_mode)

      case ('md', 'min', 'remd')

        write(MsgOut,'(A)') '[BOUNDARY]'
        write(MsgOut,'(A)') 'type          = PBC       # [NOBC,PBC]'
        write(MsgOut,'(A)') 'box_size_x    = 999999    # box size (x) in [PBC]'
        write(MsgOut,'(A)') 'box_size_y    = 999999    # box size (y) in [PBC]'
        write(MsgOut,'(A)') 'box_size_z    = 999999    # box size (z) in [PBC]'
        write(MsgOut,'(A)') ' '


      end select

    end if


    return

  end subroutine show_ctrl_boundary
  
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_ctrl_boundary
  !> @brief        read BOUNDARY section in the control file
  !! @authors      YS, TI, JJ, TM
  !! @param[in]    handle     : unit number
  !! @param[out]   bound_info : BOUNDARY section control parameters information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_ctrl_boundary(handle, bound_info)

    ! parameters
    character(*),            parameter     :: Section = 'Boundary'

    ! formal arguments
    integer,                 intent(in)    :: handle
    type(s_boundary_info),   intent(inout) :: bound_info


    ! read parameters from control file
    ! 
    call read_ctrlfile_type(handle, Section, 'type',          &
                            bound_info%type, BoundaryTypeTypes)

    select case (bound_info%type)

    case (BoundaryTypeNOBC)
      ! do nothing

    case (BoundaryTypePBC)

      ! read parameters for PBC from control file
      !
      call read_ctrlfile_real(handle, Section, 'box_size_x',  &
                              bound_info%pbc_info%box_size_x)
      call read_ctrlfile_real(handle, Section, 'box_size_y',  &
                              bound_info%pbc_info%box_size_y)
      call read_ctrlfile_real(handle, Section, 'box_size_z',  &
                              bound_info%pbc_info%box_size_z)
      call read_ctrlfile_logical(handle, Section, 'wrapall',  &
                              bound_info%pbc_info%wrap_all)

    end select


    ! write parameters to MsgOut
    !
    if (main_rank) then

      write(MsgOut,'(A)') 'Read_Ctrl_Boundary> Parameters of Boundary Condition'
      write(MsgOut,'(A20,A10)') &
            '  type            = ', BoundaryTypeTypes(bound_info%type)

      select case (bound_info%type)

      case (BoundaryTypeNOBC)
        ! do nothing

      case (BoundaryTypePBC)

        if (abs(bound_info%pbc_info%box_size_x - FakeDefault) > EPS .and. &
            abs(bound_info%pbc_info%box_size_y - FakeDefault) > EPS .and. &
            abs(bound_info%pbc_info%box_size_z - FakeDefault) > EPS) &
        write(MsgOut,'(A20,3F10.3)')                                  &
              '  box_size(x,y,z) = ', bound_info%pbc_info%box_size_x, &
                                      bound_info%pbc_info%box_size_y, &
                                      bound_info%pbc_info%box_size_z

      end select

      write(MsgOut,'(A)') ' '

    end if

    return

  end subroutine read_ctrl_boundary

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_boundary
  !> @brief        set essential variables for boundary condition
  !! @authors      YS, TM, JJ, TI
  !! @param[in]    bound_info  : BOUNDARY section control parameters information
  !! @param[in]    use_table   : flag for use table or not
  !! @param[in]    pairlistdist: pair-list distance
  !! @param[in]    rst         : restart file information
  !! @param[out]   boundary    : boundary conditions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_boundary(bound_info, use_table, pairlistdist, rst, boundary)

    ! formal arguments
    type(s_boundary_info),   intent(in)    :: bound_info
    logical,                 intent(in)    :: use_table
    real(wp),                intent(in)    :: pairlistdist
    type(s_rst),             intent(in)    :: rst
    type(s_boundary),        intent(inout) :: boundary


    ! initialize 
    !
    call init_boundary(boundary)

    ! setup variables
    !
    select case (bound_info%type)
    
    case (BoundaryTypeNOBC)

      boundary%type     = bound_info%type
      boundary%origin_x = bound_info%origin_x
      boundary%origin_y = bound_info%origin_y
      boundary%origin_z = bound_info%origin_z

      if (use_table) &
        call error_msg('Setup_Boundary> NOBC is disabled with table = YES')

    case (BoundaryTypePBC)

      boundary%type     = bound_info%type
      boundary%origin_x = bound_info%origin_x
      boundary%origin_y = bound_info%origin_y
      boundary%origin_z = bound_info%origin_z
      boundary%wrap_all = bound_info%pbc_info%wrap_all

      if (rst%rstfile_type == RstfileTypeUndef) then
        if (abs(bound_info%pbc_info%box_size_x - FakeDefault) < EPS) &
          call error_msg('Setup_Boundary> box_size_x is not specified in ctrl')
        if (abs(bound_info%pbc_info%box_size_y - FakeDefault) < EPS) &
          call error_msg('Setup_Boundary> box_size_y is not specified in ctrl')
        if (abs(bound_info%pbc_info%box_size_z - FakeDefault) < EPS) &
          call error_msg('Setup_Boundary> box_size_z is not specified in ctrl')

        boundary%box_size_x = bound_info%pbc_info%box_size_x
        boundary%box_size_y = bound_info%pbc_info%box_size_y
        boundary%box_size_z = bound_info%pbc_info%box_size_z
      else
        boundary%box_size_x = rst%box_size_x
        boundary%box_size_y = rst%box_size_y
        boundary%box_size_z = rst%box_size_z
      end if
      boundary%box_size_x_ref = boundary%box_size_x
      boundary%box_size_y_ref = boundary%box_size_y
      boundary%box_size_z_ref = boundary%box_size_z

    end select

    ! write setup info
    !
    if (main_rank) then
      write(MsgOut,'(A)') &
            'Setup_Boundary> Setup Variables for Boundary Condition'
      if (boundary%type == BoundaryTypePBC) then
        write(MsgOut,'(A20,3F10.3)')                       &
              '  box_size(x,y,z) = ', boundary%box_size_x, &
                                      boundary%box_size_y, &
                                      boundary%box_size_z
      end if
      write(MsgOut,'(A20,3F10.3)')                         &
              '  origin(x,y,z)   = ', boundary%origin_x,   &
                                      boundary%origin_y,   &
                                      boundary%origin_z
      write(MsgOut,'(A)') ' '
    end if

    ! update boundary
    !
    call update_boundary(use_table, pairlistdist, boundary)

    return

  end subroutine setup_boundary

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_boundary
  !> @brief        update variables for boundary condition
  !! @authors      TM
  !! @param[in]    use_table    : flag for use table or not
  !! @param[in]    pairlistdist : pair list distance
  !! @param[inout] boundary     : information of boundary condition
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_boundary(use_table, pairlistdist, boundary)

    ! formal arguments
    logical,                 intent(in)    :: use_table
    real(wp),                intent(in)    :: pairlistdist
    type(s_boundary),        intent(inout) :: boundary


    select case (boundary%type)

    case (BoundaryTypeNOBC)

      ! do nothing

    case (BoundaryTypePBC)

      call update_boundary_pbc(use_table, pairlistdist, boundary)

    end select

    return

  end subroutine update_boundary

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    update_boundary_pbc
  !> @brief        update variables for boundary condition
  !! @authors      TI, JJ, TM
  !! @param[in]    use_table    : flag for use table or not
  !! @param[in]    pairlistdist : pair list distance
  !! @param[inout] boundary     : information of boundary condition
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine update_boundary_pbc(use_table, pairlistdist, boundary)

    ! formal arguments
    logical,                 intent(in)    :: use_table
    real(wp),                intent(in)    :: pairlistdist
    type(s_boundary),        intent(inout) :: boundary

    ! local variables
    real(wp)                 :: bsize_x, bsize_y, bsize_z
    real(wp)                 :: csize_x, csize_y, csize_z
    integer                  :: i, j, k
    integer                  :: inb, jnb, knb, inbs, jnbs, knbs, lc, lcnb, ln
    integer                  :: ncell_x, ncell_y, ncell_z, ncell


    boundary%use_cell_linked_list = .true.
    bsize_x = boundary%box_size_x
    bsize_y = boundary%box_size_y
    bsize_z = boundary%box_size_z

    if (use_table) then

      ncell_x = int(bsize_x/(pairlistdist + TableWaterMargin))
      ncell_y = int(bsize_y/(pairlistdist + TableWaterMargin))
      ncell_z = int(bsize_z/(pairlistdist + TableWaterMargin))

      if (ncell_x < 3 .or. ncell_y < 3 .or. ncell_z < 3) &
        call error_msg('Update_Boundary_Pbc> too small boxsize/(pairdist+2). larger boxsize or shorter pairdist should be used.')

    else

      ncell_x = int(bsize_x/pairlistdist)
      ncell_y = int(bsize_y/pairlistdist)
      ncell_z = int(bsize_z/pairlistdist)

      if (ncell_x < 3 .or. ncell_y < 3 .or. ncell_z < 3) then
        if (pairlistdist < 0.5_wp*bsize_x .and. &
            pairlistdist < 0.5_wp*bsize_y .and. &
            pairlistdist < 0.5_wp*bsize_z) then
          boundary%use_cell_linked_list = .false.
          return
        else
          call error_msg('Update_Boundary_Pbc> too small boxsize/pairdist. larger boxsize or shorter pairdist should be used.')
        end if
      end if

    end if

    csize_x = bsize_x/real(ncell_x,wp)
    csize_y = bsize_y/real(ncell_y,wp)
    csize_z = bsize_z/real(ncell_z,wp)
    ncell   = ncell_x*ncell_y*ncell_z

    boundary%num_cells_x = ncell_x
    boundary%num_cells_y = ncell_y
    boundary%num_cells_z = ncell_z
    boundary%cell_size_x = csize_x
    boundary%cell_size_y = csize_y
    boundary%cell_size_z = csize_z
    boundary%num_cells   = ncell

    ! prepare cell neighbor list
    !
    call alloc_boundary(boundary, BoundaryCells, ncell)

    do k = 0, ncell_z-1
      do j = 0, ncell_y-1
        do i = 0, ncell_x-1

          lc = 1 + i + j*ncell_x + k*ncell_x*ncell_y
          ln = 0

          do knb = k-1, k+1
            if (knb == -1) then
              knbs = ncell_z - 1
            else if (knb == ncell_z) then
              knbs = 0
            else
              knbs = knb
            end if
     
            do jnb = j-1, j+1
              if (jnb == -1) then
                jnbs = ncell_y - 1
              else if (jnb == ncell_y) then
                jnbs = 0
              else
                jnbs = jnb
              end if

              do inb = i-1, i+1
                if (inb == -1) then
                  inbs = ncell_x - 1
                else if (inb == ncell_x) then
                  inbs = 0
                else
                  inbs = inb
                end if

                lcnb = 1 + inbs + jnbs*ncell_x + knbs*ncell_x*ncell_y
                ln   = ln + 1

                boundary%neighbor_cells(ln,lc) = lcnb
              end do

            end do
          end do
        end do
      end do
    end do

    return

  end subroutine update_boundary_pbc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    wrap molecules
  !> @brief        wrap molecules
  !! @authors      TM
  !! @param[in]    molecule : molecule information
  !! @param[in]    boundary : boundary conditions information
  !! @param[inout] coord    : coordinates
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine wrap_molecules(molecule, boundary, coord)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    type(s_boundary),        intent(in)    :: boundary
    real(wp),                intent(inout) :: coord(:,:)


    if (boundary%wrap_all) then
      call wrap_all(molecule, boundary, coord)
    else
      ! do nothing
    end if

    return

  end subroutine wrap_molecules

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    wrap_all
  !> @brief        wrap molecules
  !! @authors      TM
  !! @param[in]    molecule : molecule information
  !! @param[in]    boundary : boundary conditions information
  !! @param[inout] coord    : coordinates
  !! @note         both coord_ref and coord are wrapped for restart
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine wrap_all(molecule, boundary, coord)

    ! formal arguments
    type(s_molecule),        intent(in)    :: molecule
    type(s_boundary),        intent(in)    :: boundary
    real(wp),                intent(inout) :: coord(:,:)

    ! local variables
    real(wp)                 :: com(3), box(3), ori(3)
    integer                  :: i, j, initial, final


    box(1) = boundary%box_size_x
    box(2) = boundary%box_size_y
    box(3) = boundary%box_size_z
    ori(1) = boundary%origin_x
    ori(2) = boundary%origin_y
    ori(3) = boundary%origin_z

    do i = 1, molecule%num_molecules

      initial  = molecule%molecule_atom_no(i)
      if (i /= molecule%num_molecules) then
        final = molecule%molecule_atom_no(i+1) - 1
      else
        final = molecule%num_atoms
      end if

      ! compute center of mass of a molecule
      !
      com(1:3) = 0.0_wp
      do j = initial, final
        coord(1:3,j) = coord(1:3,j) - ori(1:3)
        com(1:3) = com(1:3) + coord(1:3,j)*molecule%mass(j)
      end do
      com(1:3) = com(1:3)/molecule%molecule_mass(i)

      ! move molecule into the unit cell
      !
      do j = initial, final
        coord(1:3,j) = coord(1:3,j) - box(1:3)*nint(com(1:3)/box(1:3))
        coord(1:3,j) = coord(1:3,j) + ori(1:3)
      end do

    end do

    return

  end subroutine wrap_all

end module at_boundary_mod
