!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_energy_restraints_mod
!> @brief   calculate restraints energy
!! @authors Chigusa Kobayashi (CK), Takaharu Mori (TM)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif
  
module at_energy_restraints_mod

  use at_boundary_str_mod
  use at_restraints_str_mod
  use at_enefunc_str_mod
  use dihedral_libs_mod
  use timers_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod

  implicit none
  private

  ! subroutines
  public  :: compute_energy_restraint
  private :: compute_energy_positional_restraint
  private :: compute_energy_distance_restraint
  private :: compute_energy_angle_restraint
  private :: compute_energy_dihedral_restraint
  private :: compute_energy_rmsd_restraint

contains
  
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_restraint
  !> @brief        calculate restraint energy
  !! @authors      CK, TM
  !! @param[in]    enefunc    : potential energy functions information
  !! @param[in]    coord      : coordinates of target systems
  !! @param[inout] force      : forces of target systems
  !! @param[inout] virial     : virial of target systems
  !! @param[inout] virial_ext : virial of target systems
  !! @param[inout] eneposi    : position constraint energy of target systems
  !! @param[inout] enedist    : distance restraint energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_restraint(inum, calc_force, enefunc, coord, force, &
                                      virial, virial_ext, eneval)

    ! formal arguments
    integer,                  intent(in)    :: inum
    logical,                  intent(in)    :: calc_force
    type(s_enefunc),  target, intent(in)    :: enefunc
    real(wp),                 intent(in)    :: coord(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: virial_ext(3,3)
    real(wp),                 intent(inout) :: eneval


    call timer(TimerRestraint, TimerOn)

    ! To get consistency with charmm, reference coodinates should be
    ! overwritten with initial coordinates. However, it makes a serious bug 
    ! in the case of restarting. (CK:2011/12/13)
    !   if (enefunc%step == 0) then
    !     refcoord = coord
    !   end if

    eneval = 0.0_wp
 
    select case(enefunc%restraint_kind(inum)) 

    case (RestraintsFuncPOSI) 

      ! Positional restraint
      !
      call compute_energy_positional_restraint  &
        (enefunc, coord, inum, calc_force, force, virial_ext, eneval)

    case (RestraintsFuncDIST:RestraintsFuncDISTCOM)

      ! Distance restraint
      !
      call compute_energy_distance_restraint    &
        (enefunc, coord, inum, calc_force, force, virial, eneval)

    case (RestraintsFuncANGLE:RestraintsFuncANGLECOM)

      ! Angle restraint
      !
      call compute_energy_angle_restraint       &
        (enefunc, coord, inum, calc_force, force, virial, eneval)

    case (RestraintsFuncDIHED:RestraintsFuncDIHEDCOM)

      ! Dihedral angle restraint
      !
      call compute_energy_dihedral_restraint    &
        (enefunc, coord, inum, calc_force, force, virial, eneval)

    case (RestraintsFuncRMSD:RestraintsFuncRMSDCOM)

      ! RMSD restraint
      !   In charmm c35b2, virials of rmsd is involved in internal term
      !
      call compute_energy_rmsd_restraint        &
        (enefunc, coord, inum, calc_force, .false., force, virial, eneval)

    end select


    call timer(TimerRestraint, TimerOff)

    return

  end subroutine compute_energy_restraint

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_positional_restraint
  !> @brief        calculate position restraint energy
  !! @authors      CK
  !! @param[in]    enefunc : potential energy functions information
  !! @param[in]    coord   : coordinates of target systems
  !! @param[in]    inum    : pointer for restraint function
  !! @param[in]    const   : force constants
  !! @param[inout] force   : forces of target systems
  !! @param[inout] virial  : virial of target systems
  !! @param[out]   eposi   : position constraint energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_positional_restraint(enefunc, coord, inum, &
                                              calc_force, force, virial, eposi)

    ! formal arguments
    type(s_enefunc), target, intent(in)    :: enefunc
    real(wp),                intent(in)    :: coord(:,:)
    integer,                 intent(in)    :: inum
    logical,                 intent(in)    :: calc_force
    real(wp),                intent(inout) :: force(:,:)
    real(wp),                intent(inout) :: virial(3,3)
    real(wp),                intent(out)   :: eposi

    ! local variables
    real(wp)                 :: d(1:3)
    real(wp)                 :: const(0:3), etmp, rtmp, grad_coef, vtmp
    real(wp)                 :: work(1:3)
    integer                  :: iexpo
    integer                  :: i, j, k, iatm

    real(wp),        pointer :: refcoord(:,:)
    integer,         pointer :: numatoms(:)
    integer,         pointer :: atomlist(:,:)
    integer,         pointer :: list(:,:)
    integer,         pointer :: expo(:)


    refcoord => enefunc%restraint_refcoord
    expo     => enefunc%restraint_exponent_func
    list     => enefunc%restraint_grouplist
    numatoms => enefunc%restraint_numatoms
    atomlist => enefunc%restraint_atomlist
    iexpo    =  expo(inum)


    do j = 1, 4
      const(j-1) = enefunc%restraint_const(j,inum)
    end do

    eposi = 0.0_wp

    do i = 1, numatoms(list(1,inum))

      iatm = atomlist(i,list(1,inum))

      ! Positional restraint energy: 
      !   E=(Kx(x-x0)^2+Ky(y-y0)^2+Kz(z-z0)^2)^(iexpo/2)
      !
      d(1:3) = coord(1:3,iatm) - refcoord(1:3,iatm)
      rtmp = const(1)*d(1)*d(1) + const(2)*d(2)*d(2) + const(3)*d(3)*d(3)

      if (iexpo == 2) then
        etmp      = const(0)*rtmp
        grad_coef = const(0)*2.0_wp
      else
        etmp      = const(0)*sqrt(rtmp)**iexpo
        grad_coef = const(0)*real(iexpo,wp)*(sqrt(rtmp)**(iexpo-2))
      end if

      eposi = eposi + etmp

      if (calc_force) then
        ! gradient: dE/dX
        !
        work(1:3)       = grad_coef * const(1:3) * d(1:3)

        force(1:3,iatm) = force(1:3,iatm) - work(1:3)

        ! external virial
        !
        do j = 1, 3
          do k = j+1, 3
            vtmp = - coord(k,iatm) * work(j)
            virial(k,j) = virial(k,j) + vtmp
            virial(j,k) = virial(j,k) + vtmp
          end do
          vtmp = - coord(j,iatm) * work(j)
          virial(j,j) = virial(j,j) + vtmp
        end do
      end if

    end do

    return

  end subroutine compute_energy_positional_restraint

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_distance_restraint
  !> @brief        calculate distance restraint energy
  !! @authors      CK
  !! @param[in]    enefunc   : potential energy functions information
  !! @param[in]    coord     : coordinates of target systems
  !! @param[in]    inum      : pointer for restraint function
  !! @param[in]    const     : force constants
  !! @param[in]    ref       : reference value
  !! @param[inout] force     : forces of target systems
  !! @param[inout] virial    : virial of target systems
  !! @param[out]   erestdist : distance restraint energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_distance_restraint(enefunc, coord, inum, &
                                          calc_force, force, virial, erestdist)

    ! formal arguments
    type(s_enefunc), target, intent(in)    :: enefunc
    real(wp),                intent(in)    :: coord(:,:)
    integer,                 intent(in)    :: inum
    logical,                 intent(in)    :: calc_force
    real(wp),                intent(inout) :: force(:,:)
    real(wp),                intent(inout) :: virial(3,3)
    real(wp),                intent(out)   :: erestdist

    ! local variables
    real(wp)                 :: r_sum, r_dif, coef, coefi, const(2), ref(2)
    real(wp)                 :: grad_coef, vtmp
    real(wp)                 :: work(1:3)
    integer                  :: numgrp, iexpo
    integer                  :: j, k, iatm, ind
    integer                  :: ndist, idist, id1, id2, id

    real(wp),        pointer :: wtind(:,:)
    real(wp),        pointer :: masscoef(:,:)
    real(wp),        pointer :: com(:,:), dij(:,:), rij(:)
    real(wp),        pointer :: coefgrp(:,:)
    integer,         pointer :: numatoms(:)
    integer,         pointer :: atmlist(:,:)
    integer,         pointer :: kind(:)
    integer,         pointer :: grplist(:,:)
    integer,         pointer :: funcgrp(:)
    integer,         pointer :: expo(:)
    integer,         pointer :: expoind(:,:)


    funcgrp  => enefunc%restraint_funcgrp 
    expo     => enefunc%restraint_exponent_func
    grplist  => enefunc%restraint_grouplist
    numatoms => enefunc%restraint_numatoms
    atmlist  => enefunc%restraint_atomlist
    expoind  => enefunc%restraint_exponent_dist
    wtind    => enefunc%restraint_weight_dist
    kind     => enefunc%restraint_kind
    masscoef => enefunc%restraint_masscoef
    coefgrp  => enefunc%restraint_wtmp
    com      => enefunc%restraint_wcom1
    dij     => enefunc%restraint_wcom2
    rij      => enefunc%restraint_wdrt

    numgrp   = funcgrp(inum)
    iexpo    = expo(inum)
    ndist    = int(numgrp/2)

    const(1:2) = enefunc%restraint_const(1:2,inum)
    ref(1:2)   = enefunc%restraint_ref(1:2,inum)

    ! calculation of com
    !
    coefgrp(:,:) = 0.0_wp
    if (kind(inum) == RestraintsFuncDIST) then

      do j = 1, numgrp
        do k = 1, numatoms(grplist(j,inum))
          coefgrp(k,j) = 1.0_wp/real(numatoms(grplist(j,inum)),wp)
        end do
      end do

    else

      do j = 1, numgrp
        do k = 1, numatoms(grplist(j,inum))
          coefgrp(k,j) = masscoef(k,grplist(j,inum))
        end do
      end do

    end if

    do j = 1, numgrp
      com(1:3,j) = 0.0_wp

      do k = 1, numatoms(grplist(j,inum))
        iatm = atmlist(k,grplist(j,inum))
        com(1:3,j) = com(1:3,j) + coord(1:3,iatm) * coefgrp(k,j)
      end do

    end do


    ! Calculation of distance between two group
    !     U(R) = 1/k * const(1) * (R - Ref(1))^k : R < Ref(1)
    !     U(R) = 0                         : Ref(1) <= R <= Ref(2)
    !     U(R) = 1/k * const(2) * (R - Ref(2))^k : R > Ref(2)
    !           ndist
    !       R  = sum u_n * r_n ^ e_n 
    !             n
    !     r_n  = |rcom_i - rcom_j|
    !
    !            N_i
    !   rcom_i = sum w_l * r_l ; if w_l = mass_l/totmass: mass weighted,
    !             l in group-i ;    w_l = 1     / N_i   : average
    !
    !            N_j
    !   rcom_j = sum w_m * r_m ; if w_m = mass_m/totmass: mass weighted,
    !             m in group-j ;    w_m = 1     / N_j   : average
    !

    r_sum = 0.0_wp

    do idist = 1, ndist
      id1 = idist*2-1
      id2 = idist*2

      dij(1:3,idist) = com(1:3,id1) - com(1:3,id2)

      rij(idist)     = sqrt(dij(1,idist)*dij(1,idist) +  &
                            dij(2,idist)*dij(2,idist) +  &
                            dij(3,idist)*dij(3,idist))

      if (expoind(idist,inum) == 1) then
        r_sum = r_sum + wtind(idist,inum) * rij(idist)
      else
        r_sum = r_sum + wtind(idist,inum) * rij(idist)**(expoind(idist,inum))
      end if
    end do

    if (r_sum < ref(1)) then
      ind = 1
    else if (r_sum > ref(2)) then
      ind = 2
    else 
      ! ref(1) < r_sum < ref(2) : force and energy is zero 
      return
    end if

    r_dif = r_sum - ref(ind)
    if (iexpo == 2) then
      coef      = 0.5_wp*const(ind)
      erestdist = coef*r_dif*r_dif
      grad_coef = const(ind)*r_dif

    else
      coef      = const(ind) / real(iexpo,wp)
      erestdist = coef*r_dif**(iexpo)
      grad_coef = const(ind)*r_dif

    end if
   

    ! Calculation of force
    !   coef  = const * (R - Ref)^(k-1)
    !   coefi = coef * u_i * e_n * r_n^(e_n -2)
    !   force(n,i) = coefi * dij(n) * sig(i) * coefgrp(i)   
    !
    if (calc_force) then
      do idist = 1, ndist

        if (rij(idist) < 1e-10_wp) then
          coefi = 0.0_wp
        else
          if (expoind(idist,inum) == 1) then
            coefi = grad_coef * wtind(idist,inum) / rij(idist)
          else
            coefi = expoind(idist,inum) * grad_coef  &
                   * rij(idist) ** (expoind(idist,inum)-2)
          end if
        end if

        id = 2*idist-1
        do j = 1, numatoms(grplist(id,inum))
          iatm = atmlist(j,grplist(id,inum))

          work(1:3) = - coefi * coefgrp(j,id) * dij(1:3,idist)

          force(1:3,iatm) = force(1:3,iatm) + work(1:3)

        end do

        id = 2*idist
        do j = 1, numatoms(grplist(id,inum))
          iatm = atmlist(j,grplist(id,inum))
          work(1:3) =   coefi * coefgrp(j,id) * dij(1:3,idist)
          force(1:3,iatm) = force(1:3,iatm) + work(1:3)
        end do

        ! virial
        ! 
        work(1:3) = coefi * dij(1:3,idist)

        do j = 1,3

          do k = j+1, 3
            vtmp = -work(k) * dij(j,idist)
            virial(k,j) = virial(k,j) + vtmp
            virial(j,k) = virial(j,k) + vtmp
          end do

          vtmp = -work(j) * dij(j,idist)
          virial(j,j) = virial(j,j) + vtmp
        end do

      end do
    end if

    return

  end subroutine compute_energy_distance_restraint

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_angle_restraint
  !> @brief        calculate angle restraint energy
  !! @authors      CK
  !! @param[in]    enefunc   : potential energy functions information
  !! @param[in]    coord     : coordinates of target systems
  !! @param[in]    inum      : pointer for restraint function
  !! @param[in]    const     : force constants
  !! @param[in]    ref       : reference value
  !! @param[inout] force     : forces of target systems
  !! @param[inout] virial    : virial of target systems
  !! @param[out]   erestangle: angle restraint energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_angle_restraint(enefunc, coord, inum, &
                                         calc_force, force, virial, erestangle)

    ! formal arguments
    type(s_enefunc), target, intent(in)    :: enefunc
    real(wp),                intent(in)    :: coord(:,:)
    integer,                 intent(in)    :: inum
    logical,                 intent(in)    :: calc_force
    real(wp),                intent(inout) :: force(:,:)
    real(wp),                intent(inout) :: virial(3,3)
    real(wp),                intent(out)   :: erestangle

    ! local variables
    real(wp)                 :: dij(1:3), dkj(1:3)
    real(wp)                 :: rij, rkj, inv_rij2, inv_rkj2
    real(wp)                 :: rij2, rkj2, rijrkj,inv_rijrkj, vtmp
    real(wp)                 :: cos_t, t_dif, coef, theta, sin_t, grad_coef
    real(wp)                 :: v(1:3,1:3)
    real(wp)                 :: const(2), ref(2), work(1:6)
    real(wp)                 :: radref(1:2)
    integer                  :: numgrp, iexpo
    integer                  :: i, j, k, iatm, ind

    real(wp),        pointer :: masscoef(:,:)
    real(wp),        pointer :: com(:,:)
    real(wp),        pointer :: coefgrp(:,:)
    integer,         pointer :: numatoms(:)
    integer,         pointer :: atmlist(:,:)
    integer,         pointer :: kind(:)
    integer,         pointer :: grplist(:,:)
    integer,         pointer :: funcgrp(:)
    integer,         pointer :: expo(:)
    logical                  :: anglecheck


    funcgrp  => enefunc%restraint_funcgrp 
    expo     => enefunc%restraint_exponent_func
    grplist  => enefunc%restraint_grouplist
    numatoms => enefunc%restraint_numatoms
    atmlist  => enefunc%restraint_atomlist
    kind     => enefunc%restraint_kind
    masscoef => enefunc%restraint_masscoef
    coefgrp  => enefunc%restraint_wtmp
    com      => enefunc%restraint_wcom1

    numgrp   = funcgrp(inum)
    iexpo    = expo   (inum)

    const(1:2) = enefunc%restraint_const(1:2,inum)
    ref(1:2)   = enefunc%restraint_ref(1:2,inum)

    coefgrp = 0.0_wp

    if (kind(inum) == RestraintsFuncANGLE) then
      do j = 1, numgrp
        do k = 1, numatoms(grplist(j,inum))
          coefgrp(k,j) = 1.0_wp/real(numatoms(grplist(j,inum)),wp)
        end do
      end do
    else
      do j = 1, numgrp
        do k = 1, numatoms(grplist(j,inum))
          coefgrp(k,j) = masscoef(k,grplist(j,inum))
        end do
      end do
    end if


    ! calculation of com
    !    
    com(1:3,1:3)=0.0_wp

    do i = 1,3 
      do j = 1,numatoms(grplist(i,inum))
        iatm = atmlist(j,grplist(i,inum))
        com(1:3,i) = com(1:3,i) + coord(1:3,iatm) * coefgrp(j,i)
      end do
    end do
    
    radref(1:2) = ref(1:2)*RAD

    dij(1:3) = com(1:3,1) - com(1:3,2)
    dkj(1:3) = com(1:3,3) - com(1:3,2)

    rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
    rkj2  = dkj(1)*dkj(1) + dkj(2)*dkj(2) + dkj(3)*dkj(3)

    rijrkj = sqrt(rij2*rkj2)

    inv_rijrkj = 1.0_wp /rijrkj
    inv_rij2 = 1.0_wp /rij2
    inv_rkj2 = 1.0_wp /rkj2

    cos_t = (dij(1)*dkj(1) + dij(2)*dkj(2) + dij(3)*dkj(3) )*inv_rijrkj
    cos_t  = min(  1.0_wp, cos_t )
    cos_t  = max( -1.0_wp, cos_t )
    theta = acos(cos_t)

    if (theta < radref(1)) then
      ind = 1
    else if (theta > radref(2)) then
      ind = 2
    else 
      return
    end if

    anglecheck = .true.
    do while(anglecheck)
      if (theta > pi)  theta = theta - 2.0_wp * pi
      if (theta < -pi) theta = theta + 2.0_wp * pi
      if (theta >= -pi  .and. theta <= pi)  anglecheck = .false.
    end do

    t_dif = theta - radref(ind)

    anglecheck = .true.
    do while(anglecheck)
      if (t_dif > pi)  t_dif = t_dif - 2.0_wp * pi
      if (t_dif < -pi) t_dif = t_dif + 2.0_wp * pi
      if (t_dif >= -pi  .and. t_dif <= pi)  anglecheck = .false.
    end do

    if (iexpo == 2) then
      coef       = 0.5_wp * const(ind) 
      erestangle = coef * t_dif * t_dif
      grad_coef  = const(ind) * t_dif
    else
      coef       = const(ind) / real(iexpo,wp)
      erestangle = coef * t_dif**(iexpo)
      grad_coef  = const(ind) * t_dif
    end if


    ! compute force
    !
    if (calc_force) then    
      sin_t = (1.0_wp-cos_t*cos_t)
      sin_t = sqrt(sin_t)
      sin_t  = max( EPS, sin_t )
      grad_coef = -grad_coef / sin_t

      work(1:3) = grad_coef * ( dkj(1:3) * inv_rijrkj -  &
                                dij(1:3) * cos_t * inv_rij2 )

      work(4:6) = grad_coef * ( dij(1:3) * inv_rijrkj -  &
                                dkj(1:3) * cos_t * inv_rkj2 )

      ! virial from angle
      !
      do j = 1, 3

        do k = j+1, 3
          vtmp = -(dij(k)*work(j) + dkj(k)*work(j+3))
          virial(k,j) = virial(k,j) + vtmp
          virial(j,k) = virial(j,k) + vtmp
        end do

        vtmp = -(dij(j)*work(j) + dkj(j)*work(j+3))
        virial(j,j) = virial(j,j) + vtmp
      end do

      do j = 1, numatoms(grplist(1,inum))
        iatm = atmlist(j,grplist(1,inum))
        force(1:3,iatm) = force(1:3,iatm) - work(1:3)*coefgrp(j,1)
      end do

      do j = 1, numatoms(grplist(2,inum))
        iatm = atmlist(j,grplist(2,inum))
        force(1:3,iatm) = force(1:3,iatm)  &
                          + (work(1:3) + work(4:6)) *coefgrp(j,2)
      end do

      do j = 1, numatoms(grplist(3,inum))
        iatm = atmlist(j,grplist(3,inum))
        force(1:3,iatm) = force(1:3,iatm) - work(4:6)*coefgrp(j,3)
      end do
    end if

    return

  end subroutine compute_energy_angle_restraint

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_dihedral_restraint
  !> @brief        calculate dihedral restraint energy
  !! @authors      CK
  !! @param[in]    enefunc   : potential energy functions information
  !! @param[in]    coord     : coordinates of target systems
  !! @param[in]    inum      : pointer for restraint function
  !! @param[in]    const     : force constants
  !! @param[in]    ref       : reference value
  !! @param[inout] force     : forces of target systems
  !! @param[inout] virial    : virial of target systems
  !! @param[out]   erestdihed: dihed restraint energy of target systems
  !! @note         Blondel and Karplus, J. Comput. Chem., 17, 1132-1141 (1996)
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_dihedral_restraint(enefunc, coord, inum, &
                                         calc_force, force, virial, erestdihed)

    ! formal arguments
    type(s_enefunc), target, intent(in)    :: enefunc
    real(wp),                intent(in)    :: coord(:,:)
    integer,                 intent(in)    :: inum
    logical,                 intent(in)    :: calc_force
    real(wp),                intent(inout) :: force(:,:)
    real(wp),                intent(inout) :: virial(3,3)
    real(wp),                intent(out)   :: erestdihed

    ! local variables
    real(wp)                 :: cos_dih, sin_dih, coef
    real(wp)                 :: cospha, sinpha, grad_coef, diffphi
    real(wp)                 :: cosdif, sindif, theta, vtmp
    real(wp)                 :: const(2), ref(2), work(1:9)
    real(wp)                 :: radref(1:2)
    integer                  :: aindex(1:4)
    integer                  :: numgrp, iexpo
    integer                  :: i, j, k, iatm, ind
    real(wp)                 :: grad(1:9), v(1:3,1:3)

    real(wp),        pointer :: masscoef(:,:)
    real(wp),        pointer :: com(:,:)
    real(wp),        pointer :: coefgrp(:,:)
    integer,         pointer :: numatoms(:)
    integer,         pointer :: atmlist(:,:)
    integer,         pointer :: kind(:)
    integer,         pointer :: grplist(:,:)
    integer,         pointer :: funcgrp(:)
    integer,         pointer :: expo(:)


    funcgrp  => enefunc%restraint_funcgrp 
    expo     => enefunc%restraint_exponent_func
    grplist  => enefunc%restraint_grouplist
    numatoms => enefunc%restraint_numatoms
    atmlist  => enefunc%restraint_atomlist
    kind     => enefunc%restraint_kind
    masscoef => enefunc%restraint_masscoef
    coefgrp  => enefunc%restraint_wtmp
    com      => enefunc%restraint_wcom1

    numgrp   = funcgrp(inum)
    iexpo    = expo   (inum)

    const(1:2) = enefunc%restraint_const(1:2,inum)
    ref(1:2)   = enefunc%restraint_ref(1:2,inum)

    coefgrp = 0.0_wp

    if (kind(inum) == RestraintsFuncDIHED) then
      do j = 1, numgrp
        do k = 1, numatoms(grplist(j,inum))
          coefgrp(k,j) = 1.0_wp/real(numatoms(grplist(j,inum)),wp)
        end do
      end do
    else
      do j = 1, numgrp
        do k = 1, numatoms(grplist(j,inum))
          coefgrp(k,j) = masscoef(k,grplist(j,inum))
        end do
      end do
    end if

    ! calculation of com
    !
    com(1:3,1:4)=0.0_wp

    do i = 1,4 
      do j = 1,numatoms(grplist(i,inum))
        iatm = atmlist(j,grplist(i,inum))
        com(1:3,i) = com(1:3,i) + coord(1:3,iatm) * coefgrp(j,i)
      end do
      aindex(i) = i
    end do
    radref(1:2) = ref(1:2)*RAD

    call calculate_dihedral(aindex, com, cos_dih, sin_dih, grad, v)

    if (cos_dih > 1.0E-1_wp) then
      theta = asin(sin_dih)
    else
      theta = sign(1.0_wp,sin_dih)*acos(cos_dih)
    endif
   
    if (theta < radref(1)) then
      ind = 1
    else if (theta > radref(2)) then
      ind = 2
    else 
      return
    end if
    cospha = cos(radref(ind))
    sinpha = sin(radref(ind))

    cosdif = cos_dih*cospha + sin_dih*sinpha
    sindif = cos_dih*sinpha - sin_dih*cospha

    if (cosdif > 1.0E-1_wp) then
      diffphi = asin(sindif)
    else
      diffphi = sign(1.0_wp,sindif)*acos(cosdif)
    endif

    if (iexpo == 2) then
      coef       = 0.5_wp * const(ind) 
      erestdihed = coef * diffphi * diffphi
      grad_coef  = const(ind) * diffphi
    else
      coef       = const(ind) / real(iexpo,wp)
      erestdihed = coef * diffphi **(iexpo)
      grad_coef  = const(ind) * diffphi
    end if
   

    ! compute force
    !
    if (calc_force) then

      if (abs(theta) < 1e-10_wp) then
        grad_coef = 0.0_wp
      end if
      work(1:9) = grad_coef*grad(1:9)

      do j = 1, 3
        do k = j+1, 3
          vtmp = grad_coef*v(k,j)

          virial(k,j) = virial(k,j) + vtmp
          virial(j,k) = virial(j,k) + vtmp
        end do

        vtmp = grad_coef*v(j,j)
        virial(j,j) = virial(j,j) + vtmp
      end do

      do j = 1, numatoms(grplist(1,inum))
        iatm = atmlist(j,grplist(1,inum))
        force(1:3,iatm) = force(1:3,iatm) - work(1:3)*coefgrp(j,1)
      end do

      do j = 1, numatoms(grplist(2,inum))
        iatm = atmlist(j,grplist(2,inum))
        force(1:3,iatm) = force(1:3,iatm) + (work(1:3) - work(4:6))*coefgrp(j,2)
      end do

      do j = 1, numatoms(grplist(3,inum))
        iatm = atmlist(j,grplist(3,inum))
        force(1:3,iatm) = force(1:3,iatm) + (work(4:6) + work(7:9))*coefgrp(j,3)
      end do

      do j = 1, numatoms(grplist(4,inum))
        iatm = atmlist(j,grplist(4,inum))
        force(1:3,iatm) = force(1:3,iatm) - work(7:9)*coefgrp(j,4)
      end do
    end if

    return

  end subroutine compute_energy_dihedral_restraint

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_rmsd_restraint
  !> @brief        calculate position restraint energy with RMSD
  !! @authors      CK
  !! @param[in]    enefunc : potential energy functions information
  !! @param[in]    coord   : coordinates of target systems
  !! @param[in]    inum    : pointer for restraint function
  !! @param[in]    rep_md  : flag for replica MD or not
  !! @param[in]    const   : force constants
  !! @param[in]    ref     : reference value
  !! @param[inout] force   : forces of target systems
  !! @param[inout] virial  : virial of target systems
  !! @param[out]   eposi   : position constraint energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_rmsd_restraint(enefunc, coord, inum, &
                                      calc_force, rep_md, force, virial, eposi)

    ! formal arguments
    type(s_enefunc), target, intent(in)    :: enefunc
    real(wp),                intent(in)    :: coord(:,:)
    integer,                 intent(in)    :: inum
    logical,                 intent(in)    :: calc_force
    logical,                 intent(in)    :: rep_md
    real(wp),                intent(inout) :: force(:,:)
    real(wp),                intent(inout) :: virial(3,3)
    real(wp),                intent(out)   :: eposi

    ! local variables
    real(wp)                 :: r_dif, coef, grad_coef, vtmp
    real(wp)                 :: rmsd
    real(wp)                 :: const(2), ref(2), work(1:3)
    integer                  :: i, j, k, iatm, ind
    integer                  :: natmgrp, iexpo

    real(wp),        pointer :: refcoord(:,:)
    real(wp),        pointer :: masscoef(:,:)
    real(wp),        pointer :: crd_mov(:,:),crd_rmov(:,:), dij(:,:)
    real(wp),        pointer :: coefgrp(:,:)
    integer,         pointer :: numatoms(:)
    integer,         pointer :: atmlist(:,:)
    integer,         pointer :: kind(:)
    integer,         pointer :: grplist(:,:)
    integer,         pointer :: expo(:)


    ! use pointers
    !
    expo     => enefunc%restraint_exponent_func
    grplist  => enefunc%restraint_grouplist
    numatoms => enefunc%restraint_numatoms
    atmlist  => enefunc%restraint_atomlist
    kind     => enefunc%restraint_kind
    masscoef => enefunc%restraint_masscoef
    refcoord => enefunc%restraint_refcoord
    coefgrp  => enefunc%restraint_wtmp
    crd_mov  => enefunc%restraint_wcom3
    crd_rmov => enefunc%restraint_wcom4
    dij      => enefunc%restraint_wcom5

    natmgrp  = numatoms(grplist(1,inum))
    iexpo    = expo(inum)

    do j = 1, 2
      const(j) = enefunc%restraint_const(j,inum)
      ref  (j) = enefunc%restraint_ref  (j,inum)
    end do

    coefgrp(:,:) = 0.0_wp
    if (kind(inum) == RestraintsFuncRMSD) then
      do i = 1, natmgrp
        coefgrp(i,1) = 1.0_wp/real(natmgrp,wp)
      end do
    else
      ! TODO why rep_md is here?
      if (rep_md) then
        do i = 1, natmgrp
          coefgrp(i,1) = masscoef(1,grplist(1,inum))
        end do
      else
        do i = 1, natmgrp
          coefgrp(i,1) = masscoef(i,grplist(1,inum))
        end do
      end if
    end if

    do i = 1, natmgrp
      iatm = atmlist(i,grplist(1,inum))
      crd_mov(1:3,i)  = coord(1:3,iatm)
      crd_rmov(1:3,i) = refcoord(1:3,iatm)
    end do


    ! RMSD calculation
    !   U(r) = const(1) * (rmsd - ref(1))^iexpo : rmsd < ref(1)
    !   U(r) = 0                                : ref(1) <= rmsd <= ref(2)
    !   U(r) = const(2) * (rmsd - ref(2))^iexpo : rmsd > ref(2)
    !   rmsd  = (sum w_l * ( r_l -  r_l,0)^2)^1/2
    !
    rmsd = 0.0_wp
    do i = 1, natmgrp
      dij(1:3,i) = crd_mov(1:3,i) - crd_rmov(1:3,i)
      rmsd = rmsd +  &
             coefgrp(i,1)*(dij(1,i)*dij(1,i) + &
                           dij(2,i)*dij(2,i) + &
                           dij(3,i)*dij(3,i) )
    end do
    rmsd = sqrt(rmsd)

    if (rmsd < ref(1)) then
      ind = 1
    else if (rmsd > ref(2)) then
      ind = 2
    else 
      !ref(1) < rmsd < ref(2) : force and energy is zero 
      return
    end if

    r_dif = rmsd - ref(ind)

    if (iexpo == 2) then
      coef  = const(ind) 
      eposi = coef * r_dif  * r_dif
      grad_coef  = 2.0_wp * const(ind) * r_dif
    else
      coef  = const(ind) 
      eposi = coef * r_dif ** (iexpo)
      grad_coef  = real(iexpo,wp) * const(ind) * r_dif ** (iexpo-1) 
    end if

    ! Compute gradient 
    !
    if (calc_force) then
      if (rmsd < 1e-10_wp) then
        grad_coef = 0.0_wp
      else
        grad_coef = grad_coef/rmsd
      end if

      do i = 1, natmgrp
        iatm = atmlist(i,grplist(1,inum))

        ! gradient: dE/dX
        !
        work(1:3) = grad_coef * coefgrp(i,1) * dij(1:3,i)

        force(1:3,iatm) = force(1:3,iatm) - work(1:3)

        ! external virial
        !
        do j = 1, 3
          do k = j+1, 3

            vtmp = -crd_mov(k,i) * work(j)
            virial(k,j) = virial(k,j) + vtmp
            virial(j,k) = virial(j,j) + vtmp
          end do

          vtmp = -crd_mov(j,i) * work(j)
          virial(j,j) = virial(j,j) + vtmp
        end do
      end do
    end if

    return

  end subroutine compute_energy_rmsd_restraint

end module at_energy_restraints_mod
