!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_energy_table_mod
!> @brief   calculate nonbonded energy with table
!! @authors Jaewoon Jung (JJ), Takaharu Mori (TM)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_energy_table_mod

  use at_boundary_str_mod
  use at_pairlist_str_mod
  use at_enefunc_str_mod
  use at_pairlist_mod
  use molecules_str_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod

  implicit none
  private

  ! variables
  real(wp), save, allocatable :: force_i(:,:)
  real(wp), save, allocatable :: force_j(:,:)

  ! subroutines
  public  :: compute_energy_nonbond_table
  private :: compute_energy_nonbond_table_nb14_linear
  private :: compute_energy_nonbond_table_solute_linear
  private :: compute_energy_nonbond_table_water_linear
  private :: compute_energy_nonbond_table_nb14_cubic
  private :: compute_energy_nonbond_table_solute_cubic
  private :: compute_energy_nonbond_table_water_cubic
  public  :: pme_bond_corr_table
  private :: pme_bond_corr_table_linear
  private :: pme_bond_corr_table_cubic

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_table
  !> @brief        Calculate nonbonded energy with lookup table
  !! @authors      TM
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    pairlist : pairlist information
  !! @param[in]    boundary : boundary information
  !! @param[in]    nonb_ene : flag for calculate nonbonded energy 
  !! @param[in]    coord    : coordinates of target systems
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : lennard-jones energy of target systems
  !! @note         J.Jung et al., J.Comput.Chem., 34, 2412-2420 (2013)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_table(enefunc, molecule, pairlist,  &
                                          boundary, nonb_ene, coord,    &
                                          force, virial, eelec, evdw)

    ! formal arguments
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_molecule),        intent(in)    :: molecule
    type(s_pairlist),        intent(in)    :: pairlist
    type(s_boundary),        intent(in)    :: boundary
    logical,                 intent(in)    :: nonb_ene
    real(wp),                intent(in)    :: coord(:,:)
    real(wp),                intent(inout) :: force(:,:)
    real(wp),                intent(inout) :: virial(3,3)
    real(wp),                intent(inout) :: eelec
    real(wp),                intent(inout) :: evdw


    if (enefunc%table%table_order == 1) then

      ! Compute solute-solute 1-4 interactions
      !
      call compute_energy_nonbond_table_nb14_linear(enefunc, molecule,   &
                                coord, force, virial, eelec, evdw)

      ! Compute solute-solute and solute-water interactions
      !
      call compute_energy_nonbond_table_solute_linear(enefunc, molecule, &
                                boundary, pairlist, nonb_ene, coord,     &
                                force, virial, eelec, evdw)

      ! Compute water-water interactions
      !
      call compute_energy_nonbond_table_water_linear(enefunc, molecule,  &
                                boundary, pairlist, nonb_ene, coord,     &
                                force, virial, eelec, evdw)

    else if (enefunc%table%table_order == 3) then

      ! Compute solute-solute 1-4 interactions
      !
      call compute_energy_nonbond_table_nb14_cubic(enefunc, molecule,    &
                                coord, force, virial, eelec, evdw)

      ! Compute solute-solute and solute-water interactions
      !
      call compute_energy_nonbond_table_solute_cubic(enefunc, molecule,  &
                                boundary, pairlist, nonb_ene, coord,     &
                                force, virial, eelec, evdw)

      ! Compute water-water interactions
      !
      call compute_energy_nonbond_table_water_cubic(enefunc, molecule,   &
                                boundary, pairlist, nonb_ene, coord,     &
                                force, virial, eelec, evdw)

    else

      call error_msg('Compute_Energy_Nonbond_Table> Bad table_order variable')

    end if


    return
 
  end subroutine compute_energy_nonbond_table

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_table_nb14_linear
  !> @brief        Calculate PME nonbond 1-4 term real part + Lennard-Jones 
  !!               with linear 1/R**2 table
  !! @authors      JJ
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    coord    : coordinates of target systems
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : lennard-jones energy of target systems
  !! @note         J.Jung et al., J.Comput.Chem., 34, 2412-2420 (2013)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_table_nb14_linear(enefunc, molecule, coord,&
                                        force, virial, eelec, evdw)

    ! formal arguments
    type(s_molecule), target, intent(in)    :: molecule
    type(s_enefunc),  target, intent(in)    :: enefunc
    real(wp),                 intent(in)    :: coord(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec
    real(wp),                 intent(inout) :: evdw

    ! local variables
    real(wp)                  :: dij(3)
    real(wp)                  :: rij2
    real(wp)                  :: cutoff2
    real(wp)                  :: grad_coef
    real(wp)                  :: work(3)
    real(wp)                  :: R, table(6)
    real(wp)                  :: lj6, lj12
    real(wp)                  :: term_lj12, term_lj6, term_elec
    real(wp)                  :: force_local(1:3)
    integer                   :: ii, i, j, k, m, L
    integer                   :: num_nb14, ini_nb14, fin_nb14
    integer                   :: id, my_id
    integer                   :: num_nb14_all

    real(wp),         pointer :: charge(:)
    real(wp),         pointer :: cutoff
    real(wp),         pointer :: table_ene(:), table_grad(:)
    integer,          pointer :: natom
    integer,          pointer :: atmcls(:)
    integer,          pointer :: num_nb14_calc(:), nb14_calc_list(:)

#ifdef OMP
    integer                   :: omp_get_thread_num
#endif


    ! use pointers
    !
    natom          => enefunc%table%num_solute
    charge         => molecule%charge
    atmcls         => molecule%atom_cls_no
    cutoff         => enefunc%cutoffdist
    num_nb14_calc  => enefunc%num_nb14_calc
    nb14_calc_list => enefunc%nb14_calc_list
    table_ene      => enefunc%table%table_ene
    table_grad     => enefunc%table%table_grad

    cutoff2  = cutoff * cutoff
    num_nb14 = 0


    if (.not. allocated(force_i)) then
      allocate(force_i(3,enefunc%table%num_solute), &
               force_j(3,size(enefunc%nb14_calc_list)))
    end if

    force_i(1:3,1:natom) = 0.0_wp
    force_j(1:3,1:size(enefunc%nb14_calc_list)) = 0.0_wp


    ! calculate energy and gradient
    !
    !$omp parallel                                                             &
    !$omp firstprivate(num_nb14)                                               &
    !$omp private(id, my_id, ini_nb14, fin_nb14, ii, i, k, j, m, dij, rij2,    &
    !$omp         R, L, work, grad_coef, lj6, lj12, term_lj12, term_lj6,       &
    !$omp         term_elec, table, force_local)                               &
    !$omp reduction(+:virial) reduction(+:eelec) reduction(+:evdw)              
    !
#ifdef OMP
    id      = omp_get_thread_num()
#else
    id      = 0
#endif
   my_id = my_city_rank * nthread + id

    do ii = 1, natom-1

      ini_nb14  = num_nb14 + 1
      fin_nb14  = num_nb14 + num_nb14_calc(ii)
      num_nb14  = fin_nb14

      force_local(1:3) = 0.0_wp

      i = enefunc%table%solute_list(ii)

      if (mod(ii-1,nproc_city*nthread) == my_id) then

        do k = ini_nb14, fin_nb14

          j = nb14_calc_list(k)

          ! compute distance
          !
          dij(1:3) = coord(1:3,i) - coord(1:3,j)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          ! cutoff
          !
          if (rij2 < cutoff2) then

            rij2 = cutoff2 * enefunc%table%density / rij2
            L = int(rij2)
            R = rij2 - L
            lj6  = enefunc%nb14_lj6(atmcls(i),atmcls(j))
            lj12 = enefunc%nb14_lj12(atmcls(i),atmcls(j))

            table(1:6) = table_ene(3*L-2:3*L+3)
            term_lj12 = table(1) + R*(table(4)-table(1))
            term_lj6  = table(2) + R*(table(5)-table(2))
            term_elec = table(3) + R*(table(6)-table(3))
            evdw = evdw + term_lj12*lj12 - term_lj6*lj6
            eelec = eelec + charge(i)*charge(j)*term_elec

            table(1:6) = table_grad(3*L-2:3*L+3)
            term_lj12 = table(1) + R*(table(4)-table(1))
            term_lj6  = table(2) + R*(table(5)-table(2))
            term_elec = table(3) + R*(table(6)-table(3))
            grad_coef = term_lj12*lj12 - term_lj6*lj6 &
                      + charge(i)*charge(j)*term_elec

            work(1:3) = grad_coef*dij(1:3)

            ! store force
            !
            force_local(1:3) = force_local(1:3) - work(1:3)
            force_j(1:3,k) = work(1:3)

            ! virial
            !
            do m = 1, 3
              virial(1:3,m) = virial(1:3,m) - dij(1:3)*work(m)
            end do

          end if

        end do

        force_i(1:3,ii) = force_local(1:3)

      end if

    end do

    num_nb14_all  = num_nb14
    !$omp end parallel

    do ii = 1, natom-1
       i = enefunc%table%solute_list(ii)
       force(1,i) = force(1,i) + force_i(1,ii)
       force(2,i) = force(2,i) + force_i(2,ii)
       force(3,i) = force(3,i) + force_i(3,ii)
    end do

    do k = 1, num_nb14_all
       j = nb14_calc_list(k)
       force(1,j) = force(1,j) + force_j(1,k)
       force(2,j) = force(2,j) + force_j(2,k)
       force(3,j) = force(3,j) + force_j(3,k)
    end do

    return

  end subroutine compute_energy_nonbond_table_nb14_linear

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_table_solute_linear
  !> @brief        Calculate PME real part + Lennard-Jones with lookup table
  !! @authors      JJ, TM
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    boundary : boundary information
  !! @param[in]    pairlist : pairlist information 
  !! @param[in]    coord    : coordinates of target systems
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : lennard-jones energy of target systems
  !! @note         J.Jung et al., J.Comput.Chem., 34, 2412-2420 (2013)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_table_solute_linear(enefunc, molecule,     &
                                   boundary, pairlist, nonb_ene, coord, force, &
                                   virial, eelec, evdw)

    ! formal arguments
    type(s_molecule), target, intent(in)    :: molecule
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_boundary), target, intent(in)    :: boundary
    type(s_pairlist), target, intent(in)    :: pairlist
    logical,                  intent(in)    :: nonb_ene
    real(wp),                 intent(in)    :: coord(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec
    real(wp),                 intent(inout) :: evdw

    ! local variables
    real(wp)                  :: rtmp(1:3)
    real(wp)                  :: dij(3)
    real(wp)                  :: rij2
    real(wp)                  :: cutoff2
    real(wp)                  :: grad_coef
    real(wp)                  :: work(3)
    real(wp)                  :: R, qtmp, lj12, lj6
    real(wp)                  :: term_lj12, term_lj6, term_elec
    real(wp)                  :: force_local(1:3)
    real(wp)                  :: box_size(1:3)
    integer                   :: ii, i, j, k, m, L, L1
    integer                   :: num_nb15, ini_nb15, fin_nb15
    integer                   :: id, my_id

    real(wp),         pointer :: cutoff, charge(:)
    real(wp),         pointer :: table_ene(:), table_grad(:)
    integer,          pointer :: nsolute, atmcls(:), solute_list(:)
    integer,          pointer :: num_nb15_calc(:,:), nb15_calc_list(:,:)
    integer,          pointer :: num_nb15_calcw(:,:), nb15_calc_listw(:,:)

#ifdef OMP
    integer                   :: omp_get_thread_num
#endif


    ! use pointers
    nsolute         => enefunc%table%num_solute
    charge          => molecule%charge
    atmcls          => molecule%atom_cls_no
    num_nb15_calc   => pairlist%table%num_nb15_calc
    num_nb15_calcw  => pairlist%table%num_nb15_calcw
    nb15_calc_list  => pairlist%table%nb15_calc_list
    nb15_calc_listw => pairlist%table%nb15_calc_listw
    solute_list     => enefunc%table%solute_list
    cutoff          => enefunc%cutoffdist
    table_ene       => enefunc%table%table_ene
    table_grad      => enefunc%table%table_grad
    box_size(1)     =  boundary%box_size_x
    box_size(2)     =  boundary%box_size_y
    box_size(3)     =  boundary%box_size_z

    cutoff2  = cutoff * cutoff
    num_nb15 = 0

    ! calculate energy and gradient
    !
    !$omp parallel                                                    & 
    !$omp firstprivate(num_nb15)                                      &
    !$omp private(id, my_id, ini_nb15, fin_nb15, i, ii, k, j, m, dij, &
    !$omp         rij2, R, L, L1, work, grad_coef, rtmp, qtmp,        &
    !$omp         term_lj12, term_lj6, term_elec, lj12, lj6,          &
    !$omp         force_local)                                        &
    !$omp reduction(+:force) reduction(+:virial) reduction(+:eelec)   &
    !$omp reduction(+:evdw)
    !
#ifdef OMP
    id      = omp_get_thread_num()
#else
    id      = 0
#endif
    my_id   = my_city_rank * nthread + id

    do ii = 1, nsolute-1

      ini_nb15 = num_nb15 + 1
      fin_nb15 = num_nb15 + num_nb15_calc(ii,id+1)
      num_nb15 = fin_nb15

      i = solute_list(ii)

      rtmp(1:3) = coord(1:3,i)
      qtmp = charge(i)
      force_local(1:3) = 0.0_wp

      if (mod(ii-1,nproc_city*nthread) /= my_id) ini_nb15 = fin_nb15 + 1

      do k = ini_nb15, fin_nb15

        j = nb15_calc_list(k,id+1)

        ! compute distance
        !
        dij(1:3) = rtmp(1:3) - coord(1:3,j)
        dij(1:3) = dij(1:3)-box_size(1:3)*anint(dij(1:3)/box_size(1:3))

        rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

        if (rij2 < cutoff2) then

          rij2  = cutoff2 * enefunc%table%density / rij2

          L    = int(rij2)
          R    = rij2 - L
          lj12 = enefunc%nonb_lj12(atmcls(i),atmcls(j))
          lj6  = enefunc%nonb_lj6(atmcls(i),atmcls(j))

          ! compute force
          !
          L1 = 3*L - 2
          term_lj12 = table_grad(L1  ) + R*(table_grad(L1+3)-table_grad(L1  ))
          term_lj6  = table_grad(L1+1) + R*(table_grad(L1+4)-table_grad(L1+1))
          term_elec = table_grad(L1+2) + R*(table_grad(L1+5)-table_grad(L1+2))
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(j)*term_elec

          ! comute energy
          !
          if (nonb_ene) then
            L1 = 3*L - 2
            term_lj12 = table_ene(L1  ) + R*(table_ene(L1+3)-table_ene(L1  ))
            term_lj6  = table_ene(L1+1) + R*(table_ene(L1+4)-table_ene(L1+1))
            term_elec = table_ene(L1+2) + R*(table_ene(L1+5)-table_ene(L1+2))
            evdw      = evdw + term_lj12*lj12 - term_lj6*lj6
            eelec     = eelec + qtmp*charge(j)*term_elec
          end if

          work(1:3) = grad_coef*dij(1:3)

          ! store force
          !
          force_local(1:3) = force_local(1:3) - work(1:3)
          force(1:3,j) = force(1:3,j) + work(1:3)

          ! virial
          !
          do m = 1, 3
            virial(1:3,m) = virial(1:3,m) - dij(1:3)*work(m)
          end do

        end if
      end do

      force(1:3,i) = force(1:3,i) + force_local(1:3)

    end do

    num_nb15 = 0
    do ii = 1, nsolute

      ini_nb15 = num_nb15 + 1
      fin_nb15 = num_nb15 + num_nb15_calcw(ii,id+1)
      num_nb15 = fin_nb15

      i = solute_list(ii)

      rtmp(1:3) = coord(1:3,i)
      qtmp = charge(i)
      force_local(1:3) = 0.0_wp

      if (mod(ii-1,nproc_city*nthread) /= my_id) ini_nb15 = fin_nb15 + 1

      do k = ini_nb15, fin_nb15

        j = nb15_calc_listw(k,id+1)

        ! compute distance
        !
        dij(1:3) = rtmp(1:3) - coord(1:3,j)
        dij(1:3) = dij(1:3)-box_size(1:3)*anint(dij(1:3)/box_size(1:3))

        rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

        if (rij2 < cutoff2) then

          rij2  = cutoff2 * enefunc%table%density / rij2

          L    = int(rij2)
          R    = rij2 - L
          lj12 = enefunc%nonb_lj12(atmcls(i),atmcls(j))
          lj6  = enefunc%nonb_lj6 (atmcls(i),atmcls(j))


          ! compute force
          !
          L1 = 3*L - 2
          term_lj12 = table_grad(L1  ) + R*(table_grad(L1+3)-table_grad(L1  ))
          term_lj6  = table_grad(L1+1) + R*(table_grad(L1+4)-table_grad(L1+1))
          term_elec = table_grad(L1+2) + R*(table_grad(L1+5)-table_grad(L1+2))
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(j)*term_elec

          ! compute energy
          !
          if (nonb_ene) then
            L1 = 3*L - 2
            term_lj12 = table_ene(L1  ) + R*(table_ene(L1+3)-table_ene(L1  ))
            term_lj6  = table_ene(L1+1) + R*(table_ene(L1+4)-table_ene(L1+1))
            term_elec = table_ene(L1+2) + R*(table_ene(L1+5)-table_ene(L1+2))
            evdw      = evdw  + term_lj12*lj12 - term_lj6*lj6
            eelec     = eelec + qtmp*charge(j)*term_elec
          end if

          work(1:3) = grad_coef*dij(1:3)

          ! store force
          !
          force_local(1:3) = force_local(1:3) - work(1:3)
          force(1:3,j) = force(1:3,j) + work(1:3)

          ! virial
          !
          do m = 1, 3
            virial(1:3,m) = virial(1:3,m) - dij(1:3)*work(m)
          end do

        end if
      end do

      force(1:3,i) = force(1:3,i) + force_local(1:3)

    end do

    !$omp end parallel

    return

  end subroutine compute_energy_nonbond_table_solute_linear
 
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_table_water_linear
  !> @brief        Calculate Nonbonded water-water interaction with table
  !! @authors      JJ, TM
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    boundary : boundary information
  !! @param[in]    pairlist : pairlist information 
  !! @param[in]    coord    : coordinates of target systems
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : lennard-jones energy of target systems
  !! @note         J.Jung et al., J.Comput.Chem., 34, 2412-2420 (2013)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_table_water_linear(enefunc, molecule,     &
                                         boundary, pairlist, nonb_ene, coord, &
                                         force, virial, eelec, evdw)

    ! formal arguments
    type(s_molecule), target, intent(in)    :: molecule
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_boundary), target, intent(in)    :: boundary
    type(s_pairlist), target, intent(in)    :: pairlist
    logical,                  intent(in)    :: nonb_ene
    real(wp),                 intent(in)    :: coord(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec
    real(wp),                 intent(inout) :: evdw

    ! local variables
    real(wp)                  :: rtmp(1:3,1:3)
    real(wp)                  :: force_local(1:3,1:3)
    real(wp)                  :: dij(1:3,1:3,1:3)
    real(wp)                  :: work(1:3,1:3,1:3)
    real(wp)                  :: rij2(1:3,1:3)
    real(wp)                  :: viri_local(1:3,1:3)
    real(wp)                  :: R, cutoff2
    real(wp)                  :: Delta_pbc(1:3), box_size(1:3)
    real(wp)                  :: grad_coef, term_lj, term_elec 
    integer                   :: type(1:3,1:3), water_type
    integer                   :: i, j, k, L, L1
    integer                   :: i1(1:3), j1(1:3)
    integer                   :: ii, jj, ia, ja, m
    integer                   :: num_nb15, ini_nb15, fin_nb15
    integer                   :: id, my_id

    real(wp),         pointer :: cutoff
    real(wp),         pointer :: table_ene_WW(:,:), table_de_WW(:,:)
    integer,          pointer :: nwater
    integer,          pointer :: num_nb15_calc(:,:), nb15_calc_list(:,:)

#ifdef OMP
    integer                   :: omp_get_thread_num
#endif


    ! use pointers
    nwater         => enefunc%table%num_water
    num_nb15_calc  => pairlist%table%num_nb15_calc_water
    nb15_calc_list => pairlist%table%nb15_calc_list_water
    cutoff         => enefunc%cutoffdist
    table_ene_WW   => enefunc%table%table_ene_WW
    table_de_WW    => enefunc%table%table_de_WW
    box_size(1)    =  boundary%box_size_x
    box_size(2)    =  boundary%box_size_y
    box_size(3)    =  boundary%box_size_z

    cutoff2  = cutoff * cutoff
    num_nb15 = 0

    type(1,1)     = 1
    type(1,2:3)   = 2
    type(2:3,1)   = 2
    type(2:3,2:3) = 3

    ! calculate energy and gradient
    !
    !$omp parallel                                                             & 
    !$omp firstprivate(num_nb15)                                               &
    !$omp private(id, my_id, ini_nb15, fin_nb15, i, k, j, i1, j1, ia, ja, m,   &
    !$omp         ii, jj, dij, rij2, R, L, L1, term_lj, grad_coef, term_elec,  &
    !$omp         rtmp,  force_local, viri_local, work, Delta_pbc, water_type) &
    !$omp reduction(+:force) reduction(+:virial) reduction(+:eelec)            &
    !$omp reduction(+:evdw)
    !
#ifdef OMP
    id      = omp_get_thread_num()
#else
    id      = 0
#endif
    my_id   = my_city_rank * nthread + id

    do i = 1, nwater-1

      ini_nb15 = num_nb15 + 1
      fin_nb15 = num_nb15 + num_nb15_calc(i,id+1)
      num_nb15 = fin_nb15

      i1(1:3) = enefunc%table%water_list(1:3,i)

      do ii = 1, 3
        rtmp(1:3,ii) = coord(1:3,i1(ii))
      end do

      force_local(1:3,1:3) = 0.0_wp

      if (mod(i-1,nproc_city*nthread) /= my_id) ini_nb15 = fin_nb15 + 1

      do k = ini_nb15, fin_nb15

        j = nb15_calc_list(k,id+1)

        j1(1:3) = enefunc%table%water_list(1:3,j)

        ! compute distance
        !
        dij(1:3,1,1)   = rtmp(1:3,1) - coord(1:3,j1(1))
        Delta_pbc(1:3) = box_size(1:3)*anint(dij(1:3,1,1)/box_size(1:3))
        do ii = 1, 3
          do jj = 1, 3
            ja = j1(jj)
            dij(1:3,ii,jj) = rtmp(1:3,ii) - coord(1:3,ja) - Delta_pbc(1:3) 
            rij2(ii,jj) = dij(1,ii,jj)*dij(1,ii,jj) + dij(2,ii,jj)*dij(2,ii,jj) &
                        + dij(3,ii,jj)*dij(3,ii,jj)
          end do
        end do

        viri_local(1:3,1:3) = 0.0_wp        

        do ii = 1, 3
          do jj = 1, 3

            if (rij2(ii,jj) < cutoff2) then

              rij2(ii,jj) = cutoff2 * enefunc%table%density / rij2(ii,jj)

              L    = int(rij2(ii,jj))
              R    = rij2(ii,jj) - L
              water_type = type(ii,jj)


              if (.not. nonb_ene) then
                ! compute force only
                !
                grad_coef = table_de_WW(L,water_type)       &
                          + R*(table_de_WW(L+1,water_type)  &
                             - table_de_WW(L,  water_type))
              else
                ! compute both energy and force
                !
                L1 = 6*L-5
                term_lj   =   table_ene_WW(L1,  water_type)
                term_elec =   table_ene_WW(L1+1,water_type)
                grad_coef =   table_ene_WW(L1+2,water_type)
                term_lj   = term_lj                         &
                          + R*table_ene_WW(L1+3,water_type)
                term_elec = term_elec                       &
                          + R*table_ene_WW(L1+4,water_type)
                grad_coef = grad_coef                       &
                          + R*table_ene_WW(L1+5,water_type)

                evdw  = evdw  + term_lj
                eelec = eelec + term_elec
              end if

              work(1:3,ii,jj) = grad_coef * dij(1:3,ii,jj)

              force_local(1:3,ii) = force_local(1:3,ii) - work(1:3,ii,jj)
              ja = j1(jj)
              force(1:3,ja) = force(1:3,ja) + work(1:3,ii,jj)

              do m = 1, 3
                viri_local(1:3,m) = viri_local(1:3,m)  &
                                  + dij(1:3,ii,jj)*work(m,ii,jj)
              end do

            end if

          end do
        end do

        ! virial
        !
        do m = 1, 3
          virial(1:3,m) = virial(1:3,m) - viri_local(1:3,m)
        end do

      end do

      do ii = 1, 3
        ia = i1(ii)
        force(1:3,ia) = force(1:3,ia) + force_local(1:3,ii)
      end do

    end do
    !$omp end parallel

    return

  end subroutine compute_energy_nonbond_table_water_linear

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_table_nb14_cubic
  !> @brief        Calculate PME nonbond 1-4 term real part + Lennard-Jones 
  !!               with table
  !! @authors      JJ
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    coord    : coordinates of target systems
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : lennard-jones energy of target systems
  !! @note         J.Jung et al., J.Comput.Chem., 34, 2412-2420 (2013)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_table_nb14_cubic(enefunc, molecule, coord, &
                                       force, virial, eelec, evdw)

    ! formal arguments
    type(s_molecule), target, intent(in)    :: molecule
    type(s_enefunc),  target, intent(in)    :: enefunc
    real(wp),                 intent(in)    :: coord(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec
    real(wp),                 intent(inout) :: evdw

    ! local variables
    real(wp)                  :: dij(3)
    real(wp)                  :: rij2
    real(wp)                  :: cutoff2
    real(wp)                  :: grad_coef
    real(wp)                  :: work(1:3), force_local(1:3)
    real(wp)                  :: R, h00, h01, h10, h11, table(12)
    real(wp)                  :: lj6, lj12
    real(wp)                  :: term_lj12, term_lj6, term_elec
    integer                   :: ii, i, j, k, m, L
    integer                   :: num_nb14, ini_nb14, fin_nb14
    integer                   :: id, my_id
    integer                   :: num_nb14_all

    real(wp),         pointer :: charge(:)
    real(wp),         pointer :: cutoff
    real(wp),         pointer :: table_ene(:), table_grad(:)
    integer,          pointer :: natom
    integer,          pointer :: atmcls(:)
    integer,          pointer :: num_nb14_calc(:), nb14_calc_list(:)

#ifdef OMP
    integer                   :: omp_get_thread_num
#endif

    ! use pointers
    !
    natom          => enefunc%table%num_solute
    charge         => molecule%charge
    atmcls         => molecule%atom_cls_no
    cutoff         => enefunc%cutoffdist
    num_nb14_calc  => enefunc%num_nb14_calc
    nb14_calc_list => enefunc%nb14_calc_list
    table_ene      => enefunc%table%table_ene
    table_grad     => enefunc%table%table_grad

    cutoff2  = cutoff * cutoff
    num_nb14 = 0

    if (.not. allocated(force_i)) then
      allocate(force_i(3,enefunc%table%num_solute), &
               force_j(3,size(enefunc%nb14_calc_list)))
    end if

    force_i(1:3,1:natom) = 0.0_wp
    force_j(1:3,1:size(enefunc%nb14_calc_list)) = 0.0_wp


    ! calculate energy and gradient
    !
    !$omp parallel                                                             &
    !$omp firstprivate(num_nb14)                                               &
    !$omp private(id, my_id, ini_nb14, fin_nb14, ii, i, k, j, m, dij, rij2,    &
    !$omp         R, L, work, grad_coef, lj6, lj12, term_lj12, term_lj6,       &
    !$omp         term_elec, table, h00, h10, h01, h11, force_local)           &
    !$omp reduction(+:virial) reduction(+:eelec) reduction(+:evdw) 
    !
#ifdef OMP
    id      = omp_get_thread_num()
#else
    id      = 0
#endif
    my_id = my_city_rank * nthread + id

    do ii = 1, natom-1

      ini_nb14  = num_nb14 + 1
      fin_nb14  = num_nb14 + num_nb14_calc(ii)
      num_nb14  = fin_nb14

      i = enefunc%table%solute_list(ii)

      force_local(1:3) = 0.0_wp

      if (mod(ii-1,nproc_city*nthread) == my_id) then

        do k = ini_nb14, fin_nb14

          j = nb14_calc_list(k)

          ! compute distance
          !
          dij(1:3) = coord(1:3,i) - coord(1:3,j)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          ! cutoff
          !
          if (rij2 < cutoff2) then

            rij2 = rij2 * enefunc%table%density 
            L    = int(rij2)
            R    = rij2 - L
            lj6  = enefunc%nb14_lj6(atmcls(i),atmcls(j))
            lj12 = enefunc%nb14_lj12(atmcls(i),atmcls(j))

            h00  = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
            h10  = R*(1.0_wp-R)*(1.0_wp-R)
            h01  = R*R*(3.0_wp-2.0_wp*R)
            h11  = R*R*(R-1.0_wp)

            table(1:12) = table_ene(6*L-5:6*L+6)
            term_lj12 = table(1)*h00 + table(2)*h10 + table(7 )*h01 + table(8 )*h11
            term_lj6  = table(3)*h00 + table(4)*h10 + table(9 )*h01 + table(10)*h11
            term_elec = table(5)*h00 + table(6)*h10 + table(11)*h01 + table(12)*h11
            evdw = evdw + term_lj12*lj12 - term_lj6*lj6
            eelec = eelec + charge(i)*charge(j)*term_elec

            table(1:12) = table_grad(6*L-5:6*L+6)
            term_lj12 = table(1)*h00 + table(2)*h10 + table(7 )*h01 + table(8 )*h11
            term_lj6  = table(3)*h00 + table(4)*h10 + table(9 )*h01 + table(10)*h11
            term_elec = table(5)*h00 + table(6)*h10 + table(11)*h01 + table(12)*h11
            grad_coef = term_lj12*lj12 - term_lj6*lj6 + charge(i)*charge(j)*term_elec

            work(1:3) = grad_coef*dij(1:3)

            ! store force
            !
            force_local(1:3) = force_local(1:3) - work(1:3)
            force_j(1:3,k)   = work(1:3)

            ! virial
            !
            do m = 1, 3
              virial(1:3,m) = virial(1:3,m) - dij(1:3)*work(m)
            end do

          end if

        end do

        force_i(1:3,ii) = force_local(1:3)

      end if

    end do

    num_nb14_all  = num_nb14
    !$omp end parallel

    do ii = 1, natom-1
       i = enefunc%table%solute_list(ii)
       force(1,i) = force(1,i) + force_i(1,ii)
       force(2,i) = force(2,i) + force_i(2,ii)
       force(3,i) = force(3,i) + force_i(3,ii)
    end do

    do k = 1, num_nb14_all
       j = nb14_calc_list(k)
       force(1,j) = force(1,j) + force_j(1,k)
       force(2,j) = force(2,j) + force_j(2,k)
       force(3,j) = force(3,j) + force_j(3,k)
    end do

    return

  end subroutine compute_energy_nonbond_table_nb14_cubic

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_table_solute_cubic
  !> @brief        Calculate PME real part + Lennard-Jones with lookup table
  !! @authors      JJ, TM
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    boundary : boundary information
  !! @param[in]    pairlist : pairlist information 
  !! @param[in]    coord    : coordinates of target systems
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : lennard-jones energy of target systems
  !! @note         J.Jung et al., J.Comput.Chem., 34, 2412-2420 (2013)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_table_solute_cubic(enefunc, molecule,     &
                                  boundary, pairlist, nonb_ene, coord, force, &
                                  virial, eelec, evdw)

    ! formal arguments
    type(s_molecule), target, intent(in)    :: molecule
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_boundary), target, intent(in)    :: boundary
    type(s_pairlist), target, intent(in)    :: pairlist
    logical,                  intent(in)    :: nonb_ene
    real(wp),                 intent(in)    :: coord(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec
    real(wp),                 intent(inout) :: evdw

    ! local variables
    real(wp)                  :: rtmp(1:3)
    real(wp)                  :: dij(3)
    real(wp)                  :: rij2
    real(wp)                  :: cutoff2
    real(wp)                  :: grad_coef
    real(wp)                  :: work(3)
    real(wp)                  :: R, qtmp, lj12, lj6
    real(wp)                  :: term_lj12, term_lj6, term_elec
    real(wp)                  :: force_local(1:3)
    real(wp)                  :: box_size(1:3)
    real(wp)                  :: h00, h01, h10, h11
    integer                   :: ii, i, j, k, m, L, L1
    integer                   :: num_nb15, ini_nb15, fin_nb15
    integer                   :: id, my_id

    real(wp),         pointer :: cutoff, charge(:)
    real(wp),         pointer :: table_ene(:), table_grad(:)
    real(wp),         pointer :: bsize_x, bsize_y, bsize_z
    integer,          pointer :: nsolute, atmcls(:), solute_list(:)
    integer,          pointer :: num_nb15_calc(:,:), nb15_calc_list(:,:)
    integer,          pointer :: num_nb15_calcw(:,:), nb15_calc_listw(:,:)

#ifdef OMP
    integer                   :: omp_get_thread_num
#endif


    ! use pointers
    nsolute         => enefunc%table%num_solute
    charge          => molecule%charge
    atmcls          => molecule%atom_cls_no
    num_nb15_calc   => pairlist%table%num_nb15_calc
    num_nb15_calcw  => pairlist%table%num_nb15_calcw
    nb15_calc_list  => pairlist%table%nb15_calc_list
    nb15_calc_listw => pairlist%table%nb15_calc_listw
    solute_list     => enefunc%table%solute_list
    cutoff          => enefunc%cutoffdist
    table_ene       => enefunc%table%table_ene
    table_grad      => enefunc%table%table_grad
    box_size(1)     =  boundary%box_size_x
    box_size(2)     =  boundary%box_size_y
    box_size(3)     =  boundary%box_size_z

    cutoff2  = cutoff * cutoff
    num_nb15 = 0

    ! calculate energy and gradient
    !
    !$omp parallel                                                    & 
    !$omp firstprivate(num_nb15)                                      &
    !$omp private(id, my_id, ini_nb15, fin_nb15, i, ii, k, j, m, dij, &
    !$omp         rij2, R, L, L1, work, grad_coef, rtmp, qtmp,        &
    !$omp         term_lj12, term_lj6, term_elec, lj12, lj6,          &
    !$omp         h00, h01, h10, h11, force_local)                    &
    !$omp reduction(+:force) reduction(+:virial) reduction(+:eelec)   &
    !$omp reduction(+:evdw)
    !
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif
    my_id = my_city_rank * nthread + id

    ! solute-solute
    !
    do ii = 1, nsolute-1

      ini_nb15 = num_nb15 + 1
      fin_nb15 = num_nb15 + num_nb15_calc(ii,id+1)
      num_nb15 = fin_nb15

      i = solute_list(ii)

      rtmp(1:3) = coord(1:3,i)
      qtmp = charge(i)
      force_local(1:3) = 0.0_wp

      if (mod(ii-1,nproc_city*nthread) /= my_id) ini_nb15 = fin_nb15 + 1

      do k = ini_nb15, fin_nb15

        j = nb15_calc_list(k,id+1)

        ! compute distance
        !
        dij(1:3) = rtmp(1:3) - coord(1:3,j)
        dij(1:3) = dij(1:3)-box_size(1:3)*anint(dij(1:3)/box_size(1:3))

        rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

        if (rij2 < cutoff2) then

          rij2 = rij2 * enefunc%table%density 
          L    = int(rij2)
          R    = rij2 - L
          lj12 = enefunc%nonb_lj12(atmcls(i),atmcls(j))
          lj6  = enefunc%nonb_lj6(atmcls(i),atmcls(j))

          h00  = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10  = R*(1.0_wp-R)*(1.0_wp-R)
          h01  = R*R*(3.0_wp-2.0_wp*R)
          h11  = R*R*(R-1.0_wp)

          ! compute force
          !
          L1 = 6*L - 5
          term_lj12 = table_grad(L1  )*h00 + table_grad(L1+1)*h10
          term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
          term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
          term_lj12 = term_lj12 + table_grad(L1+6 )*h01 + table_grad(L1+7 )*h11
          term_lj6  = term_lj6  + table_grad(L1+8 )*h01 + table_grad(L1+9 )*h11
          term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(j)*term_elec

          ! compute energy
          !
          if (nonb_ene) then
            L1 = 6*L - 5
            term_lj12 = table_ene(L1  )*h00 + table_ene(L1+1)*h10
            term_lj6  = table_ene(L1+2)*h00 + table_ene(L1+3)*h10
            term_elec = table_ene(L1+4)*h00 + table_ene(L1+5)*h10
            term_lj12 = term_lj12 + table_ene(L1+6 )*h01 + table_ene(L1+7 )*h11
            term_lj6  = term_lj6  + table_ene(L1+8 )*h01 + table_ene(L1+9 )*h11
            term_elec = term_elec + table_ene(L1+10)*h01 + table_ene(L1+11)*h11
            evdw      = evdw  + term_lj12*lj12 - term_lj6*lj6
            eelec     = eelec + qtmp*charge(j)*term_elec
          end if

          work(1:3) = grad_coef*dij(1:3)

          ! store force
          !
          force_local(1:3) = force_local(1:3) - work(1:3)
          force(1:3,j)     = force(1:3,j) + work(1:3)

          ! virial
          !
          do m = 1, 3
            virial(1:3,m) = virial(1:3,m) - dij(1:3)*work(m)
          end do

        end if
      end do

      force(1:3,i) = force(1:3,i) + force_local(1:3)

    end do

    ! solute-water
    !
    num_nb15 = 0
    do ii = 1, nsolute

      ini_nb15 = num_nb15 + 1
      fin_nb15 = num_nb15 + num_nb15_calcw(ii,id+1)
      num_nb15 = fin_nb15

      i = solute_list(ii)

      rtmp(1:3) = coord(1:3,i)
      qtmp = charge(i)
      force_local(1:3) = 0.0_wp

      if (mod(ii-1,nproc_city*nthread) /= my_id) ini_nb15 = fin_nb15 + 1

      do k = ini_nb15, fin_nb15

        j = nb15_calc_listw(k,id+1)

        ! compute distance
        !
        dij(1:3) = rtmp(1:3) - coord(1:3,j)
        dij(1:3) = dij(1:3)-box_size(1:3)*anint(dij(1:3)/box_size(1:3))

        rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

        if (rij2 < cutoff2) then

          rij2  = rij2 * enefunc%table%density 
          L    = int(rij2)
          R    = rij2 - L
          lj12 = enefunc%nonb_lj12(atmcls(i),atmcls(j))
          lj6  = enefunc%nonb_lj6(atmcls(i),atmcls(j))

          h00  = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10  = R*(1.0_wp-R)*(1.0_wp-R)
          h01  = R*R*(3.0_wp-2.0_wp*R)
          h11  = R*R*(R-1.0_wp)

          if (.not. nonb_ene) then

            ! compute force only
            !
            L1 = 6*L-5
            term_lj12 = table_grad(L1  )*h00 + table_grad(L1+1)*h10
            term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
            term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
            term_lj12 = term_lj12 + table_grad(L1+6 )*h01 + table_grad(L1+7 )*h11
            term_lj6  = term_lj6  + table_grad(L1+8 )*h01 + table_grad(L1+9 )*h11
            term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
            grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(j)*term_elec

          else

            ! compute both energy and force
            !
            L1 = 6*L - 5
            term_lj12 = table_ene(L1  )*h00 + table_ene(L1+1)*h10
            term_lj6  = table_ene(L1+2)*h00 + table_ene(L1+3)*h10
            term_elec = table_ene(L1+4)*h00 + table_ene(L1+5)*h10
            term_lj12 = term_lj12 + table_ene(L1+6 )*h01 + table_ene(L1+7 )*h11
            term_lj6  = term_lj6  + table_ene(L1+8 )*h01 + table_ene(L1+9 )*h11
            term_elec = term_elec + table_ene(L1+10)*h01 + table_ene(L1+11)*h11
            evdw      = evdw  + term_lj12*lj12 - term_lj6*lj6
            eelec     = eelec + qtmp*charge(j)*term_elec

            term_lj12 = table_grad(L1  )*h00 + table_grad(L1+1)*h10
            term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
            term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
            term_lj12 = term_lj12 + table_grad(L1+6 )*h01 + table_grad(L1+7 )*h11
            term_lj6  = term_lj6  + table_grad(L1+8 )*h01 + table_grad(L1+9 )*h11
            term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
            grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(j)*term_elec

          end if

          work(1:3) = grad_coef*dij(1:3)

          ! store force
          !
          force_local(1:3) = force_local(1:3) - work(1:3)
          force(1:3,j) = force(1:3,j) + work(1:3)

          ! virial
          !
          do m = 1, 3
            virial(1:3,m) = virial(1:3,m) - dij(1:3)*work(m)
          end do

        end if
      end do

      force(1:3,i) = force(1:3,i) + force_local(1:3)

    end do
    !$omp end parallel

    return

  end subroutine compute_energy_nonbond_table_solute_cubic
 
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_table_water_cubic
  !> @brief        Calculate nonbonded water-water interaction with table
  !! @authors      JJ, TM
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    boundary : boundary information
  !! @param[in]    pairlist : pairlist information 
  !! @param[in]    coord    : coordinates of target systems
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : lennard-jones energy of target systems
  !! @note         J.Jung et al., J.Comput.Chem., 34, 2412-2420 (2013)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_table_water_cubic(enefunc, molecule,     &
                                        boundary, pairlist, nonb_ene, coord, &
                                        force, virial, eelec, evdw)

    ! formal arguments
    type(s_molecule), target, intent(in)    :: molecule
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_boundary), target, intent(in)    :: boundary
    type(s_pairlist), target, intent(in)    :: pairlist
    logical,                  intent(in)    :: nonb_ene
    real(wp),                 intent(in)    :: coord(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec
    real(wp),                 intent(inout) :: evdw

    ! local variables
    real(wp)                  :: rtmp(1:3,1:3)
    real(wp)                  :: force_local(1:3,1:3)
    real(wp)                  :: dij(1:3,1:3,1:3)
    real(wp)                  :: work(1:3,1:3,1:3)
    real(wp)                  :: rij2(1:3,1:3)
    real(wp)                  :: viri_local(1:3,1:3)
    real(wp)                  :: R, cutoff2
    real(wp)                  :: Delta_pbc(1:3), box_size(1:3)
    real(wp)                  :: grad_coef, term_lj, term_elec
    real(wp)                  :: h00, h01, h10, h11
    integer                   :: type(1:3,1:3), water_type
    integer                   :: i, j, k, L, L1
    integer                   :: i1(1:3), j1(1:3)
    integer                   :: ii, jj, ia, ja, m
    integer                   :: num_nb15, ini_nb15, fin_nb15
    integer                   :: id, my_id

    real(wp),         pointer :: cutoff
    real(wp),         pointer :: table_ene_WW(:,:), table_de_WW(:,:)
    integer,          pointer :: nwater
    integer,          pointer :: num_nb15_calc(:,:), nb15_calc_list(:,:)

#ifdef OMP
    integer                   :: omp_get_thread_num
#endif


    ! use pointers
    nwater         => enefunc%table%num_water
    cutoff         => enefunc%cutoffdist
    table_ene_WW   => enefunc%table%table_ene_WW
    table_de_WW    => enefunc%table%table_de_WW
    num_nb15_calc  => pairlist%table%num_nb15_calc_water
    nb15_calc_list => pairlist%table%nb15_calc_list_water
    box_size(1)    =  boundary%box_size_x
    box_size(2)    =  boundary%box_size_y
    box_size(3)    =  boundary%box_size_z

    cutoff2  = cutoff * cutoff
    num_nb15 = 0

    type(1,1)     = 1
    type(1,2:3)   = 2
    type(2:3,1)   = 2
    type(2:3,2:3) = 3

    ! calculate energy and gradient  water-water
    !
    !$omp parallel                                                            & 
    !$omp firstprivate(num_nb15)                                              &
    !$omp private(id, my_id, ini_nb15, fin_nb15, i, k, j, i1, j1, ia, ja,     &
    !$omp         ii, jj, dij, rij2, R, L, L1, term_lj, grad_coef, term_elec, &
    !$omp         rtmp,  force_local, viri_local, work, Delta_pbc,            &
    !$omp         water_type, m, h00, h01, h10, h11)                          &
    !$omp reduction(+:force) reduction(+:virial) reduction(+:eelec)           &
    !$omp reduction(+:evdw) 
    !
#ifdef OMP
    id    = omp_get_thread_num()
#else
    id    = 0
#endif
    my_id = my_city_rank * nthread + id

    do i = 1, nwater-1

      ini_nb15 = num_nb15 + 1
      fin_nb15 = num_nb15 + num_nb15_calc(i,id+1)
      num_nb15 = fin_nb15

      i1(1:3) = enefunc%table%water_list(1:3,i)

      do ii = 1, 3
        rtmp(1:3,ii) = coord(1:3,i1(ii))
      end do

      force_local(1:3,1:3) = 0.0_wp

      if (mod(i-1,nproc_city*nthread) /= my_id) ini_nb15 = fin_nb15 + 1

      do k = ini_nb15, fin_nb15

        j = nb15_calc_list(k,id+1)

        j1(1:3) = enefunc%table%water_list(1:3,j)

        ! compute distance
        !
        dij(1:3,1,1)   = rtmp(1:3,1) - coord(1:3,j1(1))
        Delta_pbc(1:3) = box_size(1:3)*anint(dij(1:3,1,1)/box_size(1:3))
        do ii = 1, 3
          do jj = 1, 3
            ja = j1(jj)
            dij(1:3,ii,jj) = rtmp(1:3,ii) - coord(1:3,ja) - Delta_pbc(1:3)
            rij2(ii,jj) = dij(1,ii,jj)*dij(1,ii,jj) + dij(2,ii,jj)*dij(2,ii,jj) &
                        + dij(3,ii,jj)*dij(3,ii,jj)
          end do
        end do

        viri_local(1:3,1:3) = 0.0_wp

        do ii = 1, 3
          do jj = 1, 3

            if (rij2(ii,jj) < cutoff2) then

              rij2(ii,jj) = rij2(ii,jj) * enefunc%table%density 

              L   = int(rij2(ii,jj))
              R   = rij2(ii,jj) - L
              water_type = type(ii,jj)
              h00 = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
              h10 = R*(1.0_wp-R)*(1.0_wp-R)
              h01 = R*R*(3.0_wp-2.0_wp*R)
              h11 = R*R*(R-1.0_wp)

              if (.not. nonb_ene) then
                ! compute force only
                !
                L  = 2*L - 1
                grad_coef = table_de_WW(L,   water_type)*h00   &
                          + table_de_WW(L+1, water_type)*h10   &
                          + table_de_WW(L+2, water_type)*h01   &
                          + table_de_WW(L+3, water_type)*h11
              else
                ! compute both energy and force
                !
                L1 = 6*L - 5
                term_lj   = table_ene_WW(L1,   water_type)*h00 &
                          + table_ene_WW(L1+1, water_type)*h10
                term_elec = table_ene_WW(L1+2, water_type)*h00 &
                          + table_ene_WW(L1+3, water_type)*h10
                grad_coef = table_ene_WW(L1+4, water_type)*h00 &
                          + table_ene_WW(L1+5, water_type)*h10
                term_lj   = term_lj                            &
                          + table_ene_WW(L1+6, water_type)*h01 &
                          + table_ene_WW(L1+7, water_type)*h11
                term_elec = term_elec                          &
                          + table_ene_WW(L1+8, water_type)*h01 &
                          + table_ene_WW(L1+9, water_type)*h11
                grad_coef = grad_coef                          &
                          + table_ene_WW(L1+10,water_type)*h01 &
                          + table_ene_WW(L1+11,water_type)*h11

                evdw  = evdw  + term_lj
                eelec = eelec + term_elec
              end if

              work(1:3,ii,jj) = grad_coef * dij(1:3,ii,jj)

              ! store force
              !
              force_local(1:3,ii) = force_local(1:3,ii) - work(1:3,ii,jj)
              ja = j1(jj)
              force(1:3,ja) = force(1:3,ja) + work(1:3,ii,jj)

              ! virial
              !
              do m = 1, 3
                viri_local(1:3,m) = viri_local(1:3,m)  &
                                  + dij(1:3,ii,jj)*work(m,ii,jj)
              end do

            end if

          end do
        end do

        do m = 1, 3
          virial(1:3,m) = virial(1:3,m) - viri_local(1:3,m)
        end do

      end do

      do ii = 1, 3
        ia = i1(ii)
        force(1:3,ia) = force(1:3,ia) + force_local(1:3,ii)
      end do

    end do
    !$omp end parallel

    return

  end subroutine compute_energy_nonbond_table_water_cubic

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pme_bond_corr_table
  !> @brief        Calculate PME bonded energy correction with lookup table
  !! @authors      TM
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    coord    : coordinates 
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_bond_corr_table(enefunc, molecule, coord, force, virial, eelec)

    ! formal arguments
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_molecule), target, intent(in)    :: molecule
    real(wp),                 intent(in)    :: coord(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec

    if (enefunc%table%table_order == 1) then

      call pme_bond_corr_table_linear(enefunc, molecule, coord, &
                                      force, virial, eelec)

    else if (enefunc%table%table_order == 3) then

      call pme_bond_corr_table_cubic(enefunc, molecule, coord,  &
                                     force, virial, eelec)

    end if

    return

  end subroutine pme_bond_corr_table

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pme_bond_corr_table_linear
  !> @brief        Calculate PME bonded energy correction with lookup table
  !! @authors      JJ, TM
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    r        : coordinates 
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_bond_corr_table_linear(enefunc, molecule, r, force, virial, &
                                        eelec)

    ! formal arguments
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_molecule), target, intent(in)    :: molecule
    real(wp),                 intent(in)    :: r(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec

    ! local variables
    real(wp)                  :: u, el_fact, alpha, alpha2m, alpha2sp
    real(wp)                  :: rij(3), rij2, qij, uij, fij(3)
    real(wp)                  :: vxx,vyx,vzx,vyy,vzy,vzz
    real(wp)                  :: Res, cutoff2
    integer                   :: i, j, k, jatm
    integer                   :: num_excl,ini_excl,fin_excl
    integer                   :: iatm, L

    real(wp),         pointer :: q(:), force_wk(:,:)
    real(wp),         pointer :: table_elec(:), table_delec(:)
    integer,          pointer :: natm, num_nonb_excl(:), nonb_excl_list(:)
    integer,          pointer :: i_list(:)

    ! use pointers
    !
    natm           => molecule%num_atoms
    q              => molecule%charge
    num_nonb_excl  => enefunc%num_nonb_excl
    nonb_excl_list => enefunc%nonb_excl_list
    table_elec     => enefunc%table%table_ecor
    table_delec    => enefunc%table%table_decor
    force_wk       => enefunc%pme%force_wk
    i_list         => enefunc%pme%i_list
    el_fact        =  ELECOEF/enefunc%dielec_const
    alpha          =  enefunc%pme%alpha
    alpha2m        =  enefunc%pme%alpha2m
    alpha2sp       =  enefunc%pme%alpha2sp
    cutoff2        =  enefunc%cutoffdist * enefunc%cutoffdist

    ! Initializing the energy and force 
    !
    u        = 0.0_wp
    num_excl = 0
    i_list(:)     = 0
    force_wk(:,:) = 0.0_wp

    ! Running only the 1-2 and 1-3 connected atoms
    !
    !$omp parallel default (shared)                                       &
    !$omp firstprivate(num_excl)                                          &
    !$omp private(i, ini_excl, fin_excl, j, jatm, rij2, k, rij, qij,      &
    !$omp         uij, fij, vxx, vyx, vzx, vyy, vzy, vzz, Res, L)         &
    !$omp shared(natm, alpha, alpha2sp, alpha2m, r, q, el_fact,           &
    !$omp        num_nonb_excl, nonb_excl_list, nproc_city, my_city_rank, &
    !$omp        i_list, force_wk)                                        &
    !$omp reduction(+:eelec) reduction(+:virial)
    !
    do i = 1, natm-1

      ini_excl = num_excl + 1
      fin_excl = num_excl + num_nonb_excl(i)
      num_excl = fin_excl

      if (mod(i-1,nproc_city) == my_city_rank) then

        !$omp do
        do j = ini_excl, fin_excl

          jatm      = nonb_excl_list(j)
          i_list(j) = i

          ! Calculating R_ij
          rij2 = 0.0_wp

          do k = 1, 3
            rij(k) = r(k,i) - r(k,jatm)
            rij2 = rij2 + rij(k)**2
          end do

          qij  = q(i) * q(jatm)

          rij2 = cutoff2 * enefunc%table%density /rij2
          L    = int(rij2)
          Res  = rij2 - L

          uij = table_elec(L) + Res*(table_elec(L+1)-table_elec(L))
          eelec = eelec + uij*qij

          ! Calculating F_ij and adding it to F_i (symmetry considered)
          !
          uij = table_delec(L) + Res*(table_delec(L+1)-table_delec(L))

          do k = 1, 3
            fij(k) = uij * rij(k) * qij
            force_wk(k,j) = fij(k)
          end do

          ! virial
          !
          vxx = rij(1) * fij(1)
          vyx = rij(2) * fij(1)
          vzx = rij(3) * fij(1)
          vyy = rij(2) * fij(2)
          vzy = rij(3) * fij(2)
          vzz = rij(3) * fij(3)
          virial(1,1) = virial(1,1) - vxx
          virial(2,1) = virial(2,1) - vyx
          virial(3,1) = virial(3,1) - vzx
          virial(1,2) = virial(1,2) - vyx
          virial(2,2) = virial(2,2) - vyy
          virial(3,2) = virial(3,2) - vzy
          virial(1,3) = virial(1,3) - vzx
          virial(2,3) = virial(2,3) - vzy
          virial(3,3) = virial(3,3) - vzz

        end do
        !$omp end do

      end if

    end do
    !$omp end parallel

    do j = 1, size(enefunc%nonb_excl_list(:))

      iatm = i_list(j)
      jatm = nonb_excl_list(j)

      if (iatm > 0) then
        do k = 1, 3
          force(k,iatm) = force(k,iatm) - force_wk(k,j)
          force(k,jatm) = force(k,jatm) + force_wk(k,j)
        end do
      end if

    end do

    return

  end subroutine pme_bond_corr_table_linear

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    pme_bond_corr_table_cubic
  !> @brief        Calculate PME bonded energy correction with lookup table
  !! @authors      JJ, TM
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    molecule : molecule information
  !! @param[in]    r        : coordinates 
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine pme_bond_corr_table_cubic(enefunc, molecule, r, force, virial, &
                                       eelec)

    ! formal arguments
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_molecule), target, intent(in)    :: molecule
    real(wp),                 intent(in)    :: r(:,:)
    real(wp),                 intent(inout) :: force(:,:)
    real(wp),                 intent(inout) :: virial(3,3)
    real(wp),                 intent(inout) :: eelec

    ! local variables
    real(wp)                  :: u, el_fact, alpha, alpha2m, alpha2sp
    real(wp)                  :: rij(3), rij2, qij, uij, fij(3)
    real(wp)                  :: vxx,vyx,vzx,vyy,vzy,vzz
    real(wp)                  :: Res, h00, h01, h10, h11
    integer                   :: i, j, k, jatm, density
    integer                   :: num_excl,ini_excl,fin_excl
    integer                   :: iatm, L

    real(wp),         pointer :: q(:), force_wk(:,:)
    real(wp),         pointer :: table_elec(:), table_delec(:)
    real(wp),         pointer :: dtable_elec(:), dtable_delec(:)
    integer,          pointer :: natm, num_nonb_excl(:), nonb_excl_list(:)
    integer,          pointer :: i_list(:)

    ! use pointers
    !
    natm           => molecule%num_atoms
    q              => molecule%charge
    num_nonb_excl  => enefunc%num_nonb_excl
    nonb_excl_list => enefunc%nonb_excl_list
    table_elec     => enefunc%table%table_ecor
    table_delec    => enefunc%table%table_decor
    dtable_elec    => enefunc%table%dtable_ecor
    dtable_delec   => enefunc%table%dtable_decor
    force_wk       => enefunc%pme%force_wk
    i_list         => enefunc%pme%i_list
    el_fact        =  ELECOEF/enefunc%dielec_const
    alpha          =  enefunc%pme%alpha
    alpha2m        =  enefunc%pme%alpha2m
    alpha2sp       =  enefunc%pme%alpha2sp
    density        =  enefunc%table%density

    ! Initializing the energy and force 
    !
    u        = 0.0_wp
    num_excl = 0
    i_list(:)     = 0
    force_wk(:,:) = 0.0_wp

    ! Running only the 1-2 and 1-3 connected atoms
    !
    !$omp parallel default (none)                                         &
    !$omp firstprivate(num_excl)                                          &
    !$omp private(i, ini_excl, fin_excl, j, jatm, rij2, k, rij, qij,      &
    !$omp         uij, fij, vxx, vyx, vzx, vyy, vzy, vzz,                 &
    !$omp         Res, h00, h01, h10, h11, L)                             &
    !$omp shared(natm, alpha, alpha2sp, alpha2m, r, q, el_fact,           &
    !$omp        num_nonb_excl, nonb_excl_list, nproc_city, my_city_rank, &
    !$omp        i_list, force_wk, table_elec, dtable_elec, table_delec,  &
    !$omp        dtable_delec, density)  &
    !$omp reduction(+:eelec) reduction(+:virial)
    !
    do i = 1, natm-1

      ini_excl = num_excl + 1
      fin_excl = num_excl + num_nonb_excl(i)
      num_excl = fin_excl

      if (mod(i-1,nproc_city) == my_city_rank) then

        !$omp do
        do j = ini_excl, fin_excl

          jatm      = nonb_excl_list(j)
          i_list(j) = i

          ! Calculating R_ij
          rij2 = 0.0_wp

          do k = 1, 3
            rij(k) = r(k,i) - r(k,jatm)
            rij2 = rij2 + rij(k)**2
          end do

          qij  = q(i) * q(jatm)

          rij2 = rij2 * density
          L    = int(rij2)
          Res  = rij2 - L
          h00  = (1.0_wp + 2.0_wp*Res)*(1.0_wp-Res)*(1.0_wp-Res)
          h10  = Res*(1.0_wp-Res)*(1.0_wp-Res)
          h01  = Res*Res*(3.0_wp-2.0_wp*Res)
          h11  = Res*Res*(Res-1.0_wp)

          uij = table_elec(L)*h00  + table_elec(L+1)*h01                    &
              + dtable_elec(L)*h10 + dtable_elec(L+1)*h11
          eelec = eelec + uij*qij

          ! Calculating F_ij and adding it to F_i (symmetry considered)
          !
          uij = table_delec(L)*h00  + table_delec(L+1)*h01                  &
              + dtable_delec(L)*h10 + dtable_delec(L+1)*h11

          do k = 1, 3
            fij(k) = uij * rij(k) * qij
            force_wk(k,j) = fij(k)
          end do

          ! virial
          !
          vxx = rij(1) * fij(1)
          vyx = rij(2) * fij(1)
          vzx = rij(3) * fij(1)
          vyy = rij(2) * fij(2)
          vzy = rij(3) * fij(2)
          vzz = rij(3) * fij(3)
          virial(1,1) = virial(1,1) - vxx
          virial(2,1) = virial(2,1) - vyx
          virial(3,1) = virial(3,1) - vzx
          virial(1,2) = virial(1,2) - vyx
          virial(2,2) = virial(2,2) - vyy
          virial(3,2) = virial(3,2) - vzy
          virial(1,3) = virial(1,3) - vzx
          virial(2,3) = virial(2,3) - vzy
          virial(3,3) = virial(3,3) - vzz

        end do
        !$omp end do

      end if

    end do
    !$omp end parallel

    do j = 1, size(enefunc%nonb_excl_list(:))

      iatm = i_list(j)
      jatm = nonb_excl_list(j)

      if (iatm > 0) then
        do k = 1, 3
          force(k,iatm) = force(k,iatm) - force_wk(k,j)
          force(k,jatm) = force(k,jatm) + force_wk(k,j)
        end do
      end if

    end do

    return

  end subroutine pme_bond_corr_table_cubic

end module at_energy_table_mod
