!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_minimize_mod
!> @brief   perform energy minimization
!! @authors Takaharu Mori (TM), Chigusa Kobayashi (CK), Norio Takase (NT)
!!          Yuji Sugita (YS)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_minimize_mod

  use at_output_mod
  use at_dynvars_mod
  use at_boundary_mod
  use at_pairlist_mod
  use at_energy_mod
  use at_output_str_mod
  use at_minimize_str_mod
  use at_dynvars_str_mod
  use at_boundary_str_mod
  use at_pairlist_str_mod
  use at_enefunc_str_mod
  use molecules_str_mod
  use fileio_rst_mod
  use timers_mod
  use fileio_control_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod

  implicit none
  private

  ! structures
  type, public :: s_min_info
    integer          :: method           = MinimizeMethodSD
    integer          :: nsteps           = 100
    integer          :: eneout_period    =  10
    integer          :: crdout_period    =   0
    integer          :: rstout_period    =   0
    integer          :: nbupdate_period  =  10
  end type s_min_info

  ! subroutines
  public  :: show_ctrl_minimize
  public  :: read_ctrl_minimize
  public  :: setup_minimize
  public  :: run_min
  public  :: steepest_descent

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_minimize
  !> @brief        show usage of MINIMIZE section
  !! @authors      NT
  !! @param[in]    show_all : show all usage or not
  !! @param[in]    run_mode : run mode string : "md", "min", "remd"
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_minimize(show_all, run_mode)

    ! formal arguments
    logical,                 intent(in)    :: show_all
    character(*),            intent(in)    :: run_mode


    if (show_all) then

      select case (run_mode)

      case ('min')

        write(MsgOut,'(A)') '[MINIMIZE]'
        write(MsgOut,'(A)') 'method        = SD        # [SD]'
        write(MsgOut,'(A)') 'nsteps        = 100       # number of minimization steps'
        write(MsgOut,'(A)') '# eneout_period = 10        # energy output period'
        write(MsgOut,'(A)') '# crdout_period = 0         # coordinates output period'
        write(MsgOut,'(A)') '# rstout_period = 0         # restart output period'
        write(MsgOut,'(A)') '# nbupdate_period = 10      # nonbond update period'
        write(MsgOut,'(A)') ' '


      end select

    else

      select case (run_mode)

      case ('min')

        write(MsgOut,'(A)') '[MINIMIZE]'
        write(MsgOut,'(A)') 'method        = SD        # [SD]'
        write(MsgOut,'(A)') 'nsteps        = 100       # number of minimization steps'
        write(MsgOut,'(A)') ' '

      end select


    end if

    return

  end subroutine show_ctrl_minimize
  
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_ctrl_minimize
  !> @brief        read MINIMIZE section in the control file
  !! @authors      TM
  !! @param[in]    handle   : unit number of control file
  !! @param[out]   min_info : MINIMIZE section in control parameters
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_ctrl_minimize(handle, min_info)

    ! parameters
    character(*),            parameter     :: Section = 'Minimize'

    ! formal arguments
    integer,                 intent(in)    :: handle
    type(s_min_info),        intent(inout) :: min_info


    ! read parameters from control file
    !
    call read_ctrlfile_type   (handle, Section, 'method',        &
                               min_info%method, MinimizeMethodTypes)
    call read_ctrlfile_integer(handle, Section, 'nsteps',        &
                               min_info%nsteps)
    call read_ctrlfile_integer(handle, Section, 'eneout_period', &
                               min_info%eneout_period)
    call read_ctrlfile_integer(handle, Section, 'crdout_period', &
                               min_info%crdout_period)
    call read_ctrlfile_integer(handle, Section, 'rstout_period', &
                               min_info%rstout_period)
    call read_ctrlfile_integer(handle, Section, 'nbupdate_period', &
                               min_info%nbupdate_period)


    ! write parameters to MsgOut
    !
    if (main_rank) then
      write(MsgOut,'(A)') 'Read_Ctrl_Minimize> Parameters of MIN'

      write(MsgOut,'(A20,A10,A20,I10)')                        &
            '  method          = ', MinimizeMethodTypes(min_info%method), &
            '  nsteps          = ', min_info%nsteps
      write(MsgOut,'(A20,I10,A20,I10)')                        &
            '  eneout_period   = ', min_info%eneout_period,    &
            '  crdout_period   = ', min_info%crdout_period
      write(MsgOut,'(A20,I10,A20,I10)')                        &
            '  rstout_period   = ', min_info%rstout_period,    &
            '  nbupdate_period = ', min_info%nbupdate_period
      write(MsgOut,'(A)') ' '
    end if


    ! error check
    !
    if (main_rank) then
      if (min_info%eneout_period > 0 .and.                        &
          mod(min_info%nsteps, min_info%eneout_period) /= 0) then
        write(MsgOut,'(A)') 'Read_Ctrl_Minimize> Error in eneout_period'
        write(MsgOut,'(A)') '  mod(nsteps, eneout_period) is not ZERO'
        call error_msg
      end if

      if (min_info%crdout_period > 0 .and.                        &
          mod(min_info%nsteps, min_info%crdout_period) /= 0) then
        write(MsgOut,'(A)') 'Read_Ctrl_Minimize> Error in crdout_period'
        write(MsgOut,'(A)') '  mod(nsteps, crdout_period) is not ZERO'
        call error_msg
      end if

      if (min_info%rstout_period > 0 .and.                        &
          mod(min_info%nsteps, min_info%rstout_period) /= 0) then
        write(MsgOut,'(A)') 'Read_Ctrl_Minimize> Error in rstout_period'
        write(MsgOut,'(A)') '  mod(nsteps, rstout_period) is not ZERO'
        call error_msg
      end if

      if (min_info%nbupdate_period > 0 .and.                      &
          mod(min_info%nsteps, min_info%nbupdate_period) /= 0) then
        write(MsgOut,'(A)') 'Read_Ctrl_Minimize> Error in nbupdate_period'
        write(MsgOut,'(A)') '  mod(nsteps, nbupdate_period) is not ZERO'
        call error_msg
      end if

    end if


    return

  end subroutine read_ctrl_minimize

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_minimize
  !> @brief        setup minimize information
  !> @authors      TM
  !! @param[in]    min_info : MINIMIZE section in control parameters
  !! @param[out]   minimize : minimize information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_minimize(min_info, out_info, minimize)

    ! formal arguments
    type(s_min_info),        intent(in)    :: min_info
    type(s_out_info),        intent(in)    :: out_info
    type(s_minimize),        intent(inout) :: minimize


    ! initialize variables in minimize
    !
    call init_minimize(minimize)


    ! setup minimize
    !
    minimize%method           = min_info%method
    minimize%nsteps           = min_info%nsteps
    minimize%eneout_period    = min_info%eneout_period
    minimize%crdout_period    = min_info%crdout_period
    minimize%rstout_period    = min_info%rstout_period
    minimize%nbupdate_period  = min_info%nbupdate_period

    if (out_info%dcdfile /= '' .and. minimize%crdout_period <= 0) &
      minimize%crdout_period = minimize%nsteps
    if (out_info%rstfile /= '' .and. minimize%rstout_period <= 0) &
      minimize%rstout_period = minimize%nsteps

    return

  end subroutine setup_minimize

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    run_min
  !> @brief        perform energy minimization
  !! @authors      YS
  !! @param[inout] output      : output information
  !! @param[inout] molecule    : molecule information
  !! @param[inout] enefunc     : potential energy functions
  !! @param[inout] dynvars     : dynamic variables
  !! @param[inout] minimize    : minimize information
  !! @param[inout] pairlist    : non-bond pair list
  !! @param[inout] boundary    : boundary condition
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine run_min(output, molecule, enefunc, dynvars, minimize, &
                     pairlist, boundary)

    ! formal arguments
    type(s_output),           intent(inout) :: output
    type(s_molecule), target, intent(inout) :: molecule
    type(s_enefunc),          intent(inout) :: enefunc
    type(s_dynvars),  target, intent(inout) :: dynvars
    type(s_minimize),         intent(inout) :: minimize
    type(s_pairlist),         intent(inout) :: pairlist
    type(s_boundary),         intent(inout) :: boundary

    ! Open output files
    !
    call open_output(output)

    ! MIN main loop
    !
    select case (minimize%method)

    case (MinimizeMethodSD)

      call steepest_descent (output, molecule, enefunc, dynvars, minimize, &
                             pairlist, boundary)

    end select


    return

  end subroutine run_min

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    steepest_descent
  !> @brief        steepest descent integrator
  !> @authors      TM, CK
  !! @param[inout] output   : output information
  !! @param[inout] molecule : molecule information
  !! @param[inout] enefunc  : potential energy functions information
  !! @param[inout] dynvars  : dynamic variables information
  !! @param[inout] minimize : minimize information
  !! @param[inout] pairlist : non-bond pair list information
  !! @param[inout] boundary : boundary conditions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine steepest_descent(output, molecule, enefunc, dynvars, minimize, &
                              pairlist, boundary)

    ! formal arguments
    type(s_output),           intent(inout) :: output
    type(s_molecule), target, intent(inout) :: molecule
    type(s_enefunc),          intent(inout) :: enefunc
    type(s_dynvars),  target, intent(inout) :: dynvars
    type(s_minimize),         intent(inout) :: minimize
    type(s_pairlist),         intent(inout) :: pairlist
    type(s_boundary),         intent(inout) :: boundary

    ! local variables
    real(wp)                  :: energy_ref, delta_energy
    real(wp)                  :: delta_r, delta_rini, delta_rmax
    real(wp)                  :: rmsg
    integer                   :: i, j, natom, nsteps

    real(wp),         pointer :: coord(:,:), coord_ref(:,:), force(:,:)


    ! use pointers
    !
    natom      =  molecule%num_atoms
    coord      => dynvars%coord
    coord_ref  => dynvars%coord_ref
    force      => dynvars%force
    nsteps     =  minimize%nsteps

    delta_rmax = 0.0001_wp
    delta_rini = 0.00005_wp
    delta_r    = delta_rini


    ! Compute energy of the initial structure
    !
    call compute_energy(molecule, enefunc, pairlist, boundary, .true., &
                        dynvars%coord,     &
                        dynvars%energy,    &
                        dynvars%temporary, &
                        dynvars%force,     &
                        dynvars%virial,    &
                        dynvars%virial_extern)

    rmsg = 0.0_wp
    do j = 1, natom
      rmsg = rmsg + force(1,j)**2 + force(2,j)**2 + force(3,j)**2
    end do
    rmsg = sqrt(rmsg/real(3*natom,wp))

    coord(1:3,1:natom) = coord(1:3,1:natom) + delta_r*force(1:3,1:natom)/rmsg

    dynvars%step = 0
    dynvars%total_pene = dynvars%energy%total

    call output_dynvars(output, enefunc, dynvars)


    ! Main loop of the Steepest descent method
    !
    do i = 1, nsteps

      dynvars%step = i

      ! save old energy and coordinates
      !
      coord_ref(1:3,1:natom) = coord(1:3,1:natom)
      energy_ref = dynvars%energy%total


      ! Compute energy and forces
      !
      call compute_energy(molecule, enefunc, pairlist, boundary,  &
                          mod(i, minimize%eneout_period) == 0, &
                          dynvars%coord,         &
                          dynvars%energy,        &
                          dynvars%temporary,     &
                          dynvars%force,         &
                          dynvars%virial,        &
                          dynvars%virial_extern)

      delta_energy = dynvars%energy%total - energy_ref
      if (delta_energy > 0) then
        delta_r = 0.5_wp*delta_r
      else
        delta_r = min(delta_rmax, 1.2_wp*delta_r)
      end if

      rmsg = 0.0_wp
      do j = 1, natom
        rmsg = rmsg + force(1,j)**2 + force(2,j)**2 + force(3,j)**2
      end do
      rmsg = sqrt(rmsg/real(3*natom,wp))
      dynvars%rms_gradient = rmsg

      coord(1:3,1:natom) = coord(1:3,1:natom) + delta_r*force(1:3,1:natom)/rmsg


      ! update nonbond pairlist
      !
      if (minimize%nbupdate_period > 0) then
        if (mod(i,minimize%nbupdate_period) == 0 .and. real_calc) then

          call update_pairlist(enefunc, boundary, coord, pairlist)

        end if
      end if


      ! output energy and dynamical variables
      !
     call output_min(output, molecule, enefunc, minimize, boundary, &
                     delta_r, dynvars)


    end do


    ! Close output files
    !
    call close_output(output)


    return

  end subroutine steepest_descent

end module at_minimize_mod
