!--------1---------2---------3---------4---------5---------6---------7---------8
! 
!> Program  ATDYN
!! @brief   Molecular Dynamics Simulation of Biomolecules using 
!!          Atom Decomposition Scheme
!! @authors Takaharu Mori (TM), Jaewoon Jung (JJ), Chigusa Kobayashi (CK),
!!          Yasuhiro Matsunaga (YM), Takashi Imai (TI), Takao Yoda (TY),
!!          Yuji Sugita (YS)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

program atdyn

  use at_setup_mpi_mod
  use at_remd_mod
  use at_minimize_mod
  use at_dynamics_mod
  use at_setup_atdyn_mod
  use at_control_mod
  use at_energy_mod
  use at_energy_pme_mod
  use at_minimize_str_mod
  use at_dynamics_str_mod
  use at_dynvars_str_mod
  use at_ensemble_str_mod
  use at_output_str_mod
  use at_remd_str_mod
  use at_constraints_str_mod
  use at_boundary_str_mod
  use at_pairlist_str_mod
  use at_enefunc_str_mod
  use at_energy_str_mod
  use molecules_str_mod
  use fileio_control_mod
  use timers_mod
  use string_mod
  use messages_mod
  use mpi_parallel_mod
#ifdef MPI
  use mpi
#endif

  implicit none

  integer             :: genesis_run_mode
  integer, parameter  :: GenesisMD    = 1
  integer, parameter  :: GenesisMIN   = 2
  integer, parameter  :: GenesisREMD  = 3

  ! local variables
  character(MaxFilename)      :: ctrl_filename
  type(s_ctrl_data)           :: ctrl_data
  type(s_molecule)            :: molecule
  type(s_enefunc)             :: enefunc
  type(s_dynvars)             :: dynvars
  type(s_pairlist)            :: pairlist
  type(s_boundary)            :: boundary
  type(s_constraints)         :: constraints
  type(s_ensemble)            :: ensemble
  type(s_dynamics)            :: dynamics
  type(s_minimize)            :: minimize
  type(s_remd)                :: remd
  type(s_output)              :: output
  integer                     :: omp_get_max_threads


#ifdef MPI
  call mpi_init(ierror)
  call mpi_comm_rank(mpi_comm_world, my_world_rank, ierror)
  call mpi_comm_size(mpi_comm_world, nproc_world,   ierror)
  main_rank = (my_world_rank == 0)
#else
  my_world_rank = 0
  nproc_world   = 1
  main_rank = .true.
#endif

#ifdef OMP
  nthread = omp_get_max_threads()
#else
  nthread = 1
#endif
 
 
  ! show usage
  !
  call usage(ctrl_filename)


  ! get run mode from control file
  !
  call get_genesis_mode(ctrl_filename, genesis_run_mode)


  ! run genesis
  !
  call atomic_decomposition_genesis(ctrl_filename, genesis_run_mode)


#ifdef MPI
  call mpi_finalize(ierror)
#endif

  stop

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    get_genesis_mode
  !> @brief        get genesis run mode
  !! @authors      TM
  !! @param[in]    ctrl_filename    : control file name
  !! @param[out]   genesis_run_mode : run MD, MIN, or REMD
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine get_genesis_mode(ctrl_filename, genesis_run_mode)

    ! formal arguments
    character(*),            intent(in)    :: ctrl_filename
    integer,                 intent(inout) :: genesis_run_mode


    if (find_ctrlfile_section(ctrl_filename, 'REMD')) then
      genesis_run_mode = GenesisREMD

    else if (find_ctrlfile_section(ctrl_filename, 'DYNAMICS')) then
      genesis_run_mode = GenesisMD

    else if (find_ctrlfile_section(ctrl_filename, 'MINIMIZE')) then
      genesis_run_mode = GenesisMIN

    else
      genesis_run_mode = GenesisMD

    end if

    return

  end subroutine get_genesis_mode

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    atomic_decomposition_genesis
  !> @brief        simulation/minimization using atomic decomposition scheme
  !! @authors      TM
  !! @param[in]    ctrl_filename : control file name
  !! @param[in]    genesis_mod   : run MD, MIN, or REMD
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine atomic_decomposition_genesis(ctrl_filename, genesis_run_mode)

    ! formal arguments
    character(*),            intent(in)    :: ctrl_filename
    integer,                 intent(in)    :: genesis_run_mode


    ! set timer
    !
    call timer(TimerTotal, TimerOn)


    ! [Step1] Read control file
    !
    if (main_rank) then
      write(MsgOut,'(A)') '[STEP1] Read Control Parameters'
      write(MsgOut,'(A)') ' '
    end if

    select case (genesis_run_mode)

    case (GenesisMD)

      call control_md  (ctrl_filename, ctrl_data)

    case (GenesisMIN)

      call control_min (ctrl_filename, ctrl_data)

    case (GenesisREMD)

      call control_remd(ctrl_filename, ctrl_data)

    end select


    ! [Step2] Setup MPI
    !
    if (main_rank) then
      write(MsgOut,'(A)') '[STEP2] Setup MPI'
      write(MsgOut,'(A)') ' '
    end if

    select case (genesis_run_mode)

    case (GenesisMD,GenesisMIN)

      call setup_mpi_md  (ctrl_data%ene_info)

    case (GenesisREMD)

      call setup_mpi_remd(ctrl_data%ene_info, ctrl_data%rep_info)
   
    end select


    ! [Step3] Set relevant variables and structures 
    !
    if (main_rank) then
      write(MsgOut,'(A)') '[STEP3] Set Relevant Variables and Structures'
      write(MsgOut,'(A)') ' '
    end if

    select case (genesis_run_mode)

    case (GenesisMD)

      call setup_atdyn_md  (ctrl_data, output, molecule, enefunc,     &
                            pairlist, dynvars, dynamics, constraints, &
                            ensemble, boundary)

    case (GenesisMIN)

      call setup_atdyn_min (ctrl_data, output, molecule, enefunc,     &
                            pairlist, dynvars, minimize, boundary)

    case (GenesisREMD)

      call setup_atdyn_remd(ctrl_data, output, molecule, enefunc,     &
                            pairlist, dynvars, dynamics, constraints, &
                            ensemble, boundary, remd)

    end select


    ! [Step4] Compute single point energy for molecules
    !
    if (main_rank) then
      write(MsgOut,'(A)') '[STEP4] Compute Single Point Energy for Molecules'
      write(MsgOut,'(A)') ' '
    end if

    call compute_energy(molecule, enefunc, pairlist, boundary, .true., &
                        dynvars%coord,     &
                        dynvars%energy,    &
                        dynvars%temporary, &
                        dynvars%force,     &
                        dynvars%virial,    &
                        dynvars%virial_extern)

    call output_energy (dynvars%step, enefunc, dynvars%energy)


    ! [Step5] Perform MD/REMD simulation or Energy minimization
    !

    call timer(TimerDynamics, TimerOn)

    select case (genesis_run_mode)

    case (GenesisMD)

      if (main_rank) then
        write(MsgOut,'(A)') '[STEP5] Perform Molecular Dynamics Simulation'
        write(MsgOut,'(A)') ' '
      end if

      call run_md(output, molecule, enefunc, dynvars, dynamics,    &
                  pairlist, boundary, constraints, ensemble)

    case (GenesisMIN)

      if (main_rank) then
        write(MsgOut,'(A)') '[STEP5] Perform Energy Minimization'
        write(MsgOut,'(A)') ' '
      end if

      call run_min(output, molecule, enefunc, dynvars, minimize,   &
                   pairlist, boundary)

    case (GenesisREMD)

      if (main_rank) then
        write(MsgOut,'(A)') '[STEP5] Perform Replica-Exchange MD Simulation'
        write(MsgOut,'(A)') ' '
      end if

      call run_remd(output, molecule, enefunc, dynvars, dynamics,  &
                    pairlist, boundary, constraints, ensemble, remd)

    end select

    call timer(TimerDynamics, TimerOff)


    ! [Step6] Deallocate arrays
    !
    if (main_rank) then
      write(MsgOut,'(A)') ' '
      write(MsgOut,'(A)') '[STEP6] Deallocate Arrays'
      write(MsgOut,'(A)') ' '
    end if

    call dealloc_remd_all       (remd)
    call dealloc_constraints_all(constraints)
    call dealloc_boundary_all   (boundary)
    call dealloc_pairlist_all   (pairlist)
    call dealloc_energy_all     (dynvars%energy)
    call dealloc_dynvars_all    (dynvars)
    call dealloc_enefunc_all    (enefunc)
    call dealloc_molecules_all  (molecule)


    call timer(TimerTotal, TimerOff)


    ! output process time
    !
    call output_time


    return

  end subroutine atomic_decomposition_genesis

end program atdyn
