!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   fft3d_mod
!> @brief   routines for fast fourier transform
!! @authors Jaewoon Jung (JJ)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module fft3d_mod

  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    fft3d
  !> @brief        calculate 3D fast fourier transform 
  !! @authors      JJ
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine fft3d(qdf, qdfxyz, qdfyxz, qdfzyx, qdfxyz_work,        &
                   qdfyxz_work, qdfzyx_work, work1, work2,          &
                   nlocalx, nlocaly, nlocalz, nlocalx1, x_local1,   &
                   nx, ny, nz, nprocx, nprocy, nprocz, id, nthread, &
                   my_x_rank, x_start1, x_end1, y_start, y_end,     &
                   z_start, z_end, grid_commx, grid_commy,          &
                   grid_commz)

#ifdef FFTW
    include 'fftw3.f'
#endif

    ! formal arguments
    real(wp),         intent(inout) :: qdf(nlocalx,nlocaly,nlocalz,*)
    complex(wp),      intent(inout) :: qdfxyz(nlocalx1,nlocaly,*)
    complex(wp),      intent(inout) :: qdfyxz(nlocaly,nlocalx1,*)
    complex(wp),      intent(inout) :: qdfzyx(nlocalz,nlocaly,*)
    complex(wp),      intent(inout) :: qdfxyz_work(nlocalx1,nlocaly,nlocalz,*)
    complex(wp),      intent(inout) :: qdfyxz_work(nlocaly,nlocalx1,nlocalz,*)
    complex(wp),      intent(inout) :: qdfzyx_work(nlocalz,nlocaly,nlocalx1,*)
    complex(wp),      intent(inout) :: work1(*), work2(*)
    integer,          intent(in)    :: nlocalx, nlocaly, nlocalz
    integer,          intent(in)    :: nlocalx1
    integer,          intent(in)    :: x_local1
    integer,          intent(in)    :: nx, ny, nz
    integer,          intent(in)    :: nprocx, nprocy, nprocz
    integer,          intent(in)    :: id, nthread
    integer,          intent(in)    :: my_x_rank
    integer,          intent(in)    :: x_start1, x_end1
    integer,          intent(in)    :: y_start, y_end
    integer,          intent(in)    :: z_start, z_end
    integer,          intent(in)    :: grid_commx, grid_commy, grid_commz

    ! local variables
    integer*8         :: plan
    integer           :: ix, iy, iz, ixs, iys
    integer           :: i, k, iproc, ierror


    ! x direction
    !

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_plan_dft_1d(plan,nx,work1,work2,FFTW_FORWARD,FFTW_MEASURE)
#else
    call dfftw_plan_dft_1d(plan,nx,work1,work2,FFTW_FORWARD,FFTW_MEASURE)
#endif

#else
    call zfft1d(work1, nx, 0, work2)
#endif

    do iz = id+1, nlocalz, nthread
      do iy = 1, nlocaly, 2

        do iproc = 1, nprocx
          k = (iproc-1)*nlocalx
          do i = 1, nlocalx, 16
!dir$ vector aligned
            do ix = i, min0(i+15,nlocalx)
              work1(k+ix) = dcmplx(qdf(ix,iy,iz,iproc), &
                                   qdf(ix,iy+1,iz,iproc)) 
            end do
          end do
        end do

#ifdef FFTW

#ifdef _SINGLE
        call sfftw_execute_dft(plan, work1, work2)
#else
        call dfftw_execute_dft(plan, work1, work2)
#endif

        if (my_x_rank == 0) then
          qdfyxz(iy,1,iz) = real(work2(1),wp)
          qdfyxz(iy+1,1,iz) = imag(work2(1))
          do ix = 2, x_local1
            qdfyxz(iy,ix,iz) = 0.5_wp                 &
                             *(work2(ix)+conjg(work2(nx-ix+2)))
            qdfyxz(iy+1,ix,iz) = (0.0_wp,-0.5_wp)      &
                             *(work2(ix)-conjg(work2(nx-ix+2)))
          end do
        else
          do ix = x_start1, x_end1
            ixs = ix - x_start1 + 1
            qdfyxz(iy,ixs,iz) = 0.5_wp                &
                             *(work2(ix)+conjg(work2(nx-ix+2)))
            qdfyxz(iy+1,ixs,iz) = (0.0_wp,-0.5_wp)     &
                             *(work2(ix)-conjg(work2(nx-ix+2)))
          end do
        end if
#else
        call zfft1d(work1, nx, -1, work2)

        if (my_x_rank == 0) then
          qdfyxz(iy,1,iz) = real(work1(1),wp)
          qdfyxz(iy+1,1,iz) = imag(work1(1))
          do ix = 2, x_local1
            qdfyxz(iy,ix,iz) = 0.5_wp                 &
                             *(work1(ix)+conjg(work1(nx-ix+2)))
            qdfyxz(iy+1,ix,iz) = (0.0_wp,-0.5_wp)      &
                             *(work1(ix)-conjg(work1(nx-ix+2)))
          end do
        else
          do ix = x_start1, x_end1
            ixs = ix - x_start1 + 1
            qdfyxz(iy,ixs,iz) = 0.5_wp                &
                             *(work1(ix)+conjg(work1(nx-ix+2)))
            qdfyxz(iy+1,ixs,iz) = (0.0_wp,-0.5_wp)     &
                             *(work1(ix)-conjg(work1(nx-ix+2)))
          end do
        end if
#endif
      end do
    end do

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_destroy_plan(plan)
#else
    call dfftw_destroy_plan(plan)
#endif

#endif

    !$omp barrier
    !$omp master

#ifdef MPI
    call mpi_allgather(qdfyxz, nlocalx1*nlocaly*nlocalz,         &
                       mpi_wp_complex, qdfyxz_work,              &
                       nlocalx1*nlocaly*nlocalz, mpi_wp_complex, &
                       grid_commy, ierror)
#else
    qdfyxz_work(1:nlocaly,1:nlocalx1,1:nlocalz,1)  &
     = qdfyxz(1:nlocaly,1:nlocalx1,1:nlocalz)
#endif

    !$omp end master
    !$omp barrier

    ! y direction
    !
#ifdef FFTW

#ifdef _SINGLE
    call sfftw_plan_dft_1d(plan,ny,work1,work2,FFTW_FORWARD,FFTW_MEASURE)
#else
    call dfftw_plan_dft_1d(plan,ny,work1,work2,FFTW_FORWARD,FFTW_MEASURE)
#endif

#else
    call zfft1d(work1, ny, 0, work2)
#endif

    do iz = id+1, nlocalz, nthread
      do ix = 1, x_local1

        do iproc = 1, nprocy
          k = (iproc-1)*nlocaly
          do i = 1, nlocaly, 16
            do iy = i, min0(i+15,nlocaly)
!dir$ vector aligned
              work1(k+iy) = qdfyxz_work(iy,ix,iz,iproc)
            end do
          end do
        end do

#ifdef FFTW

#ifdef _SINGLE
        call sfftw_execute_dft(plan, work1, work2)
#else
        call dfftw_execute_dft(plan, work1, work2)
#endif

        do i = y_start, y_end, 16
          do iy = i, min0(i+15,y_end)
            iys = iy - y_start + 1
            qdfzyx(iz,iys,ix) = work2(iy)
          end do
        end do
#else
        call zfft1d(work1, ny, -1, work2)

        do i = y_start, y_end, 16
          do iy = i, min0(i+15,y_end)
            iys = iy - y_start + 1
            qdfzyx(iz,iys,ix) = work1(iy)
          end do
        end do
#endif
      end do
    end do

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_destroy_plan(plan)
#else
    call dfftw_destroy_plan(plan)
#endif

#endif

    !$omp barrier
    !$omp master

#ifdef MPI
    call mpi_allgather(qdfzyx, nlocalx1*nlocaly*nlocalz,         &
                       mpi_wp_complex, qdfzyx_work,              &
                       nlocalx1*nlocaly*nlocalz, mpi_wp_complex, &
                       grid_commz, ierror)
#else
    qdfzyx_work(1:nlocalz,1:nlocaly,1:nlocalx1,1)  &
     = qdfzyx(1:nlocalz,1:nlocaly,1:nlocalx1)
#endif

    !$omp end master
    !$omp barrier


    ! z direction
    !
#ifdef FFTW

#ifdef _SINGLE
    call sfftw_plan_dft_1d(plan,nz,work1,work2,FFTW_FORWARD,FFTW_MEASURE)
#else
    call dfftw_plan_dft_1d(plan,nz,work1,work2,FFTW_FORWARD,FFTW_MEASURE)
#endif

#else
    call zfft1d(work1, nz, 0, work2)
#endif

    do ix = 1, x_local1
      do iy = id+1, nlocaly, nthread

        do iproc = 1, nprocz
          k = (iproc-1)*nlocalz
          do i = 1, nlocalz, 16
!dir$ vector aligned
            do iz = 1, min0(i+15,nlocalz)
              work1(k+iz) = qdfzyx_work(iz,iy,ix,iproc)
            end do
          end do
        end do

#ifdef FFTW

#ifdef _SINGLE
        call sfftw_execute_dft(plan, work1, work2)
#else
        call dfftw_execute_dft(plan, work1, work2)
#endif

        do iproc = 1, nprocz
          k = (iproc-1)*nlocalz
          qdfzyx_work(1:nlocalz,iy,ix,iproc) = work2(k+1:k+nlocalz)
        end do
#else
        call zfft1d(work1, nz, -1, work2)

        do iproc = 1, nprocz
          k = (iproc-1)*nlocalz
          qdfzyx_work(1:nlocalz,iy,ix,iproc) = work1(k+1:k+nlocalz)
        end do
#endif
      end do
    end do

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_destroy_plan(plan)
#else
    call dfftw_destroy_plan(plan)
#endif

#endif

    return

  end subroutine fft3d

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    bfft3d
  !> @brief        calculate inverse 3D fast fourier transform 
  !! @authors      JJ
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine bfft3d(qdf, qdf_real, qdfxyz, qdfyxz, qdfzyx,        &
                    qdfxyz_work, qdfyxz_work, qdfzyx_work, work1, &
                    work2, nlocalx, nlocaly, nlocalz, nlocalx1,   &
                    x_local1, nx, ny, nz, nprocx, nprocy, nprocz, &
                    id, nthread, my_x_rank, x_start1, x_end1,     &
                    y_start, y_end, z_start, z_end, grid_commx,   &
                    grid_commy, grid_commz)

#ifdef FFTW
    include 'fftw3.f'
#endif

    ! formal arguments
    real(wp),         intent(inout) :: qdf(nlocalx,nlocaly,nlocalz,*)
    real(wp),         intent(inout) :: qdf_real(*)
    complex(wp),      intent(inout) :: qdfxyz(nlocalx1,nlocaly,*)
    complex(wp),      intent(inout) :: qdfyxz(nlocaly,nlocalx1,*)
    complex(wp),      intent(inout) :: qdfzyx(nlocalz,nlocaly,*)
    complex(wp),      intent(inout) :: qdfxyz_work(nlocalx1,nlocaly,nlocalz,*)
    complex(wp),      intent(inout) :: qdfyxz_work(nlocaly,nlocalx1,nlocalz,*)
    complex(wp),      intent(inout) :: qdfzyx_work(nlocalz,nlocaly,nlocalx1,*)
    complex(wp),      intent(inout) :: work1(*), work2(*)
    integer,          intent(in)    :: nlocalx, nlocaly, nlocalz
    integer,          intent(in)    :: nlocalx1
    integer,          intent(in)    :: x_local1
    integer,          intent(in)    :: nx, ny, nz
    integer,          intent(in)    :: nprocx, nprocy, nprocz
    integer,          intent(in)    :: id, nthread
    integer,          intent(in)    :: my_x_rank
    integer,          intent(in)    :: x_start1, x_end1
    integer,          intent(in)    :: y_start, y_end
    integer,          intent(in)    :: z_start, z_end
    integer,          intent(in)    :: grid_commx, grid_commy, grid_commz

    ! local variables
    real(wp)          :: work3(nx)
    integer*8         :: plan
    integer           :: ix, iy, iz, iys, izs
    integer           :: i, k, k1, iproc, ierror


    ! z direction
    !
#ifdef FFTW

#ifdef _SINGLE
    call sfftw_plan_dft_1d(plan,nz,work1,work2,FFTW_BACKWARD,FFTW_MEASURE)
#else
    call dfftw_plan_dft_1d(plan,nz,work1,work2,FFTW_BACKWARD,FFTW_MEASURE)
#endif

#else
    call zfft1d(work1, nz, 0, work2)
#endif

    do ix = 1, x_local1
      do iy = id+1, nlocaly, nthread

        do iproc = 1, nprocz
          k = (iproc-1)*nlocalz
          do i = 1, nlocalz, 16
!dir$ vector aligned
            do iz = i, min0(i+15,nlocalz)
              work1(k+iz) = qdfzyx_work(iz,iy,ix,iproc)
            end do
          end do
        end do

#ifdef FFTW

#ifdef _SINGLE
        call sfftw_execute_dft(plan, work1, work2)
#else
        call dfftw_execute_dft(plan, work1, work2)
#endif

        do iz = z_start, z_end
          izs = iz - z_start + 1
          qdfzyx(izs,iy,ix) = work2(iz)
        end do
#else
        call zfft1d(work1, nz, 1, work2)

        do iz = z_start, z_end
          izs = iz - z_start + 1
          qdfzyx(izs,iy,ix) = work1(iz)
        end do
#endif
      end do
    end do

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_destroy_plan(plan)
#else
    call dfftw_destroy_plan(plan)
#endif

#endif

    !$omp barrier
    !$omp master
#ifdef MPI
    call mpi_allgather(qdfzyx, nlocalx1*nlocaly*nlocalz,         &
                       mpi_wp_complex, qdfzyx_work,              &
                       nlocalx1*nlocaly*nlocalz, mpi_wp_complex, &
                       grid_commy, ierror)
#else
    qdfzyx_work(1:nlocalz,1:nlocaly,1:nlocalx1,1)  &
     = qdfzyx(1:nlocalz,1:nlocaly,1:nlocalx1)
#endif
    !$omp end master
    !$omp barrier


    ! y direction
    !

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_plan_dft_1d(plan,ny,work1,work2,FFTW_BACKWARD,FFTW_MEASURE)
#else
    call dfftw_plan_dft_1d(plan,ny,work1,work2,FFTW_BACKWARD,FFTW_MEASURE)
#endif

#else
    call zfft1d(work1, ny, 0, work2)
#endif

    do iz = id+1, nlocalz, nthread
      do ix = 1, x_local1

        do iproc = 1, nprocy
          k = (iproc-1)*nlocaly
          do i = 1, nlocaly, 16
!dir$ vector aligned
            do iy = i, min0(i+15,nlocaly)
              work1(k+iy) = qdfzyx_work(iz,iy,ix,iproc)
            end do
          end do
        end do

#ifdef FFTW

#ifdef _SINGLE
        call sfftw_execute_dft(plan, work1, work2)
#else
        call dfftw_execute_dft(plan, work1, work2)
#endif

        do iy = y_start, y_end
          iys = iy - y_start + 1
          qdfyxz(iys,ix,iz) = work2(iy)
        end do
#else
        call zfft1d(work1, ny, 1, work2)

        do iy = y_start, y_end
          iys = iy - y_start + 1
          qdfyxz(iys,ix,iz) = work1(iy)
        end do
#endif
      end do
    end do

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_destroy_plan(plan)
#else
    call dfftw_destroy_plan(plan)
#endif

#endif

    !$omp barrier
    !$omp master
#ifdef MPI
    call mpi_allgather(qdfyxz, nlocalx1*nlocaly*nlocalz,         &
                       mpi_wp_complex, qdfyxz_work,              &
                       nlocalx1*nlocaly*nlocalz, mpi_wp_complex, &
                       grid_commx, ierror)
#else
    qdfyxz_work(1:nlocaly,1:nlocalx1,1:nlocalz,1)  &
     = qdfyxz(1:nlocaly,1:nlocalx1,1:nlocalz)
#endif
    !$omp end master
    !$omp barrier


    ! x direction
    !

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_plan_dft_c2r_1d(plan, nx, work1, work3, FFTW_MEASURE)
#else
    call dfftw_plan_dft_c2r_1d(plan, nx, work1, work3, FFTW_MEASURE)
#endif

#else
    call zfft1d(work1, nx, 0, work2)
#endif

    do iz = id+1, nlocalz, nthread
      do iy = 1, nlocaly

        do i = 1, nlocalx1, 16
!dir$ vector aligned
          do ix = i, min0(i+15,nlocalx1)
            work1(ix) = qdfyxz_work(iy,ix,iz,1)
          end do
        end do

        do iproc = 2, nprocx
          k = nlocalx1 + (iproc-2)*(nlocalx1-1)
          do i = 1, nlocalx1-1, 16
!dir$ vector aligned
            do ix = 1, min0(i+15,nlocalx1-1)
              work1(k+ix) = qdfyxz_work(iy,ix,iz,iproc)
            end do
          end do
        end do

#ifdef FFTW

#ifdef _SINGLE
        call sfftw_execute_dft_c2r(plan, work1, work3)
#else
        call dfftw_execute_dft_c2r(plan, work1, work3)
#endif

        k = (iy-1)*nlocalx + (iz-1)*nlocalx*nlocaly
        k1 = my_x_rank*nlocalx
        qdf_real(k+1:k+nlocalx) = work3(k1+1:k1+nlocalx)
#else 
        call zfft1d(work1, nx, 2, work2)

        k = (iy-1)*nlocalx + (iz-1)*nlocalx*nlocaly
        k1 = my_x_rank*nlocalx
        qdf_real(k+1:k+nlocalx) = work1(k1+1:k1+nlocalx)
#endif
      end do
    end do

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_destroy_plan(plan)
#else
    call dfftw_destroy_plan(plan)
#endif

#endif

    return

  end subroutine bfft3d

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    fft3d_tuned
  !> @brief        calculate 3D fast fourier transform 
  !! @authors      JJ
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine fft3d_tuned(qdf, qdfxyz, qdfyxz, qdfzyx, qdfzxy, qdfyxz2,       &
                         qdfxyz_work, qdfyxz_work, qdfzyx_work, qdfzxy_work, &
                         qdfzxy_work2, qdfzxy_work3, work1, work2,           &
                         nlocalx, nlocaly, nlocalz, nlocalx1, x_local1,      &
                         nx, ny, nz, nprocx, nprocy, nprocz, id, nthread,    &
                         my_x_rank, my_y_rank, my_z_rank,                    &
                         x_start1, x_end1, y_start, y_end,                   &
                         z_start, z_end, grid_commx, grid_commy,             &
                         grid_commz)

#ifdef FFTW
    include 'fftw3.f'
#endif

    ! formal arguments
    real(wp),         intent(inout) :: qdf(nlocalx,nlocaly,nlocalz,*)
    complex(wp),      intent(inout) :: qdfxyz(nlocalx1,nlocaly,*)
    complex(wp),      intent(inout) :: qdfyxz(nlocaly,nlocalx1,*)
    complex(wp),      intent(inout) :: qdfzyx(nlocalz,nlocaly,*)
    complex(wp),      intent(inout) :: qdfzxy(nlocalz,nlocalx1,*)
    complex(wp),      intent(inout) :: qdfyxz2(nlocaly,nlocalx1,*)
    complex(wp),      intent(inout) :: qdfxyz_work(nlocalx1,nlocaly,nlocalz,*)
    complex(wp),      intent(inout) :: qdfyxz_work(nlocaly,nlocalx1,nlocalz,*)
    complex(wp),      intent(inout) :: qdfzyx_work(nlocalz,nlocaly,nlocalx1,*)
    complex(wp),      intent(inout) :: qdfzxy_work(nlocalz,nlocalx1,nlocaly,*)
    complex(wp),      intent(inout) :: qdfzxy_work2(nz,nlocalx1,nlocaly)
    complex(wp),      intent(inout) :: qdfzxy_work3(nz,nlocalx1,nlocaly)
    complex(wp),      intent(inout) :: work1(*), work2(*)
    integer,          intent(in)    :: nlocalx, nlocaly, nlocalz
    integer,          intent(in)    :: nlocalx1
    integer,          intent(in)    :: x_local1
    integer,          intent(in)    :: nx, ny, nz
    integer,          intent(in)    :: nprocx, nprocy, nprocz
    integer,          intent(in)    :: id, nthread
    integer,          intent(in)    :: my_x_rank, my_y_rank, my_z_rank
    integer,          intent(in)    :: x_start1, x_end1
    integer,          intent(in)    :: y_start, y_end
    integer,          intent(in)    :: z_start, z_end
    integer,          intent(in)    :: grid_commx, grid_commy, grid_commz

    ! local variables
    integer*8         :: plan
    integer           :: ix, iy, iz, ixs, ixe, iys, izs, niz, niy, izss, iyss
    integer           :: ix_start1, ix_end1, iy_start, iy_end, iz_start, iz_end
    integer           :: i, k, k1, iproc, ierror


    ! x direction
    !
#ifdef FFTW

#ifdef _SINGLE
    call sfftw_plan_dft_1d(plan,nx,work1,work2,FFTW_FORWARD,FFTW_MEASURE)
#else
    call dfftw_plan_dft_1d(plan,nx,work1,work2,FFTW_FORWARD,FFTW_MEASURE)
#endif

#else
    call zfft1d(work1, nx, 0, work2)
#endif

    niz      = nlocalz/nprocx
    iz_start = niz * my_x_rank + 1
    iz_end   = iz_start + niz - 1

    do iz = iz_start+id, iz_end, nthread
      izs = iz - iz_start + 1
      do iy = 1, nlocaly, 2

        do iproc = 1, nprocx
          k = (iproc-1)*nlocalx
          izss = (iproc-1)*niz + izs
          do ix = 1, nlocalx
            work1(k+ix) = dcmplx(qdf(ix,iy,izss,1), &
                                 qdf(ix,iy+1,izss,1)) 
          end do
        end do
#ifdef FFTW

#ifdef _SINGLE
        call sfftw_execute_dft(plan, work1, work2)
#else
        call dfftw_execute_dft(plan, work1, work2)
#endif
        ixe = nlocalx/2 + 1
        qdfyxz_work(iy,1,izs,1) = real(work2(1),wp)
        qdfyxz_work(iy+1,1,izs,1) = imag(work2(1))

        do ix = 2, ixe
          qdfyxz_work(iy,ix,izs,1)   = 0.5_wp          &
                             *(work2(ix)+conjg(work2(nx-ix+2)))
          qdfyxz_work(iy+1,ix,izs,1) = (0.0_wp,-0.5_wp) &
                             *(work2(ix)-conjg(work2(nx-ix+2)))
        end do

        do iproc = 2, nprocx

          ix_start1 = (iproc-1)*nlocalx/2 + 2
          ix_end1   = ix_start1 + nlocalx/2 -1
          izss = (iproc-1)*niz + izs

          do ix = ix_start1, ix_end1
            ixs = ix - ix_start1 + 1
            qdfyxz_work(iy,ixs,izss,1) = 0.5_wp            &
                             *(work2(ix)+conjg(work2(nx-ix+2)))
            qdfyxz_work(iy+1,ixs,izss,1) = (0.0_wp,-0.5_wp) &
                             *(work2(ix)-conjg(work2(nx-ix+2)))
          end do
        end do

#else
        call zfft1d(work1, nx, -1, work2)

        ixe = nlocalx/2 + 1
        qdfyxz_work(iy,1,izs,1) = real(work1(1),wp)
        qdfyxz_work(iy+1,1,izs,1) = imag(work1(1))

        do ix = 2, ixe
          qdfyxz_work(iy,ix,izs,1)   = 0.5_wp          &
                             *(work1(ix)+conjg(work1(nx-ix+2)))
          qdfyxz_work(iy+1,ix,izs,1) = (0.0_wp,-0.5_wp) &
                             *(work1(ix)-conjg(work1(nx-ix+2)))
        end do

        do iproc = 2, nprocx 

          ix_start1 = (iproc-1)*nlocalx/2 + 2
          ix_end1   = ix_start1 + nlocalx/2 -1
          izss = (iproc-1)*niz + izs 

          do ix = ix_start1, ix_end1
            ixs = ix - ix_start1 + 1
            qdfyxz_work(iy,ixs,izss,1) = 0.5_wp            &
                             *(work1(ix)+conjg(work1(nx-ix+2)))
            qdfyxz_work(iy+1,ixs,izss,1) = (0.0_wp,-0.5_wp) &
                             *(work1(ix)-conjg(work1(nx-ix+2)))
          end do
        end do
#endif
      end do
    end do

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_destroy_plan(plan)
#else
    call dfftw_destroy_plan(plan)
#endif

#endif

    !$omp barrier
    !$omp master

#ifdef MPI
    call mpi_alltoall(qdfyxz_work(1,1,1,1), &
                      nlocalx1*nlocaly*niz, mpi_wp_complex, &
                      qdfyxz, &
                      nlocalx1*nlocaly*niz, mpi_wp_complex, &
                      grid_commx, ierror)
#endif

#ifdef MPI
    call mpi_alltoall(qdfyxz, &
                      nlocalx1*nlocaly*nlocalz/nprocy, mpi_wp_complex, &
                      qdfyxz_work, &
                      nlocalx1*nlocaly*nlocalz/nprocy, mpi_wp_complex, &
                      grid_commy, ierror)
#endif

    !$omp end master
    !$omp barrier


    ! y direction
    !
!$omp master
!$omp end master

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_plan_dft_1d(plan,ny,work1,work2,FFTW_FORWARD,FFTW_MEASURE)
#else
    call dfftw_plan_dft_1d(plan,ny,work1,work2,FFTW_FORWARD,FFTW_MEASURE)
#endif

#else
    call zfft1d(work1, ny, 0, work2)
#endif

    niz      = nlocalz/nprocy
    iz_start = niz * my_y_rank + 1
    iz_end   = iz_start + niz - 1

    do iz = iz_start+id, iz_end, nthread
      izs = iz - iz_start + 1
      do ix = 1, x_local1

        do iproc = 1, nprocy
          k = (iproc-1)*nlocaly
          izss = (iproc-1)*niz + izs
          do iy = 1, nlocaly
            work1(k+iy) = qdfyxz_work(iy,ix,izss,1)
          end do
        end do

#ifdef FFTW

#ifdef _SINGLE
        call sfftw_execute_dft(plan, work1, work2)
#else
        call dfftw_execute_dft(plan, work1, work2)
#endif

        do iproc = 1, nprocy
          k = (iproc-1)*nlocaly
          izss = (iproc-1)*niz + izs
          do iy = 1, nlocaly
            qdfyxz2(iy,ix,izss) = work2(k+iy)
          end do
        end do
#else
        call zfft1d(work1, ny, -1, work2)

        do iproc = 1, nprocy
          k = (iproc-1)*nlocaly
          izss = (iproc-1)*niz + izs
          do iy = 1, nlocaly
            qdfyxz2(iy,ix,izss) = work1(k+iy)
          end do
        end do
#endif
      end do
    end do

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_destroy_plan(plan)
#else
    call dfftw_destroy_plan(plan)
#endif

#endif

    !$omp barrier
    !$omp master

#ifdef MPI
    call mpi_alltoall(qdfyxz2, &
                      nlocaly*nlocalx1*niz, mpi_wp_complex, &
                      qdfyxz,  &
                      nlocaly*nlocalx1*niz, mpi_wp_complex, &
                      grid_commy, ierror)
#endif

    !$omp end master
    !$omp barrier

    !$omp do
    do iz = 1, nlocalz
      do ix = 1, nlocalx1
        do iy = 1, nlocaly
          qdfzxy(iz,ix,iy) = qdfyxz(iy,ix,iz)
        end do
      end do
    end do

    !$omp master

#ifdef MPI
    call mpi_alltoall(qdfzxy, &
                      nlocalx1*nlocaly*nlocalz/nprocz,mpi_wp_complex, &
                      qdfzxy_work, &
                      nlocalx1*nlocaly*nlocalz/nprocz,mpi_wp_complex, &
                      grid_commz, ierror)
#endif
    !$omp end master
    !$omp barrier


    ! z direction
    !

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_plan_dft_1d(plan,nz,work1,work2,FFTW_FORWARD,FFTW_MEASURE)
#else
    call dfftw_plan_dft_1d(plan,nz,work1,work2,FFTW_FORWARD,FFTW_MEASURE)
#endif

#else
    call zfft1d(work1, nz, 0, work2)
#endif

    niy      = nlocaly/nprocz
    iy_start = niy * my_z_rank + 1
    iy_end   = iy_start + niy - 1

    do ix = 1, x_local1
      do iy = iy_start+id, iy_end, nthread
        iys = iy - iy_start +1

        do iproc = 1, nprocz
          k = (iproc-1)*nlocalz
          iyss = (iproc-1)*niy + iys
          do iz = 1, nlocalz
            qdfzxy_work2(k+iz,ix,iy) = qdfzxy_work(iz,ix,iyss,1)
          end do
        end do
#ifdef FFTW

#ifdef _SINGLE
        call sfftw_execute_dft(plan, qdfzxy_work2(1,ix,iy), work2)
#else
        call dfftw_execute_dft(plan, qdfzxy_work2(1,ix,iy), work2)
#endif 
        do k = 1, nz
          qdfzxy_work2(k,ix,iy) = work2(k)
        end do
#else
        call zfft1d(qdfzxy_work2(1,ix,iy), nz, -1, work2)
#endif
      end do
    end do

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_destroy_plan(plan)
#else
    call dfftw_destroy_plan(plan)
#endif

#endif

    !$omp barrier
    !$omp master

#ifdef MPI
    call mpi_allgather(qdfzxy_work2(1,1,iy_start), nz*nlocalx1*niy, &
                       mpi_wp_complex, qdfzxy_work3,    &
                       nz*nlocalx1*niy, mpi_wp_complex, &
                       grid_commz, ierror)
#endif

    !$omp end master
    !$omp barrier

    return

  end subroutine fft3d_tuned

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    bfft3d_tuned
  !> @brief        calculate inverse 3D fast fourier transform 
  !! @authors      JJ
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine bfft3d_tuned(qdf, qdf_real, qdfzxy, qdfyxz, qdfzyx,        &
                          qdfzxy_work, qdfyxz_work, qdfzyx_work, work1, &
                          work2, nlocalx, nlocaly, nlocalz, nlocalx1,   &
                          x_local1, nx, ny, nz, nprocx, nprocy, nprocz, &
                          id, nthread, my_x_rank, my_y_rank, my_z_rank, &
                          x_start1, x_end1,                             &
                          y_start, y_end, z_start, z_end, grid_commx,   &
                          grid_commy, grid_commz)

#ifdef FFTW
    include 'fftw3.f'
#endif

    ! formal arguments
    real(wp),         intent(inout) :: qdf(nlocalx,nlocaly,nlocalz,*)
    real(wp),         intent(inout) :: qdf_real(*)
    complex(wp),      intent(inout) :: qdfzxy(nlocalz,nlocalx1,*)
    complex(wp),      intent(inout) :: qdfyxz(nlocaly,nlocalx1,*)
    complex(wp),      intent(inout) :: qdfzyx(nlocalz,nlocaly,*)
    complex(wp),      intent(inout) :: qdfzxy_work(nlocalz,nlocalx1,nlocaly,*)
    complex(wp),      intent(inout) :: qdfyxz_work(nlocaly,nlocalx1,nlocalz,*)
    complex(wp),      intent(inout) :: qdfzyx_work(nlocalz,nlocaly,nlocalx1,*)
    complex(wp),      intent(inout) :: work1(*), work2(*)
    integer,          intent(in)    :: nlocalx, nlocaly, nlocalz
    integer,          intent(in)    :: nlocalx1
    integer,          intent(in)    :: x_local1
    integer,          intent(in)    :: nx, ny, nz
    integer,          intent(in)    :: nprocx, nprocy, nprocz
    integer,          intent(in)    :: id, nthread
    integer,          intent(in)    :: my_x_rank, my_y_rank, my_z_rank
    integer,          intent(in)    :: x_start1, x_end1
    integer,          intent(in)    :: y_start, y_end
    integer,          intent(in)    :: z_start, z_end
    integer,          intent(in)    :: grid_commx, grid_commy, grid_commz

    ! local variables
    real(wp)          :: work3(nx)
    integer*8         :: plan
    integer           :: ix, iy, iz, iys, iyss, izs, izss, k1
    integer           :: niy, niz, iy_start, iy_end, iz_start, iz_end
    integer           :: i, k, iproc, ierror
    

    ! z direction
    !

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_plan_dft_1d(plan,nz,work1,work2,FFTW_BACKWARD,FFTW_MEASURE)
#else
    call dfftw_plan_dft_1d(plan,nz,work1,work2,FFTW_BACKWARD,FFTW_MEASURE)
#endif

#else
    call zfft1d(work1, nz, 0, work2)
#endif

    niy      = nlocaly/nprocz
    iy_start = niy * my_z_rank + 1
    iy_end   = iy_start + niy - 1

    do ix = 1, x_local1
      do iy = iy_start+id, iy_end, nthread

        do iproc = 1, nprocz
          k = (iproc-1)*nlocalz
          do i = 1, nlocalz, 16
!dir$ vector aligned
            do iz = i, min0(i+15,nlocalz)
              work1(k+iz) = qdfzyx_work(iz,iy,ix,iproc)
            end do
          end do
        end do

#ifdef FFTW

#ifdef _SINGLE
        call sfftw_execute_dft(plan, work1, work2)
#else
        call dfftw_execute_dft(plan, work1, work2)
#endif
        iys = iy - iy_start + 1
        do iproc = 1, nprocz
          k = (iproc-1)*nlocalz
          iyss = (iproc-1)*niy + iys
          do iz = 1, nlocalz
            qdfzxy_work(iz,ix,iyss,1) = work2(k+iz)
          end do
        end do
#else
        call zfft1d(work1, nz, 1, work2)

        iys = iy - iy_start + 1
        do iproc = 1, nprocz
          k = (iproc-1)*nlocalz
          iyss = (iproc-1)*niy + iys
          do iz = 1, nlocalz
            qdfzxy_work(iz,ix,iyss,1) = work1(k+iz)
          end do
        end do
#endif

      end do
    end do

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_destroy_plan(plan)
#else
    call dfftw_destroy_plan(plan)
#endif

#endif

    !$omp barrier
    !$omp master

#ifdef MPI
    call mpi_alltoall(qdfzxy_work, &
                      nlocalz*nlocalx1*niy, mpi_wp_complex, &
                      qdfzxy, &
                      nlocalz*nlocalx1*niy, mpi_wp_complex, &
                      grid_commz, ierror)
#endif

    !$omp end master
    !$omp barrier

    !$omp do
    do iz = 1, nlocalz
      do ix = 1, nlocalx1
        do iy = 1, nlocaly
          qdfyxz(iy,ix,iz) = qdfzxy(iz,ix,iy)
        end do
      end do
    end do

    !$omp master

#ifdef MPI
    call mpi_alltoall(qdfyxz, &
                      nlocalx1*nlocaly*nlocalz/nprocy, mpi_wp_complex, &
                      qdfyxz_work, &
                      nlocalx1*nlocaly*nlocalz/nprocy, mpi_wp_complex, &
                      grid_commy, ierror)
#endif

    !$omp end master
    !$omp barrier


    ! y direction
    !

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_plan_dft_1d(plan,ny,work1,work2,FFTW_BACKWARD,FFTW_MEASURE)
#else
    call dfftw_plan_dft_1d(plan,ny,work1,work2,FFTW_BACKWARD,FFTW_MEASURE)
#endif

#else
    call zfft1d(work1, ny, 0, work2)
#endif

    niz      = nlocalz/nprocy
    iz_start = niz * my_y_rank + 1
    iz_end   = iz_start + niz - 1

    do iz = iz_start+id, iz_end, nthread
      izs = iz - iz_start + 1
      do ix = 1, x_local1
        do iproc = 1, nprocy
          k = (iproc-1)*nlocaly
          izss = (iproc-1)*niz + izs
          do i = 1, nlocaly, 16
!dir$ vector aligned
            do iy = i, min0(i+15,nlocaly)
              work1(k+iy) = qdfyxz_work(iy,ix,izss,1)
            end do
          end do
        end do

#ifdef FFTW

#ifdef _SINGLE
        call sfftw_execute_dft(plan, work1, work2)
#else
        call dfftw_execute_dft(plan, work1, work2)
#endif

        do iproc = 1, nprocy
          k = (iproc-1)*nlocaly
          izss = (iproc-1)*niz + izs
          do iy = 1, nlocaly
            qdfyxz_work(iy,ix,izss,1) = work2(k+iy)
          end do
        end do
#else
        call zfft1d(work1, ny, 1, work2)

        do iproc = 1, nprocy
          k = (iproc-1)*nlocaly
          izss = (iproc-1)*niz + izs
          do iy = 1, nlocaly
            qdfyxz_work(iy,ix,izss,1) = work1(k+iy)
          end do
        end do
#endif
      end do
    end do

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_destroy_plan(plan)
#else
    call dfftw_destroy_plan(plan)
#endif

#endif

    !$omp barrier
    !$omp master

#ifdef MPI
    call mpi_alltoall(qdfyxz_work, &
                      nlocaly*nlocalx1*niz, mpi_wp_complex, &
                      qdfyxz, &
                      nlocaly*nlocalx1*niz, mpi_wp_complex, &
                      grid_commy, ierror)
#endif

#ifdef MPI
    call mpi_alltoall(qdfyxz, &
                      nlocalx1*nlocaly*nlocalz/nprocx, mpi_wp_complex, &
                      qdfyxz_work, &
                      nlocalx1*nlocaly*nlocalz/nprocx, mpi_wp_complex, &
                      grid_commx, ierror)
#endif

    !$omp end master
    !$omp barrier


    ! x direction
    !

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_plan_dft_c2r_1d(plan, nx, work1, work3, FFTW_MEASURE)
#else
    call dfftw_plan_dft_c2r_1d(plan, nx, work1, work3, FFTW_MEASURE)
#endif

#else
    call zfft1d(work1, nx, 0, work2)
#endif

    niz      = nlocalz/nprocx
    iz_start = niz * my_x_rank + 1
    iz_end   = iz_start + niz - 1

    do iz = iz_start+id, iz_end, nthread
      izs = iz - iz_start + 1
      do iy = 1, nlocaly

        do i = 1, nlocalx1, 16
!dir$ vector aligned
          do ix = i, min0(i+15,nlocalx1)
            work1(ix) = qdfyxz_work(iy,ix,izs,1)
          end do
        end do

        do iproc = 2, nprocx
          k = nlocalx1 + (iproc-2)*(nlocalx1-1)
          izss = (iproc-1)*niz + izs
          do i = 1, nlocalx1-1, 16
!dir$ vector aligned
            do ix = i, min0(i+15,nlocalx1-1)
              work1(k+ix) = qdfyxz_work(iy,ix,izss,1)
            end do
          end do
        end do

#ifdef FFTW

#ifdef _SINGLE
        call sfftw_execute_dft_c2r(plan, work1, work3)
#else
        call dfftw_execute_dft_c2r(plan, work1, work3)
#endif

        do iproc = 1, nprocx
          k = (iproc-1)*nlocalx
          izss = (iproc-1)*niz + izs
          do ix = 1, nlocalx
            qdf(ix,iy,izss,1) = work3(k+ix)
          end do
        end do
#else 
        call zfft1d(work1, nx, 2, work2)

        do iproc = 1, nprocx
          k = (iproc-1)*nlocalx
          izss = (iproc-1)*niz + izs
          do ix = 1, nlocalx
            qdf(ix,iy,izss,1) = work1(k+ix)
          end do
        end do
#endif

      end do
    end do

#ifdef FFTW

#ifdef _SINGLE
    call sfftw_destroy_plan(plan)
#else
    call dfftw_destroy_plan(plan)
#endif

#endif

    !$omp barrier
    !$omp master
#ifdef MPI
    call mpi_alltoall(qdf, &
                      nlocalx*nlocaly*niz, mpi_wp_real, &
                      qdf_real, &
                      nlocalx*nlocaly*niz, mpi_wp_real, &
                      grid_commx, ierror)
#endif
    !$omp end master
    !$omp barrier

    return

  end subroutine bfft3d_tuned

end module fft3d_mod
