!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   random_mod
!> @brief   random number generator
!! @authors Takaharu Mori (TM), Jaewoon Jung (JJ)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module random_mod

  use messages_mod
  use constants_mod
  use mpi_parallel_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! parameters in the Mersenne Twister method
  integer, parameter :: defaultsd = 314159
  integer, parameter :: N  = 624
  integer, parameter :: N1 = N + 1
  integer, save, dimension(0:N-1) :: mt
  integer, save                   :: mti = N1

  ! functions
  public  :: random
  public  :: setup_mersenne_twister
  public  :: setup_mersenne_twister_nobcast
  public  :: grnd  !< uniform random number
  public  :: gauss !< gaussian random number

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      random
  !> @brief        random number generator (uniform distribution [0,1])
  !! @authors      Takaharu Mori (TM)
  !! @param[inout] iseed : random number seed
  !! @note         P.A.W.Lewis et al., BM Systems Journal, 8, 136 (1969).
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  function random(iseed)

    ! formal arguments
    integer,                 intent(inout) :: iseed

    ! local variables
    real(dp)                 :: dseed
    real(dp)                 :: random


    if (iseed <= 1) iseed = 314159

    dseed = 16807.0_dp * iseed
    dseed = mod(dseed, 2147483647.0_dp)

    random = dseed / 2147483711.0_dp
    iseed = dseed

    return

  end function random

  !____________________________________________________________________________
  ! A C-program for MT19937: Real number version
  !   genrand() generates one pseudorandom real number (double)
  ! which is uniformly distributed on [0,1]-interval, for each
  ! call. sgenrand(seed) set initial values to the working area
  ! of 624 words. Before genrand(), sgenrand(seed) must be
  ! called once. (seed is any 32-bit integer except for 0).
  ! Integer generator is obtained by modifying two lines.
  !   Coded by Takuji Nishimura, considering the suggestions by
  ! Topher Cooper and Marc Rieffel in July-Aug. 1997.
  !
  ! This library is free software; you can redistribute it and/or
  ! modify it under the terms of the GNU Library General Public
  ! License as published by the Free Software Foundation; either
  ! version 2 of the License, or (at your option) any later
  ! version.
  ! This library is distributed in the hope that it will be useful,
  ! but WITHOUT ANY WARRANTY; without even the implied warranty of
  ! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  ! See the GNU Library General Public License for more details.
  ! You should have received a copy of the GNU Library General
  ! Public License along with this library; if not, write to the
  ! Free Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
  ! 02111-1307  USA
  !
  ! Copyright (C) 1997 Makoto Matsumoto and Takuji Nishimura.
  ! When you use this, send an email to: matumoto@math.keio.ac.jp
  ! with an appropriate reference to your work.
  !
  !***********************************************************************
  ! Fortran translation by Hiroshi Takano.  Jan. 13, 1999.
  !
  !   genrand()      -> real(dp)         function grnd()
  !   sgenrand(seed) -> subroutine sgrnd(seed)
  !                     integer seed
  !
  ! This program uses the following non-standard intrinsics.
  !   ishft(i,n): If n>0, shifts bits in i by n positions to left.
  !               If n<0, shifts bits in i by n positions to right.
  !   iand (i,j): Performs logical AND on corresponding bits of i and j.
  !   ior  (i,j): Performs inclusive OR on corresponding bits of i and j.
  !   ieor (i,j): Performs exclusive OR on corresponding bits of i and j.
  !
  !***********************************************************************
  ! Fortran version rewritten as an F90 module and mt state saving and getting
  ! subroutines added by Richard Woloshyn. (rwww@triumf.ca). June 30, 1999

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_mersenne_twister
  !> @brief        setup mersenne twister random number generator
  !! @authors      TM
  !! @param[in]    seed : random number seed
  !! @note         setting initial seeds to mt[N] using
  !!               the generator Line 25 of Table 1 in
  !!               [KNUTH 1981, The Art of Computer Programming
  !!                  Vol. 2 (2nd Ed.), pp102]
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_mersenne_twister(seed)

    ! formal arguments
    integer,                 intent(in)    :: seed

    ! local variables
    integer                  :: gseed, dtval(8)
    character(10)            :: time
    character(8)             :: date
    character(5)             :: zone


    ! if iseed is less than 0, date_and_time is used
    !
    if (seed <= 0) then
      
      call date_and_time(date, time, zone, dtval)
      gseed = dtval(8)*1000000+dtval(7)*10000+dtval(6)*100+dtval(5)
        
#ifdef MPI
      call mpi_bcast(gseed, 1, mpi_integer, 0, mpi_comm_world, ierror)
#endif
      if (main_rank) then
        write(MsgOut,'(A)')                                     &
          'Setup_Mersenne_Twister> Initialize the random number using the current time'
        write(MsgOut,'(A20,A10,A12,A7)')                        &
          '  date and time   = ', date, time, zone
        write(MsgOut,'(A20,I10)')                               &
          '  seed            = ', gseed
        write(MsgOut,'(A)') ''
      end if

    else

      gseed = seed

      if (main_rank) then
        write(MsgOut,'(A)') 'Setup_Mersenne_Twister> Initialize the random number'
        write(MsgOut,'(A20,I10)') '  seed            = ', gseed
        write(MsgOut,'(A)') ''
      end if

    end if

    mt(0) = iand(gseed,-1)
    do mti = 1, N-1
      mt(mti) = iand(69069 * mt(mti-1),-1)
    end do

    return

  end subroutine setup_mersenne_twister

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_mersenne_twister_nobcast
  !> @brief        setup mersenne twister random number generator with no-bcast
  !! @authors      TM, JJ
  !! @param[in]    seed : random number seed
  !! @note         setting initial seeds to mt[N] using
  !!               the generator Line 25 of Table 1 in
  !!                [KNUTH 1981, The Art of Computer Programming
  !!                   Vol. 2 (2nd Ed.), pp102]
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_mersenne_twister_nobcast(seed)

    ! formal arguments
    integer,                 intent(in)    :: seed

    ! local variables
    integer                  :: i, gseed, dtval(8)
    character(10)            :: time
    character(8)             :: date
    character(5)             :: zone


    ! if iseed is less than 0, date_and_time is used
    !
    if (seed <= 0) then

      call date_and_time(date, time, zone, dtval)
      gseed = my_world_rank*1000000+dtval(8)*10000+dtval(7)*1000+ &
              dtval(6)*100+dtval(5)*10 

      if (main_rank) then
        write(MsgOut,'(A)')                                     &
          'Setup_Mersenne_Twister> Initialize the random number using the current time'
        write(MsgOut,'(A20,A10,A12,A7)')                        &
          '  date and time   = ', date, time, zone
        write(MsgOut,'(A20,I10)')                               &
          '  seed            = ', gseed
        write(MsgOut,'(A)') ''
      end if

    else

      gseed = seed+my_world_rank

      if (main_rank) then
        write(MsgOut,'(A)') 'Setup_Mersenne_Twister> Initialize the random number'
        write(MsgOut,'(A)') 'Setup_Mersenne_Twister> Seed is different at MPI rank'
        write(MsgOut,'(A20,I10)') '  seed            = ', gseed
        write(MsgOut,'(A)') ''

      end if

    endif

    mt(0) = iand(gseed,-1)
    do mti = 1, N-1
      mt(mti) = iand(69069 * mt(mti-1),-1)
    end do

    return

  end subroutine setup_mersenne_twister_nobcast

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      grnd
  !> @brief        random number generator
  !! @authors      TM
  !! @return       random number
  !! @note         Mersenne Twister method
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  real(dp)         function grnd()

    implicit integer(a-z)

    ! Period parameters
    integer, parameter :: M = 397, MATA  = -1727483681
    ! constant vector a
    integer, parameter :: LMASK =  2147483647
    ! least significant r bits
    integer, parameter :: UMASK = -LMASK - 1
    ! most significant w-r bits
    ! Tempering parameters
    integer, parameter :: TMASKB= -1658038656, TMASKC= -272236544

    dimension mag01(0:1)
    data mag01/0, MATA/
    save mag01
    ! mag01(x) = x * MATA for x=0,1

    TSHFTU(y) = ishft(y,-11)
    TSHFTS(y) = ishft(y,7)
    TSHFTT(y) = ishft(y,15)
    TSHFTL(y) = ishft(y,-18)


    if (mti .ge. N) then

      ! generate N words at one time
      !
      if (mti .eq. N+1) then
        ! if sgrnd() has not been called, a default initial seed is used
        !
        call setup_mersenne_twister(defaultsd)
      end if

      do kk = 0, N-M-1
        y = ior(iand(mt(kk),UMASK),iand(mt(kk+1),LMASK))
        mt(kk) = ieor(ieor(mt(kk+M),ishft(y,-1)),mag01(iand(y,1)))
      end do

      do kk = N-M, N-2
        y = ior(iand(mt(kk),UMASK),iand(mt(kk+1),LMASK))
        mt(kk) = ieor(ieor(mt(kk+(M-N)),ishft(y,-1)),mag01(iand(y,1)))
      end do

      y = ior(iand(mt(N-1),UMASK),iand(mt(0),LMASK))
      mt(N-1) = ieor(ieor(mt(M-1),ishft(y,-1)),mag01(iand(y,1)))
      mti = 0
    end if


    y = mt(mti)
    mti = mti + 1
    y = ieor(y,TSHFTU(y))
    y = ieor(y,iand(TSHFTS(y),TMASKB))
    y = ieor(y,iand(TSHFTT(y),TMASKC))
    y = ieor(y,TSHFTL(y))

    if (y .lt. 0) then
      grnd = (real(y,dp)+2.0_dp**32)/(2.0_dp**32-1.0_dp)
    else
      grnd = real(y,dp)/(2.0_dp**32-1.0_dp)
    end if

    return

  end function grnd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      gauss
  !> @brief        Gaussian random number generator
  !! @authors      TM
  !! @return       random number
  !! @note         Polar form of the Box-Muller transformation
  !!               Because Mersenne twister generates 0 of grnd,
  !!               basic form of the transformation (using log) cannnot be used
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  real(dp)         function gauss()

    ! local variables
    real(dp)                 :: v1, v2, rsq


    rsq = 2.0_dp
    do while (rsq >= 1.0_dp)
      v1  = 2.0_dp*grnd() - 1.0_dp
      v2  = 2.0_dp*grnd() - 1.0_dp
      rsq = v1*v1 + v2*v2
    end do
    rsq   = sqrt(-2.0_dp * log(rsq) / rsq)
    gauss = rsq * v1

    return

  end function gauss

end module random_mod
