!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_dynvars_mod
!> @brief   define dynamic variables
!! @authors Jaewoon Jung (JJ), Takaharu Mori (TM), Yuji Sugita (YS), 
!!          Chigusa Kobayashi (CK)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_dynvars_mod

  use sp_output_str_mod
  use sp_dynamics_str_mod
  use sp_dynvars_str_mod
  use sp_ensemble_str_mod
  use sp_boundary_str_mod
  use sp_enefunc_str_mod
  use sp_domain_str_mod
  use molecules_str_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif
  
  implicit none
  private

  ! variables
  integer, public, save :: DynvarsOut = 6

  logical,         save :: etitle = .true.

  ! subroutines
  public  :: setup_dynvars
  public  :: output_dynvars
  public  :: compute_dynvars
  private :: output_dynvars_genesis
  private :: output_dynvars_charmm
  private :: output_dynvars_namd
  private :: compute_pressure
  private :: reduce_property

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_dynvars
  !> @brief        copy molecule information into dynvars
  !! @authors      JJ
  !! @param[out]   dynvars  : dynamic variables
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_dynvars(dynvars)

    ! formal arguments
    type(s_dynvars),         intent(inout) :: dynvars


    ! initialize variables in dynvars
    !
    call init_dynvars(dynvars)

    return

  end subroutine setup_dynvars

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    output_dynvars
  !> @brief        output dynamical variables
  !! @authors      YS
  !! @param[in]    output  : output information
  !! @param[in]    enefunc : potential energy functions information
  !! @param[in]    dynvars : dynamical variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_dynvars(output, enefunc, dynvars)

    ! formal arguments
    type(s_output),          intent(in)    :: output
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_dynvars),         intent(in)    :: dynvars


    ! If MD, output is done on main_rank
    !
    if (output%replica) then
      if (.not. replica_main_rank) return
    else
      if (.not. main_rank) return
    end if

    select case (enefunc%output_style)

    case (OutputStyleGENESIS)

      call output_dynvars_genesis(enefunc, dynvars)

    case (OutputStyleCHARMM)

      call output_dynvars_charmm (enefunc, dynvars)

    case (OutputStyleNAMD)

      call output_dynvars_namd   (dynvars)

    end select

    return

  end subroutine output_dynvars

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_dynvars
  !> @brief        calculate other dynamical variables with domain
  !! @authors      JJ
  !! @param[in]    dynamics : dynamics information
  !! @param[in]    boundary : boundary information
  !! @param[in]    ensemble : ensemble information
  !! @param[inout] domain   : domain information
  !! @param[inout] dynvars  : dynamical variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_dynvars(dynamics, boundary, ensemble, domain, dynvars)

    ! formal arguments
    type(s_dynamics),        intent(in)    :: dynamics
    type(s_boundary),        intent(in)    :: boundary
    type(s_ensemble),        intent(in)    :: ensemble
    type(s_domain),  target, intent(inout) :: domain
    type(s_dynvars), target, intent(inout) :: dynvars

    ! local variables
    real(wp)                 :: kex, key, kez
    real(wp)                 :: ekin, ekin_old, ekin_new
    real(wp)                 :: vel_oldx, vel_oldy, vel_oldz
    real(wp)                 :: rmsg, viri_ke, viri_ke_ext
    real(wp)                 :: gamma_t, dt, scale_v
    real(wp)                 :: virial_sum(1:3,1:3)
    integer                  :: i, ix

    real(wp),        pointer :: mass(:,:)
    real(wp),        pointer :: vel(:,:,:), vel_ref(:,:,:), vel_full(:,:,:)
    real(wp),        pointer :: force(:,:,:)
    real(wp),        pointer :: virial(:,:), virial_ext(:,:)
    real(wp),        pointer :: viri_const(:,:)
    integer,         pointer :: ncell, natom(:)


    ncell       => domain%num_cell_local
    natom       => domain%num_atom
    mass        => domain%mass
    vel         => domain%velocity
    vel_ref     => domain%velocity_ref
    vel_full    => domain%velocity_full
    force       => domain%force

    virial      => dynvars%virial
    virial_ext  => dynvars%virial_extern
    viri_const  => dynvars%virial_const

    gamma_t     =  ensemble%gamma_t*AKMA_PS
    dt          =  dynamics%timestep/AKMA_PS

    ! Compute velocities(t + dt)
    !   v(t+dt) = 0.5*(v(t+3/2dt) + v(t+1/2dt))
    !
    if (dynamics%integrator == IntegratorLEAP) then

      if (ensemble%tpcontrol == TpcontrolLangevin) then
        scale_v = sqrt(1.0_wp+0.5_wp*gamma_t*dt)
      else
        scale_v = 1.0_wp
      end if

      do i = 1, ncell
        do ix = 1, natom(i)
          vel_full(1,ix,i) = 0.5_wp*(vel(1,ix,i) + vel_ref(1,ix,i))*scale_v
          vel_full(2,ix,i) = 0.5_wp*(vel(2,ix,i) + vel_ref(2,ix,i))*scale_v
          vel_full(3,ix,i) = 0.5_wp*(vel(3,ix,i) + vel_ref(3,ix,i))*scale_v
        end do
      end do

    else

      do i = 1, ncell
        do ix = 1, natom(i)
          vel_full(1,ix,i) = vel(1,ix,i)
          vel_full(2,ix,i) = vel(2,ix,i)
          vel_full(3,ix,i) = vel(3,ix,i)
        end do
      end do

    end if

    ! kinetic energy(t+dt) and temperature(t+dt)
    ! 
    kex = 0.0_wp
    key = 0.0_wp
    kez = 0.0_wp

    do i = 1, ncell
      do ix = 1, natom(i)
        kex = kex + mass(ix,i)*vel_full(1,ix,i)*vel_full(1,ix,i)
        key = key + mass(ix,i)*vel_full(2,ix,i)*vel_full(2,ix,i)
        kez = kez + mass(ix,i)*vel_full(3,ix,i)*vel_full(3,ix,i)
      end do
    end do

    ekin  = 0.5_wp * (kex + key + kez)

    ! other dynvars(t+dt)
    !
    ekin_old = 0.0_wp
    ekin_new = 0.0_wp
    rmsg     = 0.0_wp

    if (dynamics%integrator == IntegratorLEAP) then

      do i = 1, ncell
        do ix = 1, natom(i)

          vel_oldx = 2.0_wp*vel_full(1,ix,i)-vel(1,ix,i)
          vel_oldy = 2.0_wp*vel_full(2,ix,i)-vel(2,ix,i)
          vel_oldz = 2.0_wp*vel_full(3,ix,i)-vel(3,ix,i)
          ekin_old = ekin_old + &
                     (vel_oldx**2+vel_oldy**2+vel_oldz**2)*mass(ix,i)
          ekin_new = ekin_new + &
                     (vel(1,ix,i)**2+ vel(2,ix,i)**2+vel(3,ix,i)**2)*mass(ix,i)
          rmsg     = rmsg+ force(1,ix,i)**2+force(2,ix,i)**2+force(3,ix,i)**2

         end do
      end do

      ekin_old = 0.5_wp * ekin_old
      ekin_new = 0.5_wp * ekin_new

    else

      do i = 1, ncell
        do ix = 1, natom(i)
          rmsg = rmsg + force(1,ix,i)**2 + force(2,ix,i)**2 + force(3,ix,i)**2
        end do
      end do

    end if

    ! virial (virial + virial_const)
    !
    if (dynamics%integrator /= IntegratorLEAP) &
      virial_sum(1:3,1:3) = virial(1:3,1:3) + viri_const(1:3,1:3)

#ifdef MPI

    if (dynamics%integrator == IntegratorLEAP) then

      call reduce_property(dynvars%energy%bond,               &
                           dynvars%energy%angle,              &
                           dynvars%energy%urey_bradley,       &
                           dynvars%energy%dihedral,           &
                           dynvars%energy%improper,           &
                           dynvars%energy%cmap,               &
                           dynvars%energy%electrostatic,      &
                           dynvars%energy%van_der_waals,      &
                           dynvars%energy%restraint_position, &
                           dynvars%energy%total,              &
                           virial(1,1),     virial(2,2),     virial(3,3),     &
                           virial_ext(1,1), virial_ext(2,2), virial_ext(3,3), &
                           kex, key, kez, ekin, ekin_old, ekin_new, rmsg)

    else

      call reduce_property(dynvars%energy%bond,               &
                           dynvars%energy%angle,              &
                           dynvars%energy%urey_bradley,       &
                           dynvars%energy%dihedral,           &
                           dynvars%energy%improper,           &
                           dynvars%energy%cmap,               &
                           dynvars%energy%electrostatic,      &
                           dynvars%energy%van_der_waals,      &
                           dynvars%energy%restraint_position, &           
                           dynvars%energy%total,              &
                           virial_sum(1,1), virial_sum(2,2), virial_sum(3,3), &
                           virial_ext(1,1), virial_ext(2,2), virial_ext(3,3), &
                           kex, key, kez, ekin, ekin_old, ekin_new, rmsg)
    end if

#endif

    dynvars%total_kene  = ekin
    dynvars%temperature = 2.0_wp*ekin/(real(domain%num_deg_freedom,wp)*KBOLTZ)

    ! potential energy(t+dt) and total energy(t+dt)
    !
    dynvars%total_pene = &
               dynvars%energy%bond               + &
               dynvars%energy%angle              + &
               dynvars%energy%urey_bradley       + &
               dynvars%energy%dihedral           + &
               dynvars%energy%improper           + &
               dynvars%energy%cmap               + &
               dynvars%energy%electrostatic      + &
               dynvars%energy%van_der_waals      + &
               dynvars%energy%restraint_distance + &
               dynvars%energy%restraint_position

    dynvars%total_energy = &
               dynvars%total_pene + dynvars%total_kene


    ! volume(t+dt) and pressure(t+dt)
    !
    dynvars%box_size_x = boundary%box_size_x_ref
    dynvars%box_size_y = boundary%box_size_y_ref
    dynvars%box_size_z = boundary%box_size_z_ref

    dynvars%volume =  boundary%box_size_x_ref &
                    * boundary%box_size_y_ref &
                    * boundary%box_size_z_ref

    if (dynamics%integrator == IntegratorLEAP) then

      call compute_pressure(domain%num_deg_freedom, dynvars%volume, &
                            dynvars%temperature, virial,            &
                            dynvars%internal_pressure)

    else

      call compute_pressure(domain%num_deg_freedom, dynvars%volume, &
                          dynvars%temperature, virial_sum,          &
                          dynvars%internal_pressure)
    end if

    call compute_pressure(domain%num_deg_freedom, dynvars%volume,   &
                          dynvars%temperature, virial_ext,          &
                          dynvars%external_pressure)


    if (dynamics%integrator == IntegratorLEAP) then

      dynvars%pressure_xx = P_ATMOS*(kex + virial(1,1))/dynvars%volume
      dynvars%pressure_yy = P_ATMOS*(key + virial(2,2))/dynvars%volume
      dynvars%pressure_zz = P_ATMOS*(kez + virial(3,3))/dynvars%volume

    else

      dynvars%pressure_xx = P_ATMOS*(kex + virial_sum(1,1))/dynvars%volume
      dynvars%pressure_yy = P_ATMOS*(key + virial_sum(2,2))/dynvars%volume
      dynvars%pressure_zz = P_ATMOS*(kez + virial_sum(3,3))/dynvars%volume

    end if


    dynvars%hfc_kene     = 0.5_wp*(ekin_old + ekin_new)
    dynvars%rms_gradient = sqrt(rmsg/real(3*domain%num_atom_all,wp))


    if (dynamics%integrator == IntegratorLEAP) then

      viri_ke =  virial(1,1) + virial(2,2) + virial(3,3)

    else

      viri_ke =  virial_sum(1,1) + virial_sum(2,2) + virial_sum(3,3)

    end if

    dynvars%internal_virial = viri_ke/3.0_wp
    dynvars%virial_kene     = -0.5_wp*viri_ke

    viri_ke_ext =virial_ext(1,1) + virial_ext(2,2) + virial_ext(3,3)
    dynvars%external_virial = viri_ke_ext/3.0_wp

    return

  end subroutine compute_dynvars

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    output_dynvars_genesis
  !> @brief        output dynvars in GENESIS style
  !! @authors      YS, TM
  !! @param[in]    enefunc : potential energy functions information
  !! @param[in]    dynvars : dynamical variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_dynvars_genesis(enefunc, dynvars)

    ! formal arguments
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_dynvars),         intent(in)    :: dynvars

    ! local variables
    real(wp)                 :: ene_misc1, ene_misc2, ene_restraint
    integer                  :: i
    character                :: cmisc1*16, cmisc2*16


    if (enefunc%forcefield == ForcefieldCHARMM) then

      ene_misc1 = dynvars%energy%urey_bradley
      ene_misc2 = dynvars%energy%cmap
      cmisc1    = '    UREY-BRADLEY'
      cmisc2    = '            CMAP'

    end if

    ene_restraint =   dynvars%energy%restraint_distance &
                    + dynvars%energy%restraint_position

    if (etitle) then

      write(DynvarsOut,'(A,A16,A16,A)') &
                             'INFO:       STEP            TIME       TOTAL_ENE   POTENTIAL_ENE     KINETIC_ENE' &
                          // '            RMSG            BOND           ANGLE        DIHEDRAL        IMPROPER' &
                          // '         VDWAALS           ELECT',       cmisc1,        cmisc2,'       RESTRAINT' &
                          // '            BOXX            BOXY            BOXZ          VOLUME     TEMPERATURE' &
                          // '         PRESSXX         PRESSYY         PRESSZZ          VIRIAL        PRESSURE'
      write(DynvarsOut,'(A80)') ' --------------- --------------- --------------- --------------- ---------------'
      etitle = .false.
    end if

    write(DynvarsOut,'(A5,1x,i10,24(1x,F15.4))')'INFO:', & 
                           dynvars%step,                 &
                           dynvars%time,                 &
                           dynvars%total_energy,         &
                           dynvars%total_pene,           &
                           dynvars%total_kene,           &
                           dynvars%rms_gradient,         &
                           dynvars%energy%bond,          &
                           dynvars%energy%angle,         &
                           dynvars%energy%dihedral,      &
                           dynvars%energy%improper,      &
                           dynvars%energy%van_der_waals, &
                           dynvars%energy%electrostatic, &
                           ene_misc1,                    &
                           ene_misc2,                    &
                           ene_restraint,                &
                           dynvars%box_size_x,           &
                           dynvars%box_size_y,           &
                           dynvars%box_size_z,           &
                           dynvars%volume,               &
                           dynvars%temperature,          &
                           dynvars%pressure_xx,          &
                           dynvars%pressure_yy,          &
                           dynvars%pressure_zz,          &
                           dynvars%internal_virial,      &
                           dynvars%internal_pressure
    write(DynvarsOut,'(A)') ''

    return

  end subroutine output_dynvars_genesis

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    output_dynvars_charmm
  !> @brief        output dynvars in CHARMM style
  !! @authors      YS, CK
  !! @param[in]    enefunc : potential energy functions information
  !! @param[in]    dynvars : dynamical variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_dynvars_charmm(enefunc, dynvars)

    ! formal arguments
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_dynvars), target, intent(in)    :: dynvars

    ! local variables
    real(wp)                 :: time, totener, totke, energy, temperature
    real(wp)                 :: grms, hfctote, hfcke, ehfcor, virke
    real(wp)                 ::                hbonds, asp, user
    real(wp)                 :: vire, viri, presse, pressi, volume
    real(wp)                 :: boxx, boxy, boxz, pixx, piyy, pizz
    real(wp)                 :: cdihe, cintcr, noe
    integer                  :: step

    real(wp),        pointer :: bonds, angles, urey_b, dihedrals, impropers
    real(wp),        pointer :: cmaps
    real(wp),        pointer :: vdwaals, elec
    real(wp),        pointer :: posicon, restdist


    step        = 0
    time        = 0.0_wp
    hbonds      = 0.0_wp
    totener     = 0.0_wp
    totke       = 0.0_wp
    energy      = 0.0_wp
    temperature = 0.0_wp
    asp         = 0.0_wp
    user        = 0.0_wp
    vire        = 0.0_wp
    viri        = 0.0_wp
    presse      = 0.0_wp
    pressi      = 0.0_wp
    volume      = 0.0_wp
    boxx        = 0.0_wp
    boxy        = 0.0_wp
    boxz        = 0.0_wp
    pixx        = 0.0_wp
    piyy        = 0.0_wp
    pizz        = 0.0_wp
    hfctote     = 0.0_wp
    ehfcor      = 0.0_wp
    volume      = 0.0_wp


    step        = dynvars%step
    time        = dynvars%time
    energy      = dynvars%total_pene
    totke       = dynvars%total_kene
    totener     = dynvars%total_energy
    temperature = dynvars%temperature
    grms        = dynvars%rms_gradient
    hfcke       = dynvars%hfc_kene
    virke       = dynvars%virial_kene

    cdihe       = 0.0_wp
    cintcr      = 0.0_wp
    noe         = 0.0_wp

    viri        = dynvars%internal_virial
    vire        = dynvars%external_virial
    pressi      = dynvars%internal_pressure
    presse      = dynvars%external_pressure
    volume      = dynvars%volume
    boxx        = dynvars%box_size_x
    boxy        = dynvars%box_size_y
    boxz        = dynvars%box_size_z
    pixx        = dynvars%pressure_xx 
    piyy        = dynvars%pressure_yy
    pizz        = dynvars%pressure_zz

    bonds      => dynvars%energy%bond
    angles     => dynvars%energy%angle
    urey_b     => dynvars%energy%urey_bradley
    dihedrals  => dynvars%energy%dihedral
    impropers  => dynvars%energy%improper
    elec       => dynvars%energy%electrostatic
    vdwaals    => dynvars%energy%van_der_waals
    cmaps      => dynvars%energy%cmap
    restdist   => dynvars%energy%restraint_distance
    posicon    => dynvars%energy%restraint_position


    ! write title if necessary
    !
    if (etitle) then
      write(DynvarsOut,'(A)') 'Output_Energy> CHARMM_Style is used'
      write(DynvarsOut,'(A)') ' '
      write(DynvarsOut,'(A79)') 'DYNA DYN: Step         Time      TOTEner        TOTKe       ENERgy  TEMPerature'
      write(DynvarsOut,'(A79)') 'DYNA PROP:             GRMS      HFCTote        HFCKe       EHFCor        VIRKe'
      write(DynvarsOut,'(A79)') 'DYNA INTERN:          BONDs       ANGLes       UREY-b    DIHEdrals    IMPRopers'
      if (enefunc%num_cmap_all > 0) then
        write(DynvarsOut,'(A79)') 'DYNA CROSS:           CMAPs                                                    '
      end if
      write(DynvarsOut,'(A79)') 'DYNA EXTERN:        VDWaals         ELEC       HBONds          ASP         USER'
      if (enefunc%restraint) then
        write(DynvarsOut,'(A79)') 'DYNA CONSTR:       HARMonic    CDIHedral        CIC     RESDistance       NOE'
      end if
      write(DynvarsOut,'(A79)') 'DYNA PRESS:            VIRE         VIRI       PRESSE       PRESSI       VOLUme'
      write(DynvarsOut,'(A79)') ' ----------       ---------    ---------    ---------    ---------    ---------'
      etitle = .false.
    end if


    ! write energy in CHARMM-style
    !
    write(DynvarsOut,'(A5,I9,5F13.5)') 'DYNA>', step, time, totener, totke, energy, temperature
    write(DynvarsOut,'(A14,5F13.5)')   'DYNA PROP>    ', grms, hfctote, hfcke, ehfcor, virke
    write(DynvarsOut,'(A14,5F13.5)')   'DYNA INTERN>  ', bonds, angles, urey_b, dihedrals, impropers

    if (enefunc%num_cmap_all > 0) then
      write(DynvarsOut,'(A14,F13.5)')    'DYNA CROSS>   ', cmaps
    end if

    write(DynvarsOut,'(A14,5F13.5)')   'DYNA EXTERN>  ', vdwaals, elec, hbonds, asp, user

    if (enefunc%restraint) then
      write(DynvarsOut,'(A14,5F13.5)')   'DYNA CONSTR>  ', posicon, cdihe, cintcr, restdist, noe
    end if

    write(DynvarsOut,'(A14,5F13.5)')   'DYNA PRESS>   ', vire, viri, presse, pressi, volume
    write(DynvarsOut,'(A79)') ' ----------       ---------    ---------    ---------    ---------    ---------'

    return

  end subroutine output_dynvars_charmm

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    output_dynvars_namd
  !> @brief        output dynvars in NAMD style
  !! @authors      YS, CK
  !! @param[in]    dynvars : dynamical variables information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_dynvars_namd(dynvars)

    ! formal arguments
    type(s_dynvars), target, intent(in)    :: dynvars

    ! local variables
    real(wp)                 :: ebond, eangle
    real(wp)                 ::               misc, kinetic
    real(wp)                 :: total, temp, total2, total3, tempavg
    real(wp)                 :: pressure, gpressure, volume
    real(wp)                 :: pressavg, gpressavg
    integer                  :: nstep

    real(wp),        pointer :: edihed, eimprp
    real(wp),        pointer :: eelect, evdw
    real(wp),        pointer :: eboundary


    nstep     =  0
    ebond     =  dynvars%energy%bond  + dynvars%energy%restraint_distance
    eangle    =  dynvars%energy%angle + dynvars%energy%urey_bradley

    edihed    => dynvars%energy%dihedral
    eimprp    => dynvars%energy%improper
    eelect    => dynvars%energy%electrostatic
    evdw      => dynvars%energy%van_der_waals
    eboundary => dynvars%energy%restraint_position


    ! write title if necessary
    !
    if (etitle) then
      write(DynvarsOut,'(A)') 'Output_Energy> NAMD_Style is used'
      write(DynvarsOut,'(A)') ' '
      write(DynvarsOut,'(A75)') 'ETITLE:      TS           BOND          ANGLE          DIHED          IMPRP'
      write(DynvarsOut,'(A75)') '          ELECT            VDW       BOUNDARY           MISC        KINETIC'
      write(DynvarsOut,'(A75)') '          TOTAL           TEMP         TOTAL2         TOTAL3        TEMPAVG'
      write(DynvarsOut,'(A75)') '       PRESSURE      GPRESSURE         VOLUME       PRESSAVG      GPRESSAVG'
      write(DynvarsOut,'(A)') ' '
      etitle = .false.
    end if


    ! write energy in NAMD-style
    !
    write(DynvarsOut,'(A7,I8,4F15.4)') &
                                 'ENERGY:', nstep, ebond, eangle, edihed, eimprp
    write(DynvarsOut,'(5F15.4)') eelect, evdw, eboundary, misc, kinetic
    write(DynvarsOut,'(5F15.4)') total, temp, total2, total3, tempavg
    write(DynvarsOut,'(5F15.4)') pressure, gpressure, volume, pressavg,gpressavg
    write(DynvarsOut,'(A)') ' '


    return

  end subroutine output_dynvars_namd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_pressure
  !> @brief        compute interal pressure
  !! @authors      TM
  !! @param[in]    num_degree  : number of degrees of freedom
  !! @param[in]    volume      : volume (A^3)
  !! @param[in]    temperature : temperature (K)
  !! @param[in]    virial      : internal virial
  !! @param[out]   pressure    : pressure (atm)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_pressure(num_degree, volume, temperature, virial, pressure)

    ! formal arguments
    integer,                 intent(in)    :: num_degree
    real(wp),                intent(in)    :: volume
    real(wp),                intent(in)    :: temperature
    real(wp),                intent(in)    :: virial(3,3)
    real(wp),                intent(out)   :: pressure

    ! local variables
    real(wp)                 :: viri_ke


    ! compute internal pressure
    !
    viri_ke  = virial(1,1) + virial(2,2) + virial(3,3)
    pressure = P_ATMOS / (3.0_wp * volume)                      &
              * (num_degree * KBOLTZ * temperature + viri_ke)


    return

  end subroutine compute_pressure

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    reduce_property
  !> @brief        reduce property
  !! @authors      JJ
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine reduce_property(val1, val2, val3, val4, val5, val6, val7, val8,   &
                             val9, val10, val11, val12, val13, val14, val15,   &
                             val16, val17, val18, val19, val20, val21, val22,  &
                             val23)

    ! formal arguments
    real(wp),                intent(inout) ::  val1,  val2,  val3
    real(wp),                intent(inout) ::  val4,  val5,  val6
    real(wp),                intent(inout) ::  val7,  val8,  val9
    real(wp),                intent(inout) :: val10, val11, val12
    real(wp),                intent(inout) :: val13, val14, val15
    real(wp),                intent(inout) :: val16, val17, val18
    real(wp),                intent(inout) :: val19, val20, val21
    real(wp),                intent(inout) :: val22, val23

#ifdef MPI
    ! local variables
    real(wp)                 :: before_reduce(23), after_reduce(23)


    before_reduce(1)  = val1
    before_reduce(2)  = val2
    before_reduce(3)  = val3
    before_reduce(4)  = val4
    before_reduce(5)  = val5
    before_reduce(6)  = val6
    before_reduce(7)  = val7
    before_reduce(8)  = val8
    before_reduce(9)  = val9
    before_reduce(10) = val10
    before_reduce(11) = val11
    before_reduce(12) = val12
    before_reduce(13) = val13
    before_reduce(14) = val14
    before_reduce(15) = val15
    before_reduce(16) = val16
    before_reduce(17) = val17
    before_reduce(18) = val18
    before_reduce(19) = val19
    before_reduce(20) = val20
    before_reduce(21) = val21
    before_reduce(22) = val22
    before_reduce(23) = val23

    call mpi_reduce(before_reduce, after_reduce, 23, mpi_wp_real, &
                    mpi_sum, 0, mpi_comm_country, ierror)

    val1  = after_reduce(1)
    val2  = after_reduce(2)
    val3  = after_reduce(3)
    val4  = after_reduce(4)
    val5  = after_reduce(5)
    val6  = after_reduce(6)
    val7  = after_reduce(7)
    val8  = after_reduce(8)
    val9  = after_reduce(9)
    val10 = after_reduce(10)
    val11 = after_reduce(11)
    val12 = after_reduce(12)
    val13 = after_reduce(13)
    val14 = after_reduce(14)
    val15 = after_reduce(15)
    val16 = after_reduce(16)
    val17 = after_reduce(17)
    val18 = after_reduce(18)
    val19 = after_reduce(19)
    val20 = after_reduce(20)
    val21 = after_reduce(21)
    val22 = after_reduce(22)
    val23 = after_reduce(23)

#endif
    return

  end subroutine reduce_property

end module sp_dynvars_mod
