!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_enefunc_table_mod
!> @brief   define lookup table for energy calculation
!! @authors Jaewoon Jung (JJ), Chigusa Kobayashi (CK)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_enefunc_table_mod

  use sp_boundary_mod
  use sp_energy_pme_mod
  use sp_energy_mod
  use sp_pairlist_str_mod
  use sp_enefunc_str_mod
  use molecules_str_mod
  use mpi_parallel_mod
  use messages_mod
  use constants_mod

  implicit none
  private

  ! variables
  real(wp),         save      :: el_fact    ! e^2/(4*pi*eps) A*kcal/mol
  real(wp),         save      :: alpha      ! alpha
  real(wp),         save      :: alpha2m    ! -alpha^2
  real(wp),         save      :: alpha2sp   ! 2*alpha/sqrt(pi)

  public  :: setup_enefunc_table
  private :: setup_table_general
  private :: setup_table_general_linear
  private :: setup_table_general_cutoff
  private :: setup_table_water
  private :: setup_table_water_linear
  private :: setup_table_water_cutoff

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_enefunc_table
  !> @brief        define lookup table of potential energy function
  !! @authors      JJ     
  !! @param[in]    ene_info : ENERGY section control parameters information
  !! @param[in]    molecule : molecule information
  !! @param[inout] enefunc  : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_enefunc_table(ene_info, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    type(s_enefunc),         intent(inout) :: enefunc

    ! local variables
    real(wp)                 :: cutoff, cutoff2
    real(wp)                 :: cutoff2_water, cutoff2_water1, density
    integer                  :: cutoff_int, cutoff_int2, cutoff_int1


    ! set cutoff distance etc.
    !
    enefunc%table%table       = ene_info%table
    enefunc%table%table_order = ene_info%table_order
    enefunc%table%water_model = ene_info%water_model
    enefunc%table%density     = ene_info%table_density

    ! setting common parameters
    !
    el_fact  = ELECOEF/ene_info%dielec_const
    alpha    = ene_info%pme_alpha
    alpha2m  = -alpha**2
    alpha2sp = 2.0_wp*alpha/sqrt(PI)

    ! set parameters for table
    !
    cutoff   = ene_info%cutoffdist
    density  = ene_info%table_density

    cutoff2        = cutoff * cutoff
    cutoff2_water  = (cutoff+4.5_wp) * (cutoff+4.5_wp)
    cutoff2_water1 = (cutoff+1.5_wp) * (cutoff+1.5_wp)
    cutoff_int     = int(cutoff2*density)
    cutoff_int2    = int(cutoff2_water*density)
    cutoff_int1    = int(cutoff2_water1*density)

    if (enefunc%table%table_order == 1 .or. &
        enefunc%table%table_order == 0) then

      cutoff_int = int(2.1_wp*cutoff2_water1*density)
      cutoff_int2 = cutoff_int+1
      table_maxpoint = cutoff_int

    end if 

    if (enefunc%table%water_table) then
      call alloc_enefunc(enefunc, EneFuncTblWatDomain, cutoff_int2, 1)

    else
      call alloc_enefunc(enefunc, EneFuncTableDomain, cutoff_int2, 1)

    end if

    if (enefunc%pme_use) then

      if (enefunc%table%table_order == 3) then

        call setup_table_general(ene_info, cutoff2, density, &
                                 cutoff_int, cutoff_int2, enefunc)

        if (enefunc%table%water_table) then

          call setup_table_water(ene_info, cutoff2, density, &
                                 cutoff_int, cutoff_int2, enefunc)

        end if

      else if (enefunc%table%table_order == 1 .or. &
               enefunc%table%table_order == 0) then

        call setup_table_general_linear(ene_info, cutoff2, density, &
                                        cutoff_int, cutoff_int2, enefunc)

        if (enefunc%table%water_table) then

          call setup_table_water_linear(ene_info, cutoff2, density, &
                                        cutoff_int, cutoff_int2, enefunc)

        end if

      end if

    else

      if (enefunc%table%table_order == 3) then

        call setup_table_general_cutoff(ene_info, cutoff2, density, &
                                        cutoff_int, cutoff_int2, enefunc)

        if (enefunc%table%water_table) then

          call setup_table_water_cutoff(ene_info, cutoff2, density, &
                                        cutoff_int, cutoff_int2, enefunc)
        end if

      else if (enefunc%table%table_order == 1 .or. &
               enefunc%table%table_order == 0) then

         call error_msg( &
           'Setup_Enefunc_Table> Linear interpolation is not suitable')

      end if

    end if

    return

  end subroutine setup_enefunc_table

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_table_general
  !> @brief        define lookup table of potential energy function
  !! @authors      JJ     
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[in]    cutoff2     : square cutoff distance
  !! @param[in]    density     : table density
  !! @param[in]    cutoff_int  : integer cutoff distance
  !! @param[in]    cutoff_int1 : square integer cutoff distance
  !! @param[inout] enefunc     : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_table_general(ene_info, cutoff2, density, &
                                 cutoff_int, cutoff_int1, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    real(wp),                intent(in)    :: cutoff2
    real(wp),                intent(in)    :: density
    integer,                 intent(in)    :: cutoff_int
    integer,                 intent(in)    :: cutoff_int1
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    real(wp)                 :: rij, rij2
    real(wp)                 :: inv_rij, inv_r2, inv_r3, inv_r6, inv_r12
    real(wp)                 :: switch, dswitch, switchdist
    real(wp)                 :: switchdist2
    real(wp)                 :: lj1, lj2, lj4, lj12, lj6
    integer                  :: i, switch_int

    real(wp),        pointer :: table_ene(:), table_grad(:)
    real(wp),        pointer :: table_ecor(:), table_decor(:)
    real(wp),        pointer :: dtable_ecor(:), dtable_decor(:)

    real(wp),    allocatable :: elec_array(:), delec_array(:)
    real(wp),    allocatable :: ecor_array(:), decor_array(:)
    real(wp),    allocatable :: lj12_array(:), dlj12_array(:)
    real(wp),    allocatable :: lj6_array(:), dlj6_array(:)


    table_ene    => enefunc%table%table_ene
    table_ecor   => enefunc%table%table_ecor
    table_decor  => enefunc%table%table_decor
    dtable_ecor  => enefunc%table%dtable_ecor
    dtable_decor => enefunc%table%dtable_decor
    table_grad   => enefunc%table%table_grad

    ! set switch distance etc.
    !
    switchdist  = ene_info%switchdist

    ! set lennard-jones parameters
    !
    switchdist2 = switchdist * switchdist
    switch_int  = int(switchdist2*density)
 
    if (.not. ene_info%vdw_force_switch) then
      lj1 = 1.0_wp/(cutoff2-switchdist2)
      lj1 = lj1*lj1*lj1
    end if

    allocate(elec_array(cutoff_int), delec_array(cutoff_int))
    allocate(ecor_array(cutoff_int), decor_array(cutoff_int))
    allocate(lj12_array(cutoff_int), dlj12_array(cutoff_int))
    allocate(lj6_array (cutoff_int), dlj6_array (cutoff_int))

    table_ene(1:6*cutoff_int1)  = 0.0_wp
    table_ecor(1:cutoff_int1)   = 0.0_wp
    table_decor(1:cutoff_int1)  = 0.0_wp
    dtable_ecor(1:cutoff_int1)  = 0.0_wp
    dtable_decor(1:cutoff_int1) = 0.0_wp
    table_grad(1:6*cutoff_int1) = 0.0_wp

    if (ene_info%vdw_force_switch) then

      lj1 = 1.0_wp / (cutoff2*switchdist2)
      lj1 = lj1*lj1*lj1
      lj2 = sqrt(lj1)

      do i = 2, switch_int

        rij2    = real(i,wp)/density
        rij     = sqrt(rij2)
        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        table_ene(6*i-5)  = inv_r12 - lj1
        table_ene(6*i-3)  = inv_r6 - lj2
        table_grad(6*i-5) = -12.0_wp*inv_r12*inv_r2
        table_grad(6*i-3) = -6.0_wp*inv_r6*inv_r2
        table_ecor(i)     = -erf_n(alpha*rij)*inv_rij*el_fact
        table_ene(6*i-1)  = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(6*i-1) = -(table_ene(6*i-1)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2
        table_decor(i)    = -(table_ecor(i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))/rij2

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
        ecor_array(i-1)  = table_ecor(i)     - table_ecor(i-1)
        decor_array(i-1) = table_decor(i)    - table_decor(i-1)

      end do

      lj1  = cutoff2*cutoff2*cutoff2 
      lj2  = cutoff2*sqrt(cutoff2)
      lj12 = lj1 / (lj1-switchdist2*switchdist2*switchdist2)
      lj6  = lj2 / (lj2-switchdist2*sqrt(switchdist2)) 
      lj1  = 1.0_wp / lj1
      lj2  = 1.0_wp / lj2

      do i = switch_int+1, cutoff_int

        rij2    = real(i,wp)/density
        rij     = sqrt(rij2)
        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r3  = sqrt(inv_r6)

        table_ene(6*i-5)  = lj12*(inv_r6-lj1)**2
        table_ene(6*i-3)  = lj6*(inv_r3-lj2)**2
        table_grad(6*i-5) = -12.0_wp*lj12*(inv_r6-lj1)*inv_r6*inv_r2
        table_grad(6*i-3) = -6.0_wp*lj6*(inv_r3-lj2)*inv_r3*inv_r2
        table_ene(6*i-1)  = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(6*i-1) = -(table_ene(6*i-1)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        ecor_array(i-1)  = table_ecor(i)     - table_ecor(i-1)
        decor_array(i-1) = table_decor(i)    - table_decor(i-1)
      end do

    else

      do i = 2, switch_int

        rij2     = real(i,wp)/density
        rij      = sqrt(rij2)
        switch   = 1.0_wp
        dswitch  = 0.0_wp
        inv_r2   = 1.0_wp / rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r12  = inv_r6 * inv_r6

        table_ene(6*i-5)  = switch*inv_r12
        table_ene(6*i-3)  = switch*inv_r6
        table_grad(6*i-5) = dswitch*inv_r12*inv_rij - switch*12.0_wp*inv_r12*inv_r2
        table_grad(6*i-3) = dswitch*inv_r6*inv_rij  - switch*6.0_wp*inv_r6*inv_r2
        table_ecor(i)     = -erf_n(alpha*rij)*inv_rij*el_fact
        table_ene(6*i-1)  = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(6*i-1) = -(table_ene(6*i-1)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2
        table_decor(i)    = -(table_ecor(i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))/rij2

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
        ecor_array(i-1)  = table_ecor(i)     - table_ecor(i-1)
        decor_array(i-1) = table_decor(i)    - table_decor(i-1)

      end do

      do i = switch_int+1, cutoff_int

        rij2     = real(i,wp)/density
        rij      = sqrt(rij2)
        lj2      = cutoff2 - rij2
        lj4      = lj2*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
        switch   = lj2*lj4*lj1
        dswitch  = 4.0_wp*lj1*rij  *(lj2*lj2 - lj4)
        inv_r2   = 1.0_wp / rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r12  = inv_r6 * inv_r6

        table_ene(6*i-5)  = switch*inv_r12
        table_ene(6*i-3)  = switch*inv_r6
        table_grad(6*i-5) = dswitch*inv_r12*inv_rij - switch*12.0_wp*inv_r12*inv_r2
        table_grad(6*i-3) = dswitch*inv_r6*inv_rij  - switch*6.0_wp*inv_r6*inv_r2
        table_ene(6*i-1)  = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(6*i-1) = -(table_ene(6*i-1)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        ecor_array(i-1)  = table_ecor(i)     - table_ecor(i-1)
        decor_array(i-1) = table_decor(i)    - table_decor(i-1)
      end do

    end if

    table_ecor(cutoff_int)   = 0.0_wp
    table_decor(cutoff_int)  = 0.0_wp
    dtable_ecor(cutoff_int)  = 0.0_wp
    dtable_decor(cutoff_int) = 0.0_wp
    table_ene(6*cutoff_int-5:6*cutoff_int)  = 0.0_wp
    table_grad(6*cutoff_int-5:6*cutoff_int) = 0.0_wp

    do i = 2, cutoff_int-1
      table_ene(6*i-4)  = (lj12_array(i-1) + lj12_array(i))   / 2.0_wp
      table_grad(6*i-4) = (dlj12_array(i-1) + dlj12_array(i)) / 2.0_wp
      table_ene(6*i-2)  = (lj6_array(i-1) + lj6_array(i))     / 2.0_wp
      table_grad(6*i-2) = (dlj6_array(i-1) + dlj6_array(i))   / 2.0_wp
      table_ene(6*i)    = (elec_array(i-1) + elec_array(i))   / 2.0_wp
      table_grad(6*i)   = (delec_array(i-1) + delec_array(i)) / 2.0_wp
      dtable_ecor(i)    = (ecor_array(i-1) + ecor_array(i))   / 2.0_wp
      dtable_decor(i)   = (decor_array(i-1) + decor_array(i)) / 2.0_wp
    end do

    table_ene(2)               = lj12_array(1)
    table_ene(4)               = lj6_array(1)
    table_ene(6)               = elec_array(1)
    table_ene(6*cutoff_int-4)  = lj12_array(cutoff_int-1)
    table_ene(6*cutoff_int-2)  = lj6_array(cutoff_int-1)
    table_ene(6*cutoff_int)    = elec_array(cutoff_int-1)
    table_grad(2)              = dlj12_array(1)
    table_grad(4)              = dlj6_array(1)
    table_grad(6)              = delec_array(1)
    table_grad(6*cutoff_int-4) = dlj12_array(cutoff_int-1)
    table_grad(6*cutoff_int-2) = dlj6_array(cutoff_int-1)
    table_grad(6*cutoff_int)   = delec_array(cutoff_int-1)

    deallocate(elec_array, delec_array)
    deallocate(ecor_array, decor_array)
    deallocate(lj12_array, dlj12_array)
    deallocate(lj6_array, dlj6_array)

    return

  end subroutine setup_table_general

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_table_general_linear
  !> @brief        define lookup table of potential energy function
  !!               (linear)
  !! @authors      JJ     
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[in]    cutoff2     : square cutoff distance
  !! @param[in]    density     : table density
  !! @param[in]    cutoff_int  : integer cutoff distance
  !! @param[in]    cutoff_int1 : square integer cutoff distance
  !! @param[inout] enefunc     : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_table_general_linear(ene_info, cutoff2, density, &
                                        cutoff_int, cutoff_int1, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    real(wp),                intent(in)    :: cutoff2
    real(wp),                intent(in)    :: density
    integer,                 intent(in)    :: cutoff_int
    integer,                 intent(in)    :: cutoff_int1
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    real(wp)                 :: rij, rij2
    real(wp)                 :: inv_r2, inv_rij, inv_r6, inv_r3, inv_r12
    real(wp)                 :: switch, dswitch, switchdist
    real(wp)                 :: switchdist2
    real(wp)                 :: lj1, lj2, lj4, dist, lj12, lj6
    integer                  :: i, switch_int

    real(wp),        pointer :: table_ene(:), table_grad(:)
    real(wp),        pointer :: table_ecor(:), table_decor(:)


    table_ene   => enefunc%table%table_ene
    table_ecor  => enefunc%table%table_ecor
    table_decor => enefunc%table%table_decor
    table_grad  => enefunc%table%table_grad

    ! set switch distance etc.
    !
    switchdist  = ene_info%switchdist

    ! set lennard-jones parameters
    !
    switchdist2 = switchdist * switchdist
    switch_int  = int(1.0_wp/switchdist2*cutoff2*density)

    table_ene  (1:3*cutoff_int1) = 0.0_wp
    table_ecor (1:cutoff_int1)   = 0.0_wp
    table_decor(1:cutoff_int1)   = 0.0_wp
    table_grad (1:3*cutoff_int1) = 0.0_wp

    if (.not. ene_info%vdw_force_switch) then
      lj1 = 1.0_wp/(cutoff2-switchdist2)
      lj1 = lj1*lj1*lj1
    end if

    if (ene_info%vdw_force_switch) then

      lj1 = 1.0_wp / (cutoff2*switchdist2)
      lj1 = lj1*lj1*lj1
      lj2 = sqrt(lj1)

      do i = switch_int+1, cutoff_int
        rij2    = cutoff2*density/real(i,wp)
        rij     = sqrt(rij2)
        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        table_ene(3*i-2)  = inv_r12 - lj1
        table_ene(3*i-1)  = inv_r6 - lj2
        table_grad(3*i-2) = -12.0_wp*inv_r12*inv_r2
        table_grad(3*i-1) = -6.0_wp*inv_r6*inv_r2
        table_ene(3*i)    = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(3*i)   = -(table_ene(3*i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2
        table_ecor(i)     = -erf_n(alpha*rij)*inv_rij*el_fact
        table_decor(i)    = -(table_ecor(i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))/rij2
      end do

      lj1  = cutoff2*cutoff2*cutoff2
      lj2  = cutoff2*sqrt(cutoff2)
      lj12 = lj1 / (lj1-switchdist2*switchdist2*switchdist2)
      lj6  = lj2 / (lj2-switchdist2*sqrt(switchdist2))
      lj1  = 1.0_wp / lj1
      lj2  = 1.0_wp / lj2

      do i = density, switch_int

        rij2    = cutoff2*density/real(i,wp)
        rij     = sqrt(rij2)
        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r3  = sqrt(inv_r6)

        table_ene(3*i-2)  = lj12*(inv_r6-lj1)**2
        table_ene(3*i-1)  = lj6*(inv_r3-lj2)**2
        table_grad(3*i-2) = -12.0_wp*lj12*(inv_r6-lj1)*inv_r6*inv_r2
        table_grad(3*i-1) = -6.0_wp*lj6*(inv_r3-lj2)*inv_r3*inv_r2
        table_ene(3*i)    = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(3*i)   = -(table_ene(3*i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2
      end do

    else

      do i = switch_int+1, cutoff_int
        rij2    = cutoff2*density/real(i,wp)
        rij     = sqrt(rij2)
        switch  = 1.0_wp
        dswitch = 0.0_wp
        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        table_ene(3*i-2)  = switch*inv_r12
        table_ene(3*i-1)  = switch*inv_r6
        table_grad(3*i-2) = dswitch*inv_r12*inv_rij - switch*12.0_wp*inv_r12*inv_r2
        table_grad(3*i-1) = dswitch*inv_r6*inv_rij  - switch*6.0_wp*inv_r6*inv_r2
        table_ene(3*i)    = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(3*i)   = -(table_ene(3*i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2
        table_ecor(i)     = -erf_n(alpha*rij)*inv_rij*el_fact
        table_decor(i)    = -(table_ecor(i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))/rij2
      end do

      do i = density, switch_int

        rij2     = cutoff2*density/real(i,wp)
        rij      = sqrt(rij2)
        lj2      = cutoff2 - rij2
        lj4      = lj2*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
        switch   = lj2*lj4*lj1
        dswitch  = 4.0_wp*lj1*rij  *(lj2*lj2 - lj4)
        inv_r2   = 1.0_wp / rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r12  = inv_r6 * inv_r6
  
        table_ene(3*i-2)  = switch*inv_r12
        table_ene(3*i-1)  = switch*inv_r6
        table_grad(3*i-2) = dswitch*inv_r12*inv_rij - switch*12.0_wp*inv_r12*inv_r2
        table_grad(3*i-1) = dswitch*inv_r6*inv_rij  - switch*6.0_wp*inv_r6*inv_r2
        table_ene(3*i)    = el_fact*erfc04(alpha*rij)*inv_rij
        table_grad(3*i)   = -(table_ene(3*i)+ &
                              alpha2sp*el_fact*exp(alpha2m*rij2))*inv_r2
      end do

    end if

    return

  end subroutine setup_table_general_linear

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_table_general_cutoff
  !> @brief        define lookup table for cutoff
  !! @authors      JJ     
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[in]    cutoff2     : square cutoff distance
  !! @param[in]    density     : table density
  !! @param[in]    cutoff_int  : integer cutoff distance
  !! @param[in]    cutoff_int1 : square integer cutoff distance
  !! @param[inout] enefunc     : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_table_general_cutoff(ene_info, cutoff2, density, &
                                        cutoff_int, cutoff_int1, enefunc)

    ! formal arguments
    type(s_ene_info),        intent(in)    :: ene_info
    real(wp),                intent(in)    :: cutoff2
    real(wp),                intent(in)    :: density
    integer,                 intent(in)    :: cutoff_int
    integer,                 intent(in)    :: cutoff_int1
    type(s_enefunc), target, intent(inout) :: enefunc

    ! local variables
    real(wp)                 :: rij, rij2
    real(wp)                 :: inv_r2, inv_rij, inv_r3, inv_r6, inv_r12
    real(wp)                 :: switch, dswitch, switchdist
    real(wp)                 :: switchdist2, cutoff2_inv, shift, dshift
    real(wp)                 :: lj1, lj2, lj4, el_fact, lj12, lj6
    integer                  :: i, switch_int

    real(wp),        pointer :: table_ene(:)
    real(wp),        pointer :: table_grad(:)

    real(wp),    allocatable :: elec_array(:), delec_array(:)
    real(wp),    allocatable :: lj12_array(:), dlj12_array(:)
    real(wp),    allocatable :: lj6_array(:), dlj6_array(:)


    table_ene  => enefunc%table%table_ene
    table_grad => enefunc%table%table_grad

    cutoff2_inv = 1.0_wp/cutoff2
    el_fact = ELECOEF/enefunc%dielec_const

    ! set switch distance etc.
    !
    switchdist  = ene_info%switchdist

    ! set lennard-jones parameters
    !
    switchdist2 = switchdist * switchdist
    switch_int  = int(switchdist2*density)

    if (.not. ene_info%vdw_force_switch) then
      lj1 = 1.0_wp/(cutoff2-switchdist2)
      lj1 = lj1*lj1*lj1
    end if

    allocate(elec_array(cutoff_int1), delec_array(cutoff_int1))
    allocate(lj12_array(cutoff_int1), dlj12_array(cutoff_int1))
    allocate(lj6_array (cutoff_int1), dlj6_array (cutoff_int1))

    table_ene(1:6*cutoff_int1)  = 0.0_wp
    table_grad(1:6*cutoff_int1) = 0.0_wp

    if (ene_info%vdw_force_switch) then

      lj1 = 1.0_wp / (cutoff2*switchdist2)
      lj1 = lj1*lj1*lj1
      lj2 = sqrt(lj1)

      do i = 2, switch_int
        rij2    = real(i,wp)/density
        rij     = sqrt(rij2)
        shift   = (1.0_wp - (rij2*cutoff2_inv))
        dshift  = -4.0_wp*shift*rij*cutoff2_inv
        shift   = shift * shift
        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        table_ene(6*i-5)  = inv_r12 - lj1
        table_ene(6*i-3)  = inv_r6 - lj2
        table_grad(6*i-5) = -12.0_wp*inv_r12*inv_r2
        table_grad(6*i-3) = -6.0_wp*inv_r6*inv_r2
        table_ene(6*i-1)  = el_fact * shift * inv_rij
        table_grad(6*i-1) = el_fact * (dshift*inv_rij - shift*inv_r2) * inv_rij

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
      end do

      lj1  = cutoff2*cutoff2*cutoff2
      lj2  = cutoff2*sqrt(cutoff2)
      lj12 = lj1 / (lj1-switchdist2*switchdist2*switchdist2)
      lj6  = lj2 / (lj2-switchdist2*sqrt(switchdist2))
      lj1  = 1.0_wp / lj1
      lj2  = 1.0_wp / lj2

      do i = switch_int+1, cutoff_int
        rij2     = real(i,wp)/density
        rij      = sqrt(rij2)
        shift    = (1.0_wp - (rij2*cutoff2_inv))
        dshift   = -4.0_wp*shift*rij*cutoff2_inv
        shift    = shift * shift
        inv_r2   = 1.0_wp / rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r3   = sqrt(inv_r6)

        table_ene(6*i-5)  = lj12*(inv_r6-lj1)**2
        table_ene(6*i-3)  = lj6*(inv_r3-lj2)**2
        table_grad(6*i-5) = -12.0_wp*lj12*(inv_r6-lj1)*inv_r6*inv_r2
        table_grad(6*i-3) = -6.0_wp*lj6*(inv_r3-lj2)*inv_r3*inv_r2
        table_ene(6*i-1)  = el_fact * shift * inv_rij
        table_grad(6*i-1) = el_fact * (dshift*inv_rij - shift*inv_r2) * inv_rij

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
      end do

    else

      do i = 2, switch_int

        rij2     = real(i,wp)/density
        rij      = sqrt(rij2)
        switch   = 1.0_wp
        dswitch  = 0.0_wp
        shift    = (1.0_wp - (rij2*cutoff2_inv))
        dshift   = -4.0_wp*shift*rij*cutoff2_inv
        shift    = shift * shift
        inv_r2   = 1.0_wp / rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r12  = inv_r6 * inv_r6
  
        table_ene(6*i-5)  = switch*inv_r12
        table_ene(6*i-3)  = switch*inv_r6
        table_grad(6*i-5) = dswitch*inv_r12*inv_rij - switch*12.0_wp*inv_r12*inv_r2
        table_grad(6*i-3) = dswitch*inv_r6*inv_rij  - switch*6.0_wp*inv_r6*inv_r2
        table_ene(6*i-1)  = el_fact * shift * inv_rij
        table_grad(6*i-1) = el_fact * (dshift*inv_rij - shift*inv_r2) * inv_rij

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
      end do

      do i = switch_int+1, cutoff_int

        rij2     = real(i,wp)/density
        rij      = sqrt(rij2)
        lj2      = cutoff2 - rij2
        lj4      = lj2*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
        switch   = lj2*lj4*lj1
        dswitch  = 4.0_wp*lj1*rij*(lj2*lj2 - lj4)
        shift    = (1.0_wp - (rij2*cutoff2_inv))
        dshift   = -4.0_wp*shift*rij*cutoff2_inv
        shift    = shift * shift
        inv_r2   = 1.0_wp / rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r12  = inv_r6 * inv_r6
  
        table_ene(6*i-5)  = switch*inv_r12
        table_ene(6*i-3)  = switch*inv_r6
        table_grad(6*i-5) = dswitch*inv_r12*inv_rij - switch*12.0_wp*inv_r12*inv_r2
        table_grad(6*i-3) = dswitch*inv_r6*inv_rij  - switch*6.0_wp*inv_r6*inv_r2
        table_ene(6*i-1)  = el_fact * shift * inv_rij
        table_grad(6*i-1) = el_fact * (dshift*inv_rij - shift*inv_r2) * inv_rij

        lj12_array(i-1)  = table_ene(6*i-5)  - table_ene(6*i-11)
        lj6_array(i-1)   = table_ene(6*i-3)  - table_ene(6*i-9)
        dlj12_array(i-1) = table_grad(6*i-5) - table_grad(6*i-11)
        dlj6_array(i-1)  = table_grad(6*i-3) - table_grad(6*i-9)
        elec_array(i-1)  = table_ene(6*i-1)  - table_ene(6*i-7)
        delec_array(i-1) = table_grad(6*i-1) - table_grad(6*i-7)
      end do

    end if

    table_ene (6*cutoff_int-5:6*cutoff_int) = 0.0_wp
    table_grad(6*cutoff_int-5:6*cutoff_int) = 0.0_wp

    do i = 2, cutoff_int-1
      table_ene(6*i-4)  = (lj12_array(i-1) + lj12_array(i))   / 2.0_wp
      table_grad(6*i-4) = (dlj12_array(i-1) + dlj12_array(i)) / 2.0_wp
      table_ene(6*i-2)  = (lj6_array(i-1) + lj6_array(i))     / 2.0_wp
      table_grad(6*i-2) = (dlj6_array(i-1) + dlj6_array(i))   / 2.0_wp
      table_ene(6*i)    = (elec_array(i-1) + elec_array(i))   / 2.0_wp
      table_grad(6*i)   = (delec_array(i-1) + delec_array(i)) / 2.0_wp
    end do

    table_ene(2)               = lj12_array(1)
    table_ene(4)               = lj6_array(1)
    table_ene(6)               = elec_array(1)
    table_ene(6*cutoff_int-4)  = lj12_array(cutoff_int-1)
    table_ene(6*cutoff_int-2)  = lj6_array(cutoff_int-1)
    table_ene(6*cutoff_int)    = elec_array(cutoff_int-1)
    table_grad(2)              = dlj12_array(1)
    table_grad(4)              = dlj6_array(1)
    table_grad(6)              = delec_array(1)
    table_grad(6*cutoff_int-4) = dlj12_array(cutoff_int-1)
    table_grad(6*cutoff_int-2) = dlj6_array(cutoff_int-1)
    table_grad(6*cutoff_int)   = delec_array(cutoff_int-1)

    deallocate(elec_array, delec_array)
    deallocate(lj12_array, dlj12_array)
    deallocate(lj6_array, dlj6_array)

    return

  end subroutine setup_table_general_cutoff

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_table_water
  !> @brief        define lookup table of water
  !! @authors      JJ     
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[in]    cutoff2     : square cutoff distance
  !! @param[in]    density     : table density
  !! @param[in]    cutoff_int  : integer cutoff distance
  !! @param[in]    cutoff_int1 : square integer cutoff distance
  !! @param[inout] enefunc     : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_table_water(ene_info, cutoff2, &
                               density, cutoff_int, cutoff_int1, enefunc)

    ! formal arguments
    type(s_ene_info),         intent(in)    :: ene_info
    real(wp),                 intent(in)    :: cutoff2
    real(wp),                 intent(in)    :: density
    integer,                  intent(in)    :: cutoff_int
    integer,                  intent(in)    :: cutoff_int1
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variables
    real(wp)                  :: lj6_OO, lj6_OH, lj6_HH
    real(wp)                  :: lj12_OO, lj12_OH, lj12_HH
    real(wp)                  :: term_lj12_OO, term_lj12_OH, term_lj12_HH
    real(wp)                  :: term_lj6_OO, term_lj6_OH, term_lj6_HH
    real(wp)                  :: term_lj_OO, term_lj_OH, term_lj_HH
    real(wp)                  :: cc_OO, cc_OH, cc_HH, charge_O, charge_H
    real(wp)                  :: rij, rij2
    real(wp)                  :: inv_rij, inv_r2, inv_r3, inv_r6, inv_r12
    real(wp)                  :: switch, dswitch, switchdist
    real(wp)                  :: switchdist2
    real(wp)                  :: lj1, lj2, lj4, lj12, lj6, term_lj12, term_lj6
    integer                   :: i, switch_int, atmcls_O, atmcls_H

    real(wp),         pointer :: table_de_WW(:,:)
    real(wp),         pointer :: table_ene_WW(:,:)

    real(wp),         allocatable :: elec_WW(:,:), evdw_WW(:,:), de_WW(:,:)


    table_ene_WW => enefunc%table%table_ene_WW
    table_de_WW  => enefunc%table%table_de_WW

    ! set switch distance etc.
    !
    switchdist   = ene_info%switchdist

    ! set lennard-jones parameters
    !
    switchdist2 = switchdist * switchdist
    lj1 = 1.0_wp/(cutoff2-switchdist2)
    lj1 = lj1*lj1*lj1
    switch_int = int(switchdist2*density)

    allocate(elec_WW(cutoff_int1,3), &
             evdw_WW(cutoff_int1,3), &
             de_WW  (cutoff_int1,3))

    table_ene_WW(1:6*cutoff_int1,1:3) = 0.0_wp
    table_de_WW(1:2*cutoff_int1,1:3)  = 0.0_wp

    ! atom class and charge of water oxygen and hydrogen
    !
    atmcls_O = enefunc%table%atom_cls_no_O
    atmcls_H = enefunc%table%atom_cls_no_H
    charge_O = enefunc%table%charge_O
    charge_H = enefunc%table%charge_H

    if (ene_info%vdw_force_switch) then

      lj1 = 1.0_wp / (cutoff2*switchdist2)
      lj1 = lj1*lj1*lj1
      lj2 = sqrt(lj1)

      do i = 2, switch_int
 
        rij2   = real(i,wp)/density
        rij    = sqrt(rij2)

        lj6_OO = enefunc%nonb_lj6(atmcls_O,atmcls_O)
        lj6_OH = enefunc%nonb_lj6(atmcls_O,atmcls_H)
        lj6_HH = enefunc%nonb_lj6(atmcls_H,atmcls_H)

        lj12_OO = enefunc%nonb_lj12(atmcls_O,atmcls_O)
        lj12_OH = enefunc%nonb_lj12(atmcls_O,atmcls_H)
        lj12_HH = enefunc%nonb_lj12(atmcls_H,atmcls_H)

        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        term_lj12_OO = lj12_OO * (inv_r12-lj1)
        term_lj12_OH = lj12_OH * (inv_r12-lj1)
        term_lj12_HH = lj12_HH * (inv_r12-lj1)
        term_lj6_OO   = lj6_OO * (inv_r6-lj2)
        term_lj6_OH   = lj6_OH * (inv_r6-lj2)
        term_lj6_HH   = lj6_HH * (inv_r6-lj2)

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = term_lj_OO
        table_ene_WW(6*i-5,2) = term_lj_OH
        table_ene_WW(6*i-5,3) = term_lj_HH

        term_lj12_OO = lj12_OO * inv_r12
        term_lj12_OH = lj12_OH * inv_r12
        term_lj12_HH = lj12_HH * inv_r12
        term_lj6_OO  = lj6_OO * inv_r6
        term_lj6_OH  = lj6_OH * inv_r6
        term_lj6_HH  = lj6_HH * inv_r6

        table_de_WW(2*i-1,1)   = -(12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO)*inv_r2
        table_de_WW(2*i-1,2)   = -(12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH)*inv_r2
        table_de_WW(2*i-1,3)   = -(12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH)*inv_r2

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-3,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,3) = cc_HH*erfc04(alpha*rij)*inv_rij

        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1)   &
                             - (table_ene_WW(6*i-3,1) &
                             + alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2)   &
                             - (table_ene_WW(6*i-3,2) &
                             + alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3)   &
                             - (table_ene_WW(6*i-3,3) &
                             + alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2

        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)

        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW(i-1,1:3)   = table_de_WW (2*i-1,1:3) - table_de_WW (2*i-3,1:3)

      end do
 
      lj1  = cutoff2*cutoff2*cutoff2
      lj2  = cutoff2*sqrt(cutoff2)
      lj12 = lj1 / (lj1-switchdist2*switchdist2*switchdist2)
      lj6  = lj2 / (lj2-switchdist2*sqrt(switchdist2))
      lj1  = 1.0_wp / lj1
      lj2  = 1.0_wp / lj2

      do i = switch_int+1, cutoff_int

        rij2      = real(i,wp)/density
        rij       = sqrt(rij2)
        inv_r2    = 1.0_wp / rij2
        inv_rij   = 1.0_wp / rij
        inv_r6    = inv_r2 * inv_r2 * inv_r2
        inv_r3    = sqrt(inv_r6)
        term_lj12 = lj12*(inv_r6-lj1)**2
        term_lj6  = lj6*(inv_r3-lj2)**2

        term_lj12_OO = lj12_OO * term_lj12
        term_lj12_OH = lj12_OH * term_lj12
        term_lj12_HH = lj12_HH * term_lj12
        term_lj6_OO  = lj6_OO * term_lj6
        term_lj6_OH  = lj6_OH * term_lj6
        term_lj6_HH  = lj6_HH * term_lj6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = term_lj_OO
        table_ene_WW(6*i-5,2) = term_lj_OH
        table_ene_WW(6*i-5,3) = term_lj_HH

        term_lj12 = -12.0_wp*lj12*(inv_r6-lj1)*inv_r6*inv_r2
        term_lj6  = -6.0_wp*lj6*(inv_r3-lj2)*inv_r3*inv_r2

        term_lj12_OO = lj12_OO * term_lj12
        term_lj12_OH = lj12_OH * term_lj12
        term_lj12_HH = lj12_HH * term_lj12
        term_lj6_OO  = lj6_OO * term_lj6
        term_lj6_OH  = lj6_OH * term_lj6
        term_lj6_HH  = lj6_HH * term_lj6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH
 
        table_de_WW(2*i-1,1) = term_lj_OO
        table_de_WW(2*i-1,2) = term_lj_OH
        table_de_WW(2*i-1,3) = term_lj_HH

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-3,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,3) = cc_HH*erfc04(alpha*rij)*inv_rij

        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1)   &
                             - (table_ene_WW(6*i-3,1) &
                             + alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2)   &
                             - (table_ene_WW(6*i-3,2) &
                             + alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3)   &
                             - (table_ene_WW(6*i-3,3) &
                             + alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2

        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)

        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW(i-1,1:3)   = table_de_WW (2*i-1,1:3) - table_de_WW (2*i-3,1:3)

      end do

    else

      do i = 2, switch_int
  
        rij2    = real(i,wp)/density
        rij     = sqrt(rij2)
        switch  = 1.0_wp
        dswitch = 0.0_wp

        lj6_OO  = enefunc%nonb_lj6(atmcls_O,atmcls_O)
        lj6_OH  = enefunc%nonb_lj6(atmcls_O,atmcls_H)
        lj6_HH  = enefunc%nonb_lj6(atmcls_H,atmcls_H)

        lj12_OO = enefunc%nonb_lj12(atmcls_O,atmcls_O)
        lj12_OH = enefunc%nonb_lj12(atmcls_O,atmcls_H)
        lj12_HH = enefunc%nonb_lj12(atmcls_H,atmcls_H)

        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        term_lj12_OO = lj12_OO * inv_r12
        term_lj12_OH = lj12_OH * inv_r12
        term_lj12_HH = lj12_HH * inv_r12
        term_lj6_OO  = lj6_OO * inv_r6
        term_lj6_OH  = lj6_OH * inv_r6
        term_lj6_HH  = lj6_HH * inv_r6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = switch*term_lj_OO
        table_ene_WW(6*i-5,2) = switch*term_lj_OH
        table_ene_WW(6*i-5,3) = switch*term_lj_HH

        table_de_WW(2*i-1,1) = (dswitch * term_lj_OO * rij  &
                             - switch * (12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO)) * inv_r2
        table_de_WW(2*i-1,2) = (dswitch * term_lj_OH * rij  &
                             - switch * (12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH)) * inv_r2
        table_de_WW(2*i-1,3) = (dswitch * term_lj_HH * rij  &
                             - switch * (12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH)) * inv_r2

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-3,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,3) = cc_HH*erfc04(alpha*rij)*inv_rij

        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1)   &
                             - (table_ene_WW(6*i-3,1) &
                             + alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2)   &
                             - (table_ene_WW(6*i-3,2) &
                             + alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3)   &
                             - (table_ene_WW(6*i-3,3) &
                             + alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2

        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)

        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW(i-1,1:3)   = table_de_WW(2*i-1,1:3)  - table_de_WW (2*i-3,1:3)

      end do

      do i = switch_int+1, cutoff_int

        rij2     = real(i,wp)/density
        rij      = sqrt(rij2)
        lj2      = cutoff2 - rij2
        lj4      = lj2*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
        switch   = lj2*lj4*lj1
        dswitch  = 4.0_wp*lj1*rij*(lj2*lj2 - lj4)
        inv_r2   = 1.0_wp / rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r12  = inv_r6 * inv_r6

        term_lj12_OO = lj12_OO * inv_r12
        term_lj12_OH = lj12_OH * inv_r12
        term_lj12_HH = lj12_HH * inv_r12
        term_lj6_OO  = lj6_OO * inv_r6
        term_lj6_OH  = lj6_OH * inv_r6
        term_lj6_HH  = lj6_HH * inv_r6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_ene_WW(6*i-5,1) = switch*term_lj_OO
        table_ene_WW(6*i-5,2) = switch*term_lj_OH
        table_ene_WW(6*i-5,3) = switch*term_lj_HH

        table_de_WW(2*i-1,1) = (dswitch * term_lj_OO * rij &
                             - switch * (12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO)) * inv_r2
        table_de_WW(2*i-1,2) = (dswitch * term_lj_OH * rij &
                             - switch * (12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH)) * inv_r2
        table_de_WW(2*i-1,3) = (dswitch * term_lj_HH * rij &
                             - switch * (12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH)) * inv_r2

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-3,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-3,3) = cc_HH*erfc04(alpha*rij)*inv_rij

        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1)   &
                             - (table_ene_WW(6*i-3,1) &
                             + alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2)   &
                             - (table_ene_WW(6*i-3,2) &
                             + alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3)   &
                             - (table_ene_WW(6*i-3,3) &
                             + alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2

        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)

        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW(i-1,1:3)   = table_de_WW (2*i-1,1:3) - table_de_WW (2*i-3,1:3)

      end do

    end if

    table_ene_WW(6*cutoff_int-1,1:3) = 0.0_wp
    table_ene_WW(6*cutoff_int-3,1:3) = 0.0_wp
    table_ene_WW(6*cutoff_int-5,1:3) = 0.0_wp
    table_de_WW(2*cutoff_int-1,1:3)  = 0.0_wp
    table_ene_WW(6*cutoff_int,1:3)   = 0.0_wp
    table_ene_WW(6*cutoff_int-2,1:3) = 0.0_wp
    table_ene_WW(6*cutoff_int-4,1:3) = 0.0_wp
    table_de_WW(2*cutoff_int,1:3)    = 0.0_wp

    do i = 2, cutoff_int-1

      table_ene_WW(6*i-2,1:3) = (elec_WW(i-1,1:3) + elec_WW(i,1:3)) / 2.0_wp
      table_ene_WW(6*i-4,1:3) = (evdw_WW(i-1,1:3) + evdw_WW(i,1:3)) / 2.0_wp
      table_de_WW(2*i,1:3)    = (de_WW(i-1,1:3) + de_WW(i,1:3)) / 2.0_wp
      table_ene_WW(6*i,1:3)   = table_de_WW(2*i,1:3)

    end do

    table_ene_WW(4,1:3) = elec_WW(1,1:3)
    table_ene_WW(2,1:3) = evdw_WW(1,1:3)
    table_de_WW(2,1:3)  = de_WW(1,1:3)
    table_ene_WW(6,1:3) = de_WW(1,1:3)

    table_ene_WW(6*cutoff_int-2,1:3) = elec_WW(cutoff_int-1,1:3)
    table_ene_WW(6*cutoff_int-4,1:3) = evdw_WW(cutoff_int-1,1:3)
    table_de_WW(2*cutoff_int,1:3)    = de_WW(cutoff_int-1,1:3)
    table_ene_WW(6*cutoff_int,1:3)   = de_WW(cutoff_int-1,1:3)

    deallocate(elec_WW, evdw_WW, de_WW)

    return

  end subroutine setup_table_water

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_table_water_linear
  !> @brief        define lookup table of water (linear interpolation)
  !! @authors      JJ     
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[in]    cutoff2     : square cutoff distance
  !! @param[in]    density     : table density
  !! @param[in]    cutoff_int  : integer cutoff distance
  !! @param[in]    cutoff_int1 : square integer cutoff distance
  !! @param[inout] enefunc     : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_table_water_linear(ene_info, cutoff2, &
                                      density, cutoff_int, cutoff_int1, enefunc)

    ! formal arguments
    type(s_ene_info),         intent(in)    :: ene_info
    real(wp),                 intent(in)    :: cutoff2
    real(wp),                 intent(in)    :: density
    integer,                  intent(in)    :: cutoff_int
    integer,                  intent(in)    :: cutoff_int1
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variables
    real(wp)                  :: lj6_OO, lj6_OH, lj6_HH
    real(wp)                  :: lj12_OO, lj12_OH, lj12_HH
    real(wp)                  :: term_lj12_OO, term_lj12_OH, term_lj12_HH
    real(wp)                  :: term_lj6_OO, term_lj6_OH, term_lj6_HH
    real(wp)                  :: term_lj_OO, term_lj_OH, term_lj_HH
    real(wp)                  :: cc_OO, cc_OH, cc_HH, charge_O, charge_H
    real(wp)                  :: rij, rij2
    real(wp)                  :: inv_r2, inv_rij, inv_r3, inv_r6, inv_r12
    real(wp)                  :: switch, dswitch, switchdist
    real(wp)                  :: switchdist2
    real(wp)                  :: lj1, lj2, lj4, lj12, lj6, term_lj12, term_lj6
    integer                   :: i, switch_int, atmcls_O, atmcls_H

    real(wp),         pointer :: table_de_WW(:,:)
    real(wp),         pointer :: table_ene_WW(:,:)


    table_ene_WW => enefunc%table%table_ene_WW
    table_de_WW  => enefunc%table%table_de_WW

    ! set switch distance etc.
    !
    switchdist  = ene_info%switchdist

    ! set lennard-jones parameters
    !
    switchdist2 = switchdist * switchdist
    lj1 = 1.0_wp/(cutoff2-switchdist2)
    lj1 = lj1*lj1*lj1
    switch_int = int(1.0_wp/switchdist2*cutoff2*density)

    table_ene_WW(1:6*cutoff_int1,1:3) = 0.0_wp
    table_de_WW(1:2*cutoff_int1,1:3)  = 0.0_wp

    ! atom class and charge of water oxygen and hydrogen
    !
    atmcls_O = enefunc%table%atom_cls_no_O
    atmcls_H = enefunc%table%atom_cls_no_H
    charge_O = enefunc%table%charge_O
    charge_H = enefunc%table%charge_H

    if (ene_info%vdw_force_switch) then

      lj1 = 1.0_wp / (cutoff2*switchdist2)
      lj1 = lj1*lj1*lj1
      lj2 = sqrt(lj1)

      do i = switch_int+1, cutoff_int
  
        rij2     = cutoff2*density/real(i,wp)
        rij    = sqrt(rij2)

        lj6_OO = enefunc%nonb_lj6(atmcls_O,atmcls_O)
        lj6_OH = enefunc%nonb_lj6(atmcls_O,atmcls_H)
        lj6_HH = enefunc%nonb_lj6(atmcls_H,atmcls_H)

        lj12_OO = enefunc%nonb_lj12(atmcls_O,atmcls_O)
        lj12_OH = enefunc%nonb_lj12(atmcls_O,atmcls_H)
        lj12_HH = enefunc%nonb_lj12(atmcls_H,atmcls_H)

        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        term_lj12_OO = lj12_OO * (inv_r12-lj1)
        term_lj12_OH = lj12_OH * (inv_r12-lj1)
        term_lj12_HH = lj12_HH * (inv_r12-lj1)
        term_lj6_OO  = lj6_OO * (inv_r6-lj2)
        term_lj6_OH  = lj6_OH * (inv_r6-lj2)
        term_lj6_HH  = lj6_HH * (inv_r6-lj2)

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH
 
        table_ene_WW(6*i-5,1) = term_lj_OO
        table_ene_WW(6*i-5,2) = term_lj_OH
        table_ene_WW(6*i-5,3) = term_lj_HH

        term_lj12_OO = lj12_OO * inv_r12
        term_lj12_OH = lj12_OH * inv_r12
        term_lj12_HH = lj12_HH * inv_r12
        term_lj6_OO  = lj6_OO * inv_r6
        term_lj6_OH  = lj6_OH * inv_r6
        term_lj6_HH  = lj6_HH * inv_r6

        table_de_WW(i,1)   = -(12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO)*inv_r2
        table_de_WW(i,2)   = -(12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH)*inv_r2
        table_de_WW(i,3)   = -(12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH)*inv_r2
  
        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact
  
        table_ene_WW(6*i-4,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,3) = cc_HH*erfc04(alpha*rij)*inv_rij

        table_de_WW(i,1)   = table_de_WW(i,1)                      &
                           - (table_ene_WW(6*i-4,1)                &
                           + alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,2)   = table_de_WW(i,2)                      &
                           - (table_ene_WW(6*i-4,2)                &
                           + alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,3)   = table_de_WW(i,3)                      &
                           - (table_ene_WW(6*i-4,3)                &
                           +alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2

        table_ene_WW(6*i-3,1:3) = table_de_WW(i,1:3)

      end do

      lj1  = cutoff2*cutoff2*cutoff2
      lj2  = cutoff2*sqrt(cutoff2)
      lj12 = lj1 / (lj1-switchdist2*switchdist2*switchdist2)
      lj6  = lj2 / (lj2-switchdist2*sqrt(switchdist2))
      lj1  = 1.0_wp / lj1
      lj2  = 1.0_wp / lj2

      do i = density, switch_int
        rij2      = cutoff2*density/real(i,wp)
        rij       = sqrt(rij2)
        inv_r2    = 1.0_wp / rij2
        inv_rij   = 1.0_wp / rij
        inv_r6    = inv_r2 * inv_r2 * inv_r2
        inv_r3    = sqrt(inv_r6)
        term_lj12 = lj12*(inv_r6-lj1)**2
        term_lj6  = lj6*(inv_r3-lj2)**2
  
        term_lj12_OO = lj12_OO * term_lj12
        term_lj12_OH = lj12_OH * term_lj12
        term_lj12_HH = lj12_HH * term_lj12
        term_lj6_OO  = lj6_OO * term_lj6
        term_lj6_OH  = lj6_OH * term_lj6
        term_lj6_HH  = lj6_HH * term_lj6
  
        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH
  
        table_ene_WW(6*i-5,1) = term_lj_OO
        table_ene_WW(6*i-5,2) = term_lj_OH
        table_ene_WW(6*i-5,3) = term_lj_HH
 
        term_lj12 = -12.0_wp*lj12*(inv_r6-lj1)*inv_r6*inv_r2
        term_lj6  = -6.0_wp*lj6*(inv_r3-lj2)*inv_r3*inv_r2

        term_lj12_OO = lj12_OO * term_lj12
        term_lj12_OH = lj12_OH * term_lj12
        term_lj12_HH = lj12_HH * term_lj12
        term_lj6_OO  = lj6_OO * term_lj6
        term_lj6_OH  = lj6_OH * term_lj6
        term_lj6_HH  = lj6_HH * term_lj6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH
 
        table_de_WW(i,1)   = term_lj_OO
        table_de_WW(i,2)   = term_lj_OH
        table_de_WW(i,3)   = term_lj_HH

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-4,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,3) = cc_HH*erfc04(alpha*rij)*inv_rij
  
        table_de_WW(i,1) = table_de_WW(i,1) &
                  - (table_ene_WW(6*i-4,1)+alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,2) = table_de_WW(i,2) &
                 - (table_ene_WW(6*i-4,2)+alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,3) = table_de_WW(i,3) &
                 - (table_ene_WW(6*i-4,3)+alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2
  
        table_ene_WW(6*i-3,1:3) = table_de_WW(i,1:3)

      end do

      do i = density, cutoff_int
        table_ene_WW(6*i-2,1:3) = table_ene_WW(6*(i+1)-5,1:3) - &
                                  table_ene_WW(6*i-5,1:3)
        table_ene_WW(6*i-1,1:3) = table_ene_WW(6*(i+1)-4,1:3) - &
                                  table_ene_WW(6*i-4,1:3)
        table_ene_WW(6*i,1:3)   = table_ene_WW(6*(i+1)-3,1:3) - &
                                  table_ene_WW(6*i-3,1:3)
      end do

    else

      do i = switch_int+1, cutoff_int
  
        rij2     = cutoff2*density/real(i,wp)
        rij    = sqrt(rij2)
        switch  = 1.0_wp
        dswitch = 0.0_wp

        lj6_OO = enefunc%nonb_lj6(atmcls_O,atmcls_O)
        lj6_OH = enefunc%nonb_lj6(atmcls_O,atmcls_H)
        lj6_HH = enefunc%nonb_lj6(atmcls_H,atmcls_H)

        lj12_OO = enefunc%nonb_lj12(atmcls_O,atmcls_O)
        lj12_OH = enefunc%nonb_lj12(atmcls_O,atmcls_H)
        lj12_HH = enefunc%nonb_lj12(atmcls_H,atmcls_H)

        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        term_lj12_OO = lj12_OO * inv_r12
        term_lj12_OH = lj12_OH * inv_r12
        term_lj12_HH = lj12_HH * inv_r12
        term_lj6_OO  = lj6_OO * inv_r6
        term_lj6_OH  = lj6_OH * inv_r6
        term_lj6_HH  = lj6_HH * inv_r6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH
  
        table_ene_WW(6*i-5,1) = switch*term_lj_OO
        table_ene_WW(6*i-5,2) = switch*term_lj_OH
        table_ene_WW(6*i-5,3) = switch*term_lj_HH

        table_de_WW(i,1)   = (  dswitch * term_lj_OO * rij  &
                           - switch * (12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO) ) * inv_r2
        table_de_WW(i,2)   = (  dswitch * term_lj_OH * rij  &
                           - switch * (12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH) ) * inv_r2
        table_de_WW(i,3)   = (  dswitch * term_lj_HH * rij  &
                           - switch * (12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH) ) * inv_r2
  
        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact
  
        table_ene_WW(6*i-4,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,3) = cc_HH*erfc04(alpha*rij)*inv_rij

        table_de_WW(i,1)   = table_de_WW(i,1)       &
                           - (table_ene_WW(6*i-4,1) &
                           + alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,2)   = table_de_WW(i,2)       &
                           - (table_ene_WW(6*i-4,2) &
                           + alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,3)   = table_de_WW(i,3)       &
                           - (table_ene_WW(6*i-4,3) &
                           +alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2

        table_ene_WW(6*i-3,1:3) = table_de_WW(i,1:3)

      end do

      do i = density, switch_int
        rij2     = cutoff2*density/real(i,wp)
        rij      = sqrt(rij2)
        lj2      = cutoff2 - rij2
        lj4      = lj2*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
        switch   = lj2*lj4*lj1
        dswitch  = 4.0_wp*lj1*rij*(lj2*lj2 - lj4)
        inv_r2   = 1.0_wp/rij2
        inv_rij  = 1.0_wp / rij
        inv_r6   = inv_r2 * inv_r2 * inv_r2
        inv_r12  = inv_r6 * inv_r6
  
        term_lj12_OO  = lj12_OO * inv_r12
        term_lj12_OH  = lj12_OH * inv_r12
        term_lj12_HH  = lj12_HH * inv_r12
        term_lj6_OO   = lj6_OO * inv_r6
        term_lj6_OH   = lj6_OH * inv_r6
        term_lj6_HH   = lj6_HH * inv_r6
  
        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH
  
        table_ene_WW(6*i-5,1) = switch*term_lj_OO
        table_ene_WW(6*i-5,2) = switch*term_lj_OH
        table_ene_WW(6*i-5,3) = switch*term_lj_HH

        table_de_WW(i,1)   = (  dswitch * term_lj_OO * rij     &
                           - switch * (12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO) ) * inv_r2
        table_de_WW(i,2)   = (  dswitch * term_lj_OH * rij     &
                           - switch * (12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH) ) * inv_r2
        table_de_WW(i,3)   = (  dswitch * term_lj_HH * rij     &
                           - switch * (12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH) ) * inv_r2

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact

        table_ene_WW(6*i-4,1) = cc_OO*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,2) = cc_OH*erfc04(alpha*rij)*inv_rij
        table_ene_WW(6*i-4,3) = cc_HH*erfc04(alpha*rij)*inv_rij
  
        table_de_WW(i,1)   = table_de_WW(i,1) &
                - (table_ene_WW(6*i-4,1)+alpha2sp*cc_OO*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,2)   = table_de_WW(i,2) &
                - (table_ene_WW(6*i-4,2)+alpha2sp*cc_OH*exp(alpha2m*rij2))*inv_r2
        table_de_WW(i,3)   = table_de_WW(i,3) &
                - (table_ene_WW(6*i-4,3)+alpha2sp*cc_HH*exp(alpha2m*rij2))*inv_r2
  
        table_ene_WW(6*i-3,1:3) = table_de_WW(i,1:3)

      end do

      do i = density, cutoff_int
        table_ene_WW(6*i-2,1:3) = table_ene_WW(6*(i+1)-5,1:3) - &
                                  table_ene_WW(6*i-5,1:3)
        table_ene_WW(6*i-1,1:3) = table_ene_WW(6*(i+1)-4,1:3) - &
                                  table_ene_WW(6*i-4,1:3)
        table_ene_WW(6*i,1:3)   = table_ene_WW(6*(i+1)-3,1:3) - &
                                  table_ene_WW(6*i-3,1:3)
      end do

    end if

    return

  end subroutine setup_table_water_linear

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_table_water_cutoff
  !> @brief        define lookup table of water for cutoff
  !! @authors      JJ     
  !! @param[in]    ene_info    : ENERGY section control parameters information
  !! @param[in]    cutoff2     : square cutoff distance
  !! @param[in]    density     : table density
  !! @param[in]    cutoff_int  : integer cutoff distance
  !! @param[in]    cutoff_int1 : square integer cutoff distance
  !! @param[inout] enefunc     : potential energy functions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_table_water_cutoff(ene_info, cutoff2, &
                                      density, cutoff_int, cutoff_int1, enefunc)

    ! formal arguments
    type(s_ene_info),         intent(in)    :: ene_info
    real(wp),                 intent(in)    :: cutoff2
    real(wp),                 intent(in)    :: density
    integer,                  intent(in)    :: cutoff_int
    integer,                  intent(in)    :: cutoff_int1
    type(s_enefunc),  target, intent(inout) :: enefunc

    ! local variables
    real(wp)                  :: lj6_OO, lj6_OH, lj6_HH
    real(wp)                  :: lj12_OO, lj12_OH, lj12_HH
    real(wp)                  :: term_lj12_OO, term_lj12_OH, term_lj12_HH
    real(wp)                  :: term_lj6_OO, term_lj6_OH, term_lj6_HH
    real(wp)                  :: term_lj_OO, term_lj_OH, term_lj_HH
    real(wp)                  :: cc_OO, cc_OH, cc_HH, charge_O, charge_H
    real(wp)                  :: rij, rij2
    real(wp)                  :: inv_r2, inv_rij, inv_r3, inv_r6, inv_r12
    real(wp)                  :: switch, dswitch, switchdist
    real(wp)                  :: switchdist2, cutoff2_inv, shift, dshift
    real(wp)                  :: lj1, lj2, lj4, el_fact, lj12, lj6
    real(wp)                  :: term_lj12, term_lj6
    integer                   :: i, switch_int, atmcls_O, atmcls_H

    real(wp),         pointer :: table_de_WW(:,:)
    real(wp),         pointer :: table_ene_WW(:,:)

    real(wp),         allocatable :: elec_WW(:,:), evdw_WW(:,:), de_WW(:,:)


    table_ene_WW => enefunc%table%table_ene_WW
    table_de_WW  => enefunc%table%table_de_WW

    cutoff2_inv = 1.0_wp / cutoff2
    el_fact = ELECOEF / enefunc%dielec_const

    ! set switch distance etc.
    !
    switchdist  = ene_info%switchdist

    ! set lennard-jones parameters
    !
    switchdist2 = switchdist * switchdist
    switch_int  = int(switchdist2*density)

    if (.not.ene_info%vdw_force_switch) then
      lj1 = 1.0_wp/(cutoff2-switchdist2)
      lj1 = lj1*lj1*lj1
    end if

    allocate(elec_WW(cutoff_int1,3), &
             evdw_WW(cutoff_int1,3), &
             de_WW  (cutoff_int1,3))

    table_ene_WW(1:6*cutoff_int1,1:3) = 0.0_wp
    table_de_WW(1:2*cutoff_int1,1:3)  = 0.0_wp

    ! atom class and charge of water oxygen and hydrogen
    !
    atmcls_O = enefunc%table%atom_cls_no_O
    atmcls_H = enefunc%table%atom_cls_no_H
    charge_O = enefunc%table%charge_O
    charge_H = enefunc%table%charge_H

    if (ene_info%vdw_force_switch) then

      lj1 = 1.0_wp / (cutoff2*switchdist2)
      lj1 = lj1*lj1*lj1
      lj2 = sqrt(lj1)

      do i = 2, switch_int
  
        rij2    = real(i,wp)/density
        rij     = sqrt(rij2)
        shift   = (1.0_wp - (rij2*cutoff2_inv))
        dshift  = -4.0_wp*shift*rij*cutoff2_inv
        shift   = shift * shift

        lj6_OO = enefunc%nonb_lj6(atmcls_O,atmcls_O)
        lj6_OH = enefunc%nonb_lj6(atmcls_O,atmcls_H)
        lj6_HH = enefunc%nonb_lj6(atmcls_H,atmcls_H)
  
        lj12_OO = enefunc%nonb_lj12(atmcls_O,atmcls_O)
        lj12_OH = enefunc%nonb_lj12(atmcls_O,atmcls_H)
        lj12_HH = enefunc%nonb_lj12(atmcls_H,atmcls_H)
  
        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6
  
        term_lj12_OO = lj12_OO * (inv_r12-lj1)
        term_lj12_OH = lj12_OH * (inv_r12-lj1)
        term_lj12_HH = lj12_HH * (inv_r12-lj1)
        term_lj6_OO  = lj6_OO * (inv_r6-lj2)
        term_lj6_OH  = lj6_OH * (inv_r6-lj2)
        term_lj6_HH  = lj6_HH * (inv_r6-lj2)
  
        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH
  
        table_ene_WW(6*i-5,1) = term_lj_OO
        table_ene_WW(6*i-5,2) = term_lj_OH
        table_ene_WW(6*i-5,3) = term_lj_HH
 
        term_lj12_OO = lj12_OO * inv_r12
        term_lj12_OH = lj12_OH * inv_r12
        term_lj12_HH = lj12_HH * inv_r12
        term_lj6_OO  = lj6_OO * inv_r6
        term_lj6_OH  = lj6_OH * inv_r6
        term_lj6_HH  = lj6_HH * inv_r6
 
        table_de_WW(2*i-1,1)   = -(12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO)*inv_r2
        table_de_WW(2*i-1,2)   = -(12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH)*inv_r2
        table_de_WW(2*i-1,3)   = -(12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH)*inv_r2
  
        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact
  
        table_ene_WW(6*i-3,1) = cc_OO * inv_rij * shift
        table_ene_WW(6*i-3,2) = cc_OH * inv_rij * shift
        table_ene_WW(6*i-3,3) = cc_HH * inv_rij * shift
  
        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1) &
                             + cc_OO * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2) &
                             + cc_OH * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3) &
                             + cc_HH * (dshift*inv_rij - shift*inv_r2) * inv_rij
  
        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)
  
        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW(i-1,1:3)   = table_de_WW(2*i-1,1:3)   - table_de_WW(2*i-3,1:3)

      end do

      lj1  = cutoff2*cutoff2*cutoff2
      lj2  = cutoff2*sqrt(cutoff2)
      lj12 = lj1 / (lj1-switchdist2*switchdist2*switchdist2)
      lj6  = lj2 / (lj2-switchdist2*sqrt(switchdist2))
      lj1  = 1.0_wp / lj1
      lj2  = 1.0_wp / lj2

      do i = switch_int+1, cutoff_int

        rij2      = real(i,wp)/density
        rij       = sqrt(rij2)
        shift     = (1.0_wp - (rij2*cutoff2_inv))
        dshift    = -4.0_wp*shift*rij*cutoff2_inv
        shift     = shift * shift
        inv_r2    = 1.0_wp / rij2
        inv_rij   = 1.0_wp / rij
        inv_r6    = inv_r2 * inv_r2 * inv_r2
        inv_r3    = sqrt(inv_r6)
        term_lj12 = lj12*(inv_r6-lj1)**2
        term_lj6  = lj6*(inv_r3-lj2)**2

        term_lj12_OO = lj12_OO * term_lj12
        term_lj12_OH = lj12_OH * term_lj12
        term_lj12_HH = lj12_HH * term_lj12
        term_lj6_OO  = lj6_OO * term_lj6
        term_lj6_OH  = lj6_OH * term_lj6
        term_lj6_HH  = lj6_HH * term_lj6
  
        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH
  
        table_ene_WW(6*i-5,1) = term_lj_OO
        table_ene_WW(6*i-5,2) = term_lj_OH
        table_ene_WW(6*i-5,3) = term_lj_HH
 
        term_lj12 = -12.0_wp*lj12*(inv_r6-lj1)*inv_r6*inv_r2
        term_lj6  = -6.0_wp*lj6*(inv_r3-lj2)*inv_r3*inv_r2

        term_lj12_OO = lj12_OO * term_lj12
        term_lj12_OH = lj12_OH * term_lj12
        term_lj12_HH = lj12_HH * term_lj12
        term_lj6_OO  = lj6_OO * term_lj6
        term_lj6_OH  = lj6_OH * term_lj6
        term_lj6_HH  = lj6_HH * term_lj6

        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH

        table_de_WW(2*i-1,1) = term_lj_OO
        table_de_WW(2*i-1,2) = term_lj_OH
        table_de_WW(2*i-1,3) = term_lj_HH
 
        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact
  
        table_ene_WW(6*i-3,1) = cc_OO * inv_rij * shift
        table_ene_WW(6*i-3,2) = cc_OH * inv_rij * shift
        table_ene_WW(6*i-3,3) = cc_HH * inv_rij * shift
  
        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1) &
                             + cc_OO * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2) &
                             + cc_OH * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3) &
                             + cc_HH * (dshift*inv_rij - shift*inv_r2) * inv_rij
  
        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)
  
        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW  (i-1,1:3) = table_de_WW(2*i-1,1:3)  - table_de_WW (2*i-3,1:3)
  
      end do

    else

      do i = 2, switch_int
  
        rij2     = real(i,wp)/density
        rij    = sqrt(rij2)
        switch  = 1.0_wp
        dswitch = 0.0_wp
        shift   = (1.0_wp - (rij2*cutoff2_inv))
        dshift  = -4.0_wp*shift*rij*cutoff2_inv
        shift   = shift * shift

        lj6_OO  = enefunc%nonb_lj6(atmcls_O,atmcls_O)
        lj6_OH  = enefunc%nonb_lj6(atmcls_O,atmcls_H)
        lj6_HH  = enefunc%nonb_lj6(atmcls_H,atmcls_H)
  
        lj12_OO = enefunc%nonb_lj12(atmcls_O,atmcls_O)
        lj12_OH = enefunc%nonb_lj12(atmcls_O,atmcls_H)
        lj12_HH = enefunc%nonb_lj12(atmcls_H,atmcls_H)
  
        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6
  
        term_lj12_OO = lj12_OO * inv_r12
        term_lj12_OH = lj12_OH * inv_r12
        term_lj12_HH = lj12_HH * inv_r12
        term_lj6_OO  = lj6_OO * inv_r6
        term_lj6_OH  = lj6_OH * inv_r6
        term_lj6_HH  = lj6_HH * inv_r6
  
        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH
  
        table_ene_WW(6*i-5,1) = switch*term_lj_OO
        table_ene_WW(6*i-5,2) = switch*term_lj_OH
        table_ene_WW(6*i-5,3) = switch*term_lj_HH
  
        table_de_WW(2*i-1,1) = (dswitch * term_lj_OO * rij &
                             - switch * (12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO)) * inv_r2
        table_de_WW(2*i-1,2) = (dswitch * term_lj_OH * rij &
                             - switch * (12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH)) * inv_r2
        table_de_WW(2*i-1,3) = (dswitch * term_lj_HH * rij &
                             - switch * (12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH)) * inv_r2
  
        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact
  
        table_ene_WW(6*i-3,1) = cc_OO * inv_rij * shift
        table_ene_WW(6*i-3,2) = cc_OH * inv_rij * shift
        table_ene_WW(6*i-3,3) = cc_HH * inv_rij * shift
  
        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1) &
                             + cc_OO * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2) &
                             + cc_OH * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3) &
                             + cc_HH * (dshift*inv_rij - shift*inv_r2) * inv_rij
  
        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)
  
        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW(i-1,1:3)   = table_de_WW(2*i-1,1:3)  - table_de_WW(2*i-3,1:3)

      end do

      do i = switch_int+1, cutoff_int

        rij2     = real(i,wp)/density
        rij      = sqrt(rij2)
        lj2      = cutoff2 - rij2
        lj4      = lj2*(cutoff2 + 2.0_wp*rij2 - 3.0_wp*switchdist2)
        switch   = lj2*lj4*lj1
        dswitch  = 4.0_wp*lj1*rij*(lj2*lj2 - lj4)
        shift    = (1.0_wp - (rij2*cutoff2_inv))
        dshift   = -4.0_wp*shift*rij*cutoff2_inv
        shift    = shift * shift
  
        inv_r2  = 1.0_wp / rij2
        inv_rij = 1.0_wp / rij
        inv_r6  = inv_r2 * inv_r2 * inv_r2
        inv_r12 = inv_r6 * inv_r6

        term_lj12_OO = lj12_OO * inv_r12
        term_lj12_OH = lj12_OH * inv_r12
        term_lj12_HH = lj12_HH * inv_r12
        term_lj6_OO  = lj6_OO * inv_r6
        term_lj6_OH  = lj6_OH * inv_r6
        term_lj6_HH  = lj6_HH * inv_r6
  
        term_lj_OO = term_lj12_OO - term_lj6_OO
        term_lj_OH = term_lj12_OH - term_lj6_OH
        term_lj_HH = term_lj12_HH - term_lj6_HH
  
        table_ene_WW(6*i-5,1) = switch*term_lj_OO
        table_ene_WW(6*i-5,2) = switch*term_lj_OH
        table_ene_WW(6*i-5,3) = switch*term_lj_HH
  
        table_de_WW(2*i-1,1) = (dswitch * term_lj_OO * rij &
                             - switch * (12.0_wp*term_lj12_OO-6.0_wp*term_lj6_OO)) * inv_r2
        table_de_WW(2*i-1,2) = (dswitch * term_lj_OH * rij &
                             - switch * (12.0_wp*term_lj12_OH-6.0_wp*term_lj6_OH)) * inv_r2
        table_de_WW(2*i-1,3) = (dswitch * term_lj_HH * rij &
                             - switch * (12.0_wp*term_lj12_HH-6.0_wp*term_lj6_HH)) * inv_r2

        cc_OO = charge_O * charge_O * el_fact
        cc_OH = charge_O * charge_H * el_fact
        cc_HH = charge_H * charge_H * el_fact
  
        table_ene_WW(6*i-3,1) = cc_OO * inv_rij * shift
        table_ene_WW(6*i-3,2) = cc_OH * inv_rij * shift
        table_ene_WW(6*i-3,3) = cc_HH * inv_rij * shift
  
        table_de_WW(2*i-1,1) = table_de_WW(2*i-1,1) &
                             + cc_OO * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,2) = table_de_WW(2*i-1,2) &
                             + cc_OH * (dshift*inv_rij - shift*inv_r2) * inv_rij
        table_de_WW(2*i-1,3) = table_de_WW(2*i-1,3) &
                             + cc_HH * (dshift*inv_rij - shift*inv_r2) * inv_rij
  
        table_ene_WW(6*i-1,1:3) = table_de_WW(2*i-1,1:3)
  
        elec_WW(i-1,1:3) = table_ene_WW(6*i-3,1:3) - table_ene_WW(6*i-9,1:3)
        evdw_WW(i-1,1:3) = table_ene_WW(6*i-5,1:3) - table_ene_WW(6*i-11,1:3)
        de_WW  (i-1,1:3) = table_de_WW(2*i-1,1:3)  - table_de_WW (2*i-3,1:3)
  
      end do

    end if

    table_ene_WW(6*cutoff_int-5,1:3) = 0.0_wp
    table_ene_WW(6*cutoff_int-3,1:3) = 0.0_wp
    table_ene_WW(6*cutoff_int-1,1:3) = 0.0_wp
    table_de_WW(2*cutoff_int-1,1:3)  = 0.0_wp
    table_ene_WW(6*cutoff_int-4,1:3) = 0.0_wp
    table_ene_WW(6*cutoff_int-2,1:3) = 0.0_wp
    table_ene_WW(6*cutoff_int,1:3)   = 0.0_wp
    table_de_WW(2*cutoff_int,1:3)    = 0.0_wp

    do i = 2, cutoff_int-1

      table_ene_WW(6*i-2,1:3) = (elec_WW(i-1,1:3) + elec_WW(i,1:3)) / 2.0_wp
      table_ene_WW(6*i-4,1:3) = (evdw_WW(i-1,1:3) + evdw_WW(i,1:3)) / 2.0_wp
      table_de_WW(2*i,1:3)    = (de_WW(i-1,1:3) + de_WW(i,1:3)) / 2.0_wp
      table_ene_WW(6*i,1:3)   = table_de_WW(2*i,1:3)

    end do

    table_ene_WW(4,1:3) = elec_WW(1,1:3)
    table_ene_WW(2,1:3) = evdw_WW(1,1:3)
    table_de_WW (2,1:3) = de_WW(1,1:3)
    table_ene_WW(6,1:3) = evdw_WW(1,1:3)

    table_ene_WW(6*cutoff_int-2,1:3) = elec_WW(cutoff_int-1,1:3)
    table_ene_WW(6*cutoff_int-4,1:3) = elec_WW(cutoff_int-1,1:3)
    table_ene_WW(6*cutoff_int,1:3)   = de_WW(cutoff_int-1,1:3)
    table_de_WW (2*cutoff_int,1:3)   = de_WW(cutoff_int-1,1:3)

    deallocate(elec_WW, evdw_WW, de_WW)

    return

  end subroutine setup_table_water_cutoff

end module sp_enefunc_table_mod
