!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_energy_mod
!> @brief   compute energy
!! @authors Jaewoon Jung(JJ), Yuji Sugita (YS), Takaharu Mori (TM),
!!          Chigusa Kobayashi (CK), Norio Takase (NT)
!  
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_energy_mod

  use sp_energy_nonbonds_mod
  use sp_energy_pme_mod
  use sp_energy_restraints_mod
  use sp_energy_dihedrals_mod
  use sp_energy_angles_mod
  use sp_energy_bonds_mod
  use sp_boundary_str_mod
  use sp_pairlist_str_mod
  use sp_enefunc_str_mod
  use sp_energy_str_mod
  use sp_domain_str_mod
  use fileio_control_mod
  use timers_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  integer, parameter      :: FakeDefault      = 0

  ! structures
  type, public :: s_ene_info
    integer               :: forcefield       = ForcefieldCHARMM
    integer               :: electrostatic    = ElectrostaticPME
    real(wp)              :: switchdist       = 10.0_wp
    real(wp)              :: cutoffdist       = 12.0_wp
    real(wp)              :: pairlistdist     = 13.5_wp
    real(wp)              :: dielec_const     = 1.0_wp
    logical               :: vdw_force_switch = .false.
    logical               :: cmap_pspline     = .false.
    real(wp)              :: pme_alpha        = 0.34_wp
    integer               :: pme_ngrid_x      = FakeDefault
    integer               :: pme_ngrid_y      = FakeDefault
    integer               :: pme_ngrid_z      = FakeDefault
    integer               :: pme_nspline      = 4
    logical               :: table            = .true.
    integer               :: table_order      = 1
    real(wp)              :: table_density    = 20.0_wp
    character(5)          :: water_model      = 'TIP3'
    integer               :: output_style     = OutputStyleGENESIS
  end type s_ene_info

  ! varibles
  logical, save           :: etitle = .true.

  ! subroutines
  public  :: show_ctrl_energy
  public  :: read_ctrl_energy
  public  :: compute_energy
  public  :: compute_energy_short
  public  :: compute_energy_long 
  public  :: output_energy
  private :: compute_energy_charmm
  private :: compute_energy_charmm_short
  private :: compute_energy_charmm_long
  private :: output_energy_genesis
  private :: output_energy_charmm
  private :: output_energy_namd
  private :: reduce_ene

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_energy
  !> @brief        show ENERGY section usage
  !! @authors      NT
  !! @param[in]    show_all : show all usage or not
  !! @param[in]    run_mode : run mode string : "md", "min"
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_energy(show_all, run_mode)

    ! formal arguments
    logical,                 intent(in)    :: show_all
    character(*),            intent(in)    :: run_mode


    if (show_all) then

      select case (run_mode)

      case ('md', 'min', 'remd')

        write(MsgOut,'(A)') '[ENERGY]'
        write(MsgOut,'(A)') 'forcefield    = CHARMM    # [CHARMM]'
        write(MsgOut,'(A)') 'electrostatic = PME       # [CUTOFF,PME]'
        write(MsgOut,'(A)') 'switchdist    = 10.0       # switch distance'
        write(MsgOut,'(A)') 'cutoffdist    = 12.0      # cutoff distance'
        write(MsgOut,'(A)') 'pairlistdist  = 13.5      # pair-list distance'
        write(MsgOut,'(A)') '# dielec_const  = 1.0       # dielectric constant'
        write(MsgOut,'(A)') '# vdw_force_switch = NO     # force switch option for van der Waals'
        write(MsgOut,'(A)') '# pme_alpha     = 0.34      # width of Gaussian distribution in [PME]'
        write(MsgOut,'(A,i3,A)') 'pme_ngrid_x   = ', FakeDefault, &
        '       # grid size_x in [PME] must be set'
        write(MsgOut,'(A,i3,A)') 'pme_ngrid_y   = ', FakeDefault, &
        '       # grid size_y in [PME] must be set'
        write(MsgOut,'(A,i3,A)') 'pme_ngrid_z   = ', FakeDefault, &
        '       # grid size_z in [PME] must be set'
        write(MsgOut,'(A)') '# pme_nspline   = 4         # order of B-spline in [PME]'
        write(MsgOut,'(A)') '# table         = YES       # usage of lookup table for [PME]'
        write(MsgOut,'(A)') '# table_order   = 1         # order of lookup table'
        write(MsgOut,'(A)') '# table_density = 20.0      # number of bins used for lookup table'
        write(MsgOut,'(A)') '# output_style  = GENESIS   # format of energy output [GENESIS,CHARMM,NAMD]'
        write(MsgOut,'(A)') ' '

      end select

    else

      select case (run_mode)

      case ('md', 'min', 'remd')

        write(MsgOut,'(A)') '[ENERGY]'
        write(MsgOut,'(A)') 'forcefield    = CHARMM    # [CHARMM]'
        write(MsgOut,'(A)') 'electrostatic = PME       # [CUTOFF,PME]'
        write(MsgOut,'(A)') 'switchdist    = 10.0       # switch distance'
        write(MsgOut,'(A)') 'cutoffdist    = 12.0      # cutoff distance'
        write(MsgOut,'(A)') 'pairlistdist  = 13.5      # pair-list distance'
        write(MsgOut,'(A,i3,A)') 'pme_ngrid_x   = ', FakeDefault, &
        '       # grid size_x in [PME] must be set'
        write(MsgOut,'(A,i3,A)') 'pme_ngrid_y   = ', FakeDefault, &
        '       # grid size_y in [PME] must be set'
        write(MsgOut,'(A,i3,A)') 'pme_ngrid_z   = ', FakeDefault, &
        '       # grid size_z in [PME] must be set'
        write(MsgOut,'(A)') ' '


      end select

    end if

    return

  end subroutine show_ctrl_energy
  
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function      read_ctrl_energy
  !> @brief        read ENERGY section in the control file
  !! @authors      YS, TM, JJ
  !! @param[in]    handle   : unit number of control files
  !! @param[out]   ene_info : ENERGY section control parameters information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_ctrl_energy(handle, ene_info) 

    ! parameters
    character(*),            parameter     :: Section = 'Energy'

    ! formal arguments
    integer,                 intent(in)    :: handle
    type(s_ene_info),        intent(inout) :: ene_info 


    ! read parameters from control file
    ! 
    call read_ctrlfile_type   (handle, Section, 'forcefield',    &
                               ene_info%forcefield, ForceFieldTypes)
    call read_ctrlfile_type   (handle, Section, 'electrostatic', &
                               ene_info%electrostatic, ElectrostaticTypes)
    call read_ctrlfile_real   (handle, Section, 'switchdist',    &
                               ene_info%switchdist)
    call read_ctrlfile_real   (handle, Section, 'cutoffdist',    &
                               ene_info%cutoffdist)
    call read_ctrlfile_real   (handle, Section, 'pairlistdist',  &
                               ene_info%pairlistdist)
    call read_ctrlfile_real   (handle, Section, 'dielec_const',  &
                               ene_info%dielec_const)
    call read_ctrlfile_logical(handle, Section, 'vdw_force_switch', &
                               ene_info%vdw_force_switch)
    call read_ctrlfile_real   (handle, Section, 'pme_alpha',     &
                               ene_info%pme_alpha)
    call read_ctrlfile_integer(handle, Section, 'pme_ngrid_x',   &
                               ene_info%pme_ngrid_x)
    call read_ctrlfile_integer(handle, Section, 'pme_ngrid_y',   &
                               ene_info%pme_ngrid_y)
    call read_ctrlfile_integer(handle, Section, 'pme_ngrid_z',   &
                               ene_info%pme_ngrid_z)
    call read_ctrlfile_integer(handle, Section, 'pme_nspline',   &
                               ene_info%pme_nspline)
    call read_ctrlfile_logical(handle, Section, 'table',         &
                               ene_info%table)
    call read_ctrlfile_integer(handle, Section, 'table_order',   &
                               ene_info%table_order)
    call read_ctrlfile_real   (handle, Section, 'table_density', &
                               ene_info%table_density)
    call read_ctrlfile_string (handle, Section, 'water_model', &
                               ene_info%water_model)
    call read_ctrlfile_type   (handle, Section, 'output_style',  &
                               ene_info%output_style, OutputStyleTypes)

    ! write parameters to MsgOut
    !
    if (main_rank) then

      write(MsgOut,'(A)')  'Read_Ctrl_Energy> Parameters of Energy Calculations'
      write(MsgOut,'(A20,A10)')                            &
            '  forcefield      = ', trim(ForceFieldTypes(ene_info%forcefield))

      write(MsgOut,'(A20,F10.3,A20,F10.3)')                &
            '  switchdist      = ', ene_info%switchdist,   &
            '  cutoffdist      = ', ene_info%cutoffdist
      write(MsgOut,'(A20,F10.3,A20,F10.3)')                &
            '  pairlistdist    = ', ene_info%pairlistdist, &
            '  dielec_const    = ', ene_info%dielec_const

      if (ene_info%vdw_force_switch) then
        write(MsgOut,'(A)') '  force_switch    =        yes'
      else
        write(MsgOut,'(A)') '  force_switch    =         no'
      end if

      ! if PME
      if (ene_info%electrostatic == ElectrostaticPME) then

        write(MsgOut,'(A20,A10)')                            &
              '  electrostatic   = ',                        &
                               trim(ElectrostaticTypes(ene_info%electrostatic))
        write(MsgOut,'(A20,3I10)')                           &
              '  pme_ngrid(x,y,z)= ', ene_info%pme_ngrid_x   &
                                    , ene_info%pme_ngrid_y   &
                                    , ene_info%pme_ngrid_z
        write(MsgOut,'(A20,I10)')                            &
              '  pme_nspline     = ', ene_info%pme_nspline
        write(MsgOut,'(A20,F10.3,A20,F10.3)')                &
              '  pme_alpha       = ', ene_info%pme_alpha

      ! if CUTOFF
      else if (ene_info%electrostatic == ElectrostaticCutoff) then

        write(MsgOut,'(A20,A10)')                            &
              '  electrostatic   = ',                        &
              trim(ElectrostaticTypes(ene_info%electrostatic))

      end if

      if (ene_info%table) then
        write(MsgOut,'(A)') '  table           =        yes'
        write(MsgOut,'(A20,I10)')   '  table_order     = ', &
             ene_info%table_order
        write(MsgOut,'(A20,F10.3)') '  table_density   = ', &
             ene_info%table_density
      else
        write(MsgOut,'(A)') '  table           =         no'
      end if

      write(MsgOut,'(A20,A10)') '  water_model     = ',     &
                                trim(ene_info%water_model)

      write(MsgOut,'(A20,A10)')                             &
            '  output_style    = ',                         &
            trim(OutputStyleTypes(ene_info%output_style))

      write(MsgOut,'(A)') ' '
    end if

    ! error check
    !
    if (ene_info%switchdist >= ene_info%cutoffdist) then
      call error_msg( &
         'Read_Ctrl_Energy> switchdist must be less than cutoffdist')
    end if

    if (ene_info%cutoffdist >= ene_info%pairlistdist) then
      call error_msg( &
         'Read_Ctrl_Energy> cutoffdist must be less than pairlistdist')
    end if

    if (ene_info%pme_nspline < 3 ) then
      call error_msg( &
         'Read_Ctrl_Energy> "pme_nspline" is too small')

    else if (mod(ene_info%pme_nspline,2) == 1 ) then
      call error_msg( &
         'Read_Ctrl_Energy> "pme_nspline" should be even (in current version)')

    end if
    if (ene_info%electrostatic == ElectrostaticPME) then
      if (abs(ene_info%pme_ngrid_x - FakeDefault) < EPS) &
        call error_msg &
          ('Read_Ctrl_Energy> pme_ngrid_x is not specified in the control file')
      if (abs(ene_info%pme_ngrid_y - FakeDefault) < EPS) &
        call error_msg &
          ('Read_Ctrl_Energy> pme_ngrid_y is not specified in the control file')
      if (abs(ene_info%pme_ngrid_z - FakeDefault) < EPS) &
        call error_msg &
          ('Read_Ctrl_Energy> pme_ngrid_z is not specified in the control file')
    end if

    return
  
  end subroutine read_ctrl_energy

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy
  !> @brief        compute potential energy
  !! @authors      JJ
  !! @param[in]    domain     : domain information
  !! @param[in]    enefunc    : potential energy functions information
  !! @param[in]    pairlist   : pair list information
  !! @param[in]    boundary   : boundary information
  !! @param[in]    coord      : coordinates of target systems
  !! @param[in]    npt        : flag for NPT or not
  !! @param[in]    reduce     : flag for reduce energy and virial
  !! @param[in]    nonb_ene   : flag for calculate nonbonded energy
  !! @param[inout] energy     : energy information
  !! @param[inout] force      : forces of target systems
  !! @param[inout] force_omp  : temprary forces of target systems
  !! @param[inout] virial     : virial term of target systems
  !! @param[inout] virial_ext : extern virial term of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy(domain, enefunc, pairlist, boundary, coord,      &
                            npt, reduce, nonb_ene, energy, force, force_omp, &
                            virial, virial_ext)

    ! formal arguments
    type(s_domain),          intent(in)    :: domain
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_pairlist),        intent(in)    :: pairlist
    type(s_boundary),        intent(in)    :: boundary
    real(wp),                intent(in)    :: coord(:,:,:)
    logical,                 intent(in)    :: npt
    logical,                 intent(in)    :: reduce
    logical,                 intent(in)    :: nonb_ene
    type(s_energy),          intent(inout) :: energy
    real(wp),                intent(inout) :: force(:,:,:)
    real(wp),                intent(inout) :: force_omp(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3)
    real(wp),                intent(inout) :: virial_ext(3,3)


    call timer(TimerEnergy, TimerOn)


    select case (enefunc%forcefield)

    case (ForcefieldCHARMM)

      call compute_energy_charmm( &
                            domain, enefunc, pairlist, boundary, coord,      &
                            npt, reduce, nonb_ene, energy, force, force_omp, &
                            virial, virial_ext)

    end select


    call timer(TimerEnergy, TimerOff)

    return

  end subroutine compute_energy

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_short
  !> @brief        compute potential energy
  !! @authors      JJ
  !! @param[in]    domain     : domain information
  !! @param[in]    enefunc    : potential energy functions information
  !! @param[in]    pairlist   : pair list information
  !! @param[in]    boundary   : boundary information
  !! @param[in]    coord      : coordinates of target systems
  !! @param[in]    nonb_ene   : flag for calculate nonbonded energy
  !! @param[inout] energy     : energy information
  !! @param[inout] force      : forces of target systems
  !! @param[inout] force_omp  : temprary forces of target systems
  !! @param[inout] virial     : virial term of target systems
  !! @param[inout] virial_ext : extern virial term of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_short(domain, enefunc, pairlist, boundary, coord, &
                                  nonb_ene, energy, force, force_omp,         &
                                  virial, virial_ext)

    ! formal arguments
    type(s_domain),          intent(in)    :: domain
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_pairlist),        intent(in)    :: pairlist
    type(s_boundary),        intent(in)    :: boundary
    real(wp),                intent(in)    :: coord(:,:,:)
    logical,                 intent(in)    :: nonb_ene
    type(s_energy),          intent(inout) :: energy
    real(wp),                intent(inout) :: force(:,:,:)
    real(wp),                intent(inout) :: force_omp(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3)
    real(wp),                intent(inout) :: virial_ext(3,3)


    call timer(TimerEnergy, TimerOn)


    select case (enefunc%forcefield)

    case (ForcefieldCHARMM)

      call compute_energy_charmm_short( &
                                  domain, enefunc, pairlist, boundary, coord, &
                                  nonb_ene, energy, force, force_omp,         &
                                  virial, virial_ext)

    end select


    call timer(TimerEnergy, TimerOff)

    return

  end subroutine compute_energy_short

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_long
  !> @brief        compute potential energy
  !! @authors      JJ
  !! @param[in]    domain     : domain information
  !! @param[in]    enefunc    : potential energy functions information
  !! @param[in]    pairlist   : pair list information
  !! @param[in]    boundary   : boundary information
  !! @param[in]    coord      : coordinates of target systems
  !! @param[in]    npt        : flag for NPT or not
  !! @param[inout] energy     : energy information
  !! @param[inout] force      : forces of target systems
  !! @param[inout] force_omp  : temprary forces of target systems
  !! @param[inout] virial     : virial term of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_long(domain, enefunc, pairlist, boundary, coord,  &
                                 npt, energy, force, force_omp, virial)

    ! formal arguments
    type(s_domain),          intent(in)    :: domain
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_pairlist),        intent(in)    :: pairlist
    type(s_boundary),        intent(in)    :: boundary
    real(wp),                intent(in)    :: coord(:,:,:)
    logical,                 intent(in)    :: npt
    type(s_energy),          intent(inout) :: energy
    real(wp),                intent(inout) :: force(:,:,:)
    real(wp),                intent(inout) :: force_omp(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3)


    call timer(TimerEnergy, TimerOn)


    select case (enefunc%forcefield)

    case (ForcefieldCHARMM)

      call compute_energy_charmm_long( &
                                 domain, enefunc, pairlist, boundary, coord,  &
                                 npt, energy, force, force_omp, virial)

    end select


    call timer(TimerEnergy, TimerOff)

    return

  end subroutine compute_energy_long

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    output_energy
  !> @brief        output energy
  !! @authors      YS
  !! @param[in]    step    : step 
  !! @param[in]    enefunc : potential energy functions information
  !! @param[in]    energy  : energy information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_energy(step, enefunc, energy)

    ! formal arguments
    integer,                 intent(in)    :: step
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_energy),          intent(in)    :: energy


    if (.not. main_rank) return

    select case (enefunc%output_style)

    case (OutputStyleGENESIS)

      call output_energy_genesis(step, enefunc, energy)

    case (OutputStyleCHARMM)

      call output_energy_charmm(step, enefunc, energy)

    case (OutputStyleNAMD)

      call output_energy_namd(step, energy)

    end select

    return

  end subroutine output_energy

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_charmm
  !> @brief        compute potential energy with charmm force field
  !! @authors      JJ
  !! @param[in]    domain     : domain information
  !! @param[in]    enefunc    : potential energy functions information
  !! @param[in]    pairlist   : pair list information
  !! @param[in]    boundary   : boundary information
  !! @param[in]    coord      : coordinates of target systems
  !! @param[in]    npt        : flag for NPT or not
  !! @param[in]    reduce     : flag for reduce energy and virial
  !! @param[in]    nonb_ene   : flag for calculate nonbonded energy
  !! @param[inout] energy     : energy information
  !! @param[inout] force      : forces of target systems
  !! @param[inout] force_omp  : temprary forces of target systems
  !! @param[inout] virial     : virial term of target systems
  !! @param[inout] virial_ext : extern virial term of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_charmm(domain, enefunc, pairlist, boundary, coord, &
                              npt, reduce, nonb_ene, energy, force, force_omp, &
                              virial, virial_ext)

    ! formal arguments
    type(s_domain),          intent(in)    :: domain
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_pairlist),        intent(in)    :: pairlist
    type(s_boundary),        intent(in)    :: boundary
    real(wp),                intent(in)    :: coord(:,:,:)
    logical,                 intent(in)    :: npt
    logical,                 intent(in)    :: reduce
    logical,                 intent(in)    :: nonb_ene
    type(s_energy),          intent(inout) :: energy
    real(wp),                intent(inout) :: force(:,:,:)
    real(wp),                intent(inout) :: force_omp(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3)
    real(wp),                intent(inout) :: virial_ext(3,3)

    ! local variable
    real(wp)                 :: virial_omp(3,3,nthread)
    real(wp)                 :: virial_ext_omp(3,3,nthread)
    real(wp)                 :: elec_omp   (nthread)
    real(wp)                 :: evdw_omp   (nthread)
    real(wp)                 :: ebond_omp  (nthread)
    real(wp)                 :: eangle_omp (nthread)
    real(wp)                 :: eurey_omp  (nthread)
    real(wp)                 :: edihed_omp (nthread)
    real(wp)                 :: eimprop_omp(nthread)
    real(wp)                 :: ecmap_omp  (nthread)
    real(wp)                 :: eposi_omp  (nthread)
    integer                  :: ncell, natom, id, i, ix


    ! number of cells and atoms
    !
    ncell = domain%num_cell_local + domain%num_cell_boundary
    natom = domain%max_num_atom

    ! initialization of energy and forces
    !
    call init_energy(energy)

    force     (1:3,1:natom,1:ncell) = 0.0_wp
    virial    (1:3,1:3)             = 0.0_wp
    virial_ext(1:3,1:3)             = 0.0_wp

    force_omp     (1:3,1:natom,1:ncell,1:nthread) = 0.0_wp
    virial_omp    (1:3,1:3,1:nthread) = 0.0_wp
    virial_ext_omp(1:3,1:3,1:nthread) = 0.0_wp
    ebond_omp     (1:nthread) = 0.0_wp
    eangle_omp    (1:nthread) = 0.0_wp
    eurey_omp     (1:nthread) = 0.0_wp
    edihed_omp    (1:nthread) = 0.0_wp
    eimprop_omp   (1:nthread) = 0.0_wp
    ecmap_omp     (1:nthread) = 0.0_wp
    elec_omp      (1:nthread) = 0.0_wp
    evdw_omp      (1:nthread) = 0.0_wp
    eposi_omp     (1:nthread) = 0.0_wp
 
    if (real_calc) then
    
      ! bond energy
      !
      call compute_energy_bond(domain, enefunc, coord, &
                              force_omp, virial_omp, ebond_omp)

      ! angle energy
      !
      call compute_energy_angle(domain, enefunc, coord, &
                              force_omp, virial_omp, eangle_omp, eurey_omp)

      ! dihedral energy
      !
      if (enefunc%local_restraint) then
        call compute_energy_dihed_localres(domain, enefunc, coord, &
                              force_omp, virial_omp, edihed_omp)
      else
        call compute_energy_dihed(domain, enefunc, coord, &
                              force_omp, virial_omp, edihed_omp)
      end if
  
      ! improper energy
      !
      call compute_energy_improp(domain, enefunc, coord, &
                              force_omp, virial_omp, eimprop_omp)
 
      ! cmap energy
      ! 
      call compute_energy_cmap(domain, enefunc, coord, &
                              force_omp, virial_omp, ecmap_omp)

      ! restraint energy
      !
      if (enefunc%restraint) &
        call compute_energy_restraints(.true., .true., domain, enefunc, coord, &
                              force_omp, virial_omp, virial_ext_omp,           &
                              eposi_omp, energy%restraint_distance)
    end if

    ! nonbonded energy
    !   type 4: nonbond_ene_pbc
    !   type 5: nonbond_ene_pme
    !
    
    if (enefunc%pme_use) then   

      call compute_energy_nonbond_pme(domain, enefunc, pairlist, boundary, &
                              npt, nonb_ene, &
                              force_omp, virial_omp, elec_omp, evdw_omp)
    else

      call compute_energy_nonbond_cutoff(domain, enefunc, pairlist, &
                              nonb_ene, &
                              force_omp, virial_omp, elec_omp, evdw_omp)
    end if

    ! gather values
    !
    do id = 1, nthread

      do i = 1, ncell
        do ix = 1, domain%num_atom(i)
          force(1,ix,i) = force(1,ix,i) + force_omp(1,ix,i,id)
          force(2,ix,i) = force(2,ix,i) + force_omp(2,ix,i,id)
          force(3,ix,i) = force(3,ix,i) + force_omp(3,ix,i,id)
        end do
      end do

      virial    (1:3,1:3) = virial    (1:3,1:3) + virial_omp    (1:3,1:3,id)
      virial_ext(1:3,1:3) = virial_ext(1:3,1:3) + virial_ext_omp(1:3,1:3,id)

      energy%bond               = energy%bond               + ebond_omp(id)
      energy%angle              = energy%angle              + eangle_omp(id)
      energy%urey_bradley       = energy%urey_bradley       + eurey_omp(id)
      energy%dihedral           = energy%dihedral           + edihed_omp(id)
      energy%improper           = energy%improper           + eimprop_omp(id)
      energy%cmap               = energy%cmap               + ecmap_omp(id)
      energy%electrostatic      = energy%electrostatic      + elec_omp(id)
      energy%van_der_waals      = energy%van_der_waals      + evdw_omp(id)
      energy%restraint_position = energy%restraint_position + eposi_omp(id)

    end do
   
    ! total energy
    !
    energy%total = energy%bond          &
                 + energy%angle         &
                 + energy%urey_bradley  &
                 + energy%dihedral      &
                 + energy%cmap          &
                 + energy%improper      &
                 + energy%electrostatic &
                 + energy%van_der_waals &
                 + energy%restraint_position

    if (reduce) &
      call reduce_ene(energy, virial)

    return

  end subroutine compute_energy_charmm

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_charmm_short
  !> @brief        compute potential energy with charmm force field
  !! @authors      JJ
  !! @param[in]    domain     : domain information
  !! @param[in]    enefunc    : potential energy functions information
  !! @param[in]    pairlist   : pair list information
  !! @param[in]    boundary   : boundary information
  !! @param[in]    coord      : coordinates of target systems
  !! @param[in]    nonb_ene   : flag for calculate nonbonded energy
  !! @param[inout] energy     : energy information
  !! @param[inout] force      : forces of target systems
  !! @param[inout] force_omp  : temprary forces of target systems
  !! @param[inout] virial     : virial term of target systems
  !! @param[inout] virial_ext : extern virial term of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_charmm_short(domain, enefunc, pairlist, boundary, &
                                  coord, nonb_ene, energy, force, force_omp,  &
                                  virial, virial_ext)

    ! formal arguments
    type(s_domain),          intent(in)    :: domain
    type(s_enefunc),         intent(inout) :: enefunc
    type(s_pairlist),        intent(in)    :: pairlist
    type(s_boundary),        intent(in)    :: boundary
    real(wp),                intent(in)    :: coord(:,:,:)
    logical,                 intent(in)    :: nonb_ene
    type(s_energy),          intent(inout) :: energy
    real(wp),                intent(inout) :: force(:,:,:)
    real(wp),                intent(inout) :: force_omp(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3)
    real(wp),                intent(inout) :: virial_ext(3,3)

    ! local variable
    integer                  :: ncell, natom, id, i, ix
    real(wp)                 :: virial_omp(3,3,nthread)
    real(wp)                 :: virial_ext_omp(3,3,nthread)
    real(wp)                 :: elec_omp(nthread)
    real(wp)                 :: evdw_omp(nthread)
    real(wp)                 :: ebond_omp(nthread)
    real(wp)                 :: eangle_omp(nthread)
    real(wp)                 :: eurey_omp(nthread)
    real(wp)                 :: edihed_omp(nthread)
    real(wp)                 :: eimprop_omp(nthread)
    real(wp)                 :: ecmap_omp(nthread)
    real(wp)                 :: eposi_omp(nthread)


    ! number of cells and atoms
    !
    ncell = domain%num_cell_local + domain%num_cell_boundary
    natom = domain%max_num_atom

    ! initialization of energy and forces
    !
    call init_energy(energy)

    force     (1:3,1:natom,1:ncell) = 0.0_wp
    virial    (1:3,1:3)             = 0.0_wp
    virial_ext(1:3,1:3)             = 0.0_wp

    force_omp     (1:3,1:natom,1:ncell,1:nthread) = 0.0_wp
    virial_omp    (1:3,1:3,1:nthread) = 0.0_wp
    virial_ext_omp(1:3,1:3,1:nthread) = 0.0_wp
    ebond_omp     (1:nthread) = 0.0_wp
    eangle_omp    (1:nthread) = 0.0_wp
    eurey_omp     (1:nthread) = 0.0_wp
    edihed_omp    (1:nthread) = 0.0_wp
    eimprop_omp   (1:nthread) = 0.0_wp
    ecmap_omp     (1:nthread) = 0.0_wp
    elec_omp      (1:nthread) = 0.0_wp
    evdw_omp      (1:nthread) = 0.0_wp
    eposi_omp     (1:nthread) = 0.0_wp

    ! bond energy
    !
    call compute_energy_bond(domain, enefunc, coord, &
                              force_omp, virial_omp, ebond_omp)

    ! angle energy
    !
    call compute_energy_angle(domain, enefunc, coord, &
                              force_omp, virial_omp, eangle_omp, eurey_omp)

    ! dihedral energy
    !
    if (enefunc%local_restraint) then
      call compute_energy_dihed_localres(domain, enefunc, coord, &
                              force_omp, virial_omp, edihed_omp)
    else
      call compute_energy_dihed(domain, enefunc, coord, &
                              force_omp, virial_omp, edihed_omp)
    end if

    ! improper energy
    !
    call compute_energy_improp(domain, enefunc, coord, &
                              force_omp, virial_omp, eimprop_omp)

    ! cmap energy
    !
    call compute_energy_cmap(domain, enefunc, coord, &
                              force_omp, virial_omp, ecmap_omp)


    ! restraint energy
    !
    if (enefunc%restraint) &
      call compute_energy_restraints(.true., .true., domain, enefunc, coord, &
                              force_omp, virial_omp, virial_ext_omp,         &
                              eposi_omp, energy%restraint_distance)

    ! nonbonded energy
    !   type 4: nonbond_ene_pbc
    !   type 5: nonbond_ene_pme
    !

    if (enefunc%pme_use) then
      call compute_energy_nonbond_pme_short(domain, enefunc, pairlist, &
                              nonb_ene, force_omp, virial_omp,         &
                              elec_omp, evdw_omp)
    else
      call error_msg( &
           'Compute_Energy_Short>  Multi-step is available only with PME')
    end if


    ! gather values
    !
    do id = 1, nthread
      do i = 1, ncell
        do ix = 1, domain%num_atom(i)
          force(1,ix,i) = force(1,ix,i) + force_omp(1,ix,i,id)
          force(2,ix,i) = force(2,ix,i) + force_omp(2,ix,i,id)
          force(3,ix,i) = force(3,ix,i) + force_omp(3,ix,i,id)
        end do
      end do

      virial(1:3,1:3)     = virial(1:3,1:3) + virial_omp(1:3,1:3,id)

      energy%bond               = energy%bond               + ebond_omp(id)
      energy%angle              = energy%angle              + eangle_omp(id)
      energy%urey_bradley       = energy%urey_bradley       + eurey_omp(id)
      energy%dihedral           = energy%dihedral           + edihed_omp(id)
      energy%improper           = energy%improper           + eimprop_omp(id)
      energy%cmap               = energy%cmap               + ecmap_omp(id)
      energy%electrostatic      = energy%electrostatic      + elec_omp(id)
      energy%van_der_waals      = energy%van_der_waals      + evdw_omp(id)
      energy%restraint_position = energy%restraint_position + eposi_omp(id)

    end do

    return

  end subroutine compute_energy_charmm_short

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_charmm_long
  !> @brief        compute potential energy with charmm force field
  !! @authors      JJ
  !! @param[in]    domain     : domain information
  !! @param[in]    enefunc    : potential energy functions information
  !! @param[in]    pairlist   : pair list information
  !! @param[in]    boundary   : boundary information
  !! @param[in]    coord      : coordinates of target systems
  !! @param[in]    npt        : flag for NPT or not
  !! @param[inout] energy     : energy information
  !! @param[inout] force      : forces of target systems
  !! @param[inout] force_omp  : temprary forces of target systems
  !! @param[inout] virial     : virial term of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_charmm_long(domain, enefunc, pairlist, boundary,   &
                                        coord,  npt, energy, force, force_omp, &
                                        virial)

    ! formal arguments
    type(s_domain),          intent(in)    :: domain
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_pairlist),        intent(in)    :: pairlist
    type(s_boundary),        intent(in)    :: boundary
    real(wp),                intent(in)    :: coord(:,:,:)
    logical,                 intent(in)    :: npt
    type(s_energy),          intent(inout) :: energy
    real(wp),                intent(inout) :: force(:,:,:)
    real(wp),                intent(inout) :: force_omp(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3)

    ! local variable
    real(wp)                 :: virial_omp(3,3,nthread)
    real(wp)                 :: elec_long, elec_omp(nthread)
    integer                  :: ncell, natom, id, i, ix


    ! number of cells and atoms
    !
    ncell = domain%num_cell_local + domain%num_cell_boundary
    natom = domain%max_num_atom

    ! initialization of energy and forces
    !
    force(1:3,1:natom,1:ncell) = 0.0_wp
    elec_long                  = 0.0_wp

    force_omp (1:3,1:natom,1:ncell,1:nthread) = 0.0_wp
    virial_omp(1:3,1:3,1:nthread) = 0.0_wp
    elec_omp  (1:nthread) = 0.0_wp

    if (enefunc%pme_use) &
      call compute_energy_nonbond_pme_long(domain, enefunc, boundary, &
                              npt, force_omp, virial_omp, elec_omp)

    do id = 1, nthread
      do i = 1, ncell
        do ix = 1, domain%num_atom(i)
          force(1,ix,i) = force(1,ix,i) + force_omp(1,ix,i,id)
          force(2,ix,i) = force(2,ix,i) + force_omp(2,ix,i,id)
          force(3,ix,i) = force(3,ix,i) + force_omp(3,ix,i,id)
        end do
      end do

      virial(1:3,1:3) = virial(1:3,1:3) + virial_omp(1:3,1:3,id)

      elec_long = elec_long + elec_omp(id)

    end do

    energy%electrostatic = energy%electrostatic + elec_long

    ! total energy
    energy%total = energy%total + elec_long

    return

  end subroutine compute_energy_charmm_long

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    output_energy_genesis
  !> @brief        output energy in GENESIS style
  !! @authors      TM, CK
  !! @param[in]    step    : step 
  !! @param[in]    enefunc : information of potential functions
  !! @param[in]    energy  : energy information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_energy_genesis(step, enefunc, energy)

    ! formal arguments
    integer,                 intent(in)    :: step
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_energy),          intent(in)    :: energy

    ! local variables
    real(wp)                 :: ene_misc1, ene_misc2, ene_restraint
    integer                  :: i
    character                :: cmisc1*16, cmisc2*16


    select case (enefunc%forcefield)

    case (ForcefieldCHARMM)

      ene_misc1 = energy%urey_bradley
      ene_misc2 = energy%cmap
      cmisc1    = '    UREY-BRADLEY'
      cmisc2    = '            CMAP'

    end select

    ene_restraint =   energy%restraint_distance &
                    + energy%restraint_position

    if (etitle) then
      write(MsgOut,'(A80)')         '            STEP            BOND           ANGLE        DIHEDRAL        IMPROPER'
      write(MsgOut,'(A,A16,A16,A)') '         VDWAALS           ELECT',       cmisc1,        cmisc2,'       RESTRAINT'
      write(MsgOut,'(A80)')         ' --------------- --------------- --------------- --------------- ---------------'
      etitle = .false.
    end if

    write(MsgOut,'(A5,1x,i10,4(1x,F15.4))') '     ',  &
                                step,                 &
                                energy%bond,          &
                                energy%angle,         &
                                energy%dihedral,      &
                                energy%improper
    write(MsgOut,'(5(1x,F15.4))')                     &
                                energy%van_der_waals, &
                                energy%electrostatic, &
                                ene_misc1,            &
                                ene_misc2,            &
                                ene_restraint
    write(MsgOut,'(A)') ''

    return

  end subroutine output_energy_genesis

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    output_energy_charmm
  !> @brief        output energy in CHARMM style
  !! @authors      YS, CK
  !! @param[in]    step    : step 
  !! @param[in]    enefunc : information of potential functions
  !! @param[in]    energy  : energy information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_energy_charmm(step, enefunc, energy)

    ! formal arguments
    integer,                 intent(in)    :: step
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_energy),  target, intent(in)    :: energy

    ! local variables
    real(wp)                 :: time, totener, totke, energy_, temperature
    real(wp)                 :: grms, hfctote, hfcke, ehfcor, virke
    real(wp)                 :: hbonds, asp, user
    real(wp)                 :: imnbvdw, imelec, imhbnd, rxnfield, extelec
    real(wp)                 :: ewksum, ewself, ewexcl, ewqcor, ewutil
    real(wp)                 :: vire, viri, presse, pressi, volume
    real(wp)                 :: cdihe, cintcr, noe

    real(wp),        pointer :: bonds, angles, urey_b, dihedrals, impropers
    real(wp),        pointer :: vdwaals, elec, cmaps
    real(wp),        pointer :: posicon, restdist


    time        = 0.0_wp
    hbonds      = 0.0_wp
    totener     = 0.0_wp
    totke       = 0.0_wp
    energy_     = 0.0_wp
    temperature = 0.0_wp
    asp         = 0.0_wp
    user        = 0.0_wp
    imnbvdw     = 0.0_wp
    imelec      = 0.0_wp
    imhbnd      = 0.0_wp
    rxnfield    = 0.0_wp
    extelec     = 0.0_wp
    ewksum      = 0.0_wp
    ewself      = 0.0_wp
    ewexcl      = 0.0_wp
    ewqcor      = 0.0_wp
    ewutil      = 0.0_wp
    vire        = 0.0_wp
    viri        = 0.0_wp
    presse      = 0.0_wp
    pressi      = 0.0_wp
    volume      = 0.0_wp
    grms        = 0.0_wp
    hfctote     = 0.0_wp
    hfcke       = 0.0_wp
    ehfcor      = 0.0_wp
    virke       = 0.0_wp
    volume      = 0.0_wp
    noe         = 0.0_wp
    cdihe       = 0.0_wp
    cintcr      = 0.0_wp

    ! write title if necessary
    !
    if (etitle) then
      write(MsgOut,'(A)') 'Output_Energy> CHARMM_Style is used'
      write(MsgOut,'(A)') ' '
      write(MsgOut,'(A79)') 'DYNA DYN: Step         Time      TOTEner        TOTKe       ENERgy  TEMPerature'
      write(MsgOut,'(A79)') 'DYNA PROP:             GRMS      HFCTote        HFCKe       EHFCor        VIRKe'
      write(MsgOut,'(A79)') 'DYNA INTERN:          BONDs       ANGLes       UREY-b    DIHEdrals    IMPRopers'

      if (enefunc%num_cmap_all > 0) then
        write(MsgOut,'(A79)') 'DYNA CROSS:           CMAPs                                                    '
      end if

      write(MsgOut,'(A79)') 'DYNA EXTERN:        VDWaals         ELEC       HBONds          ASP         USER'

      write(MsgOut,'(A79)') 'DYNA IMAGES:        IMNBvdw       IMELec       IMHBnd       RXNField    EXTElec'
      write(MsgOut,'(A79)') 'DYNA EWALD:          EWKSum       EWSElf       EWEXcl       EWQCor       EWUTil'

      if (enefunc%restraint) then
        write(MsgOut,'(A79)') 'DYNA CONSTR:       HARMonic    CDIHedral          CIC     RESDistance       NOE'
      endif

      write(MsgOut,'(A79)') 'DYNA PRESS:            VIRE         VIRI       PRESSE       PRESSI       VOLUme'
      write(MsgOut,'(A79)') ' ----------       ---------    ---------    ---------    ---------    ---------'
      etitle = .false.

    end if


    bonds      => energy%bond
    angles     => energy%angle
    urey_b     => energy%urey_bradley
    dihedrals  => energy%dihedral
    impropers  => energy%improper
    elec       => energy%electrostatic
    vdwaals    => energy%van_der_waals
    cmaps      => energy%cmap
    restdist   => energy%restraint_distance
    posicon    => energy%restraint_position

    ! write energy in CHARMM-style
    !
    write(MsgOut,'(A5,I9,5F13.5)') 'DYNA>', step, time, totener, totke, energy_, temperature
    write(MsgOut,'(A14,5F13.5)')   'DYNA PROP>    ', grms, hfctote, hfcke, ehfcor, virke
    write(MsgOut,'(A14,5F13.5)')   'DYNA INTERN>  ', bonds, angles, urey_b, dihedrals, impropers

    if (enefunc%num_cmap_all > 0) then
      write(MsgOut,'(A14, F13.5)')   'DYNA CROSS>   ', cmaps
    end if

    write(MsgOut,'(A14,5F13.5)')   'DYNA EXTERN>  ', vdwaals, elec, hbonds, asp, user

    write(MsgOut,'(A14,5F13.5)')   'DYNA IMAGES>  ', imnbvdw, imelec, imhbnd, rxnfield, extelec
    write(MsgOut,'(A14,5F13.5)')   'DYNA EWALD>   ', ewksum, ewself, ewexcl, ewqcor, ewutil

    if (enefunc%restraint) then
      write(MsgOut,'(A14,5F13.5)')   'DYNA CONSTR>  ', posicon, cdihe, cintcr, restdist, noe
    end if

    write(MsgOut,'(A14,5F13.5)')   'DYNA PRESS>   ', vire, viri, presse, pressi, volume
    write(MsgOut,'(A79)') ' ----------       ---------    ---------    ---------    ---------    ---------'
    write(MsgOut,'(A)') ' '

    return

  end subroutine output_energy_charmm

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    output_energy_namd
  !> @brief        output energy in NAMD style
  !! @authors      YS, CK
  !! @param[in]    step    : step 
  !! @param[in]    energy : energy information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine output_energy_namd(step, energy)

    ! formal arguments
    integer,                 intent(in)    :: step
    type(s_energy),  target, intent(in)    :: energy

    ! local variables
    real(wp)                 :: ebond, eangle
    real(wp)                 :: misc, kinetic
    real(wp)                 :: total, temp, total2, total3, tempavg
    real(wp)                 :: pressure, gpressure, volume
    real(wp)                 :: pressavg, gpressavg

    real(wp),        pointer :: edihed, eimprp
    real(wp),        pointer :: eelect, evdw
    real(wp),        pointer :: eboundary


    ! write title if necessary
    !
    if (etitle) then
      write(MsgOut,'(A)') 'Output_Energy> NAMD_Style is used'
      write(MsgOut,'(A)') ' '
      write(MsgOut,'(A75)') 'ETITLE:      TS           BOND          ANGLE          DIHED          IMPRP'
      write(MsgOut,'(A75)') '          ELECT            VDW       BOUNDARY           MISC        KINETIC'
      write(MsgOut,'(A75)') '          TOTAL           TEMP         TOTAL2         TOTAL3        TEMPAVG'
      write(MsgOut,'(A75)') '       PRESSURE      GPRESSURE         VOLUME       PRESSAVG      GPRESSAVG'
      write(MsgOut,'(A)') ' '
      etitle = .false.
    end if


    ebond     =  energy%bond  + energy%restraint_distance
    eangle    =  energy%angle + energy%urey_bradley
    edihed    => energy%dihedral
    eimprp    => energy%improper
    eelect    => energy%electrostatic
    evdw      => energy%van_der_waals
    eboundary => energy%restraint_position

    ! write energy in NAMD-style
    !
    write(MsgOut,'(A7,I8,4F15.4)')'ENERGY:', step, ebond, eangle, edihed,eimprp
    write(MsgOut,'(5F15.4)') eelect, evdw, eboundary, misc, kinetic
    write(MsgOut,'(5F15.4)') total, temp, total2, total3, tempavg
    write(MsgOut,'(5F15.4)') pressure, gpressure, volume, pressavg, gpressavg
    write(MsgOut,'(A)') ' '

    return

  end subroutine output_energy_namd

  !======1=========2=========3=========4=========5=========6=========7=========8
  ! 
  !  Subroutine    reduce_ene
  !> @brief        reduce energy and virial
  !! @authors      JJ
  !! @param[inout] energy : energy information
  !! @param[inout] virial : virial term of 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine reduce_ene(energy, virial)

    ! formal arguments
    type(s_energy),          intent(inout) :: energy
    real(wp),                intent(inout) :: virial(3,3)
    
#ifdef MPI 

    ! local variables      
    real(wp)                 :: before_reduce(19), after_reduce(19)
    integer                  :: i, j, n


    ! Allreduce virial and energy components
    !
    n = 0
    do i = 1, 3
      do j = 1, 3
        n = n + 1
        before_reduce(n) = virial(i,j)
      end do
    end do

    before_reduce(10) = energy%bond
    before_reduce(11) = energy%angle
    before_reduce(12) = energy%urey_bradley
    before_reduce(13) = energy%dihedral
    before_reduce(14) = energy%improper
    before_reduce(15) = energy%cmap
    before_reduce(16) = energy%electrostatic
    before_reduce(17) = energy%van_der_waals
    before_reduce(18) = energy%restraint_position
    before_reduce(19) = energy%total

    call mpi_reduce(before_reduce, after_reduce, 19, mpi_wp_real,  &
                    mpi_sum, 0, mpi_comm_country, ierror)

    n = 0
    do i = 1, 3
      do j = 1, 3
        n = n + 1
        virial(i,j) = after_reduce(n)
      end do
    end do

    energy%bond               = after_reduce(10)
    energy%angle              = after_reduce(11)
    energy%urey_bradley       = after_reduce(12)
    energy%dihedral           = after_reduce(13)
    energy%improper           = after_reduce(14)
    energy%cmap               = after_reduce(15)
    energy%electrostatic      = after_reduce(16)
    energy%van_der_waals      = after_reduce(17)
    energy%restraint_position = after_reduce(18)
    energy%total              = after_reduce(19)

#endif

    return

  end subroutine reduce_ene

end module sp_energy_mod
