!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_energy_nonbonds_mod
!> @brief   calculate nonbond energy
!! @authors Jaewoon Jung(JJ)
!  
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_energy_nonbonds_mod

  use sp_energy_table_linear_mod
  use sp_energy_table_cubic_mod
  use sp_pairlist_str_mod
  use sp_enefunc_str_mod
  use sp_domain_str_mod
  use timers_mod
  use mpi_parallel_mod
  use constants_mod

  implicit none
  private

  ! subroutines
  public  :: compute_energy_nonbond_cutoff
  private :: compute_energy_nonbond14
  private :: compute_energy_nonbond_pbc

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_cutoff
  !> @brief        compute nonbond energy with cutoff 
  !! @authors      JJ
  !! @param[in]    domain   : domain information
  !! @param[in]    enefunc  : potential energy functions information
  !! @param[in]    pairlist : pair-list information
  !! @param[in]    nonb_ene : flag for calculate nonbonded energy
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : lennard-jones energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_cutoff(domain, enefunc, pairlist, &
                                           nonb_ene, force, virial, eelec, evdw)

    ! formal arguments
    type(s_domain),          intent(in)    :: domain
    type(s_enefunc),         intent(in)    :: enefunc
    type(s_pairlist),        intent(in)    :: pairlist
    logical,                 intent(in)    :: nonb_ene
    real(wp),                intent(inout) :: force(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3,nthread)
    real(wp),                intent(inout) :: eelec(nthread)
    real(wp),                intent(inout) :: evdw(nthread)


    call timer(TimerNonBond, TimerOn)
 
    ! 1-4 energy
    !
    if (enefunc%table%table) then

      if (enefunc%table%table_order == 3) then

        call compute_energy_nonbond_nb14_table(                      &
                                          domain, enefunc,           &
                                          force, virial, eelec, evdw)

      else if (enefunc%table%table_order == 1) then

        call compute_energy_nonbond_nb14_table_linear(               &
                                          domain, enefunc,           &
                                          force, virial, eelec, evdw)

      else if (enefunc%table%table_order == 0) then

        call compute_energy_nonbond_nb14_table_linear(               &
                                          domain, enefunc,           &
                                          force, virial, eelec, evdw)

      end if

    else
      call compute_energy_nonbond14(domain, enefunc, &
                                    force, virial, eelec, evdw)
    end if

    ! energy with lookup table
    !
    if (enefunc%table%table) then

      if (nonb_ene) then

        if (enefunc%table%table_order == 3) then

          if (enefunc%table%water_table) then

            call compute_energy_nonbond_table_solute(                &
                                          domain, enefunc, pairlist, &
                                          force, virial, eelec, evdw)

            call compute_energy_nonbond_table_water(                 &
                                          domain, enefunc, pairlist, &
                                          force, virial, eelec, evdw)

          else
            call compute_energy_nonbond_table(                       &
                                          domain, enefunc, pairlist, &
                                          force, virial, eelec, evdw)
          end if

        else if (enefunc%table%table_order == 1) then

          call compute_energy_nonbond_table_solute_linear(           &
                                          domain, enefunc, pairlist, &
                                          force, virial, eelec, evdw)

          call compute_energy_nonbond_table_water_linear(            &
                                          domain, enefunc, pairlist, &
                                          force, virial, eelec, evdw)

        end if

      else

        if (enefunc%table%table_order == 3) then

          if (enefunc%table%water_table) then

            call compute_force_nonbond_table_solute(                 &
                                          domain, enefunc, pairlist, &
                                          force, virial)

            call compute_force_nonbond_table_water(                  &
                                          domain, enefunc, pairlist, &
                                          force, virial)

          else

            call compute_force_nonbond_table(                        &
                                          domain, enefunc, pairlist, &
                                          force, virial)

          end if

        else if (enefunc%table%table_order == 1) then

          call compute_force_nonbond_table_solute_linear(            &
                                          domain, enefunc, pairlist, &
                                          force, virial)

          call compute_force_nonbond_table_water_linear(             &
                                          domain, enefunc, pairlist, &
                                          force, virial)

        end if

      end if

    ! energy without using table
    !
    else

      call compute_energy_nonbond_pbc(domain, enefunc, pairlist, &
                                      force, virial, eelec, evdw)

    end if

    call timer(TimerNonBond, TimerOff)
 
    return

  end subroutine compute_energy_nonbond_cutoff

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond14
  !> @brief        calculate nonbonded14 energy
  !! @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[in]    enefunc : potential energy functions information
  !! @param[inout] force   : forces of target systems
  !! @param[inout] virial  : virial term of target systems
  !! @param[inout] eelec   : electrostatic energy of target systems
  !! @param[inout] evdw    : van der Waals energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond14(domain, enefunc, force, virial, &
                                      eelec, evdw)

    ! formal arguments 
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(in)    :: enefunc
    real(wp),                intent(inout) :: force(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3,nthread)
    real(wp),                intent(inout) :: eelec(nthread)
    real(wp),                intent(inout) :: evdw(nthread)

    ! local variables
    real(wp)                 :: dij(1:3), rij, rij2, inv_rij
    real(wp)                 :: inv_r2, inv_r6, inv_r3, inv_r12
    real(wp)                 :: c6, c12, lj6, lj12, term_lj, term_lj6, term_lj12
    real(wp)                 :: c1_switchup, c2_switchup
    real(wp)                 :: c1_switchdown, c2_switchdown
    real(wp)                 :: CCOEF, term_elec, shift, dshift, switch, dswitch
    real(wp)                 :: cutoff2, switchdist2
    real(wp)                 :: c1_switch, c2_switch, c4_switch, coef
    real(wp)                 :: work(1:3), viri_local(1:3,1:3)
    integer                  :: i, j, k, m, ix, iy, ij
    integer                  :: num_nb14, ini_nb14, fin_nb14
    integer                  :: id, omp_get_thread_num
    integer                  :: ncell_local

    real(wp),        pointer :: dielec_const
    real(wp),        pointer :: cutoff, switchdist
    real(wp),        pointer :: coord(:,:,:), charge(:,:)
    integer,         pointer :: cell_pair(:,:), atmcls(:,:), natom(:)


    coord        => domain%coord
    charge       => domain%charge
    cell_pair    => domain%cell_pairlist1
    atmcls       => domain%atom_cls_no
    natom        => domain%num_atom

    dielec_const => enefunc%dielec_const
    cutoff       => enefunc%cutoffdist
    switchdist   => enefunc%switchdist

    ncell_local  = domain%num_cell_local

    CCOEF        = ELECOEF/dielec_const
    cutoff2      = cutoff * cutoff
    switchdist2  = switchdist * switchdist

    if (enefunc%force_switch) then
      c1_switchdown = 1.0_wp / (cutoff2*switchdist2)
      c1_switchdown = c1_switchdown*c1_switchdown*c1_switchdown
      c2_switchdown = sqrt(c1_switchdown)
      c1_switchup   = cutoff2*cutoff2*cutoff2
      c2_switchup   = cutoff2*cutoff
      c12 = c1_switchup / (c1_switchup-switchdist2*switchdist2*switchdist2)
      c6  = c2_switchup / (c2_switchup-switchdist2*switchdist)
      c1_switchup = 1.0_wp / c1_switchup
      c2_switchup = 1.0_wp / c2_switchup
    else
      c1_switch = 1.0_wp/(cutoff2-switchdist2)
      c1_switch = c1_switch*c1_switch*c1_switch
    end if

    !$omp parallel default(shared)                                             &
    !$omp private(id, i, j, k, m, ij, ix, iy, ini_nb14, fin_nb14, num_nb14,    &
    !$omp         dij, rij, rij2, c2_switch, c4_switch, switch, dswitch,       &
    !$omp         shift, dshift, lj6, lj12, inv_rij, inv_r2, inv_r3, inv_r6,   &
    !$omp         inv_r12, coef, term_elec, work, viri_local,                  &
    !$omp         term_lj, term_lj6, term_lj12)
    !
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    ! calculate energy and gradient
    !
    viri_local(1:3,1:3) = 0.0_wp

    do i = id+1, ncell_local, nthread
      num_nb14 = 0

      do ix = 1, natom(i) - 1
        ini_nb14 = num_nb14 + 1
        fin_nb14 = num_nb14 + enefunc%num_nb14_calc1(ix,i)
        num_nb14 = fin_nb14

        do k = ini_nb14, fin_nb14

          iy = enefunc%nb14_calc_list1(k,i)

          ! compute distance
          !
          dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,i)
          rij2     = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
          rij      = sqrt(rij2)

          lj6     = enefunc%nb14_lj6(atmcls(ix,i),atmcls(iy,i))
          lj12    = enefunc%nb14_lj12(atmcls(ix,i),atmcls(iy,i))
          inv_r2  = 1.0_wp / rij2
          inv_rij = sqrt(inv_r2)
          inv_r3  = inv_r2 * inv_rij
          inv_r6  = inv_r2 * inv_r2 * inv_r2
          inv_r12 = inv_r6 * inv_r6

          if (enefunc%force_switch) then

            if (rij2 > switchdist2) then
              term_lj12  = lj12*c12*(inv_r6-c1_switchup)**2
              term_lj6   = lj6*c6*(inv_r3-c2_switchup)**2
              term_lj    = term_lj12 - term_lj6
              evdw(id+1) = evdw(id+1) + term_lj
              coef       = -12.0_wp*lj12*c12*(inv_r6-c1_switchup)*inv_r6*inv_r2
              coef       = coef+6.0_wp*lj6*c6*(inv_r3-c2_switchup)*inv_r3*inv_r2
            else
              coef       = -12.0_wp * lj12 * inv_r12 * inv_r2
              coef       = coef + 6.0_wp*lj6*inv_r6*inv_r2
              inv_r12    = inv_r12 - c1_switchdown
              inv_r6     = inv_r6 - c2_switchdown
              term_lj12  = lj12 * inv_r12
              term_lj6   = lj6 * inv_r6
              term_lj    = term_lj12 - term_lj6
              evdw(id+1) = evdw(id+1) + term_lj
            end if

          else

            if (rij2 > switchdist2) then
              c2_switch  = cutoff2 - rij2
              c4_switch  = c2_switch*(cutoff2 + 2.0_wp*rij2 -3.0_wp*switchdist2)
              switch     = c2_switch*c4_switch*c1_switch
              dswitch    = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch-c4_switch)
            else
              switch     = 1.0_wp
              dswitch    = 0.0_wp
            end if

            term_lj12    = lj12 * inv_r12
            term_lj6     = lj6 * inv_r6
            term_lj      = term_lj12 - term_lj6
            evdw(id+1)   = evdw(id+1) + switch*term_lj
            coef         = (dswitch * term_lj * rij &
                           - switch *(12.0_wp*term_lj12-6.0_wp*term_lj6))*inv_r2
          end if
 
          ! electrostatic energy and gradient
          !
          shift  = (1.0_wp - (rij2/cutoff2))
          dshift = -4.0_wp*shift*rij/cutoff2
          shift  = shift * shift
          term_elec   = charge(ix,i) * charge(iy,i) * CCOEF
          eelec(id+1) = eelec(id+1) + shift * ( term_elec * inv_rij )
          coef        = coef + (dshift*(term_elec*inv_rij) - &
                                 shift*(term_elec*inv_r2))*inv_rij
          work(1:3) = coef*dij(1:3)

          ! store force
          !
          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + work(1:3)

          ! virial
          !
          do m = 1, 3
            viri_local(1:3,m) = viri_local(1:3,m) + dij(1:3)*work(m)
          end do

        end do
      end do
    end do


    do ij = id+1, maxcell, nthread

      i = cell_pair(1,ij)
      j = cell_pair(2,ij)

      num_nb14 = 0

      do ix = 1, natom(i)
        ini_nb14  = num_nb14 + 1
        fin_nb14  = num_nb14 + enefunc%num_nb14_calc(ix,ij)
        num_nb14  = fin_nb14

        do k = ini_nb14, fin_nb14

          iy = enefunc%nb14_calc_list(k,ij)

          ! compute distance
          !
          dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,j)
          rij2     = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
          rij      = sqrt(rij2)

          lj6      = enefunc%nb14_lj6(atmcls(ix,i),atmcls(iy,j))
          lj12     = enefunc%nb14_lj12(atmcls(ix,i),atmcls(iy,j))
          inv_r2   = 1.0_wp / rij2
          inv_rij  = sqrt(inv_r2)
          inv_r3   = inv_r2 * inv_rij
          inv_r6   = inv_r2 * inv_r2 * inv_r2
          inv_r12  = inv_r6 * inv_r6

          if (enefunc%force_switch) then

            if (rij2 > switchdist2) then
              term_lj12  = lj12*c12*(inv_r6-c1_switchup)**2
              term_lj6   = lj6*c6*(inv_r3-c2_switchup)**2
              term_lj    = term_lj12 - term_lj6
              evdw(id+1) = evdw(id+1) + term_lj
              coef       = -12.0_wp*lj12*c12*(inv_r6-c1_switchup)*inv_r6*inv_r2
              coef       = coef+6.0_wp*lj6*c6*(inv_r3-c2_switchup)*inv_r3*inv_r2
            else
              coef       = -12.0_wp * lj12 * inv_r12 * inv_r2
              coef       = coef + 6.0_wp*lj6*inv_r6*inv_r2
              inv_r12    = inv_r12 - c1_switchdown
              inv_r6     = inv_r6 - c2_switchdown
              term_lj12  = lj12 * inv_r12
              term_lj6   = lj6 * inv_r6
              term_lj    = term_lj12 - term_lj6
              evdw(id+1) = evdw(id+1) + term_lj
            end if

          else

            if (rij2 > switchdist2) then
              c2_switch  = cutoff2 - rij2
              c4_switch  = c2_switch*(cutoff2 + 2.0_wp*rij2 -3.0_wp*switchdist2)
              switch     = c2_switch*c4_switch*c1_switch
              dswitch    = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch-c4_switch)
            else
              switch     = 1.0_wp
              dswitch    = 0.0_wp
            end if

            term_lj12    = lj12 * inv_r12
            term_lj6     = lj6 * inv_r6
            term_lj      = term_lj12 - term_lj6
            evdw(id+1)   = evdw(id+1) + switch*term_lj
            coef         = (dswitch * term_lj * rij &
                           - switch *(12.0_wp*term_lj12-6.0_wp*term_lj6))*inv_r2
          end if

          ! electrostatic energy and gradient
          !
          shift  = (1.0_wp - (rij2/cutoff2))
          dshift = -4.0_wp*shift*rij/cutoff2
          shift  = shift * shift

          term_elec   = charge(ix,i) * charge(iy,j) * CCOEF
          eelec(id+1) = eelec(id+1) + shift * ( term_elec * inv_rij )
          coef        = coef + (dshift*(term_elec*inv_rij) - &
                                 shift*(term_elec*inv_r2))*inv_rij
          work(1:3) = coef*dij(1:3)

          ! store force
          !
          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
          force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) + work(1:3)

          ! virial 
          !
          do m = 1, 3
            viri_local(1:3,m) = viri_local(1:3,m) + dij(1:3)*work(m)
          end do

        end do
      end do
    end do

    do m = 1, 3
      virial(1:3,m,id+1) = virial(1:3,m,id+1) - viri_local(1:3,m)
    end do

    !$omp end parallel

    return

  end subroutine compute_energy_nonbond14

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_pbc
  !> @brief        calculate nonbonded energy with periodic boundary condition
  !! @authors      JJ 
  !! @param[in]    domain   : domain information
  !! @param[in]    enefunc  : potential energy functions
  !! @param[in]    pairlist : interaction list in each domain
  !! @param[inout] force    : forces of target systems
  !! @param[inout] virial   : virial term of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : van der Waals energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_pbc(domain, enefunc, pairlist, &
                                        force, virial, eelec, evdw)

    ! formal arguments
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_pairlist), target, intent(in)    :: pairlist
    real(wp),                 intent(inout) :: force(:,:,:,:)
    real(wp),                 intent(inout) :: virial(3,3,nthread)
    real(wp),                 intent(inout) :: eelec(nthread)
    real(wp),                 intent(inout) :: evdw(nthread)

    ! local variables
    real(wp)                  :: dij(1:3), rij, rij2, inv_rij
    real(wp)                  :: inv_r2, inv_r3, inv_r6, inv_r12, dr126
    real(wp)                  :: lj6, lj12, term_lj12, term_lj6, term_lj
    real(wp)                  :: term_elec, shift, dshift, switch, dswitch
    real(wp)                  :: CCOEF, cutoff2, cutoff2_inv, switchdist2
    real(wp)                  :: c1_switch, c2_switch, c4_switch, coef
    real(wp)                  :: c1_switchup, c2_switchup
    real(wp)                  :: c1_switchdown, c2_switchdown, c6, c12
    real(wp)                  :: work(1:3), viri_local(1:3,1:3), trans(1:3)
    real(wp)                  :: rtmp(1:3), qtmp
    real(wp)                  :: elec_temp, evdw_temp
    integer                   :: i, j, k, m, ix, iy, ij, iix
    integer                   :: num_nb15, ini_nb15, fin_nb15
    integer                   :: id, omp_get_thread_num
    integer                   :: ncell, ncell_local

    real(wp),         pointer :: dielec_const
    real(wp),         pointer :: cutoff, switchdist
    real(wp),         pointer :: coord(:,:,:), trans1(:,:,:), trans2(:,:,:)
    real(wp),         pointer :: charge(:,:), cell_move(:,:,:)
    integer,          pointer :: cell_pairlist(:,:), atmcls(:,:)
    integer,          pointer :: natom(:)


    cell_pairlist => domain%cell_pairlist1
    atmcls        => domain%atom_cls_no
    natom         => domain%num_atom
    coord         => domain%coord
    trans1        => domain%trans_vec
    trans2        => domain%translated
    charge        => domain%charge
    cell_move     => domain%cell_move

    dielec_const  => enefunc%dielec_const
    cutoff        => enefunc%cutoffdist
    switchdist    => enefunc%switchdist

    ncell         = domain%num_cell_local + domain%num_cell_boundary
    ncell_local   = domain%num_cell_local

    CCOEF         = ELECOEF/dielec_const
    cutoff2       = cutoff * cutoff
    cutoff2_inv   = 1.0_wp/cutoff2
    switchdist2   = switchdist * switchdist

    if (enefunc%force_switch) then
      c1_switchdown = 1.0_wp / (cutoff2*switchdist2)
      c1_switchdown = c1_switchdown*c1_switchdown*c1_switchdown
      c2_switchdown = sqrt(c1_switchdown)
      c1_switchup   = cutoff2*cutoff2*cutoff2
      c2_switchup   = cutoff2*cutoff
      c12 = c1_switchup / (c1_switchup-switchdist2*switchdist2*switchdist2)
      c6  = c2_switchup / (c2_switchup-switchdist2*switchdist)
      c1_switchup = 1.0_wp / c1_switchup
      c2_switchup = 1.0_wp / c2_switchup
    else
      c1_switch = 1.0_wp/(cutoff2-switchdist2)
      c1_switch = c1_switch*c1_switch*c1_switch
    end if

    ! calculate energy and gradient
    !
    do i = 1, ncell
      do ix = 1, natom(i)
        trans2(1:3,ix,i) = coord(1:3,ix,i) + trans1(1:3,ix,i)
      end do
    end do

    !$omp parallel default(shared)                                             &
    !$omp private(id, i, j, k, m, ix, iy, ij, ini_nb15, fin_nb15, num_nb15,    &
    !$omp         dij, rij, rij2, c2_switch, c4_switch, switch, dswitch,       &
    !$omp         shift, dshift, lj6, lj12, inv_rij, inv_r2, inv_r3, inv_r6,   &
    !$omp         inv_r12, dr126, coef, term_elec, work, viri_local,           &
    !$omp         rtmp, qtmp, trans, iix, elec_temp, evdw_temp,                &
    !$omp         term_lj, term_lj6, term_lj12)    
    !
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    viri_local(1:3,1:3) = 0.0_wp
    elec_temp = 0.0_wp
    evdw_temp = 0.0_wp

    do i = id+1, ncell_local, nthread
      num_nb15 = 0

      do ix = 1, natom(i) - 1

        rtmp(1:3) = trans2(1:3,ix,i)
        qtmp = charge(ix,i) * CCOEF

        ini_nb15 = num_nb15 + 1
        fin_nb15 = num_nb15 + pairlist%num_nb15_calc1(ix,i)
        num_nb15 = fin_nb15

        do k = ini_nb15, fin_nb15

          iy = pairlist%nb15_calc_list1(k,i)

          ! compute distance
          !
          dij(1:3) = rtmp(1:3) - trans2(1:3,iy,i)
          rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff2) then

            rij = sqrt(rij2)

            ! van der Waals
            !
            lj6     = enefunc%nonb_lj6 (atmcls(ix,i),atmcls(iy,i))
            lj12    = enefunc%nonb_lj12(atmcls(ix,i),atmcls(iy,i))
            inv_r2  = 1.0_wp / rij2
            inv_rij = sqrt(inv_r2)
            inv_r3  = inv_r2 * inv_rij
            inv_r6  = inv_r2 * inv_r2 * inv_r2
            inv_r12 = inv_r6 * inv_r6

            if (enefunc%force_switch) then

              if (rij2 > switchdist2) then
                term_lj12 = lj12*c12*(inv_r6 - c1_switchup)**2
                coef      = -12.0_wp*lj12*c12*(inv_r6-c1_switchup)*inv_r6*inv_r2
                term_lj6  = lj6*c6*(inv_r3-c2_switchup)**2
                term_lj   = term_lj12 - term_lj6
                evdw_temp = evdw_temp + term_lj
                coef = coef + 6.0_wp*lj6*c6*(inv_r3-c2_switchup)*inv_r3*inv_r2
              else
                coef      = -12.0_wp * lj12 * inv_r12 * inv_r2
                coef      = coef + 6.0_wp*lj6*inv_r6*inv_r2
                inv_r12   = inv_r12 - c1_switchdown
                inv_r6    = inv_r6 - c2_switchdown
                term_lj12 = lj12 * inv_r12
                term_lj6  = lj6 * inv_r6
                term_lj   = term_lj12 - term_lj6
                evdw_temp = evdw_temp + term_lj
              end if

            else

              if (rij2 > switchdist2) then
                c2_switch = cutoff2 - rij2
                c4_switch = c2_switch*(cutoff2+ 2.0_wp*rij2- 3.0_wp*switchdist2)
                switch    = c2_switch*c4_switch*c1_switch
                dswitch   = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch-c4_switch)
              else
                switch    = 1.0_wp
                dswitch   = 0.0_wp
              end if

              term_lj12 = lj12 * inv_r12
              term_lj6  = lj6 * inv_r6
              term_lj   = term_lj12 - term_lj6
              evdw_temp = evdw_temp + switch*term_lj
              coef      = (dswitch * term_lj * rij  &
                         - switch *(12.0_wp*term_lj12-6.0_wp*term_lj6)) * inv_r2
            end if

            ! electrostatic energy and gradient
            !
            shift  = (1.0_wp - (rij2*cutoff2_inv))
            dshift = -4.0_wp*shift*rij*cutoff2_inv
            shift  = shift * shift

            term_elec = qtmp * charge(iy,i)
            elec_temp = elec_temp + shift * ( term_elec * inv_rij ) 
            coef      = coef + (dshift*(term_elec*inv_rij) - &
                                 shift*(term_elec*inv_r2))*inv_rij

            work(1:3) = coef*dij(1:3)

            ! store force
            !
            force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
            force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + work(1:3)
  
            ! virial
            !
            do m = 1, 3
              viri_local(1:3,m) = viri_local(1:3,m) + dij(1:3)*work(m)
            end do

          end if
        end do
      end do
    end do

    do ij = id+1, maxcell, nthread

      i = cell_pairlist(1,ij)
      j = cell_pairlist(2,ij)

      num_nb15 = 0

      trans(1:3) = cell_move(1:3,j,i)

      do iix = 1, pairlist%nb15_cell(ij)

        ix = pairlist%nb15_list(iix,ij)

        ini_nb15 = num_nb15 + 1
        fin_nb15 = num_nb15 + pairlist%num_nb15_calc(ix,ij)
        num_nb15 = fin_nb15

        do k = ini_nb15, fin_nb15
          iy = pairlist%nb15_calc_list(k,ij)

          ! compute distance
          !
          dij(1:3) = trans2(1:3,ix,i) - trans2(1:3,iy,j) + trans(1:3)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff2) then

            rij = sqrt(rij2)

            ! van der Waals
            !
            lj6     = enefunc%nonb_lj6(atmcls(ix,i),atmcls(iy,j))
            lj12    = enefunc%nonb_lj12(atmcls(ix,i),atmcls(iy,j))
            inv_r2  = 1.0_wp / rij2
            inv_rij = sqrt(inv_r2)
            inv_r3  = inv_r2 * inv_rij
            inv_r6  = inv_r2 * inv_r2 * inv_r2
            inv_r12 = inv_r6 * inv_r6

            if (enefunc%force_switch) then

              if (rij2 > switchdist2) then
                term_lj12 = lj12*c12*(inv_r6 - c1_switchup)**2
                term_lj6  = lj6*c6*(inv_r3-c2_switchup)**2
                term_lj   = term_lj12 - term_lj6
                evdw_temp = evdw_temp + term_lj
                coef = -12.0_wp*lj12*c12 * (inv_r6 - c1_switchup)*inv_r6*inv_r2
                coef = coef + 6.0_wp*lj6*c6*(inv_r3-c2_switchup)*inv_r3*inv_r2
              else
                coef = -12.0_wp * lj12 * inv_r12 * inv_r2
                coef = coef + 6.0_wp*lj6*inv_r6*inv_r2
                inv_r12   = inv_r12 - c1_switchdown
                inv_r6    = inv_r6 - c2_switchdown
                term_lj12 = lj12 * inv_r12
                term_lj6  = lj6 * inv_r6
                term_lj   = term_lj12 - term_lj6
                evdw_temp = evdw_temp + term_lj
              end if

            else

              if (rij2 > switchdist2) then
                c2_switch = cutoff2 - rij2
                c4_switch = c2_switch*(cutoff2+ 2.0_wp*rij2- 3.0_wp*switchdist2)
                switch    = c2_switch*c4_switch*c1_switch
                dswitch   = 4.0_wp*c1_switch*rij*(c2_switch*c2_switch-c4_switch)
              else
                switch    = 1.0_wp
                dswitch   = 0.0_wp
              end if

              term_lj12   = lj12 * inv_r12
              term_lj6    = lj6 * inv_r6
              term_lj     = term_lj12 - term_lj6
              evdw_temp   = evdw_temp + switch*term_lj
              coef   = (dswitch * term_lj * rij &
                       - switch * (12.0_wp*term_lj12-6.0_wp*term_lj6) ) * inv_r2
            end if

            shift  = (1.0_wp - (rij2*cutoff2_inv))
            dshift = -4.0_wp*shift*rij*cutoff2_inv
            shift  = shift * shift

            ! electrostatic energy and gradient
            ! 
            term_elec = charge(ix,i) * charge(iy,j) * CCOEF
            elec_temp = elec_temp + shift * ( term_elec * inv_rij ) 
            coef      = coef + (dshift*(term_elec*inv_rij) - &
                                 shift*(term_elec*inv_r2))*inv_rij

            work(1:3) = coef*dij(1:3)

            ! store force
            !
            force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
            force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) + work(1:3)

            ! virial
            !
            do m = 1, 3
              viri_local(1:3,m) = viri_local(1:3,m) + dij(1:3)*work(m)
            end do

          end if

        end do
      end do
    end do

    do m = 1, 3
      virial(1:3,m,id+1) = virial(1:3,m,id+1) - viri_local(1:3,m)
    end do

    eelec(id+1) = eelec(id+1) + elec_temp
    evdw(id+1)  = evdw(id+1) + evdw_temp

    !$omp end parallel

    return

  end subroutine compute_energy_nonbond_pbc

end module sp_energy_nonbonds_mod
