!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   sp_energy_table_cubic_mod
!> @brief   calculate nonbonded energy with table and with linear interpolation
!! @authors Jaewoon Jung(JJ)
!  
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module sp_energy_table_cubic_mod

  use sp_pairlist_str_mod
  use sp_enefunc_str_mod
  use sp_domain_str_mod
  use mpi_parallel_mod
  use constants_mod

  implicit none
  private

  ! subroutines
  !
  public  :: compute_energy_nonbond_nb14_table
  public  :: compute_energy_nonbond_table_solute
  public  :: compute_force_nonbond_table_solute
  public  :: compute_energy_nonbond_table_water
  public  :: compute_force_nonbond_table_water
  public  :: compute_energy_nonbond_table
  public  :: compute_force_nonbond_table

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_nb14_table
  !> @brief        calculate nonbonded14 energy with PME and lookup table
  !!               (linear)
  !  @authors      JJ
  !! @param[in]    domain  : domain information
  !! @param[in]    enefunc : potential energy functions
  !! @param[inout] force   : forces for each cell
  !! @param[inout] virial  : virial term of target systems
  !! @param[inout] eelec   : electrostatic energy of target systems
  !! @param[inout] evdw    : van der Waals energy of target systems
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_nb14_table(domain, enefunc, &
                                               force, virial, eelec, evdw)

    ! formal arguments
    type(s_domain),  target, intent(in)    :: domain
    type(s_enefunc), target, intent(in)    :: enefunc
    real(wp),                intent(inout) :: force(:,:,:,:)
    real(wp),                intent(inout) :: virial(3,3,nthread)
    real(wp),                intent(inout) :: eelec(nthread)
    real(wp),                intent(inout) :: evdw(nthread)

    ! local variables
    real(wp)                 :: dij(1:3), rij2
    real(wp)                 :: R, h00, h10, h01, h11, table(12)
    real(wp)                 :: term_lj12, term_lj6, term_elec
    real(wp)                 :: cutoff2, grad_coef, lj12, lj6
    real(wp)                 :: work(1:3)
    real(wp)                 :: viri_local(1:3,1:3)
    integer                  :: i, ix, iy, j, k, ij, m, L
    integer                  :: num_nb14, ini_nb14, fin_nb14
    integer                  :: ncell_local, id, omp_get_thread_num

    real(wp),        pointer :: coord(:,:,:), charge(:,:)
    real(wp),        pointer :: nb14_lj12(:,:), nb14_lj6(:,:)
    real(wp),        pointer :: table_ene(:), table_grad(:)
    real(wp),        pointer :: density, cutoff
    integer,         pointer :: cell_pair(:,:), atmcls(:,:), natom(:)
    integer,         pointer :: num_nb14_calc1(:,:), nb14_calc_list1(:,:)
    integer,         pointer :: num_nb14_calc(:,:), nb14_calc_list(:,:)


    coord           => domain%coord
    charge          => domain%charge
    cell_pair       => domain%cell_pairlist1
    atmcls          => domain%atom_cls_no
    natom           => domain%num_atom

    table_ene       => enefunc%table%table_ene
    table_grad      => enefunc%table%table_grad
    density         => enefunc%table%density
    cutoff          => enefunc%cutoffdist
    nb14_lj12       => enefunc%nb14_lj12
    nb14_lj6        => enefunc%nb14_lj6
    num_nb14_calc1  => enefunc%num_nb14_calc1
    num_nb14_calc   => enefunc%num_nb14_calc
    nb14_calc_list1 => enefunc%nb14_calc_list1
    nb14_calc_list  => enefunc%nb14_calc_list

    cutoff2         =  cutoff * cutoff
    ncell_local     =  domain%num_cell_local


    !$omp parallel default(shared)                                             &
    !$omp private(id, i, j, k, ij, ix, iy, m, ini_nb14, fin_nb14, num_nb14,    &
    !$omp         dij, rij2, L, R, term_lj12, term_lj6, term_elec, grad_coef,  &
    !$omp         work, viri_local, table, lj12, lj6, h00, h10, h01, h11)
    !
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    ! calculate energy and gradient
    !
    viri_local(1:3,1:3) = 0.0_wp

    do i = id+1, ncell_local, nthread
      num_nb14 = 0

      do ix = 1, natom(i) - 1

        ini_nb14 = num_nb14 + 1
        fin_nb14 = num_nb14 + num_nb14_calc1(ix,i)
        num_nb14 = fin_nb14

        do k = ini_nb14, fin_nb14

          iy = nb14_calc_list1(k,i)

          ! compute distance
          !
          dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,i)
          rij2     = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          ! energy and gradient
          !
          rij2  = density * rij2
          L     = int(rij2)
          R     = rij2 - L
          lj6   = nb14_lj6 (atmcls(ix,i),atmcls(iy,i))
          lj12  = nb14_lj12(atmcls(ix,i),atmcls(iy,i))
          h00   = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10   = R*(1.0_wp-R)*(1.0_wp-R)
          h01   = R*R*(3.0_wp-2.0_wp*R)
          h11   = R*R*(R-1.0_wp)

          table(1:12) = table_ene(6*L-5:6*L+6)
          term_lj12 = table(1)*h00 + table(2)*h10 + table(7)*h01 + table(8)*h11
          term_lj6  = table(3)*h00 + table(4)*h10 + table(9)*h01 + table(10)*h11
          term_elec = table(5)*h00 + table(6)*h10 + table(11)*h01 +table(12)*h11
          evdw(id+1) = evdw(id+1) + term_lj12*lj12 - term_lj6*lj6
          eelec(id+1) = eelec(id+1) + charge(ix,i)*charge(iy,i)*term_elec

          table(1:12) = table_grad(6*L-5:6*L+6)
          term_lj12 = table(1)*h00 + table(2)*h10 + table(7)*h01 + table(8)*h11
          term_lj6  = table(3)*h00 + table(4)*h10 + table(9)*h01 + table(10)*h11
          term_elec = table(5)*h00 + table(6)*h10 + table(11)*h01 +table(12)*h11
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + &
                      charge(ix,i)*charge(iy,i)*term_elec

          work(1:3) = grad_coef*dij(1:3)

          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + work(1:3)

          ! virial
          !
          do m = 1, 3
            viri_local(1:3,m) = viri_local(1:3,m) + dij(1:3)*work(m)
          end do
        end do
      end do
    end do

    do ij = id+1, maxcell, nthread

      i = cell_pair(1,ij)
      j = cell_pair(2,ij)

      num_nb14 = 0

      do ix = 1, natom(i)

        ini_nb14 = num_nb14 + 1
        fin_nb14 = num_nb14 + num_nb14_calc(ix,ij)
        num_nb14 = fin_nb14

        do k = ini_nb14, fin_nb14

          iy = nb14_calc_list(k,ij)

          ! compute distance
          !
          dij(1:3) = coord(1:3,ix,i) - coord(1:3,iy,j)
          rij2   = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          ! energy and gradient
          !
          rij2  = density * rij2
          L     = int(rij2)
          R     = rij2 - L
          lj6   = nb14_lj6(atmcls(ix,i),atmcls(iy,j))
          lj12  = nb14_lj12(atmcls(ix,i),atmcls(iy,j))
          h00   = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10   = R*(1.0_wp-R)*(1.0_wp-R)
          h01   = R*R*(3.0_wp-2.0_wp*R)
          h11   = R*R*(R-1.0_wp)

          table(1:12) = table_ene(6*L-5:6*L+6)
          term_lj12 = table(1)*h00 + table(2)*h10 + table(7)*h01 + table(8)*h11
          term_lj6  = table(3)*h00 + table(4)*h10 + table(9)*h01 + table(10)*h11
          term_elec = table(5)*h00 + table(6)*h10 + table(11)*h01 +table(12)*h11
          evdw(id+1) = evdw(id+1) + term_lj12*lj12 - term_lj6*lj6
          eelec(id+1) = eelec(id+1) + charge(ix,i)*charge(iy,j)*term_elec

          table(1:12) = table_grad(6*L-5:6*L+6)
          term_lj12 = table(1)*h00 + table(2)*h10 + table(7)*h01 + table(8)*h11
          term_lj6  = table(3)*h00 + table(4)*h10 + table(9)*h01 + table(10)*h11
          term_elec = table(5)*h00 + table(6)*h10 + table(11)*h01 +table(12)*h11
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + &
                      charge(ix,i)*charge(iy,j)*term_elec

          work(1:3) = grad_coef*dij(1:3)

          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) - work(1:3)
          force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) + work(1:3)

          ! virial
          !
          do m = 1, 3
            viri_local(1:3,m) = viri_local(1:3,m) + dij(1:3)*work(m)
          end do
        end do

      end do
    end do

    do m = 1, 3
      virial(1:3,m,id+1) = virial(1:3,m,id+1) - viri_local(1:3,m)
    end do

    !$omp end parallel

    return

  end subroutine compute_energy_nonbond_nb14_table

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_table_solute
  !> @brief        calculate nonbonded energy without solvents with lookup table
  !! @authors      JJ 
  !! @param[in]    domain   : domain information
  !! @param[in]    enefunc  : potential energy functions
  !! @param[in]    pairlist : interaction list in each domain
  !! @param[inout] force    : forces for each cell
  !! @param[inout] virial   : virial term of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : van der Waals energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_table_solute(domain, enefunc, &
                                           pairlist, force, virial, eelec, evdw)

    ! formal arguments
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_pairlist), target, intent(in)    :: pairlist
    real(wp),                 intent(inout) :: force(:,:,:,:)
    real(wp),                 intent(inout) :: virial(3,3,nthread)
    real(wp),                 intent(inout) :: eelec(nthread)
    real(wp),                 intent(inout) :: evdw(nthread)

    ! local variables
    real(wp)                  :: dij(1:3), rij2
    real(wp)                  :: lj6, lj12, switch
    real(wp)                  :: R, h00, h01, h10, h11
    real(wp)                  :: term_lj12, term_lj6, term_elec
    real(wp)                  :: cutoff2, grad_coef
    real(wp)                  :: work(1:3)
    real(wp)                  :: viri_local(1:3,1:3)
    real(wp)                  :: trans_x, trans_y, trans_z
    real(wp)                  :: rtmp(1:3), qtmp
    real(wp)                  :: elec_temp, evdw_temp
    real(wp)                  :: dij_list(3,100)
    real(wp)                  :: force_local(3), rij2_list(100)
    real(wp)                  :: force_localj(3,100)
    real(wp)                  :: ieps, jeps, eps, irmin, jrmin, rmin
    integer                   :: i, ix, iy, j, k, ij, iix, iwater, list, L, L1
    integer                   :: num_nb15, ini_nb15, fin_nb15, num_count
    integer                   :: id, omp_get_thread_num, j_list(100)
    integer                   :: ncell, ncell_local
    integer                   :: iatmcls, jatmcls

    real(wp),         pointer :: coord(:,:,:), charge(:,:)
    real(wp),         pointer :: trans1(:,:,:), trans2(:,:,:)
    real(wp),         pointer :: cell_move(:,:,:)
    real(wp),         pointer :: nonb_lj12(:,:), nonb_lj6(:,:)
    real(wp),         pointer :: density, cutoff
    real(wp),         pointer :: table_ene(:), table_grad(:)
    integer,          pointer :: cell_pairlist(:,:), atmcls(:,:)
    integer,          pointer :: natom(:), nsolute(:), nwater(:)
    integer,          pointer :: nb15_list(:,:), nb15_listw(:,:)
    integer,          pointer :: nb15_cell(:), nb15_cellw(:)
    integer,          pointer :: num_nb15_calc1(:,:), num_nb15_calc(:,:)
    integer,          pointer :: num_nb15_calcw(:,:)
    integer,          pointer :: nb15_calc_list1(:,:), nb15_calc_list(:,:)


    cell_pairlist   => domain%cell_pairlist1
    atmcls          => domain%atom_cls_no
    natom           => domain%num_atom
    nsolute         => domain%num_solute
    nwater          => domain%num_water
    coord           => domain%coord
    trans1          => domain%trans_vec
    trans2          => domain%translated
    charge          => domain%charge
    cell_move       => domain%cell_move

    table_ene       => enefunc%table%table_ene
    table_grad      => enefunc%table%table_grad
    density         => enefunc%table%density
    cutoff          => enefunc%cutoffdist
    nonb_lj12       => enefunc%nonb_lj12
    nonb_lj6        => enefunc%nonb_lj6

    nb15_list       => pairlist%nb15_list
    nb15_listw      => pairlist%nb15_listw
    nb15_cell       => pairlist%nb15_cell
    nb15_cellw      => pairlist%nb15_cellw
    num_nb15_calc1  => pairlist%num_nb15_calc1
    num_nb15_calc   => pairlist%num_nb15_calc
    num_nb15_calcw  => pairlist%num_nb15_calcw
    nb15_calc_list1 => pairlist%nb15_calc_list1
    nb15_calc_list  => pairlist%nb15_calc_list


    ncell           =  domain%num_cell_local + domain%num_cell_boundary
    ncell_local     =  domain%num_cell_local
    cutoff2         =  cutoff * cutoff


    ! calculate energy and gradient
    !

    !$omp parallel default(shared)                                          &
    !$omp private(id, i, ix, num_nb15, rtmp, qtmp, ini_nb15, fin_nb15,      &
    !$omp         k, iy, dij, rij2, L, R, term_lj12, term_lj6, grad_coef,   &
    !$omp         work, term_elec, viri_local, iwater, ij, j, trans_x,      &
    !$omp         trans_y, trans_z, iix, list, num_count, j_list, dij_list, &
    !$omp         rij2_list, force_local, force_localj, switch, lj6, lj12,  &
    !$omp         L1, elec_temp, evdw_temp, iatmcls, jatmcls, ieps, jeps,   &
    !$omp         eps, irmin, jrmin, rmin, h00, h10, h01, h11)             
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif
    do i = id+1, ncell, nthread
      do ix = 1, natom(i)
        trans2(1:3,ix,i) = coord(1:3,ix,i) + trans1(1:3,ix,i)
      end do
    end do

    !$omp barrier

    viri_local(1:3,1:3) = 0.0_wp
    elec_temp = 0.0_wp
    evdw_temp = 0.0_wp

    ! energy within a cell(solute-solute)
    !
    do i = id+1, ncell_local, nthread

      num_nb15 = 0
      do ix = 1, nsolute(i) - 1

        rtmp(1:3) = trans2(1:3,ix,i)
        qtmp = charge(ix,i)
#ifdef KCOMP
        iatmcls = atmcls(ix,i)
        ieps    = lj_coef(1,iatmcls)
        irmin   = lj_coef(2,iatmcls)
#endif
        ini_nb15  = num_nb15 + 1
        fin_nb15  = num_nb15 + num_nb15_calc1(ix,i)
        num_nb15  = fin_nb15
        num_count = 0

        do k = ini_nb15, fin_nb15

          iy = nb15_calc_list1(k,i)

          ! compute distance
          !
          dij(1:3) = rtmp(1:3) - trans2(1:3,iy,i)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff2) then
            num_count = num_count + 1
            dij_list(1:3,num_count)  = dij(1:3)
            rij2_list(num_count) = rij2
            j_list(num_count)   = iy
          end if
        end do

        force_local(1:3) = 0.0_wp
!ocl norecurrence(force)
        do k = 1, num_count
          dij(1) = dij_list(1,k)
          dij(2) = dij_list(2,k)
          dij(3) = dij_list(3,k)
          iy   = j_list(k)
          rij2  = density * rij2_list(k)
          lj6  = nonb_lj6(atmcls(ix,i),atmcls(iy,i))
          lj12 = nonb_lj12(atmcls(ix,i),atmcls(iy,i))

          L    = int(rij2)
          R    = rij2 - L
          h00   = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10   = R*(1.0_wp-R)*(1.0_wp-R)
          h01   = R*R*(3.0_wp-2.0_wp*R)
          h11   = R*R*(R-1.0_wp)

          L1 = 6*L - 5
          term_lj12 = table_ene(L1)*h00 + table_ene(L1+1)*h10
          term_lj6  = table_ene(L1+2)*h00 + table_ene(L1+3)*h10
          term_elec = table_ene(L1+4)*h00 + table_ene(L1+5)*h10
          term_lj12 = term_lj12 + table_ene(L1+6)*h01 + table_ene(L1+7)*h11
          term_lj6  = term_lj6 + table_ene(L1+8)*h01 + table_ene(L1+9)*h11
          term_elec = term_elec + table_ene(L1+10)*h01 + table_ene(L1+11)*h11
          evdw_temp = evdw_temp + term_lj12*lj12 - term_lj6*lj6
          elec_temp = elec_temp + qtmp*charge(iy,i)*term_elec

          term_lj12 = table_grad(L1)*h00 + table_grad(L1+1)*h10
          term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
          term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
          term_lj12 = term_lj12 + table_grad(L1+6)*h01 + table_grad(L1+7)*h11
          term_lj6  = term_lj6 + table_grad(L1+8)*h01 + table_grad(L1+9)*h11
          term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(iy,i)*term_elec

          work(1) = grad_coef*dij(1)
          work(2) = grad_coef*dij(2)
          work(3) = grad_coef*dij(3)

          ! store force
          !
          force_local(1) = force_local(1) - work(1)
          force_local(2) = force_local(2) - work(2)
          force_local(3) = force_local(3) - work(3)
#ifdef KCOMP
          force(1,iy,i,id+1) = force(1,iy,i,id+1) + work(1)
          force(2,iy,i,id+1) = force(2,iy,i,id+1) + work(2)
          force(3,iy,i,id+1) = force(3,iy,i,id+1) + work(3)
#else
          force_localj(1,k)  = work(1)
          force_localj(2,k)  = work(2)
          force_localj(3,k)  = work(3)
#endif

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

        end do

#ifdef KCOMP
        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
#else
        do k = 1, num_count
          iy = j_list(k)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + force_localj(1:3,k)
        end do
        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
#endif
      end do

      do ix = 1, nsolute(i)

        rtmp(1:3) = trans2(1:3,ix,i)
        qtmp = charge(ix,i)
        num_count = 0

        do iy = nsolute(i)+1, natom(i)
          ! compute distance
          !
          dij(1) = rtmp(1) - trans2(1,iy,i)
          dij(2) = rtmp(2) - trans2(2,iy,i)
          dij(3) = rtmp(3) - trans2(3,iy,i)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff2) then
            num_count = num_count + 1
            dij_list(1:3,num_count)  = dij(1:3)
            rij2_list(num_count) = rij2
            j_list(num_count)   = iy
          end if
        end do

        force_local(1:3) = 0.0_wp

!ocl norecurrence(force)
        do k = 1, num_count
          dij(1) = dij_list(1,k)
          dij(2) = dij_list(2,k)
          dij(3) = dij_list(3,k)
          iy   = j_list(k)
          rij2  = density * rij2_list(k)
          lj6  = nonb_lj6 (atmcls(ix,i),atmcls(iy,i))
          lj12 = nonb_lj12(atmcls(ix,i),atmcls(iy,i))

          L    = int(rij2)
          R    = rij2 - L
          h00   = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10   = R*(1.0_wp-R)*(1.0_wp-R)
          h01   = R*R*(3.0_wp-2.0_wp*R)
          h11   = R*R*(R-1.0_wp)

          L1 = 6*L - 5
          term_lj12 = table_ene(L1)*h00 + table_ene(L1+1)*h10
          term_lj6  = table_ene(L1+2)*h00 + table_ene(L1+3)*h10
          term_elec = table_ene(L1+4)*h00 + table_ene(L1+5)*h10
          term_lj12 = term_lj12 + table_ene(L1+6)*h01 + table_ene(L1+7)*h11
          term_lj6  = term_lj6 + table_ene(L1+8)*h01 + table_ene(L1+9)*h11
          term_elec = term_elec + table_ene(L1+10)*h01 + table_ene(L1+11)*h11
          evdw_temp = evdw_temp + term_lj12*lj12 - term_lj6*lj6
          elec_temp = elec_temp + qtmp*charge(iy,i)*term_elec

          term_lj12 = table_grad(L1)*h00 + table_grad(L1+1)*h10
          term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
          term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
          term_lj12 = term_lj12 + table_grad(L1+6)*h01 + table_grad(L1+7)*h11
          term_lj6  = term_lj6 + table_grad(L1+8)*h01 + table_grad(L1+9)*h11
          term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(iy,i)*term_elec

          work(1) = grad_coef*dij(1)
          work(2) = grad_coef*dij(2)
          work(3) = grad_coef*dij(3)

          ! store force
          !
          force_local(1) = force_local(1) - work(1)
          force_local(2) = force_local(2) - work(2)
          force_local(3) = force_local(3) - work(3)
#ifdef KCOMP
          force(1,iy,i,id+1) = force(1,iy,i,id+1) + work(1)
          force(2,iy,i,id+1) = force(2,iy,i,id+1) + work(2)
          force(3,iy,i,id+1) = force(3,iy,i,id+1) + work(3)
#else
          force_localj(1,k)  = work(1)
          force_localj(2,k)  = work(2)
          force_localj(3,k)  = work(3)
#endif

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

        end do

#ifdef KCOMP
        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
#else
        do k = 1, num_count
          iy = j_list(k)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + force_localj(1:3,k)
        end do
        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
#endif
      end do
    end do


    ! interaction between different cells
    !
    do ij = id+1, maxcell, nthread

      i = cell_pairlist(1,ij)
      j = cell_pairlist(2,ij)

      num_nb15 = 0

      trans_x = cell_move(1,j,i)
      trans_y = cell_move(2,j,i)
      trans_z = cell_move(3,j,i)

      do iix = 1, nb15_cell(ij)

        ix   = nb15_list(iix,ij)
        rtmp(1:3) = trans2(1:3,ix,i)
        qtmp = charge(ix,i)

        ini_nb15 = num_nb15 + 1
        fin_nb15 = num_nb15 + num_nb15_calc(ix,ij)
        num_nb15 = fin_nb15

        num_count = 0

        do k = ini_nb15, fin_nb15
          iy   = nb15_calc_list(k,ij)
          dij(1) = rtmp(1) - trans2(1,iy,j) + trans_x
          dij(2) = rtmp(2) - trans2(2,iy,j) + trans_y
          dij(3) = rtmp(3) - trans2(3,iy,j) + trans_z
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff2) then
            num_count = num_count + 1
            dij_list(1:3,num_count)  = dij(1:3)
            rij2_list(num_count) = rij2
            j_list(num_count)   = iy
          end if
        end do

        force_local(1:3) = 0.0_wp

!ocl norecurrence(force)
        do k = 1, num_count
          dij(1) = dij_list(1,k)
          dij(2) = dij_list(2,k)
          dij(3) = dij_list(3,k)
          iy   = j_list(k)
          rij2  = density * rij2_list(k)
          lj6  = nonb_lj6 (atmcls(ix,i),atmcls(iy,j))
          lj12 = nonb_lj12(atmcls(ix,i),atmcls(iy,j))

          L    = int(rij2)
          R    = rij2 - L
          h00   = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10   = R*(1.0_wp-R)*(1.0_wp-R)
          h01   = R*R*(3.0_wp-2.0_wp*R)
          h11   = R*R*(R-1.0_wp)

          L1 = 6*L - 5
          term_lj12 = table_ene(L1)*h00 + table_ene(L1+1)*h10
          term_lj6  = table_ene(L1+2)*h00 + table_ene(L1+3)*h10
          term_elec = table_ene(L1+4)*h00 + table_ene(L1+5)*h10
          term_lj12 = term_lj12 + table_ene(L1+6)*h01 + table_ene(L1+7)*h11
          term_lj6  = term_lj6 + table_ene(L1+8)*h01 + table_ene(L1+9)*h11
          term_elec = term_elec + table_ene(L1+10)*h01 + table_ene(L1+11)*h11
          evdw_temp = evdw_temp + term_lj12*lj12 - term_lj6*lj6
          elec_temp = elec_temp + qtmp*charge(iy,j)*term_elec

          term_lj12 = table_grad(L1)*h00 + table_grad(L1+1)*h10
          term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
          term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
          term_lj12 = term_lj12 + table_grad(L1+6)*h01 + table_grad(L1+7)*h11
          term_lj6  = term_lj6 + table_grad(L1+8)*h01 + table_grad(L1+9)*h11
          term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(iy,j)*term_elec

          work(1) = grad_coef*dij(1)
          work(2) = grad_coef*dij(2)
          work(3) = grad_coef*dij(3)

          ! store force
          !
          force_local(1) = force_local(1) - work(1)
          force_local(2) = force_local(2) - work(2)
          force_local(3) = force_local(3) - work(3)
#ifdef KCOMP
          force(1,iy,j,id+1) = force(1,iy,j,id+1) + work(1)
          force(2,iy,j,id+1) = force(2,iy,j,id+1) + work(2)
          force(3,iy,j,id+1) = force(3,iy,j,id+1) + work(3)
#else
          force_localj(1,k) = work(1)
          force_localj(2,k) = work(2)
          force_localj(3,k) = work(3)
#endif

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

        end do

#ifndef KCOMP
        do k = 1, num_count
          iy = j_list(k)
          force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) + force_localj(1:3,k)
        end do
#endif
        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)

      end do


      do iix = 1, nb15_cellw(ij)

        iwater   = nb15_listw(iix,ij)
        ini_nb15 = num_nb15 + 1
        fin_nb15 = num_nb15 + num_nb15_calcw(iwater,ij)
        num_nb15 = fin_nb15

        do list = 1, 3
          ix = nsolute(i) + 3*(iwater-1) + list
          num_count = 0

          do k = ini_nb15, fin_nb15
            iy = nb15_calc_list(k,ij)

            ! compute distance
            !
            dij(1) = trans2(1,ix,i) - trans2(1,iy,j) + trans_x
            dij(2) = trans2(2,ix,i) - trans2(2,iy,j) + trans_y
            dij(3) = trans2(3,ix,i) - trans2(3,iy,j) + trans_z
            rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

            if (rij2 < cutoff2) then
              num_count = num_count + 1
              dij_list(1:3,num_count)  = dij(1:3)
              rij2_list(num_count) = rij2
              j_list(num_count)   = iy
            end if
          end do

          force_local(1:3) = 0.0_wp

!ocl norecurrence(force)
          do k = 1, num_count
            dij(1) = dij_list(1,k)
            dij(2) = dij_list(2,k)
            dij(3) = dij_list(3,k)
            iy   = j_list(k)
            rij2  = density * rij2_list(k)
            lj6  = nonb_lj6 (atmcls(ix,i),atmcls(iy,j))
            lj12 = nonb_lj12(atmcls(ix,i),atmcls(iy,j))

            L    = int(rij2)
            R    = rij2 - L
            h00   = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
            h10   = R*(1.0_wp-R)*(1.0_wp-R)
            h01   = R*R*(3.0_wp-2.0_wp*R)
            h11   = R*R*(R-1.0_wp)

            L1 = 6*L - 5
            term_lj12 = table_ene(L1)*h00 + table_ene(L1+1)*h10
            term_lj6  = table_ene(L1+2)*h00 + table_ene(L1+3)*h10
            term_elec = table_ene(L1+4)*h00 + table_ene(L1+5)*h10
            term_lj12 = term_lj12 + table_ene(L1+6)*h01 + table_ene(L1+7)*h11
            term_lj6  = term_lj6 + table_ene(L1+8)*h01 + table_ene(L1+9)*h11
            term_elec = term_elec + table_ene(L1+10)*h01 + table_ene(L1+11)*h11
            evdw_temp = evdw_temp + term_lj12*lj12 - term_lj6*lj6
            elec_temp = elec_temp + charge(ix,i)*charge(iy,j)*term_elec

            term_lj12 = table_grad(L1)*h00 + table_grad(L1+1)*h10
            term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
            term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
            term_lj12 = term_lj12 + table_grad(L1+6)*h01 + table_grad(L1+7)*h11
            term_lj6  = term_lj6 + table_grad(L1+8)*h01 + table_grad(L1+9)*h11
            term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
            grad_coef = term_lj12*lj12 - term_lj6*lj6 + charge(ix,i)*charge(iy,j)*term_elec

            work(1) = grad_coef*dij(1)
            work(2) = grad_coef*dij(2)
            work(3) = grad_coef*dij(3)

            ! store force
            !
            force_local(1) = force_local(1) - work(1)
            force_local(2) = force_local(2) - work(2)
            force_local(3) = force_local(3) - work(3)
#ifdef KCOMP
            force(1,iy,j,id+1) = force(1,iy,j,id+1) + work(1)
            force(2,iy,j,id+1) = force(2,iy,j,id+1) + work(2)
            force(3,iy,j,id+1) = force(3,iy,j,id+1) + work(3)
#else
            force_localj(1,k) = work(1)
            force_localj(2,k) = work(2)
            force_localj(3,k) = work(3)
#endif

            ! virial
            !
            viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
            viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
            viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
            viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
            viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
            viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

          end do

#ifndef KCOMP
          do k = 1, num_count
            iy = j_list(k)
            force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) + force_localj(1:3,k)
          end do
#endif
          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)

        end do
      end do

    end do

    do i = 1, 2
      do j = i+1, 3
        virial(i,j,id+1) = virial(i,j,id+1) - viri_local(j,i)
        virial(j,i,id+1) = virial(j,i,id+1) - viri_local(j,i)
      end do
    end do
    do i = 1, 3
      virial(i,i,id+1) = virial(i,i,id+1) - viri_local(i,i)
    end do

    eelec(id+1) = eelec(id+1) + elec_temp
    evdw(id+1)  = evdw(id+1) + evdw_temp

    !$omp end parallel

    return

  end subroutine compute_energy_nonbond_table_solute

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_force_nonbond_table_solute
  !> @brief        calculate nonbonded force without solvents with lookup table
  !! @authors      JJ 
  !! @param[in]    domain   : domain information
  !! @param[in]    enefunc  : potential energy functions
  !! @param[in]    pairlist : interaction list in each domain
  !! @param[inout] force    : forces for each cell
  !! @param[inout] virial   : virial term of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_force_nonbond_table_solute(domain, enefunc, &
                                                       pairlist, force, virial)

    ! formal arguments
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_pairlist), target, intent(in)    :: pairlist
    real(wp),                 intent(inout) :: force(:,:,:,:)
    real(wp),                 intent(inout) :: virial(3,3,nthread)

    ! local variables
    real(wp)                  :: dij(1:3), rij2
    real(wp)                  :: R, h00, h10, h01, h11, lj12, lj6
    real(wp)                  :: term_lj12, term_lj6, term_elec
    real(wp)                  :: cutoff2, grad_coef
    real(wp)                  :: work(1:3)
    real(wp)                  :: viri_local(1:3,1:3)
    real(wp)                  :: trans_x, trans_y, trans_z
    real(wp)                  :: rtmp(1:3), qtmp
    real(wp)                  :: dij_list(4,100)
    real(wp)                  :: force_local(3)
    real(wp)                  :: force_localj(3,100)
    real(wp)                  :: ieps, jeps, eps, irmin, jrmin, rmin
    integer                   :: i, ix, iy, j, k, ij, iix, iwater, list, L, L1
    integer                   :: num_nb15, ini_nb15, fin_nb15, num_count
    integer                   :: id, omp_get_thread_num, j_list(100)
    integer                   :: ncell, ncell_local
    integer                   :: iatmcls,jatmcls

    real(wp),         pointer :: coord(:,:,:), charge(:,:)
    real(wp),         pointer :: trans1(:,:,:), trans2(:,:,:)
    real(wp),         pointer :: cell_move(:,:,:)
    real(wp),         pointer :: nonb_lj12(:,:), nonb_lj6(:,:)
    real(wp),         pointer :: density, cutoff
    real(wp),         pointer :: table_grad(:)
    integer,          pointer :: cell_pairlist(:,:), atmcls(:,:)
    integer,          pointer :: natom(:), nsolute(:), nwater(:)
    integer,          pointer :: nb15_list(:,:), nb15_listw(:,:)
    integer,          pointer :: nb15_cell(:), nb15_cellw(:)
    integer,          pointer :: num_nb15_calc1(:,:), num_nb15_calc(:,:)
    integer,          pointer :: num_nb15_calcw(:,:)
    integer,          pointer :: nb15_calc_list1(:,:), nb15_calc_list(:,:)


    cell_pairlist   => domain%cell_pairlist1
    atmcls          => domain%atom_cls_no
    natom           => domain%num_atom
    nsolute         => domain%num_solute
    nwater          => domain%num_water
    coord           => domain%coord
    trans1          => domain%trans_vec
    trans2          => domain%translated
    charge          => domain%charge
    cell_move       => domain%cell_move

    table_grad      => enefunc%table%table_grad
    density         => enefunc%table%density
    cutoff          => enefunc%cutoffdist
    nonb_lj12       => enefunc%nonb_lj12
    nonb_lj6        => enefunc%nonb_lj6 

    nb15_list       => pairlist%nb15_list
    nb15_listw      => pairlist%nb15_listw
    nb15_cell       => pairlist%nb15_cell
    nb15_cellw      => pairlist%nb15_cellw
    num_nb15_calc1  => pairlist%num_nb15_calc1
    num_nb15_calc   => pairlist%num_nb15_calc 
    num_nb15_calcw  => pairlist%num_nb15_calcw
    nb15_calc_list1 => pairlist%nb15_calc_list1
    nb15_calc_list  => pairlist%nb15_calc_list

    ncell           =  domain%num_cell_local + domain%num_cell_boundary
    ncell_local     =  domain%num_cell_local
    cutoff2         =  cutoff * cutoff

    ! calculate energy and gradient
    !

    !$omp parallel default(shared)                                             &
    !$omp private(id, i, ix, num_nb15, rtmp, qtmp, ini_nb15, fin_nb15,         &
    !$omp         k, iy, rij2, L, R, term_lj12, term_lj6, grad_coef,           &
    !$omp         work, term_elec, viri_local, iwater, ij, j, trans_x,         &
    !$omp         trans_y, trans_z, iix, list, num_count, j_list, dij_list,    &
    !$omp         force_local, force_localj, lj12, lj6, L1, iatmcls, jatmcls,  &
    !$omp         ieps, jeps, eps, irmin, jrmin, rmin, dij, h00, h01, h10, h11)
    !
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif
    do i = id+1, ncell, nthread
      do ix = 1, natom(i)
        trans2(1:3,ix,i) = coord(1:3,ix,i) + trans1(1:3,ix,i)
      end do
    end do

    !$omp barrier

    viri_local(1:3,1:3) = 0.0_wp

    do i = id+1, ncell_local, nthread
      num_nb15 = 0

      do ix = 1, nsolute(i) - 1

        rtmp(1:3) = trans2(1:3,ix,i)
        qtmp = charge(ix,i)
        ini_nb15 = num_nb15 + 1
        fin_nb15 = num_nb15 + num_nb15_calc1(ix,i)
        num_nb15 = fin_nb15

        num_count = 0

        do k = ini_nb15, fin_nb15

          iy   = nb15_calc_list1(k,i)
          dij(1:3) = rtmp(1:3) - trans2(1:3,iy,i)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff2) then
            num_count = num_count + 1
            dij_list(1:3,num_count) = dij(1:3)
            dij_list(4,num_count) = rij2
            j_list(num_count) = iy
          end if

        end do

        force_local(1:3) = 0.0_wp

!ocl norecurrence(force)
        do k = 1, num_count

          iy = j_list(k)
          dij(1) = dij_list(1,k)
          dij(2) = dij_list(2,k)
          dij(3) = dij_list(3,k)
          rij2   = density * dij_list(4,k)
          lj12   = nonb_lj12(atmcls(ix,i),atmcls(iy,i))
          lj6    = nonb_lj6(atmcls(ix,i),atmcls(iy,i))

          L    = int(rij2)
          R    = rij2 - L
          h00  = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10  = R*(1.0_wp-R)*(1.0_wp-R)
          h01  = R*R*(3.0_wp-2.0_wp*R)
          h11  = R*R*(R-1.0_wp)

          L1 = 6*L - 5
          term_lj12 = table_grad(L1)*h00 + table_grad(L1+1)*h10
          term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
          term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
          term_lj12 = term_lj12 + table_grad(L1+6)*h01 + table_grad(L1+7)*h11
          term_lj6  = term_lj6 + table_grad(L1+8)*h01 + table_grad(L1+9)*h11
          term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(iy,i)*term_elec

          work(1) = grad_coef*dij(1)
          work(2) = grad_coef*dij(2)
          work(3) = grad_coef*dij(3)

          ! store force
          !
          force_local(1) = force_local(1) - work(1)
          force_local(2) = force_local(2) - work(2)
          force_local(3) = force_local(3) - work(3)
#ifdef KCOMP
          force(1,iy,i,id+1) = force(1,iy,i,id+1) + work(1)
          force(2,iy,i,id+1) = force(2,iy,i,id+1) + work(2)
          force(3,iy,i,id+1) = force(3,iy,i,id+1) + work(3)
#else
          force_localj(1,k) = work(1)
          force_localj(2,k) = work(2)
          force_localj(3,k) = work(3)
#endif

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)
        end do

#ifdef KCOMP
        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
#else
        do k = 1, num_count
          iy = j_list(k)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + force_localj(1:3,k)
        end do
        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
#endif
      end do

      do ix = 1, nsolute(i) 

        rtmp(1:3) = trans2(1:3,ix,i)
        qtmp = charge(ix,i)
        num_count = 0

        do iy = nsolute(i)+1, natom(i)

          ! compute distance
          !
          dij(1:3) = rtmp(1:3) - trans2(1:3,iy,i)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff2) then
            num_count = num_count + 1
            dij_list(1:3,num_count) = dij(1:3)
            dij_list(4,num_count) = rij2
            j_list(num_count) = iy
          end if
        end do

        force_local(1:3) = 0.0_wp

!ocl norecurrence(force)
        do k = 1, num_count
          iy   = j_list(k)
          dij(1) = dij_list(1,k)
          dij(2) = dij_list(2,k)
          dij(3) = dij_list(3,k)
          rij2   = density * dij_list(4,k)
          lj12   = nonb_lj12(atmcls(ix,i),atmcls(iy,i))
          lj6    = nonb_lj6(atmcls(ix,i),atmcls(iy,i))

          L     = int(rij2)
          R     = rij2 - L
          h00   = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10   = R*(1.0_wp-R)*(1.0_wp-R)
          h01   = R*R*(3.0_wp-2.0_wp*R)
          h11   = R*R*(R-1.0_wp)

          L1 = 6*L - 5
          term_lj12 = table_grad(L1)*h00 + table_grad(L1+1)*h10
          term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
          term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
          term_lj12 = term_lj12 + table_grad(L1+6)*h01 + table_grad(L1+7)*h11
          term_lj6  = term_lj6 + table_grad(L1+8)*h01 + table_grad(L1+9)*h11
          term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(iy,i)*term_elec

          work(1) = grad_coef*dij(1)
          work(2) = grad_coef*dij(2)
          work(3) = grad_coef*dij(3)

          ! store force
          !
          force_local(1) = force_local(1) - work(1)
          force_local(2) = force_local(2) - work(2)
          force_local(3) = force_local(3) - work(3)
#ifdef KCOMP
          force(1,iy,i,id+1) = force(1,iy,i,id+1) + work(1)
          force(2,iy,i,id+1) = force(2,iy,i,id+1) + work(2)
          force(3,iy,i,id+1) = force(3,iy,i,id+1) + work(3)
#else
          force_localj(1,k) = work(1)
          force_localj(2,k) = work(2)
          force_localj(3,k) = work(3)
#endif

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)
        end do 

#ifdef KCOMP
        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
#else
        do k = 1, num_count
          iy = j_list(k)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + force_localj(1:3,k)
        end do
        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
#endif
      end do
    end do

    ! interaction between different cells
    !
    do ij = id+1, maxcell, nthread

      i = cell_pairlist(1,ij)
      j = cell_pairlist(2,ij)

      num_nb15 = 0

      trans_x = cell_move(1,j,i)
      trans_y = cell_move(2,j,i)
      trans_z = cell_move(3,j,i)

      do iix = 1, nb15_cell(ij)

        ix = nb15_list(iix,ij)
        rtmp(1:3) = trans2(1:3,ix,i)
        qtmp = charge(ix,i)

        ini_nb15  = num_nb15 + 1
        fin_nb15  = num_nb15 + num_nb15_calc(ix,ij)
        num_nb15  = fin_nb15

        num_count = 0

        do k = ini_nb15, fin_nb15
          iy = nb15_calc_list(k,ij)

          dij(1) = rtmp(1) - trans2(1,iy,j) + trans_x
          dij(2) = rtmp(2) - trans2(2,iy,j) + trans_y
          dij(3) = rtmp(3) - trans2(3,iy,j) + trans_z
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff2) then
            num_count = num_count + 1
            dij_list(1:3,num_count) = dij(1:3)
            dij_list(4,num_count) = rij2
            j_list(num_count)   = iy
          end if
        end do

        force_local(1:3) = 0.0_wp

!ocl norecurrence(force)
        do k = 1, num_count

          iy   = j_list(k)
          dij(1) = dij_list(1,k)
          dij(2) = dij_list(2,k)
          dij(3) = dij_list(3,k)
          rij2   = density * dij_list(4,k)
          lj12   = nonb_lj12(atmcls(ix,i),atmcls(iy,j))
          lj6    = nonb_lj6(atmcls(ix,i),atmcls(iy,j))

          L     = int(rij2)
          R     = rij2 - L
          h00   = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10   = R*(1.0_wp-R)*(1.0_wp-R)
          h01   = R*R*(3.0_wp-2.0_wp*R)
          h11   = R*R*(R-1.0_wp)

          L1 = 6*L - 5
          term_lj12 = table_grad(L1)*h00 + table_grad(L1+1)*h10
          term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
          term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
          term_lj12 = term_lj12 + table_grad(L1+6)*h01 + table_grad(L1+7)*h11
          term_lj6  = term_lj6 + table_grad(L1+8)*h01 + table_grad(L1+9)*h11
          term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(iy,j)*term_elec

          work(1) = grad_coef*dij(1)
          work(2) = grad_coef*dij(2)
          work(3) = grad_coef*dij(3)

          ! store force
          !
          force_local(1) = force_local(1) - work(1)
          force_local(2) = force_local(2) - work(2)
          force_local(3) = force_local(3) - work(3)
#ifdef KCOMP
          force(1,iy,j,id+1) = force(1,iy,j,id+1) + work(1)
          force(2,iy,j,id+1) = force(2,iy,j,id+1) + work(2)
          force(3,iy,j,id+1) = force(3,iy,j,id+1) + work(3)
#else
          force_localj(1,k) = work(1)
          force_localj(2,k) = work(2)
          force_localj(3,k) = work(3)
#endif
          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)
        end do

        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
#ifndef KCOMP
        do k = 1, num_count
          iy = j_list(k)
          force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) + force_localj(1:3,k)
        end do
#endif
      end do

      do iix = 1, nb15_cellw(ij)

        iwater = nb15_listw(iix,ij)

        ini_nb15 = num_nb15 + 1
        fin_nb15 = num_nb15 + num_nb15_calcw(iwater,ij)
        num_nb15 = fin_nb15

        do list = 1, 3

          ix   = nsolute(i) + 3*(iwater-1) + list
          rtmp(1:3) = trans2(1:3,ix,i)
          qtmp = charge(ix,i)
          num_count = 0

          do k = ini_nb15, fin_nb15
            iy = nb15_calc_list(k,ij)

            ! compute distance
            !
            dij(1) = rtmp(1) - trans2(1,iy,j) + trans_x
            dij(2) = rtmp(2) - trans2(2,iy,j) + trans_y
            dij(3) = rtmp(3) - trans2(3,iy,j) + trans_z
            rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

            if (rij2 < cutoff2) then
              num_count = num_count + 1
              dij_list(1:3,num_count) = dij(1:3)
              dij_list(4,num_count) = rij2
              j_list(num_count) = iy
            end if
          end do

          force_local(1:3) = 0.0_wp

!ocl norecurrence(force)
          do k = 1, num_count
            iy = j_list(k)
            dij(1) = dij_list(1,k)
            dij(2) = dij_list(2,k)
            dij(3) = dij_list(3,k)
            iy = j_list(k)
            rij2 = density * dij_list(4,k)
            lj12 = nonb_lj12(atmcls(ix,i),atmcls(iy,j))
            lj6  = nonb_lj6(atmcls(ix,i),atmcls(iy,j))

            L    = int(rij2)
            R    = rij2 - L
            h00   = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
            h10   = R*(1.0_wp-R)*(1.0_wp-R)
            h01   = R*R*(3.0_wp-2.0_wp*R)
            h11   = R*R*(R-1.0_wp)

            L1 = 6*L - 5
            term_lj12 = table_grad(L1)*h00 + table_grad(L1+1)*h10
            term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
            term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
            term_lj12 = term_lj12 + table_grad(L1+6)*h01 + table_grad(L1+7)*h11
            term_lj6  = term_lj6 + table_grad(L1+8)*h01 + table_grad(L1+9)*h11
            term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
            grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(iy,j)*term_elec

            work(1) = grad_coef*dij(1)
            work(2) = grad_coef*dij(2)
            work(3) = grad_coef*dij(3)

            ! store force
            !
            force_local(1) = force_local(1) - work(1)
            force_local(2) = force_local(2) - work(2)
            force_local(3) = force_local(3) - work(3)
#ifdef KCOMP
            force(1,iy,j,id+1) = force(1,iy,j,id+1) + work(1)
            force(2,iy,j,id+1) = force(2,iy,j,id+1) + work(2)
            force(3,iy,j,id+1) = force(3,iy,j,id+1) + work(3)
#else
            force_localj(1,k) = work(1)
            force_localj(2,k) = work(2)
            force_localj(3,k) = work(3)
#endif

            ! virial
            !
            viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
            viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
            viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
            viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
            viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
            viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)
          end do

#ifdef KCOMP
          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
#else
          do k = 1, num_count
            iy = j_list(k)
            force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) + force_localj(1:3,k)
          end do
          force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
#endif

        end do

      end do

    end do

    do i = 1, 2
      do j = i+1, 3
        virial(i,j,id+1) = virial(i,j,id+1) - viri_local(j,i)
        virial(j,i,id+1) = virial(j,i,id+1) - viri_local(j,i)
      end do
    end do
    do i = 1, 3
      virial(i,i,id+1) = virial(i,i,id+1) - viri_local(i,i)
    end do

    !$omp end parallel

    return

  end subroutine compute_force_nonbond_table_solute

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_table_water
  !> @brief        calculate nonbonded energy among water molecules
  !! @authors      JJ 
  !! @param[in]    domain   : domain information
  !! @param[in]    enefunc  : potential energy functions
  !! @param[in]    pairlist : interaction list in each domain
  !! @param[inout] force    : forces for each cell
  !! @param[inout] virial   : virial term of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : van der Waals energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_table_water(domain, enefunc, &
                                           pairlist, force, virial, eelec, evdw)

    ! formal arguments
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_pairlist), target, intent(in)    :: pairlist
    real(wp),                 intent(inout) :: force(:,:,:,:)
    real(wp),                 intent(inout) :: virial(3,3,nthread)
    real(wp),                 intent(inout) :: eelec(nthread)
    real(wp),                 intent(inout) :: evdw(nthread)

    ! local variables
    real(wp)                  :: dij(1:3), rij2
    real(wp)                  :: R, h00, h10, h01, h11
    real(wp)                  :: cutoff2, cutoff_small, cutoff_large
    real(wp)                  :: term_lj, grad_coef, term_elec
    real(wp)                  :: work(1:3)
    real(wp)                  :: viri_local(1:3,1:3)
    real(wp)                  :: trans_x, trans_y, trans_z
    real(wp)                  :: rtmp(3,3)
    real(wp)                  :: force_local(3,3)
    real(wp)                  :: elec_temp, evdw_temp
    real(wp)                  :: force_localj(3,3,3,50)
    real(wp)                  :: force_localj1(3,50,3,3)
    real(wp)                  :: dij_list(4,50,3,3)
    integer                   :: i, iy, j, k, ij, iix, iwater, jwater
    integer                   :: num_nb15, ini_nb15, fin_nb15
    integer                   :: id, omp_get_thread_num
    integer                   :: i1, j1, L, L1
    integer                   :: num_count(3,3), c, j_list(50,3,3)
    integer                   :: small_count, j_small(50)
    integer                   :: ilist(3)
    integer                   :: ncell_local
    integer                   :: water_type(1:3,1:3), watertype

    real(wp),         pointer :: coord(:,:,:), cell_move(:,:,:)
    real(wp),         pointer :: trans1(:,:,:), trans2(:,:,:)
    real(wp),         pointer :: density, cutoff
    real(wp),         pointer :: table_de_WW(:,:)
    real(wp),         pointer :: table_ene_WW(:,:)
    integer,          pointer :: cell_pairlist(:,:), natom(:)
    integer,          pointer :: nwater(:), nsolute(:)
    integer,          pointer :: nb15_cellww(:), nb15_listww(:,:)
    integer,          pointer :: num_nb15_calcww(:,:), nb15_calc_water(:,:)


    cell_pairlist   => domain%cell_pairlist1
    natom           => domain%num_atom
    nsolute         => domain%num_solute
    nwater          => domain%num_water
    coord           => domain%coord
    trans1          => domain%trans_vec
    trans2          => domain%translated
    cell_move       => domain%cell_move

    density         => enefunc%table%density
    table_ene_WW    => enefunc%table%table_ene_WW
    table_de_WW     => enefunc%table%table_de_WW
    cutoff          => enefunc%cutoffdist

    nb15_cellww     => pairlist%nb15_cellww
    nb15_listww     => pairlist%nb15_listww
    num_nb15_calcww => pairlist%num_nb15_calcww
    nb15_calc_water => pairlist%nb15_calc_water

    ncell_local     =  domain%num_cell_local
    cutoff2         =  cutoff * cutoff
    cutoff_small    =  (cutoff-2.0_wp)*(cutoff-2.0_wp)
    cutoff_large    =  (cutoff+2.0_wp)*(cutoff+2.0_wp)

    water_type(1,1)     = 1
    water_type(1,2:3)   = 2
    water_type(2:3,1)   = 2
    water_type(2:3,2:3) = 3

    ! calculate energy and gradient
    !

    !$omp parallel default(shared)                                            &
    !$omp private(id, i, j, iwater, i1, ilist, rtmp, force_local, viri_local, &
    !$omp         jwater, j1, j_list, rij2, L, R, term_lj, term_elec,         &
    !$omp         grad_coef, work, ij, ini_nb15, fin_nb15, num_nb15,          &
    !$omp         trans_x, trans_y, trans_z, iix, iy, num_count, k, c,        &
    !$omp         small_count, j_small, force_localj, force_localj1, L1,      &
    !$omp         elec_temp, evdw_temp, dij, dij_list, h00, h10, h01, h11,    &
    !$omp         watertype)                     
#ifdef OMP
    id  = omp_get_thread_num()
#else
    id  = 0
#endif
    viri_local(1:3,1:3) = 0.0_wp
    elec_temp = 0.0_wp
    evdw_temp = 0.0_wp

    ! energy within a cell
    !
    do i = id+1, ncell_local, nthread

      do iwater = 1, nwater(i) - 1

        do i1 = 1,3
          ilist(i1) = nsolute(i) + 3*(iwater-1) + i1
          rtmp(1:3,i1) = trans2(1:3,ilist(i1),i)
        end do

        force_local(1:3,1:3) = 0.0_wp
        num_count(1:3,1:3) = 0
        small_count = 0

        do jwater = iwater+1, nwater(i)

          iy   = nsolute(i) + 3*(jwater-1) + 1
          dij(1:3) = rtmp(1:3,1) - trans2(1:3,iy,i)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff_small) then
            small_count = small_count + 1
            j_small(small_count) = jwater
          else if (rij2 < cutoff_large) then
            do i1 = 1, 3
              do j1 = 1, 3
                iy = nsolute(i)+3*(jwater-1) + j1
                dij(1:3) = rtmp(1:3,i1) - trans2(1:3,iy,i)
                rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
                if (rij2 < cutoff2) then
                  num_count(j1,i1) = num_count(j1,i1) + 1
                  c = num_count(j1,i1)
                  dij_list(1:3,c,j1,i1) = dij(1:3)
                  dij_list(4,c,j1,i1) = rij2
                  j_list(c,j1,i1) = iy
                end if
              end do
            end do
          end if

        end do

        do k = 1, small_count
          jwater = j_small(k)

          do i1 = 1, 3
            do j1 = 1, 3

              iy = nsolute(i) + 3*(jwater-1) + j1
              dij(1:3) = rtmp(1:3,i1) - trans2(1:3,iy,i)
              rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
              rij2 = density * rij2
              L    = int(rij2)
              R    = rij2 - L
              h00  = (1.0_wp+2.0_wp*R) * (1.0_wp-R) * (1.0_wp-R)
              h10  = R * (1.0_wp-R) * (1.0_wp-R)
              h01  = R * R * (3.0_wp-2.0_wp*R)
              h11  = R * R * (R-1.0_wp)
              watertype = water_type(j1,i1)

              L1   = 6*L-5
              term_lj    = table_ene_WW(L1,watertype)*h00 +    &
                           table_ene_WW(L1+1,watertype)*h10
              term_elec  = table_ene_WW(L1+2,watertype)*h00 +  &
                           table_ene_WW(L1+3,watertype)*h10
              grad_coef  = table_ene_WW(L1+4,watertype)*h00 +  &
                           table_ene_WW(L1+5,watertype)*h10
              term_lj    = term_lj + table_ene_WW(L1+6,watertype)*h01 +    &
                                     table_ene_WW(L1+7,watertype)*h11
              term_elec  = term_elec + table_ene_WW(L1+8,watertype)*h01 +  &
                                       table_ene_WW(L1+9,watertype)*h11
              grad_coef  = grad_coef + table_ene_WW(L1+10,watertype)*h01 +  &
                                      table_ene_WW(L1+11,watertype)*h11

              evdw_temp = evdw_temp + term_lj
              elec_temp = elec_temp + term_elec
              work(1)   = grad_coef*dij(1)
              work(2)   = grad_coef*dij(2)
              work(3)   = grad_coef*dij(3)
              force_local(1,i1) = force_local(1,i1) - work(1)
              force_local(2,i1) = force_local(2,i1) - work(2)
              force_local(3,i1) = force_local(3,i1) - work(3)
              force_localj(1,i1,j1,k) = work(1)
              force_localj(2,i1,j1,k) = work(2)
              force_localj(3,i1,j1,k) = work(3)
              viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
              viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
              viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
              viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
              viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
              viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

            end do
          end do

        end do

        do i1 = 1, 3
          do j1 = 1, 3

            do k = 1, num_count(j1,i1)
              iy = j_list(k,j1,i1)
              dij(1:3) = dij_list(1:3,k,j1,i1)
              rij2     = density * dij_list(4,k,j1,i1) 
              L    = int(rij2)
              R    = rij2 - L
              h00  = (1.0_wp+2.0_wp*R) * (1.0_wp-R) * (1.0_wp-R)
              h10  = R * (1.0_wp-R) * (1.0_wp-R)
              h01  = R * R * (3.0_wp-2.0_wp*R)
              h11  = R * R * (R-1.0_wp)
              watertype = water_type(j1,i1)

              L1   = 6*L-5
              term_lj    = table_ene_WW(L1,watertype)*h00 +    &
                           table_ene_WW(L1+1,watertype)*h10
              term_elec  = table_ene_WW(L1+2,watertype)*h00 +  &
                           table_ene_WW(L1+3,watertype)*h10
              grad_coef  = table_ene_WW(L1+4,watertype)*h00 +  &
                           table_ene_WW(L1+5,watertype)*h10
              term_lj    = term_lj + table_ene_WW(L1+6,watertype)*h01 +    &
                                     table_ene_WW(L1+7,watertype)*h11
              term_elec  = term_elec + table_ene_WW(L1+8,watertype)*h01 +  &
                                       table_ene_WW(L1+9,watertype)*h11
              grad_coef = grad_coef + table_ene_WW(L1+10,watertype)*h01 +  &
                                      table_ene_WW(L1+11,watertype)*h11

              evdw_temp = evdw_temp + term_lj
              elec_temp = elec_temp + term_elec
              work(1)   = grad_coef*dij(1)
              work(2)   = grad_coef*dij(2)
              work(3)   = grad_coef*dij(3)
              force_local(1,i1) = force_local(1,i1) - work(1)
              force_local(2,i1) = force_local(2,i1) - work(2)
              force_local(3,i1) = force_local(3,i1) - work(3)
              force_localj1(1,k,j1,i1) = work(1)
              force_localj1(2,k,j1,i1) = work(2)
              force_localj1(3,k,j1,i1) = work(3)
              viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
              viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
              viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
              viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
              viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
              viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)
            end do
  
          end do
        end do

        do k = 1, small_count
          jwater = j_small(k)
          do j1 = 1, 3
            iy   = nsolute(i) + 3*(jwater-1) + j1
            do i1 = 1, 3
              force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1)  &
                                   + force_localj(1:3,i1,j1,k)
            end do
          end do
        end do
        do i1 = 1, 3
          do j1 = 1, 3
            do k = 1, num_count(j1,i1)
              iy = j_list(k,j1,i1)
              force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) +  &
                                     force_localj1(1:3,k,j1,i1)
            end do
          end do
        end do
        do i1 = 1, 3
          iy = ilist(i1)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + force_local(1:3,i1)
        end do

      end do
    end do


    ! interaction between different cells
    !
    do ij = id+1, maxcell, nthread

      i = cell_pairlist(1,ij)
      j = cell_pairlist(2,ij)

      num_nb15 = 0

      trans_x = cell_move(1,j,i)
      trans_y = cell_move(2,j,i)
      trans_z = cell_move(3,j,i)

      do iix = 1, nb15_cellww(ij)

        iwater = nb15_listww(iix,ij)
        force_local(1:3,1:3) = 0.0_wp
        
        do i1 = 1, 3
          ilist(i1) = nsolute(i) + 3*(iwater-1) + i1
          rtmp(1,i1)  = trans2(1,ilist(i1),i) + trans_x
          rtmp(2,i1)  = trans2(2,ilist(i1),i) + trans_y
          rtmp(3,i1)  = trans2(3,ilist(i1),i) + trans_z
        end do

        ini_nb15 = num_nb15 + 1
        fin_nb15 = num_nb15 + num_nb15_calcww(iwater,ij)
        num_nb15 = fin_nb15

        num_count(1:3,1:3) = 0
        small_count = 0

        do k = ini_nb15, fin_nb15
          jwater = nb15_calc_water(k,ij)
          ! first check O-O distance
          !
          iy   = nsolute(j) + 3*(jwater-1) + 1
          dij(1:3) = rtmp(1:3,1) - trans2(1:3,iy,j)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff_small) then
            small_count = small_count + 1
            j_small(small_count) = jwater
          else if (rij2 < cutoff_large) then
            do i1 = 1, 3
              do j1 = 1, 3
                iy = nsolute(j)+3*(jwater-1)+j1
                dij(1:3) = rtmp(1:3,i1) - trans2(1:3,iy,j)
                rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
                if (rij2 < cutoff2) then
                  num_count(j1,i1) = num_count(j1,i1) + 1
                  c = num_count(j1,i1)
                  dij_list(1:3,c,j1,i1) = dij(1:3)
                  dij_list(4,c,j1,i1) = rij2
                  j_list(c,j1,i1) = iy
                end if
              end do
            end do
          end if

        end do

        do k = 1, small_count
          jwater = j_small(k)

          do i1 = 1, 3
            do j1 = 1, 3

              iy = nsolute(j) + 3*(jwater-1) + j1
              dij(1:3) = rtmp(1:3,i1) - trans2(1:3,iy,j)
              rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
              rij2 = density * rij2
              L    = int(rij2)
              R    = rij2 - L
              h00  = (1.0_wp+2.0_wp*R) * (1.0_wp-R) * (1.0_wp-R)
              h10  = R * (1.0_wp-R) * (1.0_wp-R)
              h01  = R * R * (3.0_wp-2.0_wp*R)
              h11  = R * R * (R-1.0_wp)
              watertype = water_type(j1,i1)

              L1   = 6*L-5
              term_lj    = table_ene_WW(L1,watertype)*h00 +    &
                           table_ene_WW(L1+1,watertype)*h10
              term_elec  = table_ene_WW(L1+2,watertype)*h00 +  &
                           table_ene_WW(L1+3,watertype)*h10
              grad_coef  = table_ene_WW(L1+4,watertype)*h00 +  &
                           table_ene_WW(L1+5,watertype)*h10
              term_lj    = term_lj + table_ene_WW(L1+6,watertype)*h01 +    &
                                     table_ene_WW(L1+7,watertype)*h11
              term_elec  = term_elec + table_ene_WW(L1+8,watertype)*h01 +  &
                                       table_ene_WW(L1+9,watertype)*h11
              grad_coef = grad_coef + table_ene_WW(L1+10,watertype)*h01 +  &
                                      table_ene_WW(L1+11,watertype)*h11

              evdw_temp = evdw_temp + term_lj
              elec_temp = elec_temp + term_elec
              work(1)   = grad_coef*dij(1)
              work(2)   = grad_coef*dij(2)
              work(3)   = grad_coef*dij(3)
              force_local(1,i1) = force_local(1,i1) - work(1)
              force_local(2,i1) = force_local(2,i1) - work(2)
              force_local(3,i1) = force_local(3,i1) - work(3)
              force_localj(1,i1,j1,k) = work(1)
              force_localj(2,i1,j1,k) = work(2)
              force_localj(3,i1,j1,k) = work(3)
              viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
              viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
              viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
              viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
              viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
              viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

            end do
          end do

        end do

        do i1 = 1, 3
          do j1 = 1, 3

            do k = 1, num_count(j1,i1)
              iy = j_list(k,j1,i1)
              dij(1:3) = dij_list(1:3,k,j1,i1)
              rij2     = density * dij_list(4,k,j1,i1)
              L    = int(rij2)
              R    = rij2 - L
              h00  = (1.0_wp+2.0_wp*R) * (1.0_wp-R) * (1.0_wp-R)
              h10  = R * (1.0_wp-R) * (1.0_wp-R)
              h01  = R * R * (3.0_wp-2.0_wp*R)
              h11  = R * R * (R-1.0_wp)
              watertype = water_type(j1,i1)

              L1   = 6*L-5
              term_lj    = table_ene_WW(L1,watertype)*h00 +    &
                           table_ene_WW(L1+1,watertype)*h10
              term_elec  = table_ene_WW(L1+2,watertype)*h00 +  &
                           table_ene_WW(L1+3,watertype)*h10
              grad_coef  = table_ene_WW(L1+4,watertype)*h00 +  &
                           table_ene_WW(L1+5,watertype)*h10
              term_lj    = term_lj + table_ene_WW(L1+6,watertype)*h01 +    &
                                     table_ene_WW(L1+7,watertype)*h11
              term_elec  = term_elec + table_ene_WW(L1+8,watertype)*h01 +  &
                                       table_ene_WW(L1+9,watertype)*h11
              grad_coef = grad_coef + table_ene_WW(L1+10,watertype)*h01 +  &
                                      table_ene_WW(L1+11,watertype)*h11

              evdw_temp = evdw_temp + term_lj
              elec_temp = elec_temp + term_elec
              work(1)   = grad_coef*dij(1)
              work(2)   = grad_coef*dij(2)
              work(3)   = grad_coef*dij(3)
              force_local(1,i1) = force_local(1,i1) - work(1)
              force_local(2,i1) = force_local(2,i1) - work(2)
              force_local(3,i1) = force_local(3,i1) - work(3)
              force_localj1(1,k,j1,i1) = work(1)
              force_localj1(2,k,j1,i1) = work(2)
              force_localj1(3,k,j1,i1) = work(3)
              viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
              viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
              viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
              viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
              viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
              viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)
            end do
  
          end do
        end do

        do k = 1, small_count
          jwater = j_small(k)
          do j1 = 1, 3
            iy   = nsolute(j) + 3*(jwater-1) + j1
            do i1 = 1, 3
              force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1)  &
                                   + force_localj(1:3,i1,j1,k)
            end do
          end do
        end do
        do i1 = 1, 3
          do j1 = 1, 3
            do k = 1, num_count(j1,i1)
              iy = j_list(k,j1,i1)
              force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) +  &
                                     force_localj1(1:3,k,j1,i1)
            end do
          end do
        end do
        do i1 = 1, 3
          iy = ilist(i1)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + force_local(1:3,i1)
        end do

      end do

    end do

    do i = 1, 2
      do j = i+1, 3
        virial(i,j,id+1) = virial(i,j,id+1) - viri_local(j,i)
        virial(j,i,id+1) = virial(j,i,id+1) - viri_local(j,i)
      end do
    end do
    do i = 1, 3
      virial(i,i,id+1) = virial(i,i,id+1) - viri_local(i,i)
    end do

    eelec(id+1) = eelec(id+1) + elec_temp
    evdw(id+1)  = evdw(id+1)  + evdw_temp

    !$omp end parallel

    return

  end subroutine compute_energy_nonbond_table_water

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_force_nonbond_table_water
  !> @brief        calculate nonbonded force among water molecules using lookup 
  !!               table (linear interpolation)
  !! @authors      JJ 
  !! @param[in]    domain   : domain information
  !! @param[in]    enefunc  : potential energy functions
  !! @param[in]    pairlist : interaction list in each domain
  !! @param[inout] force    : forces for each cell
  !! @param[inout] virial   : virial term of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_force_nonbond_table_water(domain, enefunc, &
                                               pairlist, force, virial)

    ! formal arguments
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_pairlist), target, intent(in)    :: pairlist
    real(wp),                 intent(inout) :: force(:,:,:,:)
    real(wp),                 intent(inout) :: virial(3,3,nthread)

    ! local variables
    real(wp)                  :: dij(1:3), rij2, R, h00, h01, h10, h11
    real(wp)                  :: cutoff2, cutoff_small, cutoff_large
    real(wp)                  :: grad_coef
    real(wp)                  :: work(1:3)
    real(wp)                  :: viri_local(1:3,1:3)
    real(wp)                  :: trans_x, trans_y, trans_z
    real(wp)                  :: rtmp(3,3)
    real(wp)                  :: force_local(1:3,1:3)
    real(wp)                  :: force_localj(3,3,3,50)
    real(wp)                  :: force_localj1(3,50,3,3)
    real(wp)                  :: dij_list(4,50,3,3)
    integer                   :: i, iy, j, k, ij, iix, iwater, jwater
    integer                   :: num_nb15, ini_nb15, fin_nb15
    integer                   :: id, omp_get_thread_num
    integer                   :: i1, j1, L
    integer                   :: num_count(3,3), c, j_list(50,3,3)
    integer                   :: small_count, j_small(50)
    integer                   :: ilist(3)
    integer                   :: ncell_local
    integer                   :: water_type(1:3,1:3), watertype

    real(wp),         pointer :: coord(:,:,:), cell_move(:,:,:)
    real(wp),         pointer :: trans1(:,:,:), trans2(:,:,:)
    real(wp),         pointer :: density, cutoff
    real(wp),         pointer :: table_de_WW(:,:)
    integer,          pointer :: cell_pairlist(:,:), natom(:)
    integer,          pointer :: nwater(:)
    integer,          pointer :: nsolute(:)
    integer,          pointer :: nb15_cellww(:), nb15_listww(:,:)
    integer,          pointer :: num_nb15_calcww(:,:), nb15_calc_water(:,:)


    cell_pairlist   => domain%cell_pairlist1
    natom           => domain%num_atom
    nwater          => domain%num_water
    nsolute         => domain%num_solute
    coord           => domain%coord
    cell_move       => domain%cell_move
    trans1          => domain%trans_vec
    trans2          => domain%translated

    density         => enefunc%table%density
    table_de_WW     => enefunc%table%table_de_WW
    cutoff          => enefunc%cutoffdist

    nb15_cellww     => pairlist%nb15_cellww
    nb15_listww     => pairlist%nb15_listww
    num_nb15_calcww => pairlist%num_nb15_calcww
    nb15_calc_water => pairlist%nb15_calc_water

    ncell_local     =  domain%num_cell_local
    cutoff2         =  cutoff * cutoff
    cutoff_small    =  (cutoff-2.0_wp)*(cutoff-2.0_wp)
    cutoff_large    =  (cutoff+2.0_wp)*(cutoff+2.0_wp)

    water_type(1,1)     = 1
    water_type(1,2:3)   = 2
    water_type(2:3,1)   = 2
    water_type(2:3,2:3) = 3

    ! calculate energy and gradient
    !
    !$omp parallel default(shared)                                             &
    !$omp private(id, i, j, iwater, i1, ilist, rtmp, force_local, viri_local,  &
    !$omp         jwater, j1, j_list, rij2, L, R, grad_coef, work, ij,         &
    !$omp         ini_nb15, fin_nb15, num_nb15, trans_x, trans_y, trans_z,     &
    !$omp         iix, iy, num_count, k, c, small_count, j_small,              &
    !$omp         force_localj, force_localj1, dij, dij_list, watertype,       &
    !$omp         h00, h01, h10, h11)
    !
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    viri_local(1:3,1:3) = 0.0_wp

    ! energy within a cell
    !
    do i = id+1, ncell_local, nthread

      do iwater = 1, nwater(i) - 1

        do i1 = 1,3
          ilist(i1) = nsolute(i) + 3*(iwater-1) + i1
          rtmp(1:3,i1) = trans2(1:3,ilist(i1),i)
        end do

        force_local(1:3,1:3) = 0.0_wp
        num_count(1:3,1:3) = 0
        small_count = 0

        do jwater = iwater+1, nwater(i)

          iy   = nsolute(i) + 3*(jwater-1) + 1
          dij(1:3) = rtmp(1:3,1) - trans2(1:3,iy,i)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff_small) then
            small_count = small_count + 1
            j_small(small_count) = jwater
          else if (rij2 < cutoff_large) then
            do i1 = 1, 3
              do j1 = 1, 3
                iy = nsolute(i)+3*(jwater-1) + j1
                dij(1:3) = rtmp(1:3,i1) - trans2(1:3,iy,i)
                rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
                if (rij2 < cutoff2) then
                  num_count(j1,i1) = num_count(j1,i1) + 1
                  c = num_count(j1,i1)
                  dij_list(1:3,c,j1,i1) = dij(1:3)
                  dij_list(4,c,j1,i1) = rij2
                  j_list(c,j1,i1) = iy
                end if
              end do
            end do
          end if

        end do

        do k = 1, small_count
          jwater = j_small(k)

          do i1 = 1, 3
            do j1 = 1, 3

              iy = nsolute(i) + 3*(jwater-1) + j1
              dij(1:3) = rtmp(1:3,i1) - trans2(1:3,iy,i)
              rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
              rij2 = density * rij2
              L    = int(rij2)
              R    = rij2 - L
              h00  = (1.0_wp+2.0_wp*R) * (1.0_wp-R) * (1.0_wp-R)
              h10  = R * (1.0_wp-R) * (1.0_wp-R)
              h01  = R * R * (3.0_wp-2.0_wp*R)
              h11  = R * R * (R-1.0_wp)
              watertype = water_type(j1,i1)

              L   = 2*L-1
              grad_coef  = table_de_WW(L,watertype)*h00   +  &
                           table_de_WW(L+1,watertype)*h10 +  &
                           table_de_WW(L+2,watertype)*h01 +  &
                           table_de_WW(L+3,watertype)*h11

              work(1)   = grad_coef*dij(1)
              work(2)   = grad_coef*dij(2)
              work(3)   = grad_coef*dij(3)
              force_local(1,i1) = force_local(1,i1) - work(1)
              force_local(2,i1) = force_local(2,i1) - work(2)
              force_local(3,i1) = force_local(3,i1) - work(3)
              force_localj(1,i1,j1,k) = work(1)
              force_localj(2,i1,j1,k) = work(2)
              force_localj(3,i1,j1,k) = work(3)
              viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
              viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
              viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
              viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
              viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
              viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

            end do
          end do

        end do

        do i1 = 1, 3
          do j1 = 1, 3

            do k = 1, num_count(j1,i1)
              iy = j_list(k,j1,i1)
              dij(1:3) = dij_list(1:3,k,j1,i1)
              rij2     = density * dij_list(4,k,j1,i1)
              L    = int(rij2)
              R    = rij2 - L
              h00  = (1.0_wp+2.0_wp*R) * (1.0_wp-R) * (1.0_wp-R)
              h10  = R * (1.0_wp-R) * (1.0_wp-R)
              h01  = R * R * (3.0_wp-2.0_wp*R)
              h11  = R * R * (R-1.0_wp)
              watertype = water_type(j1,i1)

              L   = 2*L-1
              grad_coef  = table_de_WW(L,watertype)*h00   +  &
                           table_de_WW(L+1,watertype)*h10 +  &
                           table_de_WW(L+2,watertype)*h01 +  &
                           table_de_WW(L+3,watertype)*h11

              work(1)   = grad_coef*dij(1)
              work(2)   = grad_coef*dij(2)
              work(3)   = grad_coef*dij(3)
              force_local(1,i1) = force_local(1,i1) - work(1)
              force_local(2,i1) = force_local(2,i1) - work(2)
              force_local(3,i1) = force_local(3,i1) - work(3)
              force_localj1(1,k,j1,i1) = work(1)
              force_localj1(2,k,j1,i1) = work(2)
              force_localj1(3,k,j1,i1) = work(3)
              viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
              viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
              viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
              viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
              viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
              viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)
            end do

          end do
        end do

        do k = 1, small_count
          jwater = j_small(k)
          do j1 = 1, 3
            iy   = nsolute(i) + 3*(jwater-1) + j1
            do i1 = 1, 3
              force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1)  &
                                   + force_localj(1:3,i1,j1,k)
            end do
          end do
        end do
        do i1 = 1, 3
          do j1 = 1, 3
            do k = 1, num_count(j1,i1)
              iy = j_list(k,j1,i1)
              force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) +  &
                                     force_localj1(1:3,k,j1,i1)
            end do
          end do
        end do
        do i1 = 1, 3
          iy = ilist(i1)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + force_local(1:3,i1)
        end do

      end do
    end do


    ! interaction between different cells
    !
    do ij = id+1, maxcell, nthread

      i = cell_pairlist(1,ij)
      j = cell_pairlist(2,ij)

      num_nb15 = 0

      trans_x = cell_move(1,j,i)
      trans_y = cell_move(2,j,i)
      trans_z = cell_move(3,j,i)

      do iix = 1, nb15_cellww(ij)

        iwater = nb15_listww(iix,ij)
        force_local(1:3,1:3) = 0.0_wp

        do i1 = 1, 3
          ilist(i1) = nsolute(i) + 3*(iwater-1) + i1
          rtmp(1,i1)  = trans2(1,ilist(i1),i) + trans_x
          rtmp(2,i1)  = trans2(2,ilist(i1),i) + trans_y
          rtmp(3,i1)  = trans2(3,ilist(i1),i) + trans_z
        end do

        ini_nb15 = num_nb15 + 1
        fin_nb15 = num_nb15 + num_nb15_calcww(iwater,ij)
        num_nb15 = fin_nb15

        num_count(1:3,1:3) = 0
        small_count = 0

        do k = ini_nb15, fin_nb15
          jwater = nb15_calc_water(k,ij)
          ! first check O-O distance
          !
          iy   = nsolute(j) + 3*(jwater-1) + 1
          dij(1:3) = rtmp(1:3,1) - trans2(1:3,iy,j)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff_small) then
            small_count = small_count + 1
            j_small(small_count) = jwater
          else if (rij2 < cutoff_large) then
            do i1 = 1, 3
              do j1 = 1, 3
                iy = nsolute(j)+3*(jwater-1)+j1
                dij(1:3) = rtmp(1:3,i1) - trans2(1:3,iy,j)
                rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
                if (rij2 < cutoff2) then
                  num_count(j1,i1) = num_count(j1,i1) + 1
                  c = num_count(j1,i1)
                  dij_list(1:3,c,j1,i1) = dij(1:3)
                  dij_list(4,c,j1,i1) = rij2
                  j_list(c,j1,i1) = iy
                end if
              end do
            end do
          end if

        end do

        do k = 1, small_count
          jwater = j_small(k)

          do i1 = 1, 3
            do j1 = 1, 3

              iy = nsolute(j) + 3*(jwater-1) + j1
              dij(1:3) = rtmp(1:3,i1) - trans2(1:3,iy,j)
              rij2 = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
              rij2 = density * rij2
              L    = int(rij2)
              R    = rij2 - L
              h00  = (1.0_wp+2.0_wp*R) * (1.0_wp-R) * (1.0_wp-R)
              h10  = R * (1.0_wp-R) * (1.0_wp-R)
              h01  = R * R * (3.0_wp-2.0_wp*R)
              h11  = R * R * (R-1.0_wp)
              watertype = water_type(j1,i1)

              L   = 2*L-1
              grad_coef  = table_de_WW(L,watertype)*h00   +  &
                           table_de_WW(L+1,watertype)*h10 +  &
                           table_de_WW(L+2,watertype)*h01 +  &
                           table_de_WW(L+3,watertype)*h11

              work(1)   = grad_coef*dij(1)
              work(2)   = grad_coef*dij(2)
              work(3)   = grad_coef*dij(3)
              force_local(1,i1) = force_local(1,i1) - work(1)
              force_local(2,i1) = force_local(2,i1) - work(2)
              force_local(3,i1) = force_local(3,i1) - work(3)
              force_localj(1,i1,j1,k) = work(1)
              force_localj(2,i1,j1,k) = work(2)
              force_localj(3,i1,j1,k) = work(3)
              viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
              viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
              viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
              viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
              viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
              viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)

            end do
          end do

        end do

        do i1 = 1, 3
          do j1 = 1, 3

            do k = 1, num_count(j1,i1)
              iy = j_list(k,j1,i1)
              dij(1:3) = dij_list(1:3,k,j1,i1)
              rij2     = density * dij_list(4,k,j1,i1)
              L    = int(rij2)
              R    = rij2 - L
              h00  = (1.0_wp+2.0_wp*R) * (1.0_wp-R) * (1.0_wp-R)
              h10  = R * (1.0_wp-R) * (1.0_wp-R)
              h01  = R * R * (3.0_wp-2.0_wp*R)
              h11  = R * R * (R-1.0_wp)
              watertype = water_type(j1,i1)

              L   = 2*L-1
              grad_coef  = table_de_WW(L,watertype)*h00   +  &
                           table_de_WW(L+1,watertype)*h10 +  &
                           table_de_WW(L+2,watertype)*h01 +  &
                           table_de_WW(L+3,watertype)*h11

              work(1)   = grad_coef*dij(1)
              work(2)   = grad_coef*dij(2)
              work(3)   = grad_coef*dij(3)
              force_local(1,i1) = force_local(1,i1) - work(1)
              force_local(2,i1) = force_local(2,i1) - work(2)
              force_local(3,i1) = force_local(3,i1) - work(3)
              force_localj1(1,k,j1,i1) = work(1)
              force_localj1(2,k,j1,i1) = work(2)
              force_localj1(3,k,j1,i1) = work(3)
              viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
              viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
              viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
              viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
              viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
              viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)
            end do

          end do
        end do

        do k = 1, small_count
          jwater = j_small(k)
          do j1 = 1, 3
            iy   = nsolute(j) + 3*(jwater-1) + j1
            do i1 = 1, 3
              force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1)  &
                                   + force_localj(1:3,i1,j1,k)
            end do
          end do
        end do
        do i1 = 1, 3
          do j1 = 1, 3
            do k = 1, num_count(j1,i1)
              iy = j_list(k,j1,i1)
              force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) +  &
                                     force_localj1(1:3,k,j1,i1)
            end do
          end do
        end do
        do i1 = 1, 3
          iy = ilist(i1)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + force_local(1:3,i1)
        end do

      end do

    end do

    do i = 1, 2
      do j = i+1, 3
        virial(i,j,id+1) = virial(i,j,id+1) - viri_local(j,i)
        virial(j,i,id+1) = virial(j,i,id+1) - viri_local(j,i)
      end do
    end do
    do i = 1, 3
      virial(i,i,id+1) = virial(i,i,id+1) - viri_local(i,i)
    end do

    !$omp end parallel

    return

  end subroutine compute_force_nonbond_table_water

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_energy_nonbond_table
  !> @brief        calculate nonbonded energy with lookup table
  !! @authors      JJ 
  !! @param[in]    domain   : domain information
  !! @param[in]    enefunc  : potential energy functions
  !! @param[in]    pairlist : interaction list in each domain
  !! @param[inout] force    : forces for each cell
  !! @param[inout] virial   : virial term of target systems
  !! @param[inout] eelec    : electrostatic energy of target systems
  !! @param[inout] evdw     : van der Waals energy of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_energy_nonbond_table(domain, enefunc, pairlist, &
                                                 force, virial, eelec, evdw)

    ! formal arguments
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_pairlist), target, intent(in)    :: pairlist
    real(wp),                 intent(inout) :: force(:,:,:,:)
    real(wp),                 intent(inout) :: virial(3,3,nthread)
    real(wp),                 intent(inout) :: eelec(nthread)
    real(wp),                 intent(inout) :: evdw(nthread)

    ! local variables
    real(wp)                  :: dij(1:3), rij2
    real(wp)                  :: lj6, lj12
    real(wp)                  :: R, h00, h01, h10, h11
    real(wp)                  :: term_lj12, term_lj6, term_elec
    real(wp)                  :: cutoff2, term_lj, grad_coef
    real(wp)                  :: work(1:3)
    real(wp)                  :: viri_local(1:3,1:3)
    real(wp)                  :: trans_x, trans_y, trans_z
    real(wp)                  :: rtmp(1:3), qtmp
    real(wp)                  :: elec_temp, evdw_temp
    real(wp)                  :: dij_list(3,100)
    real(wp)                  :: force_local(3), rij2_list(100)
    real(wp)                  :: force_localj(3,100)
    integer                   :: i, ix, iy, j, k, ij, iix, iwater, list, L, L1
    integer                   :: num_nb15, ini_nb15, fin_nb15, num_count
    integer                   :: id, omp_get_thread_num, j_list(100)
    integer                   :: ncell, ncell_local

    real(wp),         pointer :: coord(:,:,:), charge(:,:)
    real(wp),         pointer :: trans1(:,:,:), trans2(:,:,:)
    real(wp),         pointer :: cell_move(:,:,:)
    real(wp),         pointer :: nonb_lj12(:,:), nonb_lj6(:,:)
    real(wp),         pointer :: density, cutoff
    real(wp),         pointer :: table_ene(:), table_grad(:)
    integer,          pointer :: cell_pairlist(:,:), atmcls(:,:)
    integer,          pointer :: natom(:), nsolute(:), nwater(:)
    integer,          pointer :: nb15_list(:,:)
    integer,          pointer :: nb15_cell(:)
    integer,          pointer :: num_nb15_calc1(:,:), num_nb15_calc(:,:)
    integer,          pointer :: nb15_calc_list1(:,:), nb15_calc_list(:,:)


    cell_pairlist   => domain%cell_pairlist1
    atmcls          => domain%atom_cls_no
    natom           => domain%num_atom
    nsolute         => domain%num_solute
    nwater          => domain%num_water
    coord           => domain%coord
    trans1          => domain%trans_vec
    trans2          => domain%translated
    charge          => domain%charge
    cell_move       => domain%cell_move

    table_ene       => enefunc%table%table_ene
    table_grad      => enefunc%table%table_grad
    density         => enefunc%table%density
    cutoff          => enefunc%cutoffdist
    nonb_lj12       => enefunc%nonb_lj12
    nonb_lj6        => enefunc%nonb_lj6

    nb15_list       => pairlist%nb15_list
    nb15_cell       => pairlist%nb15_cell
    num_nb15_calc1  => pairlist%num_nb15_calc1
    num_nb15_calc   => pairlist%num_nb15_calc
    nb15_calc_list1 => pairlist%nb15_calc_list1
    nb15_calc_list  => pairlist%nb15_calc_list

    ncell           =  domain%num_cell_local + domain%num_cell_boundary
    ncell_local     =  domain%num_cell_local
    cutoff2         =  cutoff * cutoff

    ! calculate energy and gradient
    !

    !$omp parallel default(shared)                                             &
    !$omp private(id, i, ix, num_nb15, rtmp, qtmp, ini_nb15, fin_nb15, k, iy,  &
    !$omp         rij2, L, R, term_lj12, term_lj6, term_lj, grad_coef, work,   &
    !$omp         term_elec, viri_local, iwater, ij, j, trans_x, trans_y,      &
    !$omp         trans_z, iix, list, num_count, j_list, dij, dij_list, rij2_list,  &
    !$omp         force_local, force_localj, lj6, lj12, L1, elec_temp, &
    !$omp         evdw_temp, h00, h01, h10, h11)             
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif
    do i = id+1, ncell, nthread
      do ix = 1, natom(i)
        trans2(1,ix,i) = coord(1,ix,i) + trans1(1,ix,i)
        trans2(2,ix,i) = coord(2,ix,i) + trans1(2,ix,i)
        trans2(3,ix,i) = coord(3,ix,i) + trans1(3,ix,i)
      end do
    end do

    !$omp barrier

    viri_local(1:3,1:3) = 0.0_wp
    elec_temp = 0.0_wp
    evdw_temp = 0.0_wp

    ! energy within a cell(solute-solute)
    !
    do i = id+1, ncell_local, nthread

      num_nb15 = 0
      do ix = 1, natom(i) - 1

        rtmp(1:3) = trans2(1:3,ix,i)
        qtmp = charge(ix,i)
        ini_nb15  = num_nb15 + 1
        fin_nb15  = num_nb15 + num_nb15_calc1(ix,i)
        num_nb15  = fin_nb15
        num_count = 0

        do k = ini_nb15, fin_nb15

          iy = nb15_calc_list1(k,i)

          ! compute distance
          !
          dij(1) = rtmp(1) - trans2(1,iy,i)
          dij(2) = rtmp(2) - trans2(2,iy,i)
          dij(3) = rtmp(3) - trans2(3,iy,i)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff2) then
            num_count = num_count + 1
            dij_list(1:3,num_count)  = dij(1:3)
            rij2_list(num_count) = rij2
            j_list(num_count)   = iy
          end if
        end do

        force_local(1:3) = 0.0_wp

!ocl norecurrence(force)
        do k = 1, num_count
          dij(1) = dij_list(1,k)
          dij(2) = dij_list(2,k)
          dij(3) = dij_list(3,k)
          iy     = j_list(k)
          rij2   = density * rij2_list(k)
          lj6    = nonb_lj6(atmcls(ix,i),atmcls(iy,i))
          lj12   = nonb_lj12(atmcls(ix,i),atmcls(iy,i))

          L    = int(rij2)
          R    = rij2 - L
          h00  = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10  = R*(1.0_wp-R)*(1.0_wp-R)
          h01  = R*R*(3.0_wp-2.0_wp*R)
          h11  = R*R*(R-1.0_wp)

          L1 = 6*L - 5
          term_lj12 = table_ene(L1)*h00 + table_ene(L1+1)*h10
          term_lj6  = table_ene(L1+2)*h00 + table_ene(L1+3)*h10
          term_elec = table_ene(L1+4)*h00 + table_ene(L1+5)*h10
          term_lj12 = term_lj12 + table_ene(L1+6)*h01 + table_ene(L1+7)*h11
          term_lj6  = term_lj6  + table_ene(L1+8)*h01 + table_ene(L1+9)*h11
          term_elec = term_elec + table_ene(L1+10)*h01 + table_ene(L1+11)*h11
          evdw_temp = evdw_temp + term_lj12*lj12 - term_lj6*lj6
          elec_temp = elec_temp + qtmp*charge(iy,i)*term_elec

          term_lj12 = table_grad(L1)*h00 + table_grad(L1+1)*h10
          term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
          term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
          term_lj12 = term_lj12 + table_grad(L1+6)*h01 + table_grad(L1+7)*h11
          term_lj6  = term_lj6 + table_grad(L1+8)*h01 + table_grad(L1+9)*h11
          term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(iy,i)*term_elec

          work(1) = grad_coef*dij(1)
          work(2) = grad_coef*dij(2)
          work(3) = grad_coef*dij(3)

          ! store force
          !
          force_local(1) = force_local(1) - work(1)
          force_local(2) = force_local(2) - work(2)
          force_local(3) = force_local(3) - work(3)
#ifdef KCOMP
          force(1,iy,i,id+1) = force(1,iy,i,id+1) + work(1)
          force(2,iy,i,id+1) = force(2,iy,i,id+1) + work(2)
          force(3,iy,i,id+1) = force(3,iy,i,id+1) + work(3)
#else
          force_localj(1,k)  = work(1)
          force_localj(2,k)  = work(2)
          force_localj(3,k)  = work(3)
#endif

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)
        end do

#ifndef KCOMP
        do k = 1, num_count
          iy = j_list(k)
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + force_localj(1:3,k)
        end do
#endif
        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
      end do

    end do


    ! interaction between different cells
    !
    do ij = id+1, maxcell, nthread

      i = cell_pairlist(1,ij)
      j = cell_pairlist(2,ij)

      num_nb15 = 0

      trans_x = cell_move(1,j,i)
      trans_y = cell_move(2,j,i)
      trans_z = cell_move(3,j,i)

      do iix = 1, nb15_cell(ij)

        ix   = nb15_list(iix,ij)
        rtmp(1:3) = trans2(1:3,ix,i)
        qtmp = charge(ix,i)

        ini_nb15 = num_nb15 + 1
        fin_nb15 = num_nb15 + num_nb15_calc(ix,ij)
        num_nb15 = fin_nb15

        num_count = 0
        do k = ini_nb15, fin_nb15
          iy   = nb15_calc_list(k,ij)
          dij(1) = rtmp(1) - trans2(1,iy,j) + trans_x
          dij(2) = rtmp(2) - trans2(2,iy,j) + trans_y
          dij(3) = rtmp(3) - trans2(3,iy,j) + trans_z
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)

          if (rij2 < cutoff2) then
            num_count = num_count + 1
            dij_list(1:3,num_count)  = dij(1:3)
            rij2_list(num_count) = rij2
            j_list(num_count)   = iy
          end if
        end do

        force_local(1:3) = 0.0_wp

!ocl norecurrence(force)
        do k = 1, num_count
          dij(1) = dij_list(1,k)
          dij(2) = dij_list(2,k)
          dij(3) = dij_list(3,k)
          iy   = j_list(k)
          rij2  = density * rij2_list(k)
          lj6  = nonb_lj6 (atmcls(ix,i),atmcls(iy,j))
          lj12 = nonb_lj12(atmcls(ix,i),atmcls(iy,j))

          L    = int(rij2)
          R    = rij2 - L
          h00  = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10  = R*(1.0_wp-R)*(1.0_wp-R)
          h01  = R*R*(3.0_wp-2.0_wp*R)
          h11  = R*R*(R-1.0_wp)

          L1 = 6*L - 5
          term_lj12 = table_ene(L1)*h00 + table_ene(L1+1)*h10
          term_lj6  = table_ene(L1+2)*h00 + table_ene(L1+3)*h10
          term_elec = table_ene(L1+4)*h00 + table_ene(L1+5)*h10
          term_lj12 = term_lj12 + table_ene(L1+6)*h01 + table_ene(L1+7)*h11
          term_lj6  = term_lj6 + table_ene(L1+8)*h01 + table_ene(L1+9)*h11
          term_elec = term_elec + table_ene(L1+10)*h01 + table_ene(L1+11)*h11
          evdw_temp = evdw_temp + term_lj12*lj12 - term_lj6*lj6
          elec_temp = elec_temp + qtmp*charge(iy,j)*term_elec

          term_lj12 = table_grad(L1)*h00 + table_grad(L1+1)*h10
          term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
          term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
          term_lj12 = term_lj12 + table_grad(L1+6)*h01 + table_grad(L1+7)*h11
          term_lj6  = term_lj6 + table_grad(L1+8)*h01 + table_grad(L1+9)*h11
          term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(iy,j)*term_elec

          work(1) = grad_coef*dij(1)
          work(2) = grad_coef*dij(2)
          work(3) = grad_coef*dij(3)

          ! store force
          !
          force_local(1) = force_local(1) - work(1)
          force_local(2) = force_local(2) - work(2)
          force_local(3) = force_local(3) - work(3)

#ifdef KCOMP
          force(1,iy,j,id+1) = force(1,iy,j,id+1) + work(1)
          force(2,iy,j,id+1) = force(2,iy,j,id+1) + work(2)
          force(3,iy,j,id+1) = force(3,iy,j,id+1) + work(3)
#else
          force_localj(1,k) = work(1)
          force_localj(2,k) = work(2)
          force_localj(3,k) = work(3)
#endif

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)
        end do

#ifndef KCOMP
        do k = 1, num_count
          iy = j_list(k)
          force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) + force_localj(1:3,k)
        end do
#endif
        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)

      end do

    end do

    do i = 1, 2
      do j = i+1, 3
        virial(i,j,id+1) = virial(i,j,id+1) - viri_local(j,i)
        virial(j,i,id+1) = virial(j,i,id+1) - viri_local(j,i)
      end do
    end do
    do i = 1, 3
      virial(i,i,id+1) = virial(i,i,id+1) - viri_local(i,i)
    end do

    eelec(id+1) = eelec(id+1) + elec_temp
    evdw(id+1)  = evdw(id+1) + evdw_temp

    !$omp end parallel

    return

  end subroutine compute_energy_nonbond_table

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    compute_force_nonbond_table
  !> @brief        calculate nonbonded force without solvents with lookup table
  !! @authors      JJ 
  !! @param[in]    domain   : domain information
  !! @param[in]    enefunc  : potential energy functions
  !! @param[in]    pairlist : interaction list in each domain
  !! @param[inout] force    : forces for each cell
  !! @param[inout] virial   : virial term of target systems
  ! 
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compute_force_nonbond_table(domain, enefunc, pairlist, &
                                                force, virial)

    ! formal arguments
    type(s_domain),   target, intent(in)    :: domain
    type(s_enefunc),  target, intent(in)    :: enefunc
    type(s_pairlist), target, intent(in)    :: pairlist
    real(wp),                 intent(inout) :: force(:,:,:,:)
    real(wp),                 intent(inout) :: virial(3,3,nthread)

    ! local variables
    real(wp)                  :: dij(1:3), rij2
    real(wp)                  :: R, h00, h01, h10, h11, lj12, lj6
    real(wp)                  :: term_lj12, term_lj6, term_elec
    real(wp)                  :: cutoff2, grad_coef
    real(wp)                  :: work(1:3)
    real(wp)                  :: viri_local(1:3,1:3)
    real(wp)                  :: trans_x, trans_y, trans_z
    real(wp)                  :: rtmp(1:3), qtmp
    real(wp)                  :: force_local(3), force_local_iy(3,250)
    real(wp)                  :: dij_list(4,250)
    real(wp)                  :: ieps, jeps, eps, irmin, jrmin, rmin
    integer                   :: i, ix, iy, j, k, ij, iix, iwater, list, L, L1
    integer                   :: num_nb15, ini_nb15, fin_nb15, num_count
    integer                   :: id, omp_get_thread_num, j_list(250)
    integer                   :: iatmcls,jatmcls
    integer                   :: ncell, ncell_local

    real(wp),         pointer :: coord(:,:,:), charge(:,:)
    real(wp),         pointer :: trans1(:,:,:), trans2(:,:,:)
    real(wp),         pointer :: cell_move(:,:,:)
    real(wp),         pointer :: nonb_lj12(:,:), nonb_lj6(:,:)
    real(wp),         pointer :: density, cutoff
    real(wp),         pointer :: table_grad(:)
    integer,          pointer :: cell_pairlist(:,:), atmcls(:,:)
    integer,          pointer :: natom(:), nsolute(:), nwater(:)
    integer,          pointer :: nb15_list(:,:)
    integer,          pointer :: nb15_cell(:)
    integer,          pointer :: num_nb15_calc1(:,:), num_nb15_calc(:,:)
    integer,          pointer :: nb15_calc_list1(:,:), nb15_calc_list(:,:)


    cell_pairlist   => domain%cell_pairlist1
    atmcls          => domain%atom_cls_no
    natom           => domain%num_atom
    nsolute         => domain%num_solute
    nwater          => domain%num_water
    coord           => domain%coord
    trans1          => domain%trans_vec
    trans2          => domain%translated
    charge          => domain%charge
    cell_move       => domain%cell_move

    table_grad      => enefunc%table%table_grad
    density         => enefunc%table%density
    cutoff          => enefunc%cutoffdist
    nonb_lj12       => enefunc%nonb_lj12
    nonb_lj6        => enefunc%nonb_lj6 

    nb15_list       => pairlist%nb15_list
    nb15_cell       => pairlist%nb15_cell
    num_nb15_calc1  => pairlist%num_nb15_calc1
    num_nb15_calc   => pairlist%num_nb15_calc 
    nb15_calc_list1 => pairlist%nb15_calc_list1
    nb15_calc_list  => pairlist%nb15_calc_list

    ncell           =  domain%num_cell_local + domain%num_cell_boundary
    ncell_local     =  domain%num_cell_local
    cutoff2         =  cutoff * cutoff

    ! calculate energy and gradient
    !

    !$omp parallel default(shared)                                             &
    !$omp private(id, i, ix, num_nb15, rtmp, qtmp, ini_nb15, fin_nb15, k, iy,  &
    !$omp         rij2, L, R, term_lj12, term_lj6, grad_coef, work, term_elec, &
    !$omp         viri_local, iwater, ij, j, trans_x, trans_y, trans_z,        &
    !$omp         iix, list, num_count, j_list, force_local, force_local_iy,   &
    !$omp         lj12, lj6, L1, dij, dij_list, iatmcls, jatmcls, ieps, jeps, eps,  &
    !$omp         irmin, jrmin, rmin, h00, h01, h10, h11)
    !
#ifdef OMP
    id = omp_get_thread_num()
#else
    id = 0
#endif

    do i = id+1, ncell, nthread
      do ix = 1, natom(i)
        trans2(1,ix,i) = coord(1,ix,i) + trans1(1,ix,i)
        trans2(2,ix,i) = coord(2,ix,i) + trans1(2,ix,i)
        trans2(3,ix,i) = coord(3,ix,i) + trans1(3,ix,i)
      end do
    end do

    !$omp barrier

    viri_local(1:3,1:3) = 0.0_wp

    ! energy within a cell(solute-solute)
    !
    do i = id+1, ncell_local, nthread

      num_nb15 = 0

      do ix = 1, natom(i) - 1

        rtmp(1:3) = trans2(1:3,ix,i)
        qtmp = charge(ix,i)
        iatmcls = atmcls(ix,i)
        ieps    = lj_coef(1,iatmcls)
        irmin   = lj_coef(2,iatmcls)

        ini_nb15  = num_nb15 + 1
        fin_nb15  = num_nb15 + num_nb15_calc1(ix,i)
        num_nb15  = fin_nb15
        force_local(1:3) = 0.0_wp
        num_count = 0

!ocl norecurrence(force)
        do k = ini_nb15, fin_nb15

          iy = nb15_calc_list1(k,i)
          dij(1) = rtmp(1) - trans2(1,iy,i)
          dij(2) = rtmp(2) - trans2(2,iy,i)
          dij(3) = rtmp(3) - trans2(3,iy,i)
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
#ifdef KCOMP
          rij2  = density * rij2

          jatmcls = atmcls(iy,i)
          jeps    = lj_coef(1,jatmcls)
          jrmin   = lj_coef(2,jatmcls)
          eps     = sqrt(ieps*jeps)
          rmin    = irmin + jrmin
          rmin    = rmin * rmin * rmin
          rmin    = rmin * rmin
          lj12    = eps * (rmin * rmin)
          lj6     = 2.0_wp * eps * rmin

          L  = int(rij2)
          R  = rij2 - L
          h00   = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10   = R*(1.0_wp-R)*(1.0_wp-R)
          h01   = R*R*(3.0_wp-2.0_wp*R)
          h11   = R*R*(R-1.0_wp)

          L1 = 6*L - 5
          term_lj12 = table_grad(L1)*h00 + table_grad(L1+1)*h10
          term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
          term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
          term_lj12 = term_lj12 + table_grad(L1+6)*h01 + table_grad(L1+7)*h11
          term_lj6  = term_lj6 + table_grad(L1+8)*h01 + table_grad(L1+9)*h11
          term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(iy,i)*term_elec

          work(1) = grad_coef*dij(1)
          work(2) = grad_coef*dij(2)
          work(3) = grad_coef*dij(3)

          ! store force
          !
          force_local(1) = force_local(1) - work(1)
          force_local(2) = force_local(2) - work(2)
          force_local(3) = force_local(3) - work(3)
          force(1,iy,i,id+1) = force(1,iy,i,id+1) + work(1)
          force(2,iy,i,id+1) = force(2,iy,i,id+1) + work(2)
          force(3,iy,i,id+1) = force(3,iy,i,id+1) + work(3)

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)
        end do
#else
          if (rij2 < cutoff2) then
            num_count = num_count + 1
            dij_list(1,num_count) = dij(1)
            dij_list(2,num_count) = dij(2)
            dij_list(3,num_count) = dij(3)
            dij_list(4,num_count) = rij2
            j_list(num_count) = iy
          end if
        end do
        force_local(1:3) = 0.0_wp

        do k = 1, num_count
          iy   = j_list(k)
          dij(1) = dij_list(1,k)
          dij(2) = dij_list(2,k)
          dij(3) = dij_list(3,k)
          rij2  = density * dij_list(4,k)

          lj12 = nonb_lj12(atmcls(ix,i),atmcls(iy,i))
          lj6  = nonb_lj6(atmcls(ix,i),atmcls(iy,i))

          L    = int(rij2)
          R    = rij2 - L
          h00  = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10  = R*(1.0_wp-R)*(1.0_wp-R)
          h01  = R*R*(3.0_wp-2.0_wp*R)
          h11  = R*R*(R-1.0_wp)

          L1 = 6*L - 5
          term_lj12 = table_grad(L1)*h00 + table_grad(L1+1)*h10
          term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
          term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
          term_lj12 = term_lj12 + table_grad(L1+6)*h01 + table_grad(L1+7)*h11
          term_lj6  = term_lj6 + table_grad(L1+8)*h01 + table_grad(L1+9)*h11
          term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(iy,i)*term_elec

          work(1) = grad_coef*dij(1)
          work(2) = grad_coef*dij(2)
          work(3) = grad_coef*dij(3)

          ! store force
          !
          force_local(1) = force_local(1) - work(1)
          force_local(2) = force_local(2) - work(2)
          force_local(3) = force_local(3) - work(3)
          force_local_iy(1,k) = work(1)
          force_local_iy(2,k) = work(2)
          force_local_iy(3,k) = work(3)

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)
        end do
#endif

        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
#ifndef KCOMP
        do k = 1, num_count
          iy = j_list(k) 
          num_count = k - ini_nb15 + 1
          force(1:3,iy,i,id+1) = force(1:3,iy,i,id+1) + force_local_iy(1:3,k)
        end do
#endif
      end do

    end do

    ! interaction between different cells
    !
    do ij = id+1, maxcell, nthread

      i = cell_pairlist(1,ij)
      j = cell_pairlist(2,ij)

      num_nb15 = 0

      trans_x = cell_move(1,j,i)
      trans_y = cell_move(2,j,i)
      trans_z = cell_move(3,j,i)

      do iix = 1, nb15_cell(ij)

        ix = nb15_list(iix,ij)
        rtmp(1:3) = trans2(1:3,ix,i)
        qtmp = charge(ix,i)
        iatmcls = atmcls(ix,i)
        ieps    = lj_coef(1,iatmcls)
        irmin   = lj_coef(2,iatmcls)

        ini_nb15  = num_nb15 + 1
        fin_nb15  = num_nb15 + num_nb15_calc(ix,ij)
        num_nb15  = fin_nb15

        force_local(1:3) = 0.0_wp
        num_count = 0

!ocl norecurrence(force)
        do k = ini_nb15, fin_nb15
          iy = nb15_calc_list(k,ij)

          dij(1) = rtmp(1) - trans2(1,iy,j) + trans_x
          dij(2) = rtmp(2) - trans2(2,iy,j) + trans_y
          dij(3) = rtmp(3) - trans2(3,iy,j) + trans_z
          rij2  = dij(1)*dij(1) + dij(2)*dij(2) + dij(3)*dij(3)
#ifdef KCOMP
          rij2 = density * rij2 

          jatmcls = atmcls(iy,j)
          jeps    = lj_coef(1,jatmcls)
          jrmin   = lj_coef(2,jatmcls)
          eps     = sqrt(ieps*jeps)
          rmin    = irmin + jrmin
          rmin    = rmin * rmin * rmin
          rmin    = rmin * rmin
          lj12    = eps * (rmin * rmin)
          lj6     = 2.0_wp * eps * rmin

          L  = int(rij2)
          R  = rij2 - L
          h00   = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10   = R*(1.0_wp-R)*(1.0_wp-R)
          h01   = R*R*(3.0_wp-2.0_wp*R)
          h11   = R*R*(R-1.0_wp)

          L1 = 6*L - 5
          term_lj12 = table_grad(L1)*h00 + table_grad(L1+1)*h10
          term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
          term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
          term_lj12 = term_lj12 + table_grad(L1+6)*h01 + table_grad(L1+7)*h11
          term_lj6  = term_lj6 + table_grad(L1+8)*h01 + table_grad(L1+9)*h11
          term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(iy,j)*term_elec

          work(1) = grad_coef*dij(1)
          work(2) = grad_coef*dij(2)
          work(3) = grad_coef*dij(3)

          ! store force
          !
          force_local(1) = force_local(1) - work(1)
          force_local(2) = force_local(2) - work(2)
          force_local(3) = force_local(3) - work(3)
          force(1,iy,j,id+1) = force(1,iy,j,id+1) + work(1)
          force(2,iy,j,id+1) = force(2,iy,j,id+1) + work(2)
          force(3,iy,j,id+1) = force(3,iy,j,id+1) + work(3)

          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)
        end do

#else
          if (rij2 < cutoff2) then
            num_count = num_count + 1
            dij_list(1,num_count) = dij(1)
            dij_list(2,num_count) = dij(2)
            dij_list(3,num_count) = dij(3)
            dij_list(4,num_count) = rij2
            j_list(num_count) = iy
          end if
        end do

        force_local(1:3) = 0.0_wp
        do k = 1, num_count
          iy = j_list(k)
          dij(1) = dij_list(1,k)
          dij(2) = dij_list(2,k)
          dij(3) = dij_list(3,k)
          rij2  = density * dij_list(4,k)
          lj12 = nonb_lj12(atmcls(ix,i),atmcls(iy,j))
          lj6  = nonb_lj6(atmcls(ix,i),atmcls(iy,j))

          L  = int(rij2)
          R  = rij2 - L

          h00  = (1.0_wp + 2.0_wp*R)*(1.0_wp-R)*(1.0_wp-R)
          h10  = R*(1.0_wp-R)*(1.0_wp-R)
          h01  = R*R*(3.0_wp-2.0_wp*R)
          h11  = R*R*(R-1.0_wp)

          L1 = 6*L - 5
          term_lj12 = table_grad(L1)*h00 + table_grad(L1+1)*h10
          term_lj6  = table_grad(L1+2)*h00 + table_grad(L1+3)*h10
          term_elec = table_grad(L1+4)*h00 + table_grad(L1+5)*h10
          term_lj12 = term_lj12 + table_grad(L1+6)*h01 + table_grad(L1+7)*h11
          term_lj6  = term_lj6 + table_grad(L1+8)*h01 + table_grad(L1+9)*h11
          term_elec = term_elec + table_grad(L1+10)*h01 + table_grad(L1+11)*h11
          grad_coef = term_lj12*lj12 - term_lj6*lj6 + qtmp*charge(iy,j)*term_elec

          work(1) = grad_coef*dij(1)
          work(2) = grad_coef*dij(2)
          work(3) = grad_coef*dij(3)

          ! store force
          !
          force_local(1) = force_local(1) - work(1)
          force_local(2) = force_local(2) - work(2)
          force_local(3) = force_local(3) - work(3)
          force_local_iy(1,k) = work(1)
          force_local_iy(2,k) = work(2)
          force_local_iy(3,k) = work(3)
          ! virial
          !
          viri_local(1,1) = viri_local(1,1) + dij(1)*work(1)
          viri_local(2,1) = viri_local(2,1) + dij(2)*work(1)
          viri_local(2,2) = viri_local(2,2) + dij(2)*work(2)
          viri_local(3,1) = viri_local(3,1) + dij(3)*work(1)
          viri_local(3,2) = viri_local(3,2) + dij(3)*work(2)
          viri_local(3,3) = viri_local(3,3) + dij(3)*work(3)
        end do

        do k = 1, num_count
          iy = j_list(k)
          force(1:3,iy,j,id+1) = force(1:3,iy,j,id+1) + force_local_iy(1:3,k)
        end do
#endif
        force(1:3,ix,i,id+1) = force(1:3,ix,i,id+1) + force_local(1:3)
      end do

    end do

    do i = 1, 2
      do j = i+1, 3
        virial(i,j,id+1) = virial(i,j,id+1) - viri_local(j,i)
        virial(j,i,id+1) = virial(j,i,id+1) - viri_local(j,i)
      end do
    end do
    do i = 1, 3
      virial(i,i,id+1) = virial(i,i,id+1) - viri_local(i,i)
    end do

    !$omp end parallel

    return

  end subroutine compute_force_nonbond_table

end module sp_energy_table_cubic_mod
